!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
! for rotwfn
!
module arch_rotwfn
contains
  subroutine ltrovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: x(b:e,2,m),y(b:e,2,m)
    complex(kind=8):: g(m,m)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do schedule(static,1)
!POPTION CYCLIC
    do k=1,m
!SOPTION NOLOOPINTERCHANGE, UNROLL(4)
!CDIR NOLOOPCHG
       do j=k,m
          do i=b,e
             g(j,k) = dcmplx( &
                  dble(g(j,k)) &
                  + x(i,1,j)*y(i,1,k) + x(i,2,j)*y(i,2,k), &
                  aimag(g(j,k)) &
                  + x(i,1,j)*y(i,2,k) - x(i,2,j)*y(i,1,k) )
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ltrovlmat

  subroutine ovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: x(b:e,2,m),y(b:e,2,m)
    complex(kind=8):: g(m,m)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(3)
    do k=1,m
!SOPTION NOLOOPINTERCHANGE, UNROLL(3)
!CDIR NOLOOPCHG
       do j=1,m
          do i=b,e
             g(j,k) = dcmplx( &
                  dble(g(j,k)) &
                  + x(i,1,j)*y(i,1,k) + x(i,2,j)*y(i,2,k), &
                  aimag(g(j,k)) &
                  + x(i,1,j)*y(i,2,k) - x(i,2,j)*y(i,1,k) )
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotvecs(b,e,m,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    complex(kind=8),intent(in):: a(m,m)
    real(kind=8),intent(in):: x(b:e,2,m)
    real(kind=8):: y(b:e,2,m)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(6)
    do k=1,m
!SOPTION UNROLL(6)
       do j=1,m
          do i=b,e
             y(i,1,k) = y(i,1,k) &
                  + dble(a(j,k))*x(i,1,j) - aimag(a(j,k))*x(i,2,j)
             y(i,2,k) = y(i,2,k) &
                  + dble(a(j,k))*x(i,2,j) + aimag(a(j,k))*x(i,1,j)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine rotvecs
end module arch_rotwfn
!
! for rotwfni
!
module arch_rotwfni
contains
  subroutine ltrovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: x(b:e,m),y(b:e,m)
    real(kind=8):: g(m,m)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do schedule(static,1)
!POPTION CYCLIC
    do k=1,m
!SOPTION NOLOOPINTERCHANGE, UNROLL(4)
!CDIR NOLOOPCHG
       do j=k,m
          do i=b,e
             g(j,k) = g(j,k) + x(i,j)*y(i,k)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ltrovlmat

  subroutine ovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: x(b:e,m),y(b:e,m)
    real(kind=8):: g(m,m)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(5)
    do k=1,m
!SOPTION NOLOOPINTERCHANGE, UNROLL(5)
!CDIR NOLOOPCHG
       do j=1,m
          do i=b,e
             g(j,k) = g(j,k) + x(i,j)*y(i,k)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotvecs(b,e,m,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: a(m,m), x(b:e,m)
    real(kind=8):: y(b:e,m)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(5)
    do k=1,m
!SOPTION UNROLL(5)
       do j=1,m
          do i=b,e
             y(i,k) = y(i,k) + a(j,k)*x(i,j)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine rotvecs
end module arch_rotwfni
!
! for rotwfng
!
module arch_rotwfng
  integer,parameter:: gdr_dim_nm = 256, gdr_blk_nm = 5120
  integer,parameter:: gdr_blk_k = 2048
contains
  subroutine ltrovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: x(b:e,2,m),y(b:e,2,m)
    real(kind=8):: g(m,m)

    integer:: i,j,k
    real(kind=8),allocatable:: xt(:,:),yt(:,:),gt(:,:)
    integer:: b2
    integer:: stat
    integer:: nm,nn,nk
    integer:: bk,ek

    call tpf_begin(tpf_arch)

    b2 = max(b,2)

    nm = m/gdr_dim_nm
    if (mod(m,gdr_dim_nm).ne.0) nm = nm + 1
    nm = nm*gdr_dim_nm

    nn = m/gdr_dim_nm
    if (mod(m,gdr_dim_nm).ne.0) nn = nn + 1
    nn = nn*gdr_dim_nm

    allocate(gt(nm,nn))
    gt = 0.0d0
    gt(1:m,1:m) = g(1:m,1:m)

    allocate(xt(gdr_blk_k,nm),yt(gdr_blk_k,nn))
    xt = 0.0d0
    yt = 0.0d0

    bk = b2
    do while (bk.le.e)
       if (bk+gdr_blk_k-1 .le. e) then
          ek = bk+gdr_blk_k-1
          nk = gdr_blk_k
       else
          ek = e
          nk = gdr_blk_k
       end if

!$omp parallel do private(i,j)
       do j=1,m
          do i=bk,ek
             xt(i-bk+1,j) = x(i,1,j)
          end do
          do i=ek-bk+2,nk
             xt(i,j) = 0.0d0
          end do
       end do
!$omp parallel do private(i,j)
       do j=1,m
          do i=bk,ek
             yt(i-bk+1,j) = y(i,1,j)
          end do
          do i=ek-bk+2,nk
             yt(i,j) = 0.0d0
          end do
       end do

       !call dgemm('T','N',nm,nn,nk,1.0d0,xt,gdr_blk_k,yt,gdr_blk_k,1.0d0,gt,nm,stat)
       call gdr_dgemm(0,1,nm,nn,nk,1.0d0,1.0d0,gdr_blk_k,gdr_blk_k,nm,xt,yt,gt,stat)

!$omp parallel do private(i,j)
       do j=1,m
          do i=bk,ek
             xt(i-bk+1,j) = x(i,2,j)
          end do
          do i=ek-bk+2,nk
             xt(i,j) = 0.0d0
          end do
       end do
!$omp parallel do private(i,j)
       do j=1,m
          do i=bk,ek
             yt(i-bk+1,j) = y(i,2,j)
          end do
          do i=ek-bk+2,nk
             yt(i,j) = 0.0d0
          end do
       end do

       !call dgemm('T','N',nm,nn,nk,1.0d0,xt,gdr_blk_k,yt,gdr_blk_k,1.0d0,gt,nm,stat)
       call gdr_dgemm(0,1,nm,nn,nk,1.0d0,1.0d0,gdr_blk_k,gdr_blk_k,nm,xt,yt,gt,stat)
       bk = ek + 1
    enddo
    deallocate(xt,yt)

    g(1:m,1:m) = gt(1:m,1:m)
    deallocate(gt)

    g = 2.0d0*g
    if (b.eq.1) then
       do k=1,m
          do j=k,m
             g(j,k) = g(j,k) + x(1,1,j)*y(1,1,k)
          end do
       end do
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ltrovlmat

  subroutine ovlmat(b,e,m,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: x(b:e,2,m),y(b:e,2,m)
    real(kind=8):: g(m,m)

    integer:: i,j,k
    real(kind=8),allocatable:: xt(:,:),yt(:,:),gt(:,:)
    integer:: b2
    integer:: stat
    integer:: nm,nn,nk
    integer:: bk,ek

    call tpf_begin(tpf_arch)

    b2 = max(b,2)

    nm = m/gdr_dim_nm
    if (mod(m,gdr_dim_nm).ne.0) nm = nm + 1
    nm = nm*gdr_dim_nm

    nn = m/gdr_dim_nm
    if (mod(m,gdr_dim_nm).ne.0) nn = nn + 1
    nn = nn*gdr_dim_nm

    allocate(gt(nm,nn))
    gt = 0.0d0
    gt(1:m,1:m) = g(1:m,1:m)

    allocate(xt(gdr_blk_k,nm),yt(gdr_blk_k,nn))
    xt = 0.0d0
    yt = 0.0d0

    bk = b2
    do while (bk.le.e)
       if (bk+gdr_blk_k-1 .le. e) then
          ek = bk+gdr_blk_k-1
          nk = gdr_blk_k
       else
          ek = e
          nk = gdr_blk_k
       end if

!$omp parallel do private(i,j)
       do j=1,m
          do i=bk,ek
             xt(i-bk+1,j) = x(i,1,j)
          end do
          do i=ek-bk+2,nk
             xt(i,j) = 0.0d0
          end do
       end do
!$omp parallel do private(i,j)
       do j=1,m
          do i=bk,ek
             yt(i-bk+1,j) = y(i,1,j)
          end do
          do i=ek-bk+2,nk
             yt(i,j) = 0.0d0
          end do
       end do

       !call dgemm('T','N',nm,nn,nk,1.0d0,xt,gdr_blk_k,yt,gdr_blk_k,1.0d0,gt,nm,stat)
       call gdr_dgemm(0,1,nm,nn,nk,1.0d0,1.0d0,gdr_blk_k,gdr_blk_k,nm,xt,yt,gt,stat)

!$omp parallel do private(i,j)
       do j=1,m
          do i=bk,ek
             xt(i-bk+1,j) = x(i,2,j)
          end do
          do i=ek-bk+2,nk
             xt(i,j) = 0.0d0
          end do
       end do
!$omp parallel do private(i,j)
       do j=1,m
          do i=bk,ek
             yt(i-bk+1,j) = y(i,2,j)
          end do
          do i=ek-bk+2,nk
             yt(i,j) = 0.0d0
          end do
       end do

       !call dgemm('T','N',nm,nn,nk,1.0d0,xt,gdr_blk_k,yt,gdr_blk_k,1.0d0,gt,nm,stat)
       call gdr_dgemm(0,1,nm,nn,nk,1.0d0,1.0d0,gdr_blk_k,gdr_blk_k,nm,xt,yt,gt,stat)
       bk = ek + 1
    enddo
    deallocate(xt,yt)

    g(1:m,1:m) = gt(1:m,1:m)
    deallocate(gt)

    g = 2.0d0*g
    if (b.eq.1) then
       do k=1,m
          do j=1,m
             g(j,k) = g(j,k) + x(1,1,j)*y(1,1,k)
          end do
       end do
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotvecs(b,e,m,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m
    real(kind=8),intent(in):: a(m,m)
    real(kind=8),intent(in):: x(b:e,2,m)
    real(kind=8):: y(b:e,2,m)

    real(kind=8),allocatable:: xt(:,:),yt(:,:),at(:,:)
    integer:: stat
    integer:: nm,nn,nk
    integer:: i,j
    integer:: bk,ek,bm,em,nnm

    call tpf_begin(tpf_arch)

    nm = (e-b+1)/gdr_dim_nm
    if (mod(e-b+1,gdr_dim_nm).ne.0) nm = nm + 1
    nm = nm*gdr_dim_nm

    nn = m/gdr_dim_nm
    if (mod(m,gdr_dim_nm).ne.0) nn = nn + 1
    nn = nn*gdr_dim_nm

    allocate(xt(nm,gdr_blk_k),yt(nm,nn),at(gdr_blk_k,nn))

!$omp parallel do private(i,j)
    do j=1,m
       do i=b,e
          yt(i-b+1,j) = y(i,1,j)
       end do
       do i=e-b+2,nm
          yt(i,j) = 0.0d0
       end do
    end do
!$omp parallel do private(i,j)
    do j=m+1,nn
       do i=1,nm
          yt(i,j) = 0.0d0
       end do
    end do

    bk = 1
    do while (bk.le.m)
       if (bk+gdr_blk_k-1 .le. m) then
          ek = bk+gdr_blk_k-1
          nk = gdr_blk_k
       else
          ek = m
          nk = gdr_blk_k
       end if
       
       at = 0.0d0
       at(1:ek-bk+1,1:m) = a(bk:ek,1:m)

!$omp parallel do private(i,j)
       do j=bk,ek
          do i=b,e
             xt(i-b+1,j-bk+1) = x(i,1,j)
          end do
          do i=e-b+2,nm
             xt(i,j-bk+1) = 0.0d0
          end do
       end do
!$omp parallel do private(i,j)
       do j=ek-bk+2,nk
          do i=1,nm
             xt(i,j) = 0.0d0
          end do
       end do
       bm = 1
       do while (bm.le.nm)
          if (bm+gdr_blk_nm-1 .le. nm) then
             em = bm+gdr_blk_nm-1
          else
             em = nm
          end if
          nnm = em - bm + 1

          !call dgemm('N','N',nnm,nn,nk,1.0d0,xt(bm,1),nm,at,gdr_blk_k,1.0d0,yt(bm,1),nm)
          call gdr_dgemm(1,1,nnm,nn,nk,1.0d0,1.0d0,nm,gdr_blk_k,nm,xt(bm,1),at,yt(bm,1),stat)

          bm = em + 1
       end do

       bk = ek + 1
    end do
!$omp parallel do private(i,j)
    do j=1,m
       do i=b,e
          y(i,1,j) = yt(i-b+1,j)
       end do
    end do

!$omp parallel do private(i,j)
    do j=1,m
       do i=b,e
          yt(i-b+1,j) = y(i,2,j)
       end do
       do i=e-b+2,nm
          yt(i,j) = 0.0d0
       end do
    end do
!$omp parallel do private(i,j)
    do j=m+1,nn
       do i=1,nm
          yt(i,j) = 0.0d0
       end do
    end do
    bk = 1
    do while (bk.le.m)
       if (bk+gdr_blk_k-1 .le. m) then
          ek = bk+gdr_blk_k-1
          nk = gdr_blk_k
       else
          ek = m
          nk = gdr_blk_k
       end if

       at = 0.0d0
       at(1:ek-bk+1,1:m) = a(bk:ek,1:m)

!$omp parallel do private(i,j)
       do j=bk,ek
          do i=b,e
             xt(i-b+1,j-bk+1) = x(i,2,j)
          end do
          do i=e-b+2,nm
             xt(i,j-bk+1) = 0.0d0
          end do
       end do
!$omp parallel do private(i,j)
       do j=ek-bk+2,nk
          do i=1,nm
             xt(i,j) = 0.0d0
          end do
       end do
       bm = 1
       do while (bm.le.nm)
          if (bm+gdr_blk_nm-1 .le. nm) then
             em = bm+gdr_blk_nm-1
          else
             em = nm
          end if
          nnm = em - bm + 1

          !call dgemm('N','N',nnm,nn,nk,1.0d0,xt(bm,1),nm,at,gdr_blk_k,1.0d0,yt(bm,1),nm)
          call gdr_dgemm(1,1,nnm,nn,nk,1.0d0,1.0d0,nm,gdr_blk_k,nm,xt(bm,1),at,yt(bm,1),stat)

          bm = em + 1
       end do

       bk = ek + 1
    end do
!$omp parallel do private(i,j)
    do j=1,m
       do i=b,e
          y(i,2,j) = yt(i-b+1,j)
       end do
    end do

    deallocate(xt,yt,at)
    
    call tpf_end(tpf_arch)
    return
  end subroutine rotvecs
end module arch_rotwfng
!
! for nlprj
!
module arch_nlprj
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: x(b:e,2,m1),y(b:e,2,m2)
    complex(kind=8):: g(m1,m2)

    integer:: i,j,k
    real(kind=8):: tr,ti

    call tpf_begin(tpf_arch)

    g = (0.0d0, 0.0d0)
!$omp parallel do private(tr,ti)
!SOPTION UNROLL(3)
    do k=1,m2
!SOPTION NOLOOPINTERCHANGE,UNROLL(3)
!CDIR NOLOOPCHG
       do j=1,m1
          do i=b,e
             tr = dble(g(j,k)) + x(i,1,j)*y(i,1,k) + x(i,2,j)*y(i,2,k)
             ti = aimag(g(j,k)) + x(i,1,j)*y(i,2,k) - x(i,2,j)*y(i,1,k)
             g(j,k) = dcmplx( tr, ti )
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotprjs(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: a(m1,m2)
    real(kind=8),intent(in):: x(b:e,2,m1)
    real(kind=8):: y(b:e,2,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(6)
    do k=1,m2
!SOPTION UNROLL(6)
       do j=1,m1
          do i=b,e
             y(i,1,k) = y(i,1,k) &
                  + dble(a(j,k))*x(i,1,j) - aimag(a(j,k))*x(i,2,j)
             y(i,2,k) = y(i,2,k) &
                  + dble(a(j,k))*x(i,2,j) + aimag(a(j,k))*x(i,1,j)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine rotprjs
end module arch_nlprj

module arch_nlprji
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: x(b:e,2,m1),y(b:e,m2)
    complex(kind=8):: g(m1,m2)

    integer:: i,j,k
    real(kind=8):: tr,ti

    call tpf_begin(tpf_arch)

    g = (0.0d0, 0.0d0)
!$omp parallel do private(tr,ti)
!SOPTION UNROLL(4)
    do k=1,m2
!SOPTION NOLOOPINTERCHANGE, UNROLL(2)
!CDIR NOLOOPCHG
       do j=1,m1
          do i=b,e
             tr = dble(g(j,k)) + x(i,1,j)*y(i,k)
             ti = aimag(g(j,k)) - x(i,2,j)*y(i,k)
             g(j,k) = dcmplx( tr, ti )
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotprjs(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: a(m1,m2)
    real(kind=8),intent(in):: x(b:e,2,m1)
    real(kind=8):: y(b:e,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(4)
    do k=1,m2
!SOPTION UNROLL(2)
       do j=1,m1
          do i=b,e
             y(i,k) = y(i,k) &
                  + dble(a(j,k))*x(i,1,j) - aimag(a(j,k))*x(i,2,j)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine rotprjs
end module arch_nlprji

module arch_nlprjg
  integer,parameter:: gdr_dim_nm = 256, gdr_blk_nm = 5120
  integer,parameter:: gdr_blk_k = 2048
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: x(b:e,2,m1),y(b:e,2,m2)
    real(kind=8):: g(m1,m2)

    integer:: i,j,k
    real(kind=8),allocatable:: xt(:,:),yt(:,:),gt(:,:)
    integer:: b2
    integer:: stat
    integer:: nm,nn,nk
    integer:: bk,ek

    call tpf_begin(tpf_arch)

    b2 = max(b,2)

    nm = m1/gdr_dim_nm
    if (mod(m1,gdr_dim_nm).ne.0) nm = nm + 1
    nm = nm*gdr_dim_nm

    nn = m2/gdr_dim_nm
    if (mod(m2,gdr_dim_nm).ne.0) nn = nn + 1
    nn = nn*gdr_dim_nm

    allocate(gt(nm,nn))
    gt = 0.0d0

    allocate(xt(gdr_blk_k,nm),yt(gdr_blk_k,nn))
    xt = 0.0d0
    yt = 0.0d0

    bk = b2
    do while (bk.le.e)
       if (bk+gdr_blk_k-1 .le. e) then
          ek = bk+gdr_blk_k-1
          nk = gdr_blk_k
       else
          ek = e
          nk = gdr_blk_k
       end if

!$omp parallel do private(i,j)
       do j=1,m1
          do i=bk,ek
             xt(i-bk+1,j) = x(i,1,j)
          end do
          do i=ek-bk+2,nk
             xt(i,j) = 0.0d0
          end do
       end do
!$omp parallel do private(i,j)
       do j=1,m2
          do i=bk,ek
             yt(i-bk+1,j) = y(i,1,j)
          end do
          do i=ek-bk+2,nk
             yt(i,j) = 0.0d0
          end do
       end do

       !call dgemm('T','N',nm,nn,nk,1.0d0,xt,gdr_blk_k,yt,gdr_blk_k,1.0d0,gt,nm,stat)
       call gdr_dgemm(0,1,nm,nn,nk,1.0d0,1.0d0,gdr_blk_k,gdr_blk_k,nm,xt,yt,gt,stat)

!$omp parallel do private(i,j)
       do j=1,m1
          do i=bk,ek
             xt(i-bk+1,j) = x(i,2,j)
          end do
          do i=ek-bk+2,nk
             xt(i,j) = 0.0d0
          end do
       end do
!$omp parallel do private(i,j)
       do j=1,m2
          do i=bk,ek
             yt(i-bk+1,j) = y(i,2,j)
          end do
          do i=ek-bk+2,nk
             yt(i,j) = 0.0d0
          end do
       end do

       !call dgemm('T','N',nm,nn,nk,1.0d0,xt,gdr_blk_k,yt,gdr_blk_k,1.0d0,gt,nm,stat)
       call gdr_dgemm(0,1,nm,nn,nk,1.0d0,1.0d0,gdr_blk_k,gdr_blk_k,nm,xt,yt,gt,stat)
       bk = ek + 1
    end do
    deallocate(xt,yt)

    g(1:m1,1:m2) = gt(1:m1,1:m2)
    deallocate(gt)

    g = 2.0d0*g
    if (b.eq.1) then
       do k=1,m2
          do j=1,m1
             g(j,k) = g(j,k) + x(1,1,j)*y(1,1,k)
          end do
       end do
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine rotprjs(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: a(m1,m2)
    real(kind=8),intent(in):: x(b:e,2,m1)
    real(kind=8):: y(b:e,2,m2)

    real(kind=8),allocatable:: xt(:,:),yt(:,:),at(:,:)
    integer:: stat
    integer:: nm,nn,nk
    integer:: i,j
    integer:: bk,ek,bm,em,nnm

    call tpf_begin(tpf_arch)

    nm = (e-b+1)/gdr_dim_nm
    if (mod(e-b+1,gdr_dim_nm).ne.0) nm = nm + 1
    nm = nm*gdr_dim_nm

    nn = m2/gdr_dim_nm
    if (mod(m2,gdr_dim_nm).ne.0) nn = nn + 1
    nn = nn*gdr_dim_nm

    allocate(xt(nm,gdr_blk_k),yt(nm,nn),at(gdr_blk_k,nn))

!$omp parallel do private(i,j)
    do j=1,m2
       do i=b,e
          yt(i-b+1,j) = y(i,1,j)
       end do
       do i=e-b+2,nm
          yt(i,j) = 0.0d0
       end do
    end do
!$omp parallel do private(i,j)
    do j=m2+1,nn
       do i=1,nm
          yt(i,j) = 0.0d0
       end do
    end do

    bk = 1
    do while (bk.le.m1)
       if (bk+gdr_blk_k-1 .le. m1) then
          ek = bk+gdr_blk_k-1
          nk = gdr_blk_k
       else
          ek = m1
          nk = gdr_blk_k
       end if
       
       at = 0.0d0
       at(1:ek-bk+1,1:m2) = a(bk:ek,1:m2)

!$omp parallel do private(i,j)
       do j=bk,ek
          do i=b,e
             xt(i-b+1,j-bk+1) = x(i,1,j)
          end do
          do i=e-b+2,nm
             xt(i,j-bk+1) = 0.0d0
          end do
       end do
!$omp parallel do private(i,j)
       do j=ek-bk+2,nk
          do i=1,nm
             xt(i,j) = 0.0d0
          end do
       end do
       bm = 1
       do while (bm.le.nm)
          if (bm+gdr_blk_nm-1 .le. nm) then
             em = bm+gdr_blk_nm-1
          else
             em = nm
          end if
          nnm = em - bm + 1

          !call dgemm('N','N',nnm,nn,nk,1.0d0,xt(bm,1),nm,at,gdr_blk_k,1.0d0,yt(bm,1),nm)
          call gdr_dgemm(1,1,nnm,nn,nk,1.0d0,1.0d0,nm,gdr_blk_k,nm,xt(bm,1),at,yt(bm,1),stat)

          bm = em + 1
       end do

       bk = ek + 1
    end do
!$omp parallel do private(i,j)
    do j=1,m2
       do i=b,e
          y(i,1,j) = yt(i-b+1,j)
       end do
    end do

!$omp parallel do private(i,j)
    do j=1,m2
       do i=b,e
          yt(i-b+1,j) = y(i,2,j)
       end do
       do i=e-b+2,nm
          yt(i,j) = 0.0d0
       end do
    end do
!$omp parallel do private(i,j)
    do j=m2+1,nn
       do i=1,nm
          yt(i,j) = 0.0d0
       end do
    end do
    bk = 1
    do while (bk.le.m1)
       if (bk+gdr_blk_k-1 .le. m1) then
          ek = bk+gdr_blk_k-1
          nk = gdr_blk_k
       else
          ek = m1
          nk = gdr_blk_k
       end if

       at = 0.0d0
       at(1:ek-bk+1,1:m2) = a(bk:ek,1:m2)

!$omp parallel do private(i,j)
       do j=bk,ek
          do i=b,e
             xt(i-b+1,j-bk+1) = x(i,2,j)
          end do
          do i=e-b+2,nm
             xt(i,j-bk+1) = 0.0d0
          end do
       end do
!$omp parallel do private(i,j)
       do j=ek-bk+2,nk
          do i=1,nm
             xt(i,j) = 0.0d0
          end do
       end do
       bm = 1
       do while (bm.le.nm)
          if (bm+gdr_blk_nm-1 .le. nm) then
             em = bm+gdr_blk_nm-1
          else
             em = nm
          end if
          nnm = em - bm + 1
       
          !call dgemm('N','N',nnm,nn,nk,1.0d0,xt(bm,1),nm,at,gdr_blk_k,1.0d0,yt(bm,1),nm)
          call gdr_dgemm(1,1,nnm,nn,nk,1.0d0,1.0d0,nm,gdr_blk_k,nm,xt(bm,1),at,yt(bm,1),stat)

          bm = em + 1
       end do

       bk = ek + 1
    end do
!$omp parallel do private(i,j)
    do j=1,m2
       do i=b,e
          y(i,2,j) = yt(i-b+1,j)
       end do
    end do

    deallocate(xt,yt,at)

    call tpf_end(tpf_arch)
    return
  end subroutine rotprjs
end module arch_nlprjg
!
! for sorthcl
!
module arch_sorthcl
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: x(b:e,2,m1),y(b:e,2,m2)
    complex(kind=8):: g(m1,m2)

    integer:: i,j,k
    real(kind=8):: tr,ti

    call tpf_begin(tpf_arch)

    g = (0.0d0, 0.0d0)
!$omp parallel do private(tr,ti)
!SOPTION UNROLL(3)
    do k=1,m2
!SOPTION NOLOOPINTERCHANGE,UNROLL(3)
!CDIR NOLOOPCHG
       do j=1,m1
          do i=b,e
             tr = dble(g(j,k)) + x(i,1,j)*y(i,1,k) + x(i,2,j)*y(i,2,k)
             ti = aimag(g(j,k)) + x(i,1,j)*y(i,2,k) - x(i,2,j)*y(i,1,k)
             g(j,k) = dcmplx( tr, ti )
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine subovlpt(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    complex(kind=8),intent(in):: a(m1,m2)
    real(kind=8),intent(in):: x(b:e,2,m1)
    real(kind=8):: y(b:e,2,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(2)
    do k=1,m2
!SOPTION UNROLL(4)
       do j=1,m1
          do i=b,e
             y(i,1,k) = y(i,1,k) &
                  - dble(a(j,k))*x(i,1,j) + aimag(a(j,k))*x(i,2,j)
             y(i,2,k) = y(i,2,k) &
                  - dble(a(j,k))*x(i,2,j) - aimag(a(j,k))*x(i,1,j)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine subovlpt
end module arch_sorthcl
!
! for sorthcli
!
module arch_sorthcli
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: x(b:e,m1),y(b:e,m2)
    real(kind=8):: g(m1,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

    g = 0.0d0
!$omp parallel do
!SOPTION UNROLL(4)
    do k=1,m2
!SOPTION UNROLL(4)
       do j=1,m1
!SOPTION NOLOOPINTERCHANGE
!CDIR NOLOOPCHG
          do i=b,e
             g(j,k) = g(j,k) + x(i,j)*y(i,k)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine subovlpt(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: a(m1,m2)
    real(kind=8),intent(in):: x(b:e,m1)
    real(kind=8):: y(b:e,m2)

    integer:: i,j,k

    call tpf_begin(tpf_arch)

!$omp parallel do
!SOPTION UNROLL(4)
    do k=1,m2
!SOPTION UNROLL(4)
       do j=1,m1
          do i=b,e
             y(i,k) = y(i,k) - a(j,k)*x(i,j)
          end do
       end do
    end do

    call tpf_end(tpf_arch)
    return
  end subroutine subovlpt
end module arch_sorthcli
!
! for sorthclg
!
module arch_sorthclg
  integer,parameter:: gdr_dim_nm = 256, gdr_blk_nm = 5120
  integer,parameter:: gdr_blk_k = 2048
contains
  subroutine ovlmat(b,e,m1,m2,x,y,g)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: x(b:e,2,m1),y(b:e,2,m2)
    real(kind=8):: g(m1,m2)

    integer:: i,j,k
    real(kind=8),allocatable:: xt(:,:),yt(:,:),gt(:,:)
    integer:: b2
    integer:: stat
    integer:: nm,nn,nk
    integer:: bk,ek

    call tpf_begin(tpf_arch)

    b2 = max(b,2)

    nm = m1/gdr_dim_nm
    if (mod(m1,gdr_dim_nm).ne.0) nm = nm + 1
    nm = nm*gdr_dim_nm

    nn = m2/gdr_dim_nm
    if (mod(m2,gdr_dim_nm).ne.0) nn = nn + 1
    nn = nn*gdr_dim_nm

    allocate(gt(nm,nn))
    gt = 0.0d0

    allocate(xt(gdr_blk_k,nm),yt(gdr_blk_k,nn))
    xt = 0.0d0
    yt = 0.0d0

    bk = b2
    do while (bk.le.e)
       if (bk+gdr_blk_k-1 .le. e) then
          ek = bk+gdr_blk_k-1
          nk = gdr_blk_k
       else
          ek = e
          nk = gdr_blk_k
       end if

!$omp parallel do private(i,j)
       do j=1,m1
          do i=bk,ek
             xt(i-bk+1,j) = x(i,1,j)
          end do
          do i=ek-bk+2,nk
             xt(i,j) = 0.0d0
          end do
       end do
!$omp parallel do private(i,j)
       do j=1,m2
          do i=bk,ek
             yt(i-bk+1,j) = y(i,1,j)
          end do
          do i=ek-bk+2,nk
             yt(i,j) = 0.0d0
          end do
       end do

       !call dgemm('T','N',nm,nn,nk,1.0d0,xt,gdr_blk_k,yt,gdr_blk_k,1.0d0,gt,nm,stat)
       call gdr_dgemm(0,1,nm,nn,nk,1.0d0,1.0d0,gdr_blk_k,gdr_blk_k,nm,xt,yt,gt,stat)

!$omp parallel do private(i,j)
       do j=1,m1
          do i=bk,ek
             xt(i-bk+1,j) = x(i,2,j)
          end do
          do i=ek-bk+2,nk
             xt(i,j) = 0.0d0
          end do
       end do
!$omp parallel do private(i,j)
       do j=1,m2
          do i=bk,ek
             yt(i-bk+1,j) = y(i,2,j)
          end do
          do i=ek-bk+2,nk
             yt(i,j) = 0.0d0
          end do
       end do

       !call dgemm('T','N',nm,nn,nk,1.0d0,xt,gdr_blk_k,yt,gdr_blk_k,1.0d0,gt,nm,stat)
       call gdr_dgemm(0,1,nm,nn,nk,1.0d0,1.0d0,gdr_blk_k,gdr_blk_k,nm,xt,yt,gt,stat)
       bk = ek + 1
    end do
    deallocate(xt,yt)

    g(1:m1,1:m2) = gt(1:m1,1:m2)
    deallocate(gt)

    g = 2.0d0*g
    if (b.eq.1) then
       do k=1,m2
          do j=1,m1
             g(j,k) = g(j,k) + x(1,1,j)*y(1,1,k)
          end do
       end do
    end if

    call tpf_end(tpf_arch)
    return
  end subroutine ovlmat

  subroutine subovlpt(b,e,m1,m2,a,x,y)
    use timeprof
    implicit none
    integer,intent(in):: b,e,m1,m2
    real(kind=8),intent(in):: a(m1,m2)
    real(kind=8),intent(in):: x(b:e,2,m1)
    real(kind=8):: y(b:e,2,m2)

    real(kind=8),allocatable:: xt(:,:),yt(:,:),at(:,:)
    integer:: stat
    integer:: nm,nn,nk
    integer:: i,j
    integer:: bk,ek,bm,em,nnm

    call tpf_begin(tpf_arch)

    nm = (e-b+1)/gdr_dim_nm
    if (mod(e-b+1,gdr_dim_nm).ne.0) nm = nm + 1
    nm = nm*gdr_dim_nm

    nn = m2/gdr_dim_nm
    if (mod(m2,gdr_dim_nm).ne.0) nn = nn + 1
    nn = nn*gdr_dim_nm

    allocate(xt(nm,gdr_blk_k),yt(nm,nn),at(gdr_blk_k,nn))

!$omp parallel do private(i,j)
    do j=1,m2
       do i=b,e
          yt(i-b+1,j) = y(i,1,j)
       end do
       do i=e-b+2,nm
          yt(i,j) = 0.0d0
       end do
    end do
!$omp parallel do private(i,j)
    do j=m2+1,nn
       do i=1,nm
          yt(i,j) = 0.0d0
       end do
    end do

    bk = 1
    do while (bk.le.m1)
       if (bk+gdr_blk_k-1 .le. m1) then
          ek = bk+gdr_blk_k-1
          nk = gdr_blk_k
       else
          ek = m1
          nk = gdr_blk_k
       end if
       
       at = 0.0d0
       at(1:ek-bk+1,1:m2) = a(bk:ek,1:m2)

!$omp parallel do private(i,j)
       do j=bk,ek
          do i=b,e
             xt(i-b+1,j-bk+1) = x(i,1,j)
          end do
          do i=e-b+2,nm
             xt(i,j-bk+1) = 0.0d0
          end do
       end do
!$omp parallel do private(i,j)
       do j=ek-bk+2,nk
          do i=1,nm
             xt(i,j) = 0.0d0
          end do
       end do
       bm = 1
       do while (bm.le.nm)
          if (bm+gdr_blk_nm-1 .le. nm) then
             em = bm+gdr_blk_nm-1
          else
             em = nm
          end if
          nnm = em - bm + 1

          !call dgemm('N','N',nnm,nn,nk,-1.0d0,xt(bm,1),nm,at,gdr_blk_k,1.0d0,yt(bm,1),nm)
          call gdr_dgemm(1,1,nnm,nn,nk,-1.0d0,1.0d0,nm,gdr_blk_k,nm,xt(bm,1),at,yt(bm,1),stat)

          bm = em + 1
       end do

       bk = ek + 1
    end do
!$omp parallel do private(i,j)
    do j=1,m2
       do i=b,e
          y(i,1,j) = yt(i-b+1,j)
       end do
    end do

!$omp parallel do private(i,j)
    do j=1,m2
       do i=b,e
          yt(i-b+1,j) = y(i,2,j)
       end do
       do i=e-b+2,nm
          yt(i,j) = 0.0d0
       end do
    end do
!$omp parallel do private(i,j)
    do j=m2+1,nn
       do i=1,nm
          yt(i,j) = 0.0d0
       end do
    end do
    bk = 1
    do while (bk.le.m1)
       if (bk+gdr_blk_k-1 .le. m1) then
          ek = bk+gdr_blk_k-1
          nk = gdr_blk_k
       else
          ek = m1
          nk = gdr_blk_k
       end if

       at = 0.0d0
       at(1:ek-bk+1,1:m2) = a(bk:ek,1:m2)

!$omp parallel do private(i,j)
       do j=bk,ek
          do i=b,e
             xt(i-b+1,j-bk+1) = x(i,2,j)
          end do
          do i=e-b+2,nm
             xt(i,j-bk+1) = 0.0d0
          end do
       end do
!$omp parallel do private(i,j)
       do j=ek-bk+2,nk
          do i=1,nm
             xt(i,j) = 0.0d0
          end do
       end do
       bm = 1
       do while (bm.le.nm)
          if (bm+gdr_blk_nm-1 .le. nm) then
             em = bm+gdr_blk_nm-1
          else
             em = nm
          end if
          nnm = em - bm + 1
       
          !call dgemm('N','N',nnm,nn,nk,-1.0d0,xt(bm,1),nm,at,gdr_blk_k,1.0d0,yt(bm,1),nm)
          call gdr_dgemm(1,1,nnm,nn,nk,-1.0d0,1.0d0,nm,gdr_blk_k,nm,xt(bm,1),at,yt(bm,1),stat)

          bm = em + 1
       end do

       bk = ek + 1
    end do
!$omp parallel do private(i,j)
    do j=1,m2
       do i=b,e
          y(i,2,j) = yt(i-b+1,j)
       end do
    end do

    deallocate(xt,yt,at)

    call tpf_end(tpf_arch)
    return
  end subroutine subovlpt
end module arch_sorthclg

module arch_engdffr
  integer,parameter:: blkpw = 20480, blkprj = 300
  integer,parameter,private:: gdr_dim_nm = 256, gdr_blk_nm = 5120
  integer,parameter,private:: gdr_blk_k = 2048
contains
  subroutine dtngemm(m,n,k,alpha,a,la,b,lb,beta,c,lc)
    use timeprof
    implicit none
    integer,intent(in):: m,n,k,la,lb,lc
    real(kind=8):: a(la,m), b(lb,n), c(lc,n)
    real(kind=8):: alpha,beta

    integer:: i,j
    real(kind=8),allocatable:: at(:,:),bt(:,:),ct(:,:)
    integer:: stat
    integer:: nm,nn,nk
    integer:: bk,ek,bm,em,nnm

    call tpf_begin(tpf_arch)

    nm = m/gdr_dim_nm
    if (mod(m,gdr_dim_nm).ne.0) nm = nm + 1
    nm = nm*gdr_dim_nm

    nn = n/gdr_dim_nm
    if (mod(n,gdr_dim_nm).ne.0) nn = nn + 1
    nn = nn*gdr_dim_nm

    allocate(ct(nm,nn))

    if (beta.ne.0.0d0) then
       ct(1:m,1:n) = beta*c(1:m,1:n)
    else
       ct = 0.0d0
    end if

    allocate(at(gdr_blk_k,nm),bt(gdr_blk_k,nn))
    at = 0.0d0
    bt = 0.0d0

    bk = 1
    do while (bk.le.k)
       if (bk+gdr_blk_k-1 .le. k) then
          ek = bk+gdr_blk_k-1
          nk = gdr_blk_k
       else
          ek = k
          nk = gdr_blk_k
       end if

!$omp parallel do private(i,j)
       do j=1,m
          do i=bk,ek
             at(i-bk+1,j) = a(i,j)
          end do
          do i=ek-bk+2,nk
             at(i,j) = 0.0d0
          end do
       end do
!$omp parallel do private(i,j)
       do j=1,n
          do i=bk,ek
             bt(i-bk+1,j) = b(i,j)
          end do
          do i=ek-bk+2,nk
             bt(i,j) = 0.0d0
          end do
       end do

       !call dgemm('T','N',nm,nn,nk,alpha,at,gdr_blk_k,bt,gdr_blk_k,1.0d0,ct,nm,stat)
       bm = 1
       do while (bm.le.nm)
          if (bm+gdr_blk_nm-1 .le. nm) then
             em = bm+gdr_blk_nm-1
          else
             em = nm
          end if
          nnm = em - bm + 1
          call gdr_dgemm(0,1,nnm,nn,nk,alpha,1.0d0,gdr_blk_k,gdr_blk_k,nm, &
               at(bm,1),bt,ct(bm,1),stat)
          bm = em + 1
       end do
             
       bk = ek + 1
    end do
    deallocate(at,bt)

    c(1:m,1:n) = ct(1:m,1:n)
    deallocate(ct)

    call tpf_end(tpf_arch)

    return
  end subroutine dtngemm
end module arch_engdffr
