!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015-2017 Yoshihide Yoshimoto
!
!
#include "fftgrid.h"

module bandchrg
  private ychrp1,ywfgens,ywfgen,yrhsqf,ychrp1g,ywfgensg,ywfgeng,yrhsqfg
contains
  subroutine yps3(nspin,dp,rhoband,bzwk,wgtk &
       ,va,aa,nkmesh,ncomp,ndens,nrx,nry,nrz,nrxyz &
       ,ng,nnp,rg,pg,srg,nrcell,nks,nkm,kg,ldkg,nb1,nb2,nwk,idcwk,irdkpt &
       ,ifnoptg,iflg,myrank,nproc,g_count,g_displ,comm)
    use types
    use subr_kpoint
    implicit none
    integer:: nkmesh,nrx,nry,nrz,nrxyz,ng,nnp,nrcell
    integer:: rg(3,3,ng),pg(3,ng),srg(ng)
    integer:: nks,nb1,nb2
    integer:: nspin,ncomp,ndens
    real(kind=8),intent(out):: rhoband(8*nrx*nry*nrz,nspin*ndens)
    real(kind=8):: bzwk(nks),wgtk(nb1:nb2,nks,nspin)
    real(kind=8):: va,aa(3,3)
    integer:: nkm(nks),ldkg,kg(3,ldkg,nks)
    integer,intent(in):: nwk(nrxyz,3),idcwk(nrcell,2)
    type (irdkpt_struct),pointer:: irdkpt(:)
    logical:: ifnoptg
    integer:: iflg
    integer:: myrank,nproc,comm
    integer:: g_count(0:nproc-1,nks),g_displ(0:nproc-1,nks)
    !
    type (wfn_for_k):: dp(nks,nspin)

    integer:: is

    if (nspin.eq.1) then
       call yps3_core(dp,rhoband,bzwk,wgtk,va,aa,nkmesh &
            ,ncomp,ndens,nrx,nry,nrz,nrxyz &
            ,ng,nnp,rg,pg,srg,nrcell,nks,nkm,kg,ldkg,nb1,nb2,nwk,idcwk,irdkpt &
            ,ifnoptg,iflg,myrank,nproc,g_count,g_displ,comm)
    else
       do is=1,nspin
          call yps3_core(dp(1,is),rhoband(1,is),bzwk,wgtk(nb1,1,is) &
               ,va,aa,nkmesh,ncomp,ndens,nrx,nry,nrz,nrxyz &
               ,ng,nnp,rg,pg,srg,nrcell,nks,nkm,kg,ldkg,nb1,nb2,nwk,idcwk,irdkpt &
               ,ifnoptg,iflg,myrank,nproc,g_count,g_displ,comm)
       end do
    end if
    return
  end subroutine yps3

  subroutine yps3_core(dp,rhoband,bzwk,wgtk &
       ,va,aa,nkmesh,ncomp,ndens,nrx,nry,nrz,nrxyz &
       ,ng,nnp,rg,pg,srg,nrcell,nks,nkm,kg,ldkg,nb1,nb2,nwk,idcwk,irdkpt &
       ,ifnoptg,iflg,myrank,nproc,g_count,g_displ,comm)
    !
    !                     1992. JAN. 30. BY J.YAMAUCHI
    !  irf,iif are introduced at 1995-06-06
    !
    use mpi
    use xmpi
    use fft_3d
    use symchrg
    use types
    use timeprof
    use subr_kpoint
    implicit none
    integer:: nkmesh,nrx,nry,nrz,nrxyz,ng,nnp,nrcell
    integer:: rg(3,3,ng),pg(3,ng),srg(ng)
    integer:: nks,nb1,nb2
    integer:: ncomp,ndens
    real(kind=8),intent(out):: rhoband(nrx*2,nry*2,nrz*2,ndens)
    real(kind=8):: bzwk(nks),wgtk(nb1:nb2,nks)
    real(kind=8):: va,aa(3,3)
    integer:: nkm(nks),ldkg,kg(3,ldkg,nks)
    integer,intent(in):: nwk(nrxyz,3),idcwk(nrcell,2)
    type (irdkpt_struct),pointer:: irdkpt(:)
    logical:: ifnoptg
    integer:: iflg
    integer:: myrank,nproc,comm
    integer:: mpi_typ_wfn
    integer:: g_count(0:nproc-1,nks),g_displ(0:nproc-1,nks)
    !
    type (wfn_for_k):: dp(nks)
    !
    type (fft3_struct):: fs
    real(kind=8),allocatable:: rho(:,:,:,:),prho(:,:,:,:)
    integer:: ierr,ks,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz
    real(kind=8):: fac
    integer:: bkm,ekm,irnk
    integer,allocatable:: b_count(:),b_displ(:)
!soption contarea(wf)
    complex(kind=8),pointer:: wf(:,:,:)

    call tpf_begin(tpf_bchrg)
    !
    nrx2=nrx*2
    nry2=nry*2
    nrz2=nrz*2

    nlx = FFT_PAD_NX(nrx2)
    nly = FFT_PAD_NY(nry2)
    nlz = FFT_PAD_NZ(nrz2)

    nlxyz=nlx*nly*nlz
    !
    allocate(rho(nlx,nly,nlz,ndens),prho(nlx,nly,nlz,ndens),stat=ierr)
    call chkalloc('yps3','tmp',ierr)
    !
    allocate(b_count(0:nproc-1),b_displ(0:nproc-1),stat=ierr)
    call chkalloc('yps3','b_count',ierr)
    if (iflg .eq. 1) then
       b_count = nb2/nproc
       do irnk=0,mod(nb2,nproc)-1
          b_count(irnk) = b_count(irnk)+1
       end do
       b_displ(0) = 0
    else
       b_count = (nb2-nb1+1)/nproc
       do irnk=0,mod(nb2-nb1+1,nproc)-1
          b_count(irnk) = b_count(irnk)+1
       end do
       b_displ(0) = nb1-1
    end if
    do irnk=0,nproc-2
       b_displ(irnk+1) = b_displ(irnk) + b_count(irnk)
    end do
    !
    prho = 0.0d0
    !
    call fft3_init(nrx2,nry2,nrz2,nlx,nly,nlz,fs)

    call MPI_TYPE_CONTIGUOUS(ncomp,MPI_DOUBLE_COMPLEX,mpi_typ_wfn,ierr)
    call MPI_TYPE_COMMIT(mpi_typ_wfn,ierr)

    do ks=1,nks
       bkm = g_displ(myrank,ks)+1
       ekm = g_displ(myrank,ks)+g_count(myrank,ks)
       wf => dp(ks)%wf
       if ( irdkpt(ks)%sk(1).eq.0.0d0 .and. irdkpt(ks)%sk(2).eq.0.0d0 &
            .and. irdkpt(ks)%sk(3).eq.0.0d0 .and. (.not. ifnoptg) &
            .and. ncomp.eq.1 ) then
          call ychrp1g(bkm,ekm,nkm(ks),wf,kg(:,:,ks) &
               ,fs,bzwk(ks),wgtk(nb1,ks),prho &
               ,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz &
               ,nb1,nb2,myrank,nproc &
               ,b_count,b_displ,g_count(0,ks),g_displ(0,ks),comm)
       else
          call ychrp1(ncomp,ndens,bkm,ekm,nkm(ks),wf,kg(:,:,ks) &
               ,fs,bzwk(ks),wgtk(nb1,ks),prho &
               ,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz &
               ,nb1,nb2,myrank,nproc &
               ,b_count,b_displ,g_count(0,ks),g_displ(0,ks),mpi_typ_wfn,comm)
       end if
    end do

    call MPI_TYPE_FREE(mpi_typ_wfn,ierr)

    ! call MPI_ALLREDUCE(prho,rho,nlxyz,MPI_DOUBLE_PRECISION, &
    !      MPI_SUM,comm,ierr)
    !
    ! work around for SR8000
    !
    call xmpi_alldsum(prho,rho,nlxyz*ndens,comm,ierr)
    !
    ! after ysymch(), the correct charge is obtained.
    !
    call ysymch(rho(:,:,:,1),nwk,idcwk,nlx,nly,nlz,nrxyz,ng,nrcell)
    fac=1.0d0/dble(nkmesh)/va
    !
    rhoband(:,:,:,1)=rho(1:nrx2,1:nry2,1:nrz2,1)*fac
    !
    if (ndens.eq.4) then
       call symmag(rho(:,:,:,2:4),nlx,nly,nlz,nrx2,nry2,nrz2, &
            ng,nnp,rg,pg,srg,aa)
       rhoband(:,:,:,2:4)=rho(1:nrx2,1:nry2,1:nrz2,2:4)*fac
    end if
    !
    call fft3_done(fs)
    deallocate(b_count,b_displ)
    deallocate(rho,prho)
    !
    call tpf_end(tpf_bchrg)
    return
  end subroutine yps3_core

  subroutine ychrp1(ncomp,ndens,bkm,ekm,nkm,dp,kg,fs,bzwk,wgtk,rho &
       ,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz &
       ,nb1,nb2,myrank,nproc &
       ,b_count,b_displ,g_count,g_displ,mpi_typ_wfn,comm)
    use mpi
    use fft_3d
    use timeprof
    implicit none
    type (fft3_struct):: fs
    integer:: ncomp,ndens
    integer:: nrx2,nry2,nrz2,nb1,nb2
    integer:: nlx,nly,nlz,nlxyz
    real(kind=8):: rho(nlxyz,ndens),bzwk,wgtk(nb1:nb2)
    integer:: nkm,bkm,ekm,kg(3,nkm)
    complex(kind=8):: dp(ncomp,bkm:ekm,nb2)
    integer:: mpi_typ_wfn
    integer:: myrank,nproc,comm
    integer,dimension(0:nproc-1):: b_count,b_displ,g_count,g_displ
    !
    real(kind=8),allocatable:: ztmp(:,:,:),wtmp(:,:)
    integer:: ierr,ib,irnk,ig

    complex(kind=8),allocatable:: tp(:,:)
    integer,allocatable:: s_count(:),s_displ(:)
    integer,allocatable:: r_count(:),r_displ(:)
    real(kind=8):: fac
    integer:: ic
    !
    allocate(FFT_A1M_DEFSZ(ztmp,nlxyz,ncomp), &
         FFT_A1D_DEFSZ(wtmp,nlxyz),stat=ierr)
    call chkalloc('ychrp1','tmp',ierr)
    !
    allocate(s_count(0:nproc-1),s_displ(0:nproc-1), &
         r_count(0:nproc-1),r_displ(0:nproc-1), stat=ierr)
    call chkalloc('ychrp1','s_count',ierr)

    allocate(tp(ncomp,nkm*b_count(myrank)),stat=ierr)
    call chkalloc('ychrp1','tp',ierr)

    s_count = g_count(myrank)*b_count
    s_displ = g_count(myrank)*b_displ

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALL(s_count,1,MPI_INTEGER,r_count,1,MPI_INTEGER &
         ,comm,ierr)
    call tpf_end(tpf_mpi)
    r_displ(0) = 0
    do irnk=0,nproc-2
       r_displ(irnk+1) = r_displ(irnk) + r_count(irnk)
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALLV(dp,s_count,s_displ,mpi_typ_wfn, &
         tp,r_count,r_displ,mpi_typ_wfn,comm,ierr)
    call tpf_end(tpf_mpi)
    !
!CDIR RESERVE
    do ib=0,b_count(myrank)-1
       do ic=1,ncomp
!$omp parallel do
          do ig=1,nlxyz
             FFT_A1M_RL(ztmp,ig,ic)=0.0d0
             FFT_A1M_IM(ztmp,ig,ic)=0.0d0
          end do
       end do
       if (fs%ifsft) then
          call ywfgens(ztmp,kg,tp,nkm*b_count(myrank), &
               nrx2,nry2,nrz2,nlx,nly,nlz,ncomp, &
               nkm,ib,nproc,g_count,g_displ,r_displ)
       else
          call ywfgen(ztmp,kg,tp,nkm*b_count(myrank), &
               nrx2,nry2,nrz2,nlx,nly,nlz,ncomp, &
               nkm,ib,nproc,g_count,g_displ,r_displ)
       end if
       do ic=1,ncomp
          call fft3_bw(fs,FFT_A1M_RL(ztmp,1,ic),wtmp)
       end do
       if (ib+b_displ(myrank)+1 .lt. nb1) then
          fac = bzwk
       else
          fac = wgtk(ib+b_displ(myrank)+1)
       end if
       call yrhsqf(ncomp,ndens,ztmp,rho,fac,nlxyz)
    end do
!CDIR RELEASE
    deallocate(tp)
    deallocate(s_count,s_displ,r_count,r_displ)
    deallocate(ztmp,wtmp)
    return
  end subroutine ychrp1

  subroutine ywfgen(rho,kg,tp,ntp,nrx2,nry2,nrz2,nlx,nly,nlz, &
       ncomp,nkm,ib,nproc,g_count,g_displ,r_displ)
    use timeprof
    implicit none
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz,nkm,ntp,ib
    integer,intent(in):: ncomp
    integer,intent(in):: nproc
    integer,intent(in):: g_count(0:nproc-1),g_displ(0:nproc-1)
    integer,intent(in):: r_displ(0:nproc-1)
    !
    ! rho must be zero cleared before passed to this subroutine
    !
    real(kind=8):: FFT_A3M_DEFSZ(rho,nlx,nly,nlz,ncomp)
    complex(kind=8),intent(in):: tp(ncomp,ntp)
    integer,intent(in):: kg(3,nkm)

    integer:: k1,k2,k3,ig,jg,k,irnk

    call tpf_begin(tpf_scgt)
    select case (ncomp)
    case (1)
       do irnk=0,nproc-1
          jg=r_displ(irnk)+g_count(irnk)*ib
          ig=g_displ(irnk)
!$omp parallel do private(k1,k2,k3)
!OCL NOVREC(rho)
!VOCL LOOP,DISJOINT(rho)
!CDIR NODEP(rho)
!VOPTION INDEP(rho)
!POPTION INDEP(rho)
          do k=1,g_count(irnk)
             k1=kg(1,ig+k)
             k2=kg(2,ig+k)
             k3=kg(3,ig+k)
             k1=mod(nrx2+k1,nrx2)+1
             k2=mod(nry2+k2,nry2)+1
             k3=mod(nrz2+k3,nrz2)+1
             FFT_A3M_RL(rho,k1,k2,k3,1)=dble(tp(1,jg+k))
             FFT_A3M_IM(rho,k1,k2,k3,1)=aimag(tp(1,jg+k))
          end do
       end do
    case(2)
       do irnk=0,nproc-1
          jg=r_displ(irnk)+g_count(irnk)*ib
          ig=g_displ(irnk)
!$omp parallel do private(k1,k2,k3)
!OCL NOVREC(rho)
!VOCL LOOP,DISJOINT(rho)
!CDIR NODEP(rho)
!VOPTION INDEP(rho)
!POPTION INDEP(rho)
          do k=1,g_count(irnk)
             k1=kg(1,ig+k)
             k2=kg(2,ig+k)
             k3=kg(3,ig+k)
             k1=mod(nrx2+k1,nrx2)+1
             k2=mod(nry2+k2,nry2)+1
             k3=mod(nrz2+k3,nrz2)+1
             FFT_A3M_RL(rho,k1,k2,k3,1)=dble(tp(1,jg+k))
             FFT_A3M_IM(rho,k1,k2,k3,1)=aimag(tp(1,jg+k))
             FFT_A3M_RL(rho,k1,k2,k3,2)=dble(tp(2,jg+k))
             FFT_A3M_IM(rho,k1,k2,k3,2)=aimag(tp(2,jg+k))
          end do
       end do
    case default
       write(6,*) 'ywfgen(): invalid ncomp = ', ncomp
       stop
    end select
    call tpf_end(tpf_scgt)
    return
  end subroutine ywfgen

  subroutine ywfgens(rho,kg,tp,ntp,nrx2,nry2,nrz2,nlx,nly,nlz, &
       ncomp,nkm,ib,nproc,g_count,g_displ,r_displ)
    use timeprof
    implicit none
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz,nkm,ntp,ib
    integer,intent(in):: ncomp
    integer,intent(in):: nproc
    integer,intent(in):: g_count(0:nproc-1),g_displ(0:nproc-1)
    integer,intent(in):: r_displ(0:nproc-1)
    !
    ! rho must be zero cleared before passed to this subroutine
    !
    real(kind=8):: FFT_A3M_DEFSZ(rho,nlz,nlx,nly,ncomp)
    complex(kind=8),intent(in):: tp(ncomp,ntp)
    integer,intent(in):: kg(3,nkm)

    integer:: k1,k2,k3,ig,jg,k,irnk

    call tpf_begin(tpf_scgt)
    select case (ncomp)
    case (1)
       do irnk=0,nproc-1
          jg=r_displ(irnk)+g_count(irnk)*ib
          ig=g_displ(irnk)
!$omp parallel do private(k1,k2,k3)
!OCL NOVREC(rho)
!VOCL LOOP,DISJOINT(rho)
!CDIR NODEP(rho)
!VOPTION INDEP(rho)
!POPTION INDEP(rho)
          do k=1,g_count(irnk)
             k1=kg(1,ig+k)
             k2=kg(2,ig+k)
             k3=kg(3,ig+k)
             k1=mod(nrx2+k1,nrx2)+1
             k2=mod(nry2+k2,nry2)+1
             k3=mod(nrz2+k3,nrz2)+1
             FFT_A3M_RL(rho,k3,k1,k2,1)=dble(tp(1,jg+k))
             FFT_A3M_IM(rho,k3,k1,k2,1)=aimag(tp(1,jg+k))
          end do
       end do
    case (2)
       do irnk=0,nproc-1
          jg=r_displ(irnk)+g_count(irnk)*ib
          ig=g_displ(irnk)
!$omp parallel do private(k1,k2,k3)
!OCL NOVREC(rho)
!VOCL LOOP,DISJOINT(rho)
!CDIR NODEP(rho)
!VOPTION INDEP(rho)
!POPTION INDEP(rho)
          do k=1,g_count(irnk)
             k1=kg(1,ig+k)
             k2=kg(2,ig+k)
             k3=kg(3,ig+k)
             k1=mod(nrx2+k1,nrx2)+1
             k2=mod(nry2+k2,nry2)+1
             k3=mod(nrz2+k3,nrz2)+1
             FFT_A3M_RL(rho,k3,k1,k2,1)=dble(tp(1,jg+k))
             FFT_A3M_IM(rho,k3,k1,k2,1)=aimag(tp(1,jg+k))
             FFT_A3M_RL(rho,k3,k1,k2,2)=dble(tp(2,jg+k))
             FFT_A3M_IM(rho,k3,k1,k2,2)=aimag(tp(2,jg+k))
          end do
       end do
    case default
       write(6,*) 'ywfgens(): invalid ncomp = ', ncomp
       stop
    end select
    call tpf_end(tpf_scgt)
    return
  end subroutine ywfgens

  subroutine yrhsqf(ncomp,ndens,psi,rho,rw,nlxyz)
    implicit none
    integer,intent(in):: ncomp,ndens,nlxyz
    real(kind=8):: rw
    real(kind=8):: rho(nlxyz,ndens)
    real(kind=8):: FFT_A1M_DEFSZ(psi,nlxyz,ncomp)

    integer:: i
    !
    if (ncomp.eq.1) then
!CDIR INNER
       rho(:,1)=rho(:,1) &
            + (  FFT_A1M_RL(psi,:,1)*FFT_A1M_RL(psi,:,1) &
               + FFT_A1M_IM(psi,:,1)*FFT_A1M_IM(psi,:,1) )*rw
    else
       if (ndens.eq.1) then
!$omp parallel do private(i)
          do i=1,nlxyz
             rho(i,1)=rho(i,1) &
                  + (  FFT_A1M_RL(psi,i,1)*FFT_A1M_RL(psi,i,1) &
                  + FFT_A1M_IM(psi,i,1)*FFT_A1M_IM(psi,i,1) &
                  + FFT_A1M_RL(psi,i,2)*FFT_A1M_RL(psi,i,2) &
                  + FFT_A1M_IM(psi,i,2)*FFT_A1M_IM(psi,i,2) &
                  )*rw
          end do
       else if (ndens.eq.4) then
!$omp parallel do private(i)
          do i=1,nlxyz
             rho(i,1)=rho(i,1) &
                  + (  FFT_A1M_RL(psi,i,1)*FFT_A1M_RL(psi,i,1) &
                  + FFT_A1M_IM(psi,i,1)*FFT_A1M_IM(psi,i,1) &
                  + FFT_A1M_RL(psi,i,2)*FFT_A1M_RL(psi,i,2) &
                  + FFT_A1M_IM(psi,i,2)*FFT_A1M_IM(psi,i,2) &
                  )*rw
             rho(i,2)=rho(i,2) &
                  + (  FFT_A1M_RL(psi,i,1)*FFT_A1M_RL(psi,i,2) &
                     + FFT_A1M_IM(psi,i,1)*FFT_A1M_IM(psi,i,2) &
                    )*2.0d0*rw
             rho(i,3)=rho(i,3) &
                  + (  FFT_A1M_RL(psi,i,1)*FFT_A1M_IM(psi,i,2) &
                     - FFT_A1M_IM(psi,i,1)*FFT_A1M_RL(psi,i,2) &
                    )*2.0d0*rw
             rho(i,4)=rho(i,4) &
                  + (  FFT_A1M_RL(psi,i,1)*FFT_A1M_RL(psi,i,1) &
                     + FFT_A1M_IM(psi,i,1)*FFT_A1M_IM(psi,i,1) &
                     - FFT_A1M_RL(psi,i,2)*FFT_A1M_RL(psi,i,2) &
                     - FFT_A1M_IM(psi,i,2)*FFT_A1M_IM(psi,i,2) &
                    )*rw
          end do
       end if
    end if

    return
  end subroutine yrhsqf

  subroutine ychrp1g(bkm,ekm,nkm,dp,kg,fs,bzwk,wgtk,rho &
       ,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz &
       ,nb1,nb2,myrank,nproc &
       ,b_count,b_displ,g_count,g_displ,comm)
    use mpi
    use fft_3d
    use timeprof
    implicit none
    type (fft3_struct):: fs
    integer:: nrx2,nry2,nrz2,nb1,nb2
    integer:: nlx,nly,nlz,nlxyz
    real(kind=8):: rho(nlxyz),bzwk,wgtk(nb1:nb2)
    integer:: nkm,bkm,ekm,kg(3,nkm)
    complex(kind=8):: dp(bkm:ekm,nb2)
    integer:: myrank,nproc,comm
    integer,dimension(0:nproc-1):: b_count,b_displ,g_count,g_displ
    !
    real(kind=8),allocatable:: ztmp(:,:),wtmp(:,:)
    integer:: ierr,ib,ix,irnk,ibr,ig

    complex(kind=8),allocatable:: tp(:)
    integer,allocatable:: s_count(:),s_displ(:)
    integer,allocatable:: r_count(:),r_displ(:)
    real(kind=8):: fac(2)
    !
    allocate(FFT_A1D_DEFSZ(ztmp,nlxyz),FFT_A1D_DEFSZ(wtmp,nlxyz),stat=ierr)
    call chkalloc('ychrp1g','tmp',ierr)
    !
    allocate(s_count(0:nproc-1),s_displ(0:nproc-1), &
         r_count(0:nproc-1),r_displ(0:nproc-1), stat=ierr)
    call chkalloc('ychrp1g','s_count',ierr)

    allocate(tp(nkm*2*b_count(myrank)),stat=ierr)
    call chkalloc('ychrp1g','tp',ierr)

    s_count = g_count(myrank)*b_count
    s_displ = g_count(myrank)*b_displ

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALL(s_count,1,MPI_INTEGER,r_count,1,MPI_INTEGER &
         ,comm,ierr)
    call tpf_end(tpf_mpi)
    r_displ(0) = 0
    do irnk=0,nproc-2
       r_displ(irnk+1) = r_displ(irnk) + r_count(irnk)
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALLV(dp,s_count,s_displ,MPI_DOUBLE_COMPLEX, &
         tp,r_count,r_displ,MPI_DOUBLE_COMPLEX,comm,ierr)
    call tpf_end(tpf_mpi)
    !
!CDIR RESERVE
    ibr = mod(b_count(myrank),2)
    if (ibr.eq.1) then
!$omp parallel do
       do ig=1,nlxyz
          FFT_A1D_RL(ztmp,ig)=0.0d0
          FFT_A1D_IM(ztmp,ig)=0.0d0
       end do
       if (fs%ifsft) then
          call ywfgens(ztmp,kg,tp,nkm*b_count(myrank), &
               nrx2,nry2,nrz2,nlx,nly,nlz,1, &
               nkm,0,nproc,g_count,g_displ,r_displ)
       else
          call ywfgen(ztmp,kg,tp,nkm*b_count(myrank), &
               nrx2,nry2,nrz2,nlx,nly,nlz,1, &
               nkm,0,nproc,g_count,g_displ,r_displ)
       end if
       ix = 1
       call fft3_bwsx(fs,ztmp,wtmp,ix)
       if (b_displ(myrank)+1 .lt. nb1) then
          fac(1) = bzwk
       else
          fac(1) = wgtk(b_displ(myrank)+1)
       end if
       if (ix .eq. 1) then
          call yrhsqf(1,1,ztmp,rho,fac(1),nlxyz)
       else if (ix .eq. -1) then
          call yrhsqf(1,1,wtmp,rho,fac(1),nlxyz)
       else
          write(6,*) 'ychrp1g: abs(ix) != 1'
          stop
       end if
    end if
    do ib=ibr,b_count(myrank)-1,2
!$omp parallel do
       do ig=1,nlxyz
          FFT_A1D_RL(ztmp,ig)=0.0d0
          FFT_A1D_IM(ztmp,ig)=0.0d0
       end do
       if (fs%ifsft) then
          call ywfgensg(ztmp,kg,tp,nkm*b_count(myrank), &
               nrx2,nry2,nrz2,nlx,nly,nlz, &
               nkm,ib,nproc,g_count,g_displ,r_displ)
       else
          call ywfgeng(ztmp,kg,tp,nkm*b_count(myrank), &
               nrx2,nry2,nrz2,nlx,nly,nlz, &
               nkm,ib,nproc,g_count,g_displ,r_displ)
       end if
       ix = 1
       call fft3_bwsx(fs,ztmp,wtmp,ix)
       if (ib+b_displ(myrank)+1 .lt. nb1) then
          fac(1) = bzwk
       else
          fac(1) = wgtk(ib+b_displ(myrank)+1)
       end if
       if (ib+b_displ(myrank)+2 .lt. nb1) then
          fac(2) = bzwk
       else
          fac(2) = wgtk(ib+b_displ(myrank)+2)
       end if
       if (ix .eq. 1) then
          call yrhsqfg(ztmp,rho,fac,nlxyz)
       else if (ix .eq. -1) then
          call yrhsqfg(wtmp,rho,fac,nlxyz)
       else
          write(6,*) 'ychrp1g: abs(ix) != 1'
          stop
       end if
    end do
!CDIR RELEASE
    deallocate(tp)
    deallocate(s_count,s_displ,r_count,r_displ)
    deallocate(ztmp,wtmp)
    return
  end subroutine ychrp1g

  subroutine ywfgeng(rho,kg,tp,ntp,nrx2,nry2,nrz2,nlx,nly,nlz, &
       nkm,ib,nproc,g_count,g_displ,r_displ)
    use timeprof
    implicit none
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz,nkm,ntp,ib
    integer,intent(in):: nproc
    integer,intent(in):: g_count(0:nproc-1),g_displ(0:nproc-1)
    integer,intent(in):: r_displ(0:nproc-1)
    !
    ! rho must be zero cleared before passed to this subroutine
    !
    real(kind=8):: FFT_A3D_DEFSZ(rho,nlx,nly,nlz)
    complex(kind=8),intent(in):: tp(ntp)
    integer,intent(in):: kg(3,nkm)

    integer:: k1,k2,k3,ig,jg,k,irnk

    call tpf_begin(tpf_scgt)
    do irnk=0,nproc-1
       jg=r_displ(irnk)+g_count(irnk)*ib
       ig=g_displ(irnk)
!$omp parallel do private(k1,k2,k3)
!VOCL LOOP,DISJOINT(rho)
!CDIR NODEP(rho)
!VOPTION INDEP(rho)
!POPTION INDEP(rho)
       do k=1,g_count(irnk)
          k1=kg(1,ig+k)
          k2=kg(2,ig+k)
          k3=kg(3,ig+k)
          k1=mod(nrx2+k1,nrx2)+1
          k2=mod(nry2+k2,nry2)+1
          k3=mod(nrz2+k3,nrz2)+1
          FFT_A3D_RL(rho,k1,k2,k3)=dble(tp(jg+k)) - aimag(tp(jg+k+g_count(irnk)))
          FFT_A3D_IM(rho,k1,k2,k3)=aimag(tp(jg+k)) + dble(tp(jg+k+g_count(irnk)))
       end do
    end do
    call tpf_end(tpf_scgt)
    return
  end subroutine ywfgeng

  subroutine ywfgensg(rho,kg,tp,ntp,nrx2,nry2,nrz2,nlx,nly,nlz, &
       nkm,ib,nproc,g_count,g_displ,r_displ)
    use timeprof
    implicit none
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz,nkm,ntp,ib
    integer,intent(in):: nproc
    integer,intent(in):: g_count(0:nproc-1),g_displ(0:nproc-1)
    integer,intent(in):: r_displ(0:nproc-1)
    !
    ! rho must be zero cleared before passed to this subroutine
    !
    real(kind=8):: FFT_A3D_DEFSZ(rho,nlz,nlx,nly)
    complex(kind=8),intent(in):: tp(ntp)
    integer,intent(in):: kg(3,nkm)

    integer:: k1,k2,k3,ig,jg,k,irnk

    call tpf_begin(tpf_scgt)
    do irnk=0,nproc-1
       jg=r_displ(irnk)+g_count(irnk)*ib
       ig=g_displ(irnk)
!$omp parallel do private(k1,k2,k3)
!VOCL LOOP,DISJOINT(rho)
!CDIR NODEP(rho)
!VOPTION INDEP(rho)
!POPTION INDEP(rho)
       do k=1,g_count(irnk)
          k1=kg(1,ig+k)
          k2=kg(2,ig+k)
          k3=kg(3,ig+k)
          k1=mod(nrx2+k1,nrx2)+1
          k2=mod(nry2+k2,nry2)+1
          k3=mod(nrz2+k3,nrz2)+1
          FFT_A3D_RL(rho,k3,k1,k2)=dble(tp(jg+k)) - aimag(tp(jg+k+g_count(irnk)))
          FFT_A3D_IM(rho,k3,k1,k2)=aimag(tp(jg+k)) + dble(tp(jg+k+g_count(irnk)))
       end do
    end do
    call tpf_end(tpf_scgt)
    return
  end subroutine ywfgensg

  subroutine yrhsqfg(rho,rhot,rw,nlxyz)
    implicit none
    integer:: nlxyz
    real(kind=8):: rw(2)
    real(kind=8):: rhot(nlxyz)
    real(kind=8):: FFT_A1D_DEFSZ(rho,nlxyz)
    !
!CDIR INNER
    rhot(:)=rhot(:) &
         + FFT_A1D_RL(rho,:)*FFT_A1D_RL(rho,:)*rw(1) + FFT_A1D_IM(rho,:)*FFT_A1D_IM(rho,:)*rw(2)
    return
  end subroutine yrhsqfg
end module bandchrg
