!
! Copyright (c) 2014-2017 Yoshihide Yoshimoto
!
module subr_wannier
  type gb_for_atm
     integer:: ngau
     integer,pointer:: lgau(:),mgau(:)
     real(kind=8),pointer:: alpg(:)
  end type gb_for_atm
  type wannier_params_struct
     integer:: tspin
     integer:: nblw, nbup
     real(kind=8):: engl, engu
     real(kind=8):: enginl, enginu
     integer:: inibase_mode
     type (gb_for_atm),pointer:: gbkd(:)
     integer:: ngau, nigs
     integer,pointer:: lgau(:),mgau(:),cgau(:)
     real(kind=8),pointer:: taug(:,:),alpg(:)
     integer:: iwan, nocc ! nocc must be equal to ngau
     complex(kind=8),pointer:: B_MAT(:,:)
     integer:: Na1, Na2, Na3
     real(kind=8):: spll_thr, sprd_thr, sprd_mix, sprd_stpl_fac
     character(len=16):: pgm_mode
  end type wannier_params_struct
contains
  subroutine setupwanparam(ncomp,nband,nw,mgs,ee,wp,nkd,nsi,kd,asi,mssg)
    use meshgrid
    use wannier_metal
    implicit none
    integer,intent(in):: ncomp,nband,nw
    type (meshgrid_struct),intent(in):: mgs
    real(kind=8),intent(in):: ee(nband,mgs%nks)

    type (wannier_params_struct):: wp

    integer,intent(in):: nkd,nsi,kd(nsi),mssg
    real(kind=8),intent(in):: asi(3,nsi)

    integer:: i,iga,it,ic
    real(kind=8):: tmp

    if (wp%inibase_mode .eq. 2) then
       if (mssg.gt.-1) then
          write(6,*) ' mode: wannier basis by atomic kind'
       end if
       wp%ngau = 0
       do it =1,nsi
          wp%ngau = wp%ngau + ncomp*wp%gbkd(kd(it))%ngau
       end do
       if (mssg.gt.-1) write(6,*) ' wannier_number_basis => ', wp%ngau
       wp%nocc = wp%ngau
       wp%nigs = wp%ngau
       allocate(wp%lgau(wp%ngau),wp%mgau(wp%ngau),wp%alpg(wp%ngau), &
            wp%taug(3,wp%ngau),wp%cgau(wp%ngau),wp%B_MAT(wp%nigs,wp%ngau))
       i = 0
       if (mssg.gt.-1) then
          write(6,*) ' initial gaussians'
          write(6,*) ' lgau, mgau, alpg, taug, cgau'
       end if
       do it = 1,nsi
          do iga = 1,wp%gbkd(kd(it))%ngau
             do ic=1,ncomp
                i = i + 1
                wp%lgau(i) = wp%gbkd(kd(it))%lgau(iga)
                wp%mgau(i) = wp%gbkd(kd(it))%mgau(iga)
                wp%alpg(i) = wp%gbkd(kd(it))%alpg(iga)
                wp%taug(:,i) = asi(:,it)
                wp%cgau(i) = ic
                write(6,'(2(1x,i5),4(1x,f14.11),1x,i5)') &
                     wp%lgau(i),wp%mgau(i), &
                     wp%alpg(i),wp%taug(:,i),wp%cgau(i)
             end do
          end do
       end do
       wp%B_MAT = 0.0d0
       do i = 1, wp%ngau
          wp%B_MAT(i,i) = 1.0d0
       end do
       if (mssg.gt.-1) write(6,*) ' B_MAT is set to identity'
       wp%nblw = 1
       wp%nbup = wp%ngau
       wp%engl = minval(ee(wp%nblw,:)) - 0.1d0
       if (wp%nbup+1.le.nw) then
          tmp = minval(ee(wp%nbup+1,:))
       else
          tmp = 0.1d0
       end if
       wp%engu = (maxval(ee(wp%nbup,:)) + tmp)*0.5d0
       if (mssg.gt.-1) then
          write(6,*) ' wannier_band_lower => ', wp%nblw
          write(6,*) ' wannier_band_upper => ', wp%nbup
          write(6,*) ' wannier_eng_lower => ', wp%engl
          write(6,*) ' wannier_eng_upper => ', wp%engu
       end if
    end if

    return
  end subroutine setupwanparam

  subroutine wan_setup_bvec_gen(bb,nks,sirdkpt,ng,rg,srg, &
       NBmax, NB,b_LAT,VEC_b,WEIGHT_b,WEIGHT)
    use subr_kpoint
    use mpi
    implicit none
    real(kind=8),intent(in):: bb(3,3)
    integer,intent(in):: nks
    type(irdkpt_struct):: sirdkpt(nks)
    integer,intent(in):: ng,rg(3,3,ng),srg(ng)
    integer,intent(in):: NBmax
    integer:: NB
    real(kind=8):: b_LAT(3,NBmax), VEC_b(3,NBmax)
    real(kind=8):: WEIGHT_b(NBmax), WEIGHT

    real(kind=8):: csk(3),sq_sk,dif,tmp
    real(kind=8),parameter:: eps = 1.0d-8

    integer,parameter:: maxshel = 10
    integer:: lstshel(NBmax)
    real(kind=8):: amat(6,maxshel),umat(6,6),vtmat(maxshel,maxshel)
    real(kind=8):: dinv(6,6),sgv(6),wvec(maxshel),samat(6,maxshel)
    integer,parameter:: lwork = 5*maxshel*maxshel
    real(kind=8):: work(lwork)
    integer:: info
    integer:: iks,jks,jrg,ishel,nshel,iir,i,j,ibvec
    logical:: ifsuccess
    real(kind=8),parameter:: qvec(6) = &
         (/ 1.0d0, 1.0d0, 1.0d0, 0.0d0, 0.0d0, 0.0d0 /)

    
    csk = matmul(bb,sirdkpt(1)%sk)
    sq_sk = csk(1)*csk(1) + csk(2)*csk(2) + csk(3)*csk(3)
    if (abs(sq_sk).gt.eps) then
       write(6,*) 'Gamma point is not included in the irreducible k-point list'
       stop
    end if
    nshel = 0
    lstshel = 0

    ifsuccess = .false.
    do iks=2,nks
       amat = 0.0d0
       do ishel=1,nshel
          jks = lstshel(ishel)
          do iir=1,sirdkpt(jks)%ngir
             jrg = sirdkpt(jks)%irrg(iir)
             csk = matmul(bb,srg(jrg)*matmul(rg(:,:,jrg),sirdkpt(jks)%sk))
             amat(1,ishel) = amat(1,ishel) + csk(1)*csk(1)
             amat(2,ishel) = amat(2,ishel) + csk(2)*csk(2)
             amat(3,ishel) = amat(3,ishel) + csk(3)*csk(3)
             amat(4,ishel) = amat(4,ishel) + csk(2)*csk(3)
             amat(5,ishel) = amat(5,ishel) + csk(3)*csk(1)
             amat(6,ishel) = amat(6,ishel) + csk(1)*csk(2)
          end do
       end do
       ishel = nshel + 1
       do iir=1,sirdkpt(iks)%ngir
          jrg = sirdkpt(iks)%irrg(iir)
          csk = matmul(bb,srg(jrg)*matmul(rg(:,:,jrg),sirdkpt(iks)%sk))
          amat(1,ishel) = amat(1,ishel) + csk(1)*csk(1)
          amat(2,ishel) = amat(2,ishel) + csk(2)*csk(2)
          amat(3,ishel) = amat(3,ishel) + csk(3)*csk(3)
          amat(4,ishel) = amat(4,ishel) + csk(2)*csk(3)
          amat(5,ishel) = amat(5,ishel) + csk(3)*csk(1)
          amat(6,ishel) = amat(6,ishel) + csk(1)*csk(2)
       end do
       samat = amat
       call dgesvd('A','A',6,nshel+1,amat,6,sgv,umat,6, &
            vtmat,maxshel,work,lwork,info)
       if (info.ne.0) then
          write(6,'(1x,a)') 'dgesvd failed'
          stop
       end if
       if (minval(sgv(1:min(6,nshel))).lt.eps) then
          write(6,'(1x,a,i3)') 'skip iks = ', iks
          cycle
       end if
       nshel = nshel + 1
       lstshel(nshel) = iks

       dinv = 0.0d0
       do ishel=1,6
          dinv(ishel,ishel) = 1.0d0/sgv(ishel)
       end do
       wvec = matmul(transpose(vtmat(1:nshel,1:nshel)), &
            matmul(dinv,matmul(transpose(umat),qvec)))
       dif = maxval(abs(matmul(samat(:,1:nshel),wvec(1:nshel))-qvec))
       if (dif.lt.eps) then
          ifsuccess = .true.
          exit
       end if
    end do
    if (.not. ifsuccess) then
       write(6,*) 'too few irreducible sample k-points'
       stop
    end if
    NB = 0
    WEIGHT = 0.0d0
    do ishel=1,nshel
       iks = lstshel(ishel)
       do iir=1,sirdkpt(iks)%ngir
          NB = NB + 1
          jrg = sirdkpt(iks)%irrg(iir)
          b_LAT(:,NB) = matmul(srg(jrg)*rg(:,:,jrg),sirdkpt(iks)%sk)
          VEC_b(:,NB) = matmul(bb,b_LAT(:,NB))
          WEIGHT_b(NB) = wvec(ishel)
          WEIGHT = WEIGHT + WEIGHT_b(NB)
       end do
    end do

    ! check sum rule of b vectors
    !
    do i=1,3
       do j=1,3
          tmp = 0.0d0
          do ibvec=1, NB
             tmp = tmp + WEIGHT_b(ibvec)*VEC_b(i,ibvec)*VEC_b(j,ibvec)
          end do
          info = 0
          if (i.eq.j) then
             if (abs(tmp-1.0d0).gt.eps) then
                info = 1
             end if
          else if (abs(tmp).gt.eps) then
             info = 1
          end if
          if (info.eq.1) then
             write(6,*) 'sum rule of b vectors is broken'
             call MPI_ABORT(MPI_COMM_WORLD,1,info)
             stop
          end if
       end do
    end do

    write(6,*)
    write(6,'(1x,a)') 'wan_setup_bvec_gen():'
    write(6,'(2x,a,i4)') 'number of shells = ', nshel
    write(6,'(2x,a)') 'irreducible sk and weight = '
    do ishel = 1, nshel
       iks = lstshel(ishel)
       write(6,'(1x,4(1x,f12.8))') sirdkpt(iks)%sk,wvec(ishel)
    end do
    write(6,'(2x,a,i4)') 'number total b vector = ', NB
    write(6,'(2x,a)') 'b vectors = '
    do ishel = 1, nshel
       iks = lstshel(ishel)
       do iir = 1, sirdkpt(iks)%ngir
          jrg = sirdkpt(iks)%irrg(iir)
          write(6,'(2x,i5,3(1x,f12.8))') &
               ishel,srg(jrg)*matmul(rg(:,:,jrg),sirdkpt(iks)%sk)
       end do
    end do
    write(6,'(2x,a,f14.10)') 'WEIGHT = ', WEIGHT
    
    return
  end subroutine wan_setup_bvec_gen
  
  subroutine calcwannier(dp,nband,nw,ncomp &
       ,va,aa,bb,ee,mgs,ifinvsym &
       ,ng,rg,nnp,pg,srg,nkb,icell,wp & 
       ,bspotfc,nkd,bspots,nsi,kd,asi &
       ,mssg,myrank,nproc,comm)
    use types
    use mpi
    use meshgrid
    use subr_kpoint
    use wannier_metal
    use iounit
    use bsplpot
    use converter
    use nonlocal_proj
    use uspp_qg
    implicit none
    integer:: nband
    integer:: nw,ncomp
    integer:: ng,nnp
    integer:: mssg
    !
    type (meshgrid_struct):: mgs
    !
    type (wfn_for_k):: dp(mgs%nks)
    !
    real(kind=8):: va,aa(3,3),bb(3,3)
    real(kind=8):: ee(nband,mgs%nks)
    logical:: ifinvsym
    integer:: rg(3,3,ng),pg(3,ng),srg(ng)
    !
    integer:: nkb(3),icell
    type (wannier_params_struct):: wp
    !
    integer:: nkd,nsi,kd(nsi)
    real(kind=8):: asi(3,nsi)
    type (bspotfc_struct):: bspotfc
    type(bspot_for_atm):: bspots(nkd)

    integer:: myrank,nproc,comm

    integer,allocatable:: mapkpt(:,:,:,:)
    real(kind=8),allocatable:: E_EIG(:,:), sklist(:,:)

    integer:: ik,ib,jb,NB

    !real(kind=8),parameter :: au=27.21151d0
    !
    ! CODATA 2014
    real(kind=8),parameter :: au=27.21138602d0
    real(kind=8),allocatable:: WF_POSITION(:,:)
    complex(kind=8),allocatable:: H_MAT_R(:,:,:,:,:),A_MAT_O(:,:,:)

    integer:: N_BAND(mgs%nkmesh),N_BAND_BOTTOM(mgs%nkmesh)
    integer:: in_win_band(2,mgs%nkmesh)

    integer,parameter:: NBmax = 12
    real(kind=8) :: b_LAT(3,NBmax)
    real(kind=8) :: VEC_b(3,NBmax)
    real(kind=8) :: WEIGHT_b(NBmax)
    real(kind=8) :: WEIGHT

    type (wfn_for_k):: dsp(mgs%nks)
    complex(kind=8),allocatable:: OVERLAP(:,:,:,:)
    integer,allocatable:: KPT(:,:)

    complex(kind=8),allocatable:: A_MAT(:,:,:)

    integer,allocatable:: pj_count(:),pj_displ(:)
    integer:: bprj,eprj
    real(kind=8),allocatable:: akg(:),dvpj(:,:,:,:)
    complex(kind=8),allocatable:: dcprj(:,:)
    type(nlprj_struct):: nlpjs

    type(qgbl_for_atm):: qgbl(nkd,NBmax)
    real(kind=8):: vecb4(4)
    integer:: nob

    integer:: ia1,ia2,ia3,ja1,ja2,ja3
    integer:: bkm,ekm,irnk
    integer:: lpnum,maxr,ikd,bkms,ekms
    integer:: ibvec

    type(irdkpt_struct),pointer:: sirdkpt(:)

    integer:: ierr

    allocate(mapkpt(5,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk,-mgs%nk:mgs%nk))

    call calcmapkpt(ng,rg,srg,mgs%nk,mgs%nks,mgs%irdkpt,mgs%nkmesh,mapkpt)

    allocate(sklist(3,mgs%nkmesh))

    call calcsklist(ng,rg,srg,mgs%nk,mgs%nks, mgs%irdkpt,mgs%nkmesh,sklist,1)

    allocate(E_EIG(wp%nbup-wp%nblw+1,mgs%nks))
    do ik=1,mgs%nks
       do ib=wp%nblw, wp%nbup
          E_EIG(ib-wp%nblw+1,ik) = ee(ib,ik)
       end do
    end do

    lpnum = getlpnum(nkd,bspots)
    maxr = getmaxr(nkd,bspots)
    do ik=1,mgs%nks
       bkm=mgs%g_displ(myrank,ik)+1
       ekm=mgs%g_displ(myrank,ik)+mgs%g_count(myrank,ik)
       allocate(dsp(ik)%wf(ncomp,bkm:ekm,nw))
       dsp(ik)%wf = dp(ik)%wf
       allocate(akg(bkm:ekm),dvpj(bkm:ekm,maxr,lpnum,nkd))
       call kg2akg(bb,mgs%irdkpt(ik)%sk,mgs%g_count(myrank,ik), &
            mgs%kg(:,bkm:ekm,ik),akg)
       do ikd=1,nkd
          call evbetg(bspotfc,bspots(ikd),mgs%g_count(myrank,ik),akg, &
                   mgs%g_count(myrank,ik),maxr,lpnum,dvpj(:,:,:,ikd))
       end do
       call nlprj_init(lpnum,maxr,nkd,bspots,nsi,kd,asi,bkm,ekm, &
            bb,mgs%irdkpt(ik)%sk,mgs%kg(:,bkm:ekm,ik),dvpj,nlpjs)
       deallocate(akg,dvpj)

       allocate(pj_count(0:nproc-1),pj_displ(0:nproc-1))
       pj_count = nlpjs%nprj/nproc
       do irnk=0,mod(nlpjs%nprj,nproc)-1
          pj_count(irnk) = pj_count(irnk)+1
       end do
       pj_displ(0) = 0
       do irnk=0,nproc-2
          pj_displ(irnk+1) = pj_displ(irnk) + pj_count(irnk)
       end do
       bprj = pj_displ(myrank)+1
       eprj = pj_displ(myrank)+pj_count(myrank)

       allocate(dcprj(bprj:eprj,nw))

       bkms = (bkm-1)*ncomp+1
       ekms = ekm*ncomp
       call nlprj_apyas(nkd,bspots,kd,nsi,bkms,ekms, &
            va,nw,dp(ik)%wf,dsp(ik)%wf,bprj,eprj,dcprj,nlpjs,comm)
       call nlprj_done(nlpjs)
       deallocate(pj_count,pj_displ,dcprj)
    end do

    call wan_target_band(mgs,mgs%nkmesh,wp%nbup-wp%nblw+1, &
         E_EIG,wp%engl,wp%engu, &
         mapkpt,sklist,N_BAND,N_BAND_BOTTOM)
    if (minval(N_BAND).lt.wp%ngau) then
       if (mssg.gt.-1) then
          write(6,*) 'For every k-point,' // &
               ' [wannier_band_lower:wannier_band_upper] must contains' // &
               ' wannier_number_basis bands at least.'
       end if
       call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
       stop
    end if

    call wan_set_inner_win(mgs,mgs%nkmesh,wp%nbup-wp%nblw+1,E_EIG, &
       wp%enginl,wp%enginu,mapkpt,sklist,in_win_band)

    if (wp%pgm_mode .eq. 'generate_rcp') then
       allocate(WF_POSITION(3,wp%nocc), &
            H_MAT_R(wp%nocc,wp%nocc,0:nkb(1)-1,0:nkb(2)-1,0:nkb(3)-1), &
            A_MAT_O(wp%nbup-wp%nblw+1,wp%nocc,mgs%nkmesh), &
            A_MAT(wp%nbup-wp%nblw+1,wp%ngau,mgs%nkmesh))

       if (icell.eq.9) then
          call sort_irdkpt(bb,mgs%nks,mgs%irdkpt,sirdkpt)
          call wan_setup_bvec_gen(bb,mgs%nks,sirdkpt,mgs%ng,mgs%rg,mgs%srg, &
               NBmax,NB,b_LAT,VEC_b,WEIGHT_b,WEIGHT)
          call dealloc_irdkpt(mgs%nks,sirdkpt)
       else
          call wan_setup_bvec(aa,bb,icell,nkb(1),nkb(2),nkb(3), &
               NBmax,NB,b_LAT,VEC_b,WEIGHT_b,WEIGHT)
       end if

       do ibvec=1,NB
          vecb4(1:3) = VEC_b(1:3,ibvec)
          vecb4(4) = sqrt( VEC_b(1,ibvec)**2 + VEC_b(2,ibvec)**2 &
               + VEC_b(3,ibvec)**2 )
          do ikd=1,nkd
             nob = bspots(ikd)%nob
             allocate(qgbl(ikd,ibvec)%qg(nob*(nob+1)/2))
             !
             ! va = 1.0d0 will be canceled.
             !
             call evqg(bspotfc,bspots(ikd),1,vecb4,1.0d0,nob,1, &
                  qgbl(ikd,ibvec)%qg)
          end do
       end do

       allocate(OVERLAP(wp%nbup-wp%nblw+1,wp%nbup-wp%nblw+1,mgs%nkmesh,NB), &
            KPT(mgs%nkmesh,NB))

       call wan_calc_ism(dp,ncomp,aa,bb,mgs &
            ,ng,rg,nnp,pg,srg,mgs%nkmesh,maxval(mgs%nkm) &
            ,wp%nblw,wp%nbup,wp%nbup-wp%nblw+1 &
            ,N_BAND,N_BAND_BOTTOM &
            ,NB,b_LAT,mapkpt,sklist,OVERLAP,KPT &
            ,bspotfc,nkd,bspots,nsi,kd,asi,qgbl)

       do ikd=1,nkd
          do ibvec=1,NB
             deallocate(qgbl(ikd,ibvec)%qg)
          end do
       end do

       call wan_init_amat(dsp,ncomp,nw,va,aa,bb,mgs &
            ,ng,rg,nnp,pg,srg,mgs%nkmesh,maxval(mgs%nkm) &
            ,wp%nblw,wp%nbup,wp%nbup-wp%nblw+1 &
            ,N_BAND,N_BAND_BOTTOM &
            ,wp%ngau,wp%nigs,wp%lgau,wp%mgau,wp%cgau,wp%taug,wp%alpg,wp%B_MAT &
            ,mapkpt,sklist,A_MAT)

       do ik=1,mgs%nks
          deallocate(dsp(ik)%wf)
       end do

       call wannier_kazuma(OVERLAP,KPT,aa,bb,E_EIG,mgs &
            ,mgs%nkmesh,wp%spll_thr,wp%sprd_thr &
            ,wp%nblw,wp%nbup,wp%nbup-wp%nblw+1 &
            ,N_BAND,N_BAND_BOTTOM &
            ,in_win_band &
            ,wp%ngau,A_MAT,wp%sprd_mix,wp%sprd_stpl_fac &
            ,nkb,wp%nocc &
            ,NB,VEC_b,WEIGHT_b,WEIGHT,mapkpt,sklist &
            ,WF_POSITION,H_MAT_R,A_MAT_O)

       deallocate(OVERLAP,KPT,A_MAT)

       call xrewind(NFPOSWAN)
       do ib = 1,wp%nocc
          write(NFPOSWAN,'(3(1x,f19.10))') WF_POSITION(:,ib)
       end do
       
       write(6,*) ' '
       write(6,*) '-------'
       write(6,*) 'H_MAT_R'
       write(6,*) '-------'
       write(6,*) ' '
       call xrewind(NFHMATRWAN)
       write(NFHMATRWAN) nkb(1), nkb(2), nkb(3)
       write(NFHMATRWAN) wp%nocc
       do ja3 = 0,nkb(3)-1
          do ja2 = 0,nkb(2)-1
             do ja1 = 0,nkb(1)-1
                write(NFHMATRWAN) ((H_MAT_R(ib,jb,ja1,ja2,ja3), &
                     ib=1,wp%nocc),jb=1,wp%nocc)
             end do
          end do
       end do
       do ia1 = -wp%Na1, wp%Na1
          do ia2 = -wp%Na2, wp%Na2
             do ia3 = -wp%Na3, wp%Na3
                write(6,'(3(1x,i13))') ia1,ia2,ia3
                ja1 = modulo(ia1,nkb(1))
                ja2 = modulo(ia2,nkb(2))
                ja3 = modulo(ia3,nkb(3))
                do ib = 1, wp%nocc
                   write(6,'(100(1x,f13.8))') &
                        (dble(H_MAT_R(ib,jb,ja1,ja2,ja3)*au),jb=1,wp%nocc)
                end do
                write(6,*)
             end do
          end do
       end do
       close(NFHMATRWAN)

       call xrewind(NFAMATWAN)
       write(NFAMATWAN) mgs%nkmesh, wp%nbup-wp%nblw+1, wp%nocc
       write(NFAMATWAN) sklist
       write(NFAMATWAN) N_BAND
       do ik=1,mgs%nkmesh
          do ib=1,wp%nbup-wp%nblw+1
             write(NFAMATWAN) (A_MAT_O(ib,jb,ik),jb=1,wp%nocc)
          end do
       end do
       close(NFAMATWAN)

       call wan_write_recpwan(ncomp,maxval(mgs%nkm),mgs%nkmesh, &
            wp%nblw,wp%nbup,wp%nbup-wp%nblw+1,sklist,wp%nocc,mgs,mapkpt,dp, &
            N_BAND, N_BAND_BOTTOM, A_MAT_O, &
            ng,rg,nnp,pg,srg,aa)

       deallocate(WF_POSITION,H_MAT_R,A_MAT_O)
    else if (wp%pgm_mode .eq. 'recp_to_real') then
       call wan_recp2real(ncomp,va,aa,mgs &
            ,ng,rg,nnp,pg,srg,mgs%nkmesh,maxval(mgs%nkm) &
            ,wp%nbup-wp%nblw+1 &
            ,wp%iwan,wp%nocc,mapkpt,sklist)
    end if
    
    deallocate(E_EIG,sklist,mapkpt)

    return
  end subroutine calcwannier

end module subr_wannier
