!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2017 Yoshihide Yoshimoto
!

module subr_rdcgdata
  use subr_rddata
  use stropt_cnstr
contains
  subroutine rdcgdata(ax,bx,aa,bb,va,vb,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp &
       ,nkd,nsi,kd,asi &
       ,nband,ncomp,nspin,ndens,znext,dspin &
       ,nrx,nry,nrz,nrxyz &
       ,qwf,nwx,nwy,nwz &
       ,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode,scfcs,feps,eepsa,ekbt &
       ,niter0,niter1,ncycl,nrfr,most &
       ,decr,okatom,optcons,nf,mssg &
       ,syspext,stscl,cdm,axetp,udirchr)
    !
    use util
    use subr_inputparam
    use subr_inputsym
    use subr_inputatom
    use subr_inputkpoints
    use subr_inputstropt
    use subr_filemap
    use subr_inputinichr
    use scfloop
    use mpi
    implicit none
    real(kind=8),intent(out):: ax,bx,va,vb,qf,rf,qm,qc,qwf,znext,dspin
    real(kind=8),intent(out):: qbtp(3)
    type (scfctrl_struct):: scfcs
    real(kind=8),intent(out):: feps(2),eepsa,ekbt,decr,okatom(2)
    integer,intent(in):: nsymq,nopt
    integer,intent(out):: ng,nsv,nnp,nkd,nsi,ncomp,nband,nspin,ndens
    integer,intent(out):: nk,nw,nb1,nb2,nsk,ismode
    integer,intent(out):: idat(3)
    integer,intent(out):: ncycl,nrfr,most,niter0,niter1
    integer,intent(out):: nrx,nry,nrz,nrxyz,nwx,nwy,nwz
    integer,intent(out):: iopt(nopt)
    integer:: nf,mssg
    !
    real(kind=8),intent(out):: aa(3,3),bb(3,3)
    integer,intent(out):: rg(3,3,nsymq),pg(3,nsymq),srg(nsymq)
    integer,pointer:: kd(:),mmm(:,:,:)
    real(kind=8),pointer:: asi(:,:)

    real(kind=8):: syspext,stscl(6),cdm(6)
    type(optcons_struct):: optcons

    integer,intent(out):: axetp

    real(kind=8),pointer:: udirchr(:,:)

    real(kind=8):: qprcnd(6)

    integer:: info, tinfo

    real(kind=8):: eps,eeps,beta,qexat(2)
    integer:: ntt,ngx,ngy,ngz,ngxyz
    integer:: nhx,nhy,nhz
    integer:: ndiag0,ndiag1,ndx,ndy,ndz,itsb
    real(kind=8):: uptime
    real(kind=8):: qbsp
    integer:: nbsf
    real(kind=8):: fwqcom
    integer:: xctp
    real(kind=8):: thr_hybrid
    integer:: osqnrm
    real(kind=8),pointer:: zoin(:),znin(:)
    integer:: optmode
    real(kind=8):: thr_noncol,spaxs(3)

    if (mssg.gt.-1) write(6,'(/1x,a)') 'BEGIN input'

    info = 0

    ! namelist input

    call seektomagic(nf,magic_filemap,mssg,info)
    call read_filemap(nf,mssg,info)

    call seektomagic(nf,magic_inputparam,mssg,info)
    call read_tappinput(nf,ax,aa,ncomp,nspin,ndens,ismode,qf,qm,qc,beta, &
         qbtp,nkd,nsi,nw,znext,dspin,nopt,iopt,eps,eeps,ekbt, &
         uptime,niter0,niter1,itsb,ndiag0,ndiag1, &
         qbsp,nbsf,fwqcom,xctp,qexat,thr_hybrid,osqnrm,axetp, &
         qprcnd,thr_noncol,spaxs,mssg,info)
    nband = nw

    call seektomagic(nf,magic_tappsym,mssg,info)
    call read_tappsym(nf,ng,nsv,nnp,rg,pg,srg,nsymq,mssg,info)

    call seektomagic(nf,magic_tappatom,mssg,info)
    call read_tappatom(nf,nkd,zoin,znin,nsi,kd,asi,mssg,info)
    deallocate(zoin,znin)

    call seektomagic(nf,magic_tappkpoints,mssg,info)
    call read_tappkpoints(nf,nw,ax,aa,nb1,nb2,ekbt, &
         rf,ndx,ndy,ndz,nk,nsk,mmm,idat, &
         mssg,info)

    call seektomagic(nf,magic_tappstropt,mssg,info)
    call read_tappstropt(nf,ax,nsi,eepsa,feps, &
         decr,okatom,ncycl,nrfr,most,syspext,stscl,cdm,optmode,mssg,info)

    call seektomagic(nf,magic_tappstroptconstr,mssg,info)
    call read_tappstroptconstr(nf,nsi,optmode,optcons,mssg,info)

    ! optional section
    ! supress message to skip
    tinfo = 0
    call seektomagic(nf,magic_tappinichr,-1,tinfo)
    if (tinfo.eq.0) then
       call read_tappinichr(nf,nsi,udirchr,mssg,info)
    else
       allocate(udirchr(3,nsi))
       udirchr = 0.0d0
       if (ndens.eq.1) then
          udirchr(3,:) = 1.0d0
       end if
    end if

    if (info.ne.0) then
       if (mssg.gt.-1) write(6,*) 'input error'
       call MPI_ABORT(MPI_COMM_WORLD,1,info)
       stop
    end if

    if (mssg.gt.-1) write(6,'(/1x,a/)') 'END input'

    call latgen(ax,bx,aa,bb,va,vb)
    call printlat(ax,bx,aa,bb,va,vb,mssg)

    call calcmesh(bb,ng,pg,nnp,qf,qm,nhx,nhy,nhz,nrx,nry,nrz, &
         ngx,ngy,ngz,qwf,nwx,nwy,nwz,ntt,nrxyz,ngxyz,mssg)

    call set_scfctrl(niter0,ndiag0,ndiag1,itsb,ndx,ndy,ndz,beta, &
       eps,eeps,uptime,iopt(1),iopt(2),iopt(3), &
       qexat,thr_hybrid,osqnrm,qprcnd,thr_noncol,spaxs,scfcs)

    return
  end subroutine rdcgdata
end module subr_rdcgdata

program cgmrpt
  !*************************************************************
  !  CG  DIAGONALIZATION PROGRAM WITH VB POTENTIAL
  !   irf,iif,icf,ilf are introduced 1995-06-09
  !  Screw symmetry revived. 1995-10-20
  !  Introducing uniform background charge,
  !     we can calculate charged state.
  !    Check 'znel' 1996-08-19
  !     znext: extra electron per supercell
  !      Ex. znext= 2: -2 state
  !          znext=-2: +2 state
  !*************************************************************
  use mpi
#ifdef _OPENMP
  use omp_lib
#endif
  use chain_atom_cg
  use chain_wfn
  use util
  use iounit
  use scfloop
  use engdffr
  use bsplpot
  use uspp_scrdd
  use occup_num
  use types
  use wfnc_init
  use lpot_init
  use meshgrid
  use ewald_sum
  use linear_minim
  use timeprof
  use replica
  use exchange
  use chkinput
  use subr_rdcgdata
  use stropt_cnstr
#ifdef ESM
  use esm_manage
#endif
  implicit none
  !
#include "config90.h"
#ifdef ESM
#include "esm_version.h"
#endif
  real(kind=8):: bb0(3,3),bx,va0,vb0
  real(kind=8):: etotal,ewald,ef(2)
  real(kind=8):: entpysave,dif
  type(elec_energy):: eleng
  real(kind=8):: setotal(6),sewald(6)
  type(elec_senergy):: elseng
  real(kind=8):: eetrp
  real(kind=8),dimension(:,:),allocatable:: for,forel,force,dfor,dforel,dforce
  integer:: nrxyz,nks,nspin,nkmesh
  integer:: ncomp,ndens
  type (irdkpt_struct),pointer:: irdkpt(:)
  integer:: nrcell,nvx,nvy,nvz,nvxyz
  integer:: nwx,nwy,nwz
  real(kind=8):: znel

  integer:: nkr,nkfxa,nng
  integer:: laprj
  complex(kind=8),allocatable:: dd(:,:,:,:)
  !
  integer:: nkc
  !
  real(kind=8):: aa0(3,3),ax
  integer:: nsi,nkd
  integer:: nband,nw,nb1,nb2,nk
  real(kind=8),pointer:: asi(:,:)
  real(kind=8):: znext,dspin
  integer,pointer:: kd(:)
  integer:: rg(3,3,nsymq),pg(3,nsymq),srg(nsymq),ng,nsv,nnp
  integer:: iopt(nopt)
  real(kind=8):: qf,rf,qm,qc,qbtp(3)
  real(kind=8):: feps(2),eepsa,ekbt,decr,okatom(2)
  integer:: niter0,niter1,nrfr,most,ncycl
  integer:: ismode
  integer:: nsk
  integer,pointer:: mmm(:,:,:)
  integer:: idat(3)
  integer:: nrx,nry,nrz
  !
  real(kind=8):: qwf
  !
  real(kind=8):: rho(:,:),vxcr(:,:),crho(:,:)
  complex(kind=8):: lpt(:,:), pby(:,:,:,:,:)
  allocatable rho,vxcr,crho,pby,lpt
  !
  real(kind=8),allocatable:: ee(:,:,:),pe(:,:,:),wgtk(:,:,:)
  !
  type (bspotfc_struct):: bspotfc
  type (bspot_for_atm),allocatable:: bspots(:)
  !
  integer:: it, i
  integer:: ik,is
  integer:: iexit, itscf
  logical:: ifrgpbyee, ifinvsym
  integer:: ifchlin,ini_icy
  real(kind=8):: clk
  integer:: mssg,iex,ierr
  real(kind=8):: scfexit(nscfexit)

  real(kind=8):: syspext,stscl(6),aa(3,3),bb(3,3),va
  real(kind=8),dimension(:,:),allocatable:: dform,dfold,h
  real(kind=8):: cdm(6),hcdm(6),setotm(6),osetotm(6)
  real(kind=8):: hh,gh,gg,gam,temp,okstep(2),ddmax,almax,alpha,fmax(2)
  real(kind=8):: eokstp
  integer:: icflg,icy,ntcscf,iscf,ncscf,itmax(2)
  type (optcons_struct):: optcons
  !
  type (wfn_for_k),allocatable:: dp(:,:)
  !
  integer:: myrank,nproc,comm
  integer:: bkm,ekm

  type (scfctrl_struct):: scfcs
  type (meshgrid_struct):: mgs
!SOPTION CONTAREA(kqg)
  integer:: bkc,ekc
  integer,pointer:: kqg(:,:)

  type (replica_struct):: repls

  integer:: axetp

  real(kind=8),pointer:: udirchr(:,:)
  !
  call MPI_INIT(ierr)
  call MPI_COMM_RANK(MPI_COMM_WORLD,myrank,ierr)
  write(6,'(1x,a,i5)') 'ready world rank = ',myrank
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)

  call setup_replica(NFINPUT,MPI_COMM_WORLD,comm,repls)

  call MPI_COMM_RANK(comm,myrank,ierr)
  call MPI_COMM_SIZE(comm,nproc,ierr)
  call MPI_BARRIER(comm,ierr)
  if (myrank.eq.0) then
     mssg=1
  else
     mssg=-1
  end if
  !
  clk=xetime(0)
  if (mssg.gt.-1) then
     write(6,*)'CGMRPT'
     write(6,'(1x,a,i4)') 'MPI: nproc = ',nproc
#ifdef _OPENMP
     write(6,'(1x,a,i4)') 'OpenMP: nthread = ', omp_get_max_threads()
#endif
  end if
#ifdef ESM
  call esm_probe(mssg)
#endif
  !
  call rdcgdata(ax,bx,aa0,bb0,va0,vb0,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp &
       ,nkd,nsi,kd,asi &
       ,nband,ncomp,nspin,ndens,znext,dspin &
       ,nrx,nry,nrz,nrxyz &
       ,qwf,nwx,nwy,nwz &
       ,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode,scfcs,feps,eepsa,ekbt &
       ,niter0,niter1,ncycl,nrfr,most &
       ,decr,okatom,optcons,NFINPUT,mssg &
       ,syspext,stscl,cdm,axetp,udirchr)
  !
  if (iopt(5).ne.1) then
     call chkall(ng,rg,pg,srg,nnp,aa0,nsi,kd,asi,ndens,udirchr,comm)
  end if

  allocate(for(3,nsi),forel(3,nsi),force(3,nsi),dfor(3,nsi), &
       dforel(3,nsi),dforce(3,nsi),stat=ierr)
  call chkalloc('cgmrpt','for etc.',ierr)
  allocate(dform(3,nsi),dfold(3,nsi),h(3,nsi),stat=ierr)
  call chkalloc('cgmrpt','dform etc.',ierr)
  !
  call loadchkpots(nkd,nsi,aa0,qf,rf,qm,qc,ng,nnp,rg,pg,srg, &
       nk,nsk,mmm,idat,nvx,nvy,nvz,nvxyz,nks,nkmesh,irdkpt, &
       nng,nrcell,nkr,nkfxa,nkc,myrank,comm)
  !
  ifinvsym = .false.
  if (nsv.eq.1) then
     if (hasinvsym(ng,nnp,rg,pg,srg)) then
        ifinvsym = .true.
     else
        write(6,*) 'System does not have inversion symmetry'
        stop
     end if
  end if
  if (mssg.gt.-1 .and. ifinvsym) then
     write(6,*) 'Utilizing the inversion symmetry.'
  end if

  call cdm2aa(aa0,cdm,aa)
  call recp_uvec(aa,bb)
  va = cellvol(aa)
  !
  allocate(pe(nkr,nb1:nb2,nspin),wgtk(nb1:nb2,nks,nspin), &
       ee(nband,nks,nspin),stat=ierr)
  call chkalloc('cgmrpt','pe,wgtk,ee',ierr)
  !
  !     WARNING: ee must be cleared by 0 to notify
  !              ps2vcg() the SCF loop is initializing.
  !
  ee = 0.0d0
  !
  allocate(bspots(nkd),stat=ierr)
  call chkalloc('CGMRPT','bspots',ierr)
  call loadbspot(NFBSPOT,nkd,scfcs%xctp,bspotfc,bspots,myrank,comm)
  if (mssg.gt.-1) then
     call tellbspot(nkd,bspots)
  end if

  if (ncomp.ne.bspots(1)%ncomp) then
     write(6,*) 'inconsistent number_component between input and P.P.'
     stop
  end if
  znel = 0.0d0
  do it=1,nsi
     znel = znel + bspots(kd(it))%zo
  end do
  if (ncomp.eq.1) then
     if (nw.lt.0.5d0*znel) then
        write(6,*) 'too small number_band'
        write(6,*) 'number_band < 0.5*(total valence charge)'
        stop
     end if
  else
     if (nw.lt.znel) then
        write(6,*) 'too small number_band'
        write(6,*) 'number_band < total valence charge'
        stop
     end if
  end if

  call read_mesh(ng,rg,nnp,pg,srg, &
       nk,nks,nkmesh,irdkpt,nrcell,nkr,nng,nkfxa,nkc, &
       nrx,nry,nrz,nrxyz,nwx,nwy,nwz,nvx,nvy,nvz,nvxyz, &
       mgs,comm)

  kqg => mgs%kqg

  bkc = mgs%bkc
  ekc = mgs%ekc

  call potexch_setcutoff(scfcs%xctp,scfcs%qexat,aa0,nkmesh,mssg)

#ifdef ESM  
!!!! Added by Ando 10/15
  call esm_init_xTAPP(nsi,nkd,mgs,ng,aa,bb,kd,qf,va,bspots,mssg)
#endif
  allocate(lpt(nvxyz,nspin*ndens),stat=ierr)
  call chkalloc('cgmrpt','lpt',ierr)
  call initlpt(nspin,ndens,scfcs%spaxs,lpt,aa,nkd,kd,asi,nsi &
       ,iopt(3),udirchr,mgs,bspotfc,bspots,scfcs%xctp,mssg,comm)
  call prntime('INITLPT',mssg)
  !
  laprj = getlaprj(nkd,bspots)
  allocate(dd(laprj,laprj,nsi,nspin),stat=ierr)
  call chkalloc('cgmrpt','dd',ierr)
  !
  call mkinidd(ndens,bb,bkc,ekc,kqg,lpt,dd,kd,va,asi, &
       laprj,nsi,nkd,nvx,nvy,nvz,nspin, &
       bspotfc,bspots,comm)
  call prntime('MKINIDD',mssg)
  !
  allocate(dp(nks,nspin),stat=ierr)
  call chkalloc('cgmrpt','dp',ierr)
  do ik=1,nks
     bkm=mgs%g_displ(myrank,ik)+1
     ekm=mgs%g_displ(myrank,ik)+mgs%g_count(myrank,ik)
     do is=1,nspin
        allocate(dp(ik,is)%wf(ncomp,bkm:ekm,nw),stat=ierr)
        call chkalloc('cgmrpt','dp%wf',ierr)
     end do
  end do

  allocate(pby(laprj,nsi,nw,nks,nspin), &
       rho(nrxyz,nspin*ndens),vxcr(nrxyz,nspin*ndens), &
       crho(nrxyz,nspin*ndens),stat=ierr)
  call chkalloc('cgmrpt','pby,rho,vxcr',ierr)
  !
  if (iopt(5).eq.1) then
     if (mssg.gt.-1) write(6,*) 'loading atom cgdata ...'
     call load_atom_cg_data(ini_icy,dif,etotal,setotm,osetotm, &
          asi,cdm,dform,dfold,h,hcdm,myrank,comm)
     call cdm2aa(aa0,cdm,aa)
     call MPI_BARRIER(comm,ierr)
     if (mssg.gt.-1) then
        write(6,'(3(1x,e14.6))') aa(1,1),aa(2,1),aa(3,1)
        write(6,'(3(1x,e14.6))') aa(1,2),aa(2,2),aa(3,2)
        write(6,'(3(1x,e14.6))') aa(1,3),aa(2,3),aa(3,3)
        do it=1,nsi
           write(6,'(i3,'',( '',3(f9.6,'' ''),'')'')') &
                it,(asi(i,it),i=1,3)
        end do
     end if
     ifchlin=1
     if (mssg.gt.-1) then
        write(6,*) 'checking the symmetry of ', &
             'loaded atomic positions ...'
     end if
     !
     ! reset udirchr for this case
     !
     udirchr = 0.0d0
     call chkall(ng,rg,pg,srg,nnp,aa,nsi,kd,asi,ndens,udirchr,comm)
     itscf = 0
  end if

  if (iopt(2).eq.1) then
     call load_wfn(ncomp,nspin,ndens,nw,aa,ifrgpbyee, &
          dp,mgs,nb1,nb2,wgtk,laprj,nsi,kd,pby, &
          nband,ee,nkd,bspots,mssg,myrank,nproc,comm)
     if (ifrgpbyee) then
        if (scfcs%ndiag0 .le. 0) then
           write(6,*) ' davidson_number_diag_1st must be >= 1, ', &
                'when wave functions on sk are mapped or ', &
                'the number of bands is expanded.'
           stop
        end if
        if (iopt(5) .eq. 2) then
           write(6,*) 'chain_calc must be != 2', &
                'when wave functions on sk are mapped or ', &
                'the number of bands is expanded.'
           stop
        end if
     end if
     !
     !     In some architecture, two different I/O unit can not
     !     share the same file. Therefore NFWFI must be closed at this
     !     point for NFWFO.
     !
     close(NFWFI)
  else
     ! MAKE INITIAL RANDOM WAVE FN
     call initwfn(ncomp,nspin,nw,aa,dp,mgs,mssg, &
          myrank,nproc,comm)
     call prntime('INITWFN',mssg)

     ! reset for exchange-interaction
     wgtk = 0.0d0
  end if
#ifndef __GFORTRAN__
  if (myrank.eq.0) then
     ierr = 0
     call xrewind(NFLPT)
     call chknfwrt(NFLPT,'local potential',ierr)
     call xrewind(NFCHR)
     call chknfwrt(NFCHR,'charge density',ierr)
     call xrewind(NFRSL)
     call chknfwrt(NFRSL,'atomic structure data',ierr)
     call xrewind(NFPE)
     call chknfwrt(NFPE,'cosine band fit data',ierr)
     if (iopt(4).eq.1) then
        call xrewind(NFWFO)
        call chknfwrt(NFWFO,'wave function',ierr)
     end if
     if (ierr.ne.0) then
        call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
        stop
     end if
  end if
#endif
  !
  ! For charged state
  if (mssg.gt.-1) write(6,10) znext
10 format('CHARGE STATE CALCULATION:' &
       ,f8.3,' EXTRA ELECTRON IS ADDED')
  !
  if (iopt(5) .ne. 1) then
     clk=xetime(1)
     if (mssg.gt.-1) write(6,*) 'SCF START',clk
     !     
     if (iopt(5) .eq. 2) then
        scfcs%ifchscf = 1
     else
        scfcs%ifchscf = 0
     end if
     itscf=0
     call tpf_reset()
     call tpf_onoff(1)

     call ewalds(asi,aa,nsi,nkd,kd &
          ,for,dfor,ewald,sewald,bspots,myrank,nproc,comm,mssg)
     call prntime('EWALDS ',mssg)
     scfcs%niter=niter0
     call vbscf(dp,rho,pby,lpt,vxcr,crho &
          ,qbtp,ismode,iex,scfexit,ekbt,ewald,eleng &
          ,ef,eetrp,nband,laprj,nsi,nkd,ng &
          ,znext,dspin,ncomp,nspin,ndens,nw,nb1,nb2 &
          ,aa,asi,kd,ee,pe,wgtk,dd &
          ,scfcs,mgs,bspotfc,bspots,ifinvsym &
          ,mssg,myrank,nproc,comm)
     clk=xetime(1)
     if (mssg.gt.-1) write(6,*) 'SCF END TIME=',clk
     if (mssg.gt.-1) call tpf_print()
     if (iex.ge.0) then
        itscf=itscf+iex
     else if (iex .eq. -1) then
        itscf=itscf+niter0
     else if (iex .eq. -2) then
        if (mssg.gt.-1) then
           write(6,*) 'vbscf has exited because of uptime.'
        end if
        goto 9900
     else
        write(6,*) 'unknown result code from vbscf'
        write(6,*) 'iex = ',iex
        stop
     end if
     !     
     if (mssg.gt.-1) write(6,500)iex,scfexit(1),scfexit(2)
500  format(' ',' IEX=',i5,3x,'EXIT(1)=',e15.7,3x,'EXIT(2)=',e15.7)
     !     
     call vbfor(dp,lpt,rho,vxcr,crho &
          ,wgtk,eleng,forel,dforel,for,dfor,force,dforce &
          ,qbtp,elseng &
          ,aa,asi,kd,dd,ee,nband,laprj &
          ,rg,pg,ng,nnp,ncomp,nspin,ndens,nsi,nkd &
          ,nb1,nb2,mgs,bspotfc,bspots,scfcs%xctp,axetp &
          ,myrank,nproc,comm)
     call prntime('VBFOR  ',mssg)
     if (mssg.gt.-1) call tpf_print()
     !
     call dforce_cnstr(aa,nsi,dforce,dform,optcons)
     !
     call vtran(dforce,aa,nsi)
     etotal = eleng%total + ewald
     setotal = elseng%total + sewald
     setotm = setotal
     setotm(1:3) = setotm(1:3) - syspext*cellvol(aa)
     setotm = stscl*setotm/nsi

     if (mssg.gt.-1) then
        call prnatmst(aa,nsi,asi,dforce,dform, &
             etotal,setotal,eleng,nspin,ef,ewald)
     end if

     dfold = 0.0d0
     osetotm = 0.0d0
     h = 0.0d0
     hcdm = 0.0d0

     entpysave = 0.0d0
     dif = etotal + syspext*cellvol(aa) - entpysave
     call cnvch2(icflg,fmax,itmax,aa,dform,setotm, &
          dif,feps,eepsa,nsi,mssg)
     if (mssg.gt.-1) write(6,390)itmax(1),fmax(1)
390  format(' ','THE 1ST SCF IS DONE. ITMAX & FMAX=' &
          ,i5,e15.7,' Ht/A.U.')
     clk=xetime(1)
     if (mssg.gt.-1) write(6,*) 'THE 1st SCF END TIME=',clk
     if (icflg.eq.0 .or. ncycl.eq.0) then
        if (ncycl.eq.0 .and. mssg.gt.-1) then
           write(6,*)'SCF calculaton is done.'
        end if
        if (myrank .eq. 0) then
           call store_atom_cg_data(0,dif,etotal,setotm,osetotm, &
                asi,cdm,dform,dfold,h,hcdm)
        end if
        goto 9900
     end if
     ini_icy=1
     ifchlin=0
  end if

  if (myrank.eq.0) then
     mssg=0
  else
     mssg=-2
  end if

  ntcscf=1
  !
  do icy=ini_icy,ncycl
     if (myrank .eq. 0) then
        call store_atom_cg_data(icy,dif,etotal,setotm,osetotm, &
             asi,cdm,dform,dfold,h,hcdm)
     end if
     if (mssg.gt.-1) write(6,1500)icy
1500 format(' ','==================== ICY=',i5, &
          '====================')
     gam=0.0d0
     ! REFRESH CYCLE NRFR
     if (mod(icy-1,nrfr).ne.0) then
        gg =sekiaa(aa,dfold,dfold,nsi,osetotm,osetotm)
        dfold = dform - dfold
        osetotm = setotm - osetotm
        gam=sekiaa(aa,dfold,dform,nsi,osetotm,setotm)
        gam=gam/gg
     else
        if (mssg.gt.-1) write(6,*)' CG-DIRECTION REFRESHED !!!'
     end if
     h = dform + gam*h
     hcdm = setotm + gam*hcdm
     dfold = dform
     osetotm = setotm
     gh=sekiaa(aa,h,dform,nsi,hcdm,setotm)
     hh=sekiaa(aa,h,h    ,nsi,hcdm,hcdm)
     !
     entpysave = etotal + syspext*cellvol(aa)
     alpha=2.0d0*abs(dif/gh)
     hh=sqrt(hh)
     !        OKSTEP=SQRT(DBLE(NSI))*OKATOM/HH
     okstep=okatom
     if (mssg.gt.-1) write(6,2650)dif
2650 format(' ','TOTAL ENERGY DIFFERENCE IN CGMRPT',e15.7)
     if (mssg.gt.-1) write(6,2660)hh,gh,alpha,alpha*hh,okstep
2660 format(' ',' HH, GH, ALPHA =',3e15.7/ &
          ' ','ALPHA*HH,OKSTEP=',3e15.7)
     temp=alpha
     call chstep(aa,h,hcdm,nsi,okstep,temp,ddmax,almax,itmax(1),eokstp)
     if (mssg.gt.-1) then
        write(6,90)ddmax,itmax(1),eokstp,temp,almax
90   format(' ','MAXIMUM DISPLACEMENT SIZE & ATOM=',e15.7,i5/ &
          ' ','OKSTEP, ALPHA,ALMAX:',3e15.7)
     end if
     if (ddmax.gt.eokstp) then
        alpha=almax
        if (mssg.gt.-1) write(6,2670)alpha
2670    format(' ',' ALPHA IS TOO LARGE AND REPLACED BY ',e15.7)
     end if
     scfcs%niter = niter1
     call linmin(asi,cdm,h,hcdm,most,iexit,scfexit,iscf,ncscf &
          ,alpha,okstep,decr,feps,eepsa,stscl,syspext,optcons &
          ,lpt,rho,vxcr,crho,pby,dp &
          ,qbtp &
          ,nopt,iopt,ismode,ekbt,etotal,eleng,ewald,sewald &
          ,forel,dforel,for,dfor,force,dforce,dform,setotm &
          ,ef,eetrp,setotal,elseng &
          ,nband,laprj,nsi,nkd,rg,pg,ng,nnp &
          ,znext,dspin,ncomp,nspin,ndens,nw,nb1,nb2 &
          ,aa0,kd,ee,pe,wgtk,dd &
          ,bspotfc,bspots &
          ,scfcs,mgs,ifinvsym,mssg,ifchlin,axetp &
          ,myrank,nproc,comm)
     !
     if (mssg.gt.-1) write(6,2800)iexit,ncscf,iscf
2800 format(' ','IEXIT,NCSCF,ISCF=',3i5)
     !
     if (iexit .eq. -2) then
        !     if convergence of linmin is not achieved because of uptime
        goto 9900
     end if
     !
     call cdm2aa(aa0,cdm,aa)

     itscf=itscf+iscf
     ntcscf=ntcscf+ncscf
     if (mssg.gt.-1) write(6,3500)iscf,itscf,ntcscf
3500 format(' ',' ISCF=',i5,' TOTAL NUMBER OF SCF=',i5/ &
          ' ',' TOTAL NUMBER OF CALLING VBSCF',i5)
     !
     dif = etotal + syspext*cellvol(aa) - entpysave
     call cnvch2(icflg,fmax,itmax,aa,dform,setotm, &
          dif,feps,eepsa,nsi,mssg)
     if (mssg.gt.-1) then
        write(6,3600)icy
        write(6,3610)fmax(1),itmax(1)
        write(6,3620)fmax(2),itmax(2)
3600    format(1x,'########  ICY=',i5)
3610    format(1x,'   MAX FORCE =',e15.7,' a.u. (',i5,'-th)')
3620    format(1x,'   MAX STRESS =',e15.7,' a.u./atom (',i1,'-th)')
     end if
     if (icflg.eq.0) then
        if (myrank .eq. 0) then
           call store_atom_cg_data(icy+1,dif,etotal,setotm,osetotm, &
                asi,cdm,dform,dfold,h,hcdm)
        end if
        goto 9900
     end if
  end do

  if (mssg.gt.-1) write(6,*)'NOT CONVERGENT'
  !
9900 continue
  !
  ! TERMINATION OPERATION
  ! NOTE:DFORCE is in Cartesian coordinate now
  if (myrank .eq. 0) then
     call xrewind(NFRSL)
#ifdef COMPAT_OLD_STR_FORMAT
     call wrslt_old(NFRSL,kd,asi,optcons%mkd,nsi,nkd,ee,dforce,nband &
          ,nw,nb1,nb2,pe,nspin,ax,aa,etotal &
          ,ewald,eleng,setotal,mgs)
#else
     call wrslt(NFRSL,ax,aa,nkd,bspots,nsi,kd,asi,dforce,etotal,setotal, &
          nspin,ndens,ismode,ef,nrxyz,rho)
     call wreig(NFRSL,ee,nband,nw,nspin,mgs)
#endif

     call store_pe(nband,nspin,ee,nb1,nb2,mgs)
     !
     call xrewind(NFLPT)
     call recp_uvec(aa,bb)
     write(NFLPT) 'TAPP:lpt'
     write(NFLPT) bb, nspin*ndens
     write(NFLPT) nvx, nvy, nvz
     write(NFLPT) lpt
     close(NFLPT)
     call xrewind(NFCHR)
     write(NFCHR) 'TAPP:rho'
     write(NFCHR) aa, nspin*ndens
     write(NFCHR) 2*nrx, 2*nry, 2*nrz
     write(NFCHR) rho
     close(NFCHR)
  end if
  !
  if (iopt(4).eq.1) then
     if (mssg.gt.-1) write(6,*) 'storing wave function...'
     call store_wfn(aa,ncomp,nspin,nw,ifinvsym, &
          dp,mgs,nb1,nb2,wgtk,laprj,nsi,kd,pby, &
          nband,ee,nkd,bspots,myrank,nproc,comm)
  end if
  call dealloc_mesh(mgs)
  deallocate(pe,wgtk,ee)
  deallocate(lpt,rho,vxcr,crho,pby)
  deallocate(dd)
  do is=1,nspin
     do ik=1,nks
        deallocate(dp(ik,is)%wf)
     end do
  end do
  deallocate(dp)
  deallocate(dform,dfold,h)
  deallocate(for,forel,force,dfor,dforel,dforce)
  call dealloc_optcons(optcons)
  deallocate(kd,asi,mmm)
  clk=xetime(1)
  if (mssg.gt.-1) write(6,*) 'TOTAL TIME=',clk
  call MPI_FINALIZE(ierr)
  stop
contains

  subroutine wrslt(nf,ax,aa,nkd,bspots,nsi,kd,asi,cforce,etotal,setotal, &
       nspin,ndens,ismode,ef,nrxyz,rho)
    use bsplpot
    use util
    implicit none
    integer,intent(in):: nf
    real(kind=8),intent(in):: ax,aa(3,3)
    integer,intent(in):: nkd
    type (bspot_for_atm),intent(in):: bspots(nkd)
    integer,intent(in):: nsi,kd(nsi)
    real(kind=8),intent(in):: asi(3,nsi),cforce(3,nsi),etotal,setotal(6)
    integer,intent(in):: nspin,ndens,ismode
    real(kind=8),intent(in):: ef(2)
    integer,intent(in):: nrxyz
    real(kind=8),intent(in):: rho(nrxyz,nspin*ndens)

    integer:: it,ikd
    integer:: i
    real(kind=8):: va, spol,aspol,vpol(3),vapol(3)

    spol = 0.0d0
    aspol = 0.0d0
    vpol = 0.0d0
    if (ndens.eq.1) then
       if (nspin.eq.2) then
          va = cellvol(aa)
          do i=1,nrxyz
             spol = spol + rho(i,2) - rho(i,1)
             aspol = aspol + abs(rho(i,2) - rho(i,1))
          end do
          spol = spol*va/nrxyz
          aspol = aspol*va/nrxyz
       end if
    else
       va = cellvol(aa)
       call calchr2(rho(:,2),nrxyz,va,vpol(1),vapol(1))
       call calchr2(rho(:,3),nrxyz,va,vpol(2),vapol(2))
       call calchr2(rho(:,4),nrxyz,va,vpol(3),vapol(3))
       call calpol(rho(:,2:4),nrxyz,va,aspol)
       spol = vpol(3)
    end if

    write(nf,'(''&struct_data'')')
    write(nf,'(''lattice_factor = '',e18.10,'','')') ax
    write(nf,'(''lattice_list = '')')
    write(nf,'(2x,3(e18.10,'',''))') aa(:,1)/ax
    write(nf,'(2x,3(e18.10,'',''))') aa(:,2)/ax
    write(nf,'(2x,3(e18.10,'',''))') aa(:,3)/ax
    write(nf,'(''total_energy ='',e21.12,'','')') etotal
    write(nf,'(''stress_tensor ='')')
    write(nf,'(2x,3(e21.12,'',''))') setotal(1:3)
    write(nf,'(2x,3(e21.12,'',''))') setotal(4:6)
    if (nspin.eq.2 .and. ismode.eq.1) then
       write(nf,'(''fermi_energy ='',e19.10,'','',e19.10,'','')') ef
    else
       write(nf,'(''fermi_energy ='',e19.10,'','',e19.10,'','')') &
            ef(1), ef(1)
    end if
    write(nf,'(''number_element = '',i0,'','')') nkd
    write(nf,'(''number_atom = '',i0,'','')') nsi
    write(nf,'(''spin_polarization ='',e19.10,'','')') spol
    if (ndens.eq.4) then
       write(nf,'(''spin_polarization_vector ='',3(e17.8,'',''))') vpol
       write(nf,'(''abs_spin_polarization_vector ='',3(e17.8,'',''))') vapol
    end if
    write(nf,'(''abs_spin_polarization ='',e19.10)') aspol
    write(nf,'(''/'')')

    write(nf,*)
    write(nf,'(a)') '# valence_charge, nucleous_charge'
    do ikd=1,nkd
       write(nf,'(2(1x,f12.8))') bspots(ikd)%zo, bspots(ikd)%zn
    end do

    write(nf,*)
    write(nf,'(a)') '# atom_kind, atom_position by cell coordinate'
    do it=1,nsi
       write(nf,'(i3,1x,3f16.12)') &
            kd(it),asi(1,it),asi(2,it),asi(3,it)
    end do

    write(nf,*)
    write(nf,'(a)') '# force by Cartesian coordinate'
    do it=1,nsi
       write(nf,'(3e21.12)') cforce(1,it),cforce(2,it),cforce(3,it)
    end do
  end subroutine wrslt

  subroutine wrslt_old(nfrsl,kd,asi,mkd,nsi,nkd,ee,cforce,nband &
       ,nw,nb1,nb2,pe,nspin,ax,aa,etotal &
       ,ewald,eleng,setotal,mgs)
    use util
    use types
    use meshgrid
    implicit none
    integer,intent(in):: nfrsl,nsi,nkd,nband,nw,nb1,nb2,nspin
    integer,intent(in):: kd(nsi),mkd(nsi)
    type(meshgrid_struct),intent(in):: mgs
    real(kind=8),intent(in):: asi(3,nsi),ee(nband,mgs%nks,nspin)
    real(kind=8),intent(in):: ax,aa(3,3),cforce(3,nsi)
    real(kind=8),intent(in):: pe(mgs%nkr,nb1:nb2,nspin)
    real(kind=8),intent(in):: etotal,ewald,setotal(6)
    type(elec_energy),intent(in):: eleng

    integer,parameter:: ncolumn = 5
    real(kind=8),parameter:: pi2=6.28318530717958648d0
    integer:: it,i,j,k,iblk,nblk,ik,is,nlast,nbase,ndata,ib
    real(kind=8):: dforce(3,nsi)
    !
1   format('### CELL INFORMATION ###')
    write(nfrsl,1)
5   format(3e20.12)
    write(nfrsl,5)ax
    write(nfrsl,5)((aa(i,j)/ax,i=1,3),j=1,3)
8   format(/'### ATOMIC SITE INFORMATION ###')
    write(nfrsl,8)
    write(nfrsl,10) nkd,nsi
10  format(2i4)
    do it=1,nsi
       write(nfrsl,15) kd(it),asi(1,it),asi(2,it),asi(3,it),mkd(it)
    end do
15  format(i4,3e20.12,i4)
    write(nfrsl,20)
20  format(/'### TOTAL ENERGY INFORMATION ###')
    !
    !     XXX: first arg was the variable "const".
    write(nfrsl,21) 0.0,ewald,eleng%eht,eleng%exc,eleng%exa, &
         eleng%epot,eleng%epotl,eleng%epotnl,eleng%ekin
21  format(3e20.12,' #CONST,EWALD, EHT'/ &
           3e20.12,' #EXC,EXA,EPOT'/ &
           3e20.12,' #EPOTL,EPOTNL,EKIN')
    write(nfrsl,22) etotal,eleng%eteig
22  format(2e20.12,20x,' #ETOTAL,ETEIG')
31  format(/'### ATOMICE FORCE INFORMATION ###')
    write(nfrsl,31)
    ! Transform: cforce(in a.u.)->(in AA-representation)

    call cart2cell(aa,nsi,cforce,dforce)

    write(nfrsl,33)
33  format('# ATOMIC NUM. (DFORCE(I,),I=1,3)')
    do it=1,nsi
       write(nfrsl,34)it,(dforce(i,it),i=1,3)
    end do
34  format(i4,3e20.12)

    write(nfrsl,'(/''### STRESS TENSOR ###'')')
    write(nfrsl,'(3(1x,e18.10))') setotal(1),setotal(2),setotal(3)
    write(nfrsl,'(3(1x,e18.10))') setotal(4),setotal(5),setotal(6)
    !
    write(nfrsl,35)
35  format(//'### ELECTRONIC INFORMATION ###'/)
    ! eigen value
    write(nfrsl,40)
40  format('  ### EIGEN VALUE ###'/)
    nblk=(nw-1)/ncolumn+1
    nlast=nw-(nblk-1)*ncolumn
    do ik=1,mgs%nks
       write(nfrsl,65) ik,mgs%sk(:,ik)
       do is=1,nspin
          if (nspin .eq. 2) then
             write(nfrsl,'(1x,a,i2)') 'spin = ', is
          end if
          do iblk=1,nblk
             if (iblk.ne.nblk) then
                ndata=ncolumn
             else
                ndata=nlast
             end if
             nbase=(iblk-1)*ncolumn
             write(nfrsl,70)iblk,(ee(ib,ik,is),ib=nbase+1,nbase+ndata)
          end do
       end do
    end do
65  format(/i5,'-TH SK=',3e15.7)
70  format(i3,1x,10e15.7)
    ! occupation data
    write(nfrsl,90)
90  format(/'  ### OCCUPATION DATA ###')
    do ib=nb1,nb2
       write(nfrsl,150) ib
150    format('  IB=',i6)
       do k=1,nkr
          write(nfrsl,200) k,(pe(k,ib,is),is=1,nspin)
200       format(' K=',i5,3x,'PE=',e29.16,1x,e29.16)
       end do
    end do
    !
    return
  end subroutine wrslt_old

  subroutine wreig(nfrsl,ee,nband,nw,nspin,mgs)
    use meshgrid
    implicit none
    integer,intent(in):: nfrsl,nband,nw,nspin
    type(meshgrid_struct),intent(in):: mgs
    real(kind=8),intent(in):: ee(nband,mgs%nks,nspin)

    integer,parameter:: ncolumn = 5
    integer:: iblk,nblk,ik,is,nlast,nbase,ndata,ib
    !
    write(nfrsl,*)
    write(nfrsl,'(a)') '# orbital eigen value'
    nblk = nw/ncolumn
    nlast = nw - nblk*ncolumn
    do ik=1,mgs%nks
       write(nfrsl,'(a,3(1x,f10.7))') '# sk = ',mgs%sk(:,ik)
       do is=1,nspin
          if (nspin .eq. 2) then
             write(nfrsl,'(1x,a,i2)') '# spin = ', is
          end if
          do iblk=0,nblk,2
             if (iblk.ne.nblk) then
                ndata=ncolumn
             else
                ndata=nlast
             end if
             if (ndata.eq.0) exit

             nbase = iblk*ncolumn
             write(nfrsl,'(i3,1x,10e15.7)') iblk/2, &
                  (ee(ib,ik,is),ib=nbase+1,nbase+ndata)

             if (iblk+1.gt.nblk) exit

             if (iblk+1.ne.nblk) then
                ndata=ncolumn
             else
                ndata=nlast
             end if
             nbase = (iblk+1)*ncolumn
             write(nfrsl,'(4x,10e15.7)') (ee(ib,ik,is),ib=nbase+1,nbase+ndata)
          end do
       end do
       write(nfrsl,*)
    end do
    !
    return
  end subroutine wreig
end program cgmrpt
