!
! Copyright (c) 2013,2016 Yoshihide Yoshimoto
!
program cmpstr
  use subr_readstr
  implicit none
  real(kind=8):: ax1,ax2
  integer:: nkd1,nsi1,nkd2,nsi2
  integer,parameter:: nf1 = 10, nf2 = 20
  !
  real(kind=8):: aa1(3,3),aa2(3,3)
  integer,pointer:: kd1(:),kd2(:)
  real(kind=8),pointer:: asi1(:,:),cforce1(:,:),asi2(:,:),cforce2(:,:)

  real(kind=8),pointer:: zo1(:),zn1(:),zo2(:),zn2(:)

  real(kind=8):: etotal1,etotal2,stress1(6),stress2(6),ef1(2),ef2(2)
  real(kind=8):: spinpol1,spinpol2,aspinpol1,aspinpol2

  real(kind=8):: h(3)

  character(len=256):: cbuf

  integer:: ikd,it
  integer:: ios
  logical:: flag, ifdif
  real(kind=8):: delta

  real(kind=8):: epscell = 1.0d-5, epscellcoord = 1.0d-5
  real(kind=8):: epscartforce = 1.0d-5, epsetotal = 1.0d-8
  real(kind=8):: epsstress = 1.0d-5, epsef = 1.0d-5
  real(kind=8):: epsspinpol = 1.0d-5, epsaspinpol = 1.0d-5

  if (iargc().eq.0) then
     write(6,*) 'eps_cell = ', epscell
     write(6,*) 'eps_cell_coord = ', epscellcoord
     write(6,*) 'eps_cart_force = ', epscartforce
     write(6,*) 'eps_etotal = ', epsetotal
     write(6,*) 'eps_stress = ', epsstress
     write(6,*) 'eps_ef = ', epsef
     write(6,*) 'eps_spin_pol = ', epsspinpol
     write(6,*) 'eps_abs_spin_pol = ', epsaspinpol
     stop
  else if (iargc().ne.2) then
     write(6,*) 'error in args'
     stop 1
  end if

  call getarg(1, cbuf)
  open(nf1, file=cbuf, status='old', action='read', iostat=ios)
  if (ios.ne.0) then
     write(6,*) 'error in opening '//trim(cbuf)
     stop 1
  end if

  call getarg(2, cbuf)
  open(nf2, file=cbuf, status='old', action='read', iostat=ios)
  if (ios.ne.0) then
     write(6,*) 'error in opening '//trim(cbuf)
     stop 1
  end if

  call xread_strfile(nf1,nkd1,nsi1,ax1,aa1,kd1,asi1,cforce1,zo1,zn1, &
       etotal1,stress1,ef1,spinpol1,aspinpol1)

  call xread_strfile(nf2,nkd2,nsi2,ax2,aa2,kd2,asi2,cforce2,zo2,zn2, &
       etotal2,stress2,ef2,spinpol2,aspinpol2)

  ifdif = .false.

  if (nsi1.ne.nsi2) then
     write(6,*) 'nsi1 != nsi2'
     stop 1
  end if
  do it=1,nsi1
     if (kd1(it).ne.kd2(it)) then
        write(6,*) 'atomic kind'
        write(6,*) ' it, kd1, kd2 = ',it,kd1(it),kd2(it)
        ifdif = .true.
     end if
  end do

  aa1 = ax1*aa1
  aa2 = ax2*aa2

  flag = .false.
  delta = maxval(abs(aa1-aa2))/maxval(abs(aa1))
  if (delta .gt. epscell) then
     flag = .true.
  end if
  if (flag) then
     write(6,'(1x,a,e10.3)') 'cell form ', delta
     ifdif = .true.
  end if

  flag = .false.
  do it=1,nsi1
     h = modulo(asi2(:,it) - asi1(:,it) + 0.5d0, 1.0d0) - 0.5d0
     delta = sqrt(h(1)*h(1)+h(2)*h(2)+h(3)*h(3))
     if (delta .gt. epscellcoord) then
        flag = .true.
     end if
  end do
  if (flag) then
     write(6,'(1x,a,e10.3)') 'cell coordinate ', delta
     ifdif = .true.
  end if

  flag = .false.
  do it=1,nsi1
     h = cforce2(:,it) - cforce1(:,it)
     delta = sqrt(h(1)*h(1)+h(2)*h(2)+h(3)*h(3))
     if (delta .gt. epscartforce) then
        flag = .true.
     end if
  end do
  if (flag) then
     write(6,'(1x,a,e10.3)') 'Cartesian force ',delta
     ifdif = .true.
  end if

  delta = abs( (etotal2 - etotal1)/etotal1 )
  if (delta .gt. epsetotal) then
     write(6,'(1x,a,e10.3)') 'total energy ', delta
     ifdif = .true.
  end if

  delta = maxval(abs(stress2 - stress1))
  if (delta .gt. epsstress) then
     write(6,'(1x,a,e10.3)') 'total stress ', delta
     ifdif = .true.
  end if

  delta = maxval(abs(ef2 - ef1))
  if (delta .gt. epsef) then
     write(6,'(1x,a,e10.3)') 'fermi energy ', delta
     ifdif = .true.
  end if

  delta = abs(spinpol2 - spinpol1)
  if (delta .gt. epsspinpol) then
     write(6,'(1x,a,e10.3)') 'spin polarization ', delta
     ifdif = .true.
  end if

  delta = abs(aspinpol2 - aspinpol1)
  if (delta .gt. epsaspinpol) then
     write(6,'(1x,a,e10.3)') 'absolute spin polarization ', delta
     ifdif = .true.
  end if

  if (ifdif) then
     stop 1
  else
     stop
  end if

end program cmpstr
