!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2016,2017 Yoshihide Yoshimoto
!
#include "fftgrid.h"

module converter
contains
  subroutine kg2g2k(bb,sk,nk,kg,g2k)
    !
    !     convert kg to g2k = (x,y,z,r)
    !
    implicit none
    real(kind=8),intent(in):: bb(3,3),sk(3)
    integer,intent(in):: nk, kg(3,nk)
    real(kind=8),intent(out):: g2k(4,nk)
    !
    real(kind=8):: a,b,c,x1,x2,x3
    integer:: ig
    !
    do ig=1,nk
       a=sk(1)+kg(1,ig)
       b=sk(2)+kg(2,ig)
       c=sk(3)+kg(3,ig)
       x1=a*bb(1,1)+b*bb(1,2)+c*bb(1,3)
       x2=a*bb(2,1)+b*bb(2,2)+c*bb(2,3)
       x3=a*bb(3,1)+b*bb(3,2)+c*bb(3,3)
       g2k(1,ig)=x1
       g2k(2,ig)=x2
       g2k(3,ig)=x3
       g2k(4,ig)=sqrt(x1*x1+x2*x2+x3*x3)
    end do
    return
  end subroutine kg2g2k

  subroutine kqg2qgk(bb,nk,kqg,qgk)
    !
    !     convert kqg to qgk = (x,y,z,r)
    !
    implicit none
    real(kind=8),intent(in):: bb(3,3)
    integer,intent(in):: nk, kqg(3,nk)
    real(kind=8),intent(out):: qgk(4,nk)
    !
    real(kind=8):: a,b,c,x1,x2,x3
    integer:: ig
    !
    do ig=1,nk
       a=kqg(1,ig)
       b=kqg(2,ig)
       c=kqg(3,ig)
       x1=a*bb(1,1)+b*bb(1,2)+c*bb(1,3)
       x2=a*bb(2,1)+b*bb(2,2)+c*bb(2,3)
       x3=a*bb(3,1)+b*bb(3,2)+c*bb(3,3)
       qgk(1,ig)=x1
       qgk(2,ig)=x2
       qgk(3,ig)=x3
       qgk(4,ig)=sqrt(x1*x1+x2*x2+x3*x3)
    end do
    return
  end subroutine kqg2qgk

  subroutine kg2akk(bb,sk,nk,kg,akk)
    !
    !     convert kg to akk
    !
    implicit none
    real(kind=8),intent(in):: bb(3,3),sk(3)
    integer,intent(in):: nk,kg(3,nk)
    real(kind=8),intent(out):: akk(nk)
    !
    real(kind=8):: a,b,c,x1,x2,x3
    integer:: ig
    !
    do ig=1,nk
       a=sk(1)+kg(1,ig)
       b=sk(2)+kg(2,ig)
       c=sk(3)+kg(3,ig)
       x1=a*bb(1,1)+b*bb(1,2)+c*bb(1,3)
       x2=a*bb(2,1)+b*bb(2,2)+c*bb(2,3)
       x3=a*bb(3,1)+b*bb(3,2)+c*bb(3,3)
       akk(ig)=x1*x1+x2*x2+x3*x3
    end do
    return
  end subroutine kg2akk

  subroutine kg2akg(bb,sk,nk,kg,akg)
    !
    !     convert kg to akg = |k+g| = sqrt[(k+g)*(k+g)]
    !
    implicit none
    real(kind=8),intent(in):: bb(3,3),sk(3)
    integer,intent(in):: nk,kg(3,nk)
    real(kind=8),intent(out):: akg(nk)
    !
    real(kind=8):: a,b,c,x1,x2,x3
    integer:: ig
    !
    do ig=1,nk
       a=sk(1)+kg(1,ig)
       b=sk(2)+kg(2,ig)
       c=sk(3)+kg(3,ig)
       x1=a*bb(1,1)+b*bb(1,2)+c*bb(1,3)
       x2=a*bb(2,1)+b*bb(2,2)+c*bb(2,3)
       x3=a*bb(3,1)+b*bb(3,2)+c*bb(3,3)
       akg(ig)=sqrt(x1*x1+x2*x2+x3*x3)
    end do
    return
  end subroutine kg2akg

  subroutine kf2ag(bb,nk,kf,af)
    !
    !     convert kf to af = |g| = sqrt(g*g)
    !
    implicit none
    real(kind=8),intent(in):: bb(3,3)
    integer,intent(in):: nk,kf(3,nk)
    real(kind=8),intent(out):: af(nk)
    !
    real(kind=8):: x1,x2,x3
    integer:: ig
    !
    do ig=1,nk
       x1=kf(1,ig)*bb(1,1)+kf(2,ig)*bb(1,2)+kf(3,ig)*bb(1,3)
       x2=kf(1,ig)*bb(2,1)+kf(2,ig)*bb(2,2)+kf(3,ig)*bb(2,3)
       x3=kf(1,ig)*bb(3,1)+kf(2,ig)*bb(3,2)+kf(3,ig)*bb(3,3)
       af(ig)=sqrt(x1*x1+x2*x2+x3*x3)
    end do
    return
  end subroutine kf2ag

  subroutine kf2aff(bb,nk,kf,aff)
    !
    !     convert kf to aff = g*g
    !
    implicit none
    real(kind=8),intent(in):: bb(3,3)
    integer,intent(in):: nk,kf(3,nk)
    real(kind=8),intent(out):: aff(nk)
    !
    real(kind=8):: x1,x2,x3
    integer:: ig
    !
    do ig=1,nk
       x1=kf(1,ig)*bb(1,1)+kf(2,ig)*bb(1,2)+kf(3,ig)*bb(1,3)
       x2=kf(1,ig)*bb(2,1)+kf(2,ig)*bb(2,2)+kf(3,ig)*bb(2,3)
       x3=kf(1,ig)*bb(3,1)+kf(2,ig)*bb(3,2)+kf(3,ig)*bb(3,3)
       aff(ig)=x1*x1+x2*x2+x3*x3
    end do
    return
  end subroutine kf2aff

  subroutine kg2j2g(nwx2,nwy2,nwz2,nkm,kg,j2g)
    !
    !     convert kg to j2g
    !
    implicit none
    integer,intent(in):: nwx2,nwy2,nwz2,nkm
    integer,intent(in):: kg(3,nkm)
    integer,intent(out):: j2g(nkm)
    integer:: ig,k1,k2,k3
    !
    do ig=1,nkm
       k1=kg(1,ig)
       k2=kg(2,ig)
       k3=kg(3,ig)
       k1=mod(nwx2+k1,nwx2)+1
       k2=mod(nwy2+k2,nwy2)+1
       k3=mod(nwz2+k3,nwz2)+1
       j2g(ig)=(k3-1)*nwy2*nwx2+(k2-1)*nwx2+k1
    end do
    !
    return
  end subroutine kg2j2g

  subroutine kg2lj2g(nx,ny,nz,lx,ly,lz,nkm,kg,lj2g)
    !
    !     convert kg to lj2g
    !     l[xyz] : leading dimensions of [xyz]
    !
    implicit none
    integer,intent(in):: nx,ny,nz,lx,ly,lz,nkm
    integer,intent(in):: kg(3,nkm)
    integer,intent(out):: lj2g(nkm)
    integer:: ig,k1,k2,k3
    !
    if ((lx .lt. nx) .or. (ly .lt. ny) .or. (lz .lt. nz)) then
       write(6,*) 'kg2lj2g: l[xyz] < n[xyz]'
       stop
    end if
    !
    do ig=1,nkm
       k1=kg(1,ig)
       k2=kg(2,ig)
       k3=kg(3,ig)
       k1=mod(nx+k1,nx)+1
       k2=mod(ny+k2,ny)+1
       k3=mod(nz+k3,nz)+1
       lj2g(ig)=(k3-1)*ly*lx+(k2-1)*lx+k1
    end do
    !
    return
  end subroutine kg2lj2g

  subroutine kg2lj2gs(nx,ny,nz,lx,ly,lz,nkm,kg,lj2gs)
    !
    !     convert kg to lj2gs, shifted axis (y,z,x) version
    !     l[xyz] : leading dimensions of [xyz]
    !
    implicit none
    integer,intent(in):: nx,ny,nz,lx,ly,lz,nkm
    integer,intent(in):: kg(3,nkm)
    integer,intent(out):: lj2gs(nkm,2)
    integer:: ig,k1,k2,k3
    !
    if ((lx .lt. nx) .or. (ly .lt. ny) .or. (lz .lt. nz)) then
       write(6,*) 'kg2lj2gs: l[xyz] < n[xyz]'
       stop
    end if
    !
    do ig=1,nkm
       k1=kg(1,ig)
       k2=kg(2,ig)
       k3=kg(3,ig)
       k1=mod(nx+k1,nx)+1
       k2=mod(ny+k2,ny)+1
       k3=mod(nz+k3,nz)+1
       lj2gs(ig,1)=(k2-1)*lz*lx+(k1-1)*lz+k3
       lj2gs(ig,2)=(k1-1)*ly*lz+(k3-1)*ly+k2
    end do
    !
    return
  end subroutine kg2lj2gs

  subroutine kg2lj2gg(nx,ny,nz,lx,ly,lz,nkm,kg,lj2g)
    !
    !     convert kg to lj2g
    !     l[xyz] : leading dimensions of [xyz]
    !
    implicit none
    integer,intent(in):: nx,ny,nz,lx,ly,lz,nkm
    integer,intent(in):: kg(3,nkm)
    integer,intent(out):: lj2g(nkm,2)
    integer:: ig,k1,k2,k3
    !
    if ((lx .lt. nx) .or. (ly .lt. ny) .or. (lz .lt. nz)) then
       write(6,*) 'kg2lj2g: l[xyz] < n[xyz]'
       stop
    end if
    !
    do ig=1,nkm
       k1=kg(1,ig)
       k2=kg(2,ig)
       k3=kg(3,ig)
       k1=mod(nx+k1,nx)+1
       k2=mod(ny+k2,ny)+1
       k3=mod(nz+k3,nz)+1
       lj2g(ig,1)=(k3-1)*ly*lx+(k2-1)*lx+k1
       k1=-kg(1,ig)
       k2=-kg(2,ig)
       k3=-kg(3,ig)
       k1=mod(nx+k1,nx)+1
       k2=mod(ny+k2,ny)+1
       k3=mod(nz+k3,nz)+1
       lj2g(ig,2)=(k3-1)*ly*lx+(k2-1)*lx+k1
    end do
    !
    return
  end subroutine kg2lj2gg

  subroutine kg2lj2gsg(nx,ny,nz,lx,ly,lz,nkm,kg,lj2gs)
    !
    !     convert kg to lj2gs, shifted axis (y,z,x) version
    !     l[xyz] : leading dimensions of [xyz]
    !
    implicit none
    integer,intent(in):: nx,ny,nz,lx,ly,lz,nkm
    integer,intent(in):: kg(3,nkm)
    integer,intent(out):: lj2gs(nkm,4)
    integer:: ig,k1,k2,k3
    !
    if ((lx .lt. nx) .or. (ly .lt. ny) .or. (lz .lt. nz)) then
       write(6,*) 'kg2lj2gs: l[xyz] < n[xyz]'
       stop
    end if
    !
    do ig=1,nkm
       k1=kg(1,ig)
       k2=kg(2,ig)
       k3=kg(3,ig)
       k1=mod(nx+k1,nx)+1
       k2=mod(ny+k2,ny)+1
       k3=mod(nz+k3,nz)+1
       lj2gs(ig,1)=(k2-1)*lz*lx+(k1-1)*lz+k3
       lj2gs(ig,3)=(k1-1)*ly*lz+(k3-1)*ly+k2
       k1=-kg(1,ig)
       k2=-kg(2,ig)
       k3=-kg(3,ig)
       k1=mod(nx+k1,nx)+1
       k2=mod(ny+k2,ny)+1
       k3=mod(nz+k3,nz)+1
       lj2gs(ig,2)=(k2-1)*lz*lx+(k1-1)*lz+k3
       lj2gs(ig,4)=(k1-1)*ly*lz+(k3-1)*ly+k2
    end do
    !
    return
  end subroutine kg2lj2gsg

  subroutine vg2vr(vg,vr,nvx,nvy,nvz,nwx,nwy,nwz,nlx,nly,nlz)
    use fft_3d
    implicit none
    integer,intent(in):: nvx,nvy,nvz,nwx,nwy,nwz,nlx,nly,nlz
    complex(kind=8),intent(in):: vg(-nvx:nvx, -nvy:nvy, -nvz:nvz)
    real(kind=8):: vr(nlx,nly,nlz)
    !
    real(kind=8),allocatable:: ztmp(:,:),vtmp(:,:,:,:)
    integer:: k1,k2,k3,i1,i2,i3,nwx2,nwy2,nwz2,ierr
    type (fft3_struct):: fs
    !
    allocate(FFT_A1D_DEFSZ(ztmp,nlx*nly*nlz),FFT_A3D_DEFSZ(vtmp,nlx,nly,nlz),stat=ierr)
    call chkalloc('vg2vr','ztmp,vtmp',ierr)
    nwx2 = nwx*2
    nwy2 = nwy*2
    nwz2 = nwz*2
    vtmp = 0.0d0
!$omp parallel do private(i1,i2,i3,k1,k2,k3)
!POPTION INDEP(vtmp)
    do i3=-min(nvz,nwz-1),min(nvz,nwz-1)
       do i2=-min(nvy,nwy-1),min(nvy,nwy-1)
          do i1=-min(nvx,nwx-1),min(nvx,nwx-1)
             k1=mod(nwx2+i1,nwx2)+1
             k2=mod(nwy2+i2,nwy2)+1
             k3=mod(nwz2+i3,nwz2)+1
             FFT_A3D_RL(vtmp,k1,k2,k3)=dble(vg(i1,i2,i3))
             FFT_A3D_IM(vtmp,k1,k2,k3)=aimag(vg(i1,i2,i3))
          end do
       end do
    end do
    call fft3_init(nwx2,nwy2,nwz2,nlx,nly,nlz,fs)
    call fft3_bw(fs,vtmp,ztmp)
    call fft3_done(fs)
    vr = FFT_A3D_RL(vtmp,:,:,:)
    deallocate(ztmp,vtmp)
    return
  end subroutine vg2vr

  subroutine vr2vg(vr,vg,nvx,nvy,nvz,nwx,nwy,nwz,nlx,nly,nlz)
    use fft_3d
    implicit none
    integer,intent(in):: nvx,nvy,nvz,nwx,nwy,nwz,nlx,nly,nlz
    real(kind=8),intent(in):: vr(nlx,nly,nlz)
    complex(kind=8):: vg(-nvx:nvx, -nvy:nvy, -nvz:nvz)
    !
    real(kind=8),allocatable:: ztmp(:,:),vtmp(:,:,:,:)
    integer:: i1,i2,i3,k1,k2,k3,nwx2,nwy2,nwz2,ierr
    type (fft3_struct):: fs
    !
    allocate(FFT_A1D_DEFSZ(ztmp,nlx*nly*nlz),FFT_A3D_DEFSZ(vtmp,nlx,nly,nlz),stat=ierr)
    call chkalloc('vr2vg','ztmp,vtmp',ierr)
    nwx2 = nwx*2
    nwy2 = nwy*2
    nwz2 = nwz*2

    vtmp = 0.0d0
!$omp parallel do private(i1,i2,i3)
!POPTION INDEP(vtmp)
    do i3=1,nwz2
       do i2=1,nwy2
          do i1=1,nwx2
             FFT_A3D_RL(vtmp,i1,i2,i3) = vr(i1,i2,i3)
             FFT_A3D_IM(vtmp,i1,i2,i3) = 0.0d0
          end do
       end do
    end do
    call fft3_init(nwx2,nwy2,nwz2,nlx,nly,nlz,fs)
    call fft3_fw(fs,vtmp,ztmp)
    call fft3_done(fs)
    vg = (0.0d0, 0.0d0)
!$omp parallel do private(i1,i2,i3,k1,k2,k3)
    do i3=-min(nvz,nwz-1),min(nvz,nwz-1)
       do i2=-min(nvy,nwy-1),min(nvy,nwy-1)
          do i1=-min(nvx,nwx-1),min(nvx,nwx-1)
             k1 = mod(nwx2+i1,nwx2) + 1
             k2 = mod(nwy2+i2,nwy2) + 1
             k3 = mod(nwz2+i3,nwz2) + 1
             vg(i1,i2,i3) = dcmplx(FFT_A3D_RL(vtmp,k1,k2,k3), &
                                  FFT_A3D_IM(vtmp,k1,k2,k3))
          end do
       end do
    end do
    deallocate(ztmp,vtmp)
    return
  end subroutine vr2vg

  subroutine lpt2to4(nvx,nvy,nvz,nrx,nry,nrz,lpt2,rlptm3,lpt4)
    implicit none
    integer,intent(in):: nvx,nvy,nvz,nrx,nry,nrz
    complex(kind=8),intent(in):: lpt2(-nvx:nvx,-nvy:nvy,-nvz:nvz,2)
    real(kind=8),intent(in):: rlptm3(nrx*2,nry*2,nrz*2,3)
    complex(kind=8):: lpt4(-nvx:nvx,-nvy:nvy,-nvz:nvz,4)
    integer:: i1,i2,i3,nlx,nly,nlz,ic
    real(kind=8):: pol,rtmp,vtmp(3)
    real(kind=8),allocatable:: lptm(:,:,:),lptm3(:,:,:,:)

    nlx = FFT_PAD_NX(nrx*2)
    nly = FFT_PAD_NY(nry*2)
    nlz = FFT_PAD_NZ(nrz*2)
    allocate(lptm(nlx,nly,nlz),lptm3(nlx,nly,nlz,3))
    call vg2vr(lpt2(:,:,:,2),lptm,nvx,nvy,nvz,nrx,nry,nrz,nlx,nly,nlz)
!$omp parallel do private(i1,i2,i3,rtmp,vtmp,pol)
    do i3=1,nrz*2
       do i2=1,nry*2
          do i1=1,nrx*2
             rtmp = maxval(abs(rlptm3(i1,i2,i3,:)))
             if (rtmp.gt.0.0d0) then
                vtmp = rlptm3(i1,i2,i3,:)/rtmp
                pol = rtmp*sqrt(vtmp(1)**2+vtmp(2)**2+vtmp(3)**2)
             else
                pol = 0.0d0
             end if
             rtmp = lptm(i1,i2,i3)
             if (pol.ne.0.0d0) then
                lptm3(i1,i2,i3,1) = rtmp*(rlptm3(i1,i2,i3,1)/pol)
                lptm3(i1,i2,i3,2) = rtmp*(rlptm3(i1,i2,i3,2)/pol)
                lptm3(i1,i2,i3,3) = rtmp*(rlptm3(i1,i2,i3,3)/pol)
             else
                lptm3(i1,i2,i3,:) = 0.0d0
             end if
          end do
       end do
    end do
    lpt4(:,:,:,1) = lpt2(:,:,:,1)
    do ic=1,3
       call vr2vg(lptm3(:,:,:,ic),lpt4(:,:,:,ic+1),nvx,nvy,nvz, &
            nrx,nry,nrz,nlx,nly,nlz)
    end do
    deallocate(lptm,lptm3)
    return
  end subroutine lpt2to4

  subroutine lpt4to2(nvx,nvy,nvz,nrx,nry,nrz,lpt4,lpt2,rlptm3)
    implicit none
    integer,intent(in):: nvx,nvy,nvz,nrx,nry,nrz
    complex(kind=8),intent(in):: lpt4(-nvx:nvx,-nvy:nvy,-nvz:nvz,4)
    complex(kind=8):: lpt2(-nvx:nvx,-nvy:nvy,-nvz:nvz,2)
    real(kind=8),optional,intent(out):: rlptm3(nrx*2,nry*2,nrz*2,3)
    integer:: i1,i2,i3,nlx,nly,nlz,ic
    real(kind=8):: pol,rtmp,vtmp(3)
    real(kind=8),allocatable:: lptm(:,:,:),lptm3(:,:,:,:)

    nlx = FFT_PAD_NX(nrx*2)
    nly = FFT_PAD_NY(nry*2)
    nlz = FFT_PAD_NZ(nrz*2)
    allocate(lptm(nlx,nly,nlz),lptm3(nlx,nly,nlz,3))
    do ic=1,3
       call vg2vr(lpt4(:,:,:,ic+1),lptm3(:,:,:,ic),nvx,nvy,nvz, &
            nrx,nry,nrz,nlx,nly,nlz)
    end do
    if (present(rlptm3)) then
       rlptm3 = lptm3(1:nrx*2,1:nry*2,1:nrz*2,:)
    end if
!$omp parallel do private(i1,i2,i3,rtmp,vtmp,pol)
    do i3=1,nrz*2
       do i2=1,nry*2
          do i1=1,nrx*2
             rtmp = maxval(abs(lptm3(i1,i2,i3,:)))
             if (rtmp.gt.0.0d0) then
                vtmp = lptm3(i1,i2,i3,:)/rtmp
                pol = rtmp*sqrt(vtmp(1)**2+vtmp(2)**2+vtmp(3)**2)
             else
                pol = 0.0d0
             end if
             lptm(i1,i2,i3) = pol
          end do
       end do
    end do
    lpt2(:,:,:,1) = lpt4(:,:,:,1)
    call vr2vg(lptm,lpt2(:,:,:,2),nvx,nvy,nvz,nrx,nry,nrz,nlx,nly,nlz)
    deallocate(lptm,lptm3)
    return
  end subroutine lpt4to2
end module converter
