!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
! using lapack
!
#define EIGSYSTM_PROF
#define USE_MPI

module eigsystm
contains
  integer function getldm(c,n)
    implicit none
    character(len=1):: c
    integer,intent(in):: n

    integer:: p, q

    if (c .eq. 'Z') then
       p = n/128
       q = mod(n,128)
       if (q.lt.8) then
          getldm = 128*p+8+1
       else if (q.gt.120) then
          getldm = 128*(p+1)+8+1
       else
          getldm = n
       end if
    else if (c .eq. 'D') then
       p = n/256
       q = mod(n,256)
       if (q.lt.8) then
          getldm = 256*p+8+1
       else if (q.gt.248) then
          getldm = 256*(p+1)+8+1
       else
          getldm = n
       end if
    else
       write(6,*) 'getldm: unsupported type ', c
       stop
    end if
    return
  end function getldm

  subroutine zhegv1vu(n,a,b,ln,e,ntsk,comm,xinfo)
#ifdef EIGSYSTM_PROF
    use timeprof
#endif
#ifdef USE_MPI
    use mpi
#endif
    implicit none
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: xinfo
    complex(kind=8):: a(ln,n),b(ln,n)
    real(kind=8),intent(out):: e(n)

    complex(kind=8),allocatable:: wk1(:)
    real(kind=8),allocatable:: wk2(:)
    integer,allocatable:: wk3(:)
    integer:: lwork,lrwork,liwork

#ifdef EIGSYSTM_PROF
    call tpf_begin(tpf_eig)
#endif

    lwork = 2*n+n*n
    lrwork = 1+5*n+2*n*n
    liwork = 3+5*n
    allocate(wk1(lwork),wk2(lrwork),wk3(liwork))

    call zhegvd(1,'V','U',n,a,ln,b,ln,e, &
         wk1,lwork,wk2,lrwork,wk3,liwork,xinfo)

    deallocate(wk1,wk2,wk3)

#ifdef USE_MPI
    if (xinfo .eq. 0) then
       call mpi_bcast(a,ln*n,MPI_DOUBLE_COMPLEX,0,comm,xinfo)
       call mpi_bcast(e,n,MPI_DOUBLE_PRECISION,0,comm,xinfo)
    end if
#endif

#ifdef EIGSYSTM_PROF
    call tpf_end(tpf_eig)
#endif
    return
  end subroutine zhegv1vu

  subroutine zhegv1vl(n,a,b,ln,e,ntsk,comm,xinfo)
#ifdef EIGSYSTM_PROF
    use timeprof
#endif
#ifdef USE_MPI
    use mpi
#endif
    implicit none
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: xinfo
    complex(kind=8):: a(ln,n),b(ln,n)
    real(kind=8),intent(out):: e(n)

    complex(kind=8),allocatable:: wk1(:)
    real(kind=8),allocatable:: wk2(:)
    integer,allocatable:: wk3(:)
    integer:: lwork,lrwork,liwork

#ifdef EIGSYSTM_PROF
    call tpf_begin(tpf_eig)
#endif

    lwork = 2*n+n*n
    lrwork = 1+5*n+2*n*n
    liwork = 3+5*n
    allocate(wk1(lwork),wk2(lrwork),wk3(liwork))

    call zhegvd(1,'V','L',n,a,ln,b,ln,e, &
         wk1,lwork,wk2,lrwork,wk3,liwork,xinfo)

    deallocate(wk1,wk2,wk3)

#ifdef USE_MPI
    if (xinfo .eq. 0) then
       call mpi_bcast(a,ln*n,MPI_DOUBLE_COMPLEX,0,comm,xinfo)
       call mpi_bcast(e,n,MPI_DOUBLE_PRECISION,0,comm,xinfo)
    end if
#endif

#ifdef EIGSYSTM_PROF
    call tpf_end(tpf_eig)
#endif
    return
  end subroutine zhegv1vl

  subroutine zheevvl(n,a,ln,eig,info,ntsk,comm)
#ifdef EIGSYSTM_PROF
    use timeprof
#endif
#ifdef USE_MPI
    use mpi
#endif
    implicit none
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: info
    complex(kind=8):: a(ln,n)
    real(kind=8),intent(out):: eig(n)

    complex(kind=8),allocatable:: wk1(:)
    real(kind=8),allocatable:: wk2(:)
    integer,allocatable:: wk3(:)
    integer:: lwork,lrwork,liwork

#ifdef EIGSYSTM_PROF
    call tpf_begin(tpf_eig)
#endif

    lwork = 2*n+n*n
    lrwork = 1+5*n+2*n*n
    liwork = 3+5*n
    allocate(wk1(lwork),wk2(lrwork),wk3(liwork))

    call zheevd('V','L',n,a,ln,eig, &
         wk1,lwork,wk2,lrwork,wk3,liwork,info)

    deallocate(wk1,wk2,wk3)

#ifdef USE_MPI
    if (info .eq. 0) then
       call mpi_bcast(a,ln*n,MPI_DOUBLE_COMPLEX,0,comm,info)
       call mpi_bcast(eig,n,MPI_DOUBLE_PRECISION,0,comm,info)
    end if
#endif

#ifdef EIGSYSTM_PROF
    call tpf_end(tpf_eig)
#endif
    return
  end subroutine zheevvl

  subroutine dsygv1vu(n,a,b,ln,e,ntsk,comm,xinfo)
#ifdef EIGSYSTM_PROF
    use timeprof
#endif
#ifdef USE_MPI
    use mpi
#endif
    implicit none
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: xinfo
    real(kind=8):: a(ln,n),b(ln,n)
    real(kind=8),intent(out):: e(n)

    real(kind=8),allocatable:: wk1(:)
    integer,allocatable:: wk2(:)
    integer:: lwork,liwork

#ifdef EIGSYSTM_PROF
    call tpf_begin(tpf_eig)
#endif

    lwork = 1+6*n+2*n*n
    liwork = 3+5*n
    allocate(wk1(lwork),wk2(liwork))

    call dsygvd(1,'V','U',n,a,ln,b,ln,e,wk1,lwork,wk2,liwork,xinfo)

    deallocate(wk1,wk2)

#ifdef USE_MPI
    if (xinfo .eq. 0) then
       call mpi_bcast(a,ln*n,MPI_DOUBLE_PRECISION,0,comm,xinfo)
       call mpi_bcast(e,n,MPI_DOUBLE_PRECISION,0,comm,xinfo)
    end if
#endif

#ifdef EIGSYSTM_PROF
    call tpf_end(tpf_eig)
#endif
    return
  end subroutine dsygv1vu

  subroutine dsygv1vl(n,a,b,ln,e,ntsk,comm,xinfo)
#ifdef EIGSYSTM_PROF
    use timeprof
#endif
#ifdef USE_MPI
    use mpi
#endif
    implicit none
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: xinfo
    real(kind=8):: a(ln,n),b(ln,n)
    real(kind=8),intent(out):: e(n)

    real(kind=8),allocatable:: wk1(:)
    integer,allocatable:: wk2(:)
    integer:: lwork,liwork

#ifdef EIGSYSTM_PROF
    call tpf_begin(tpf_eig)
#endif

    lwork = 1+6*n+2*n*n
    liwork = 3+5*n
    allocate(wk1(lwork),wk2(liwork))

    call dsygvd(1,'V','L',n,a,ln,b,ln,e,wk1,lwork,wk2,liwork,xinfo)

    deallocate(wk1,wk2)

#ifdef USE_MPI
    if (xinfo .eq. 0) then
       call mpi_bcast(a,ln*n,MPI_DOUBLE_PRECISION,0,comm,xinfo)
       call mpi_bcast(e,n,MPI_DOUBLE_PRECISION,0,comm,xinfo)
    end if
#endif

#ifdef EIGSYSTM_PROF
    call tpf_end(tpf_eig)
#endif
    return
  end subroutine dsygv1vl

  subroutine dsyevvl(n,a,ln,eig,info,ntsk,comm)
#ifdef EIGSYSTM_PROF
    use timeprof
#endif
#ifdef USE_MPI
    use mpi
#endif
    implicit none
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: info
    real(kind=8):: a(ln,n)
    real(kind=8),intent(out):: eig(n)

    real(kind=8),allocatable:: wk1(:)
    integer,allocatable:: wk2(:)
    integer:: lwork,liwork

#ifdef EIGSYSTM_PROF
    call tpf_begin(tpf_eig)
#endif

    lwork = 1+6*n+2*n*n
    liwork = 3+5*n
    allocate(wk1(lwork),wk2(liwork))

    call dsyevd('V','L',n,a,ln,eig,wk1,lwork,wk2,liwork,info)

    deallocate(wk1,wk2)

#ifdef USE_MPI
    if (info .eq. 0) then
       call mpi_bcast(a,ln*n,MPI_DOUBLE_PRECISION,0,comm,info)
       call mpi_bcast(eig,n,MPI_DOUBLE_PRECISION,0,comm,info)
    end if
#endif

#ifdef EIGSYSTM_PROF
    call tpf_end(tpf_eig)
#endif
    return
  end subroutine dsyevvl

end module eigsystm
