!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2016 Yoshihide Yoshimoto
!
#include "fftgrid.h"

module engdffr
  private:: symfor,mkvqq,mkgmf,vblfor,vbnlf,kinet,vblstress

  integer, private:: nbk_gm_nsi = 10
contains
  !
  ! i : x,y,z
  ! r_i : atomic position in Cartesian coordinate, A_{i,j}x_j
  ! x_i : atomic position in cell coordinate, inv(A)_{i,j}r_j
  ! for: - dE_{ewald}/dx_i
  ! dfor:   - inv(A)_{i,j} dE_{ewald}/dr_j 
  !       = - inv(A)_{i,j} inv(A)_{k,j} dE_{ewald}/dx_k
  ! forel: - dE_{el}/dx_i
  ! dforel:   - inv(A)_{i,j} dE_{el}/dr_j 
  !         = - inv(A)_{i,j} inv(A)_{k,j} dE_{el}/dx_k
  ! force: - dE_{total}/dx_i
  ! dforce:   - inv(A)_{i,j} dE_{total}/dr_j 
  !         = - inv(A)_{i,j} inv(A)_{k,j} dE_{total}/dx_k
  !
  subroutine vbfor(dp,lpt,rho,vxc,crho &
       ,wgtk,eleng,forel,dforel,for,dfor,force,dforce &
       ,qbtp,elseng &
       ,aa,asi,kd,dd,ee,nband,laprj &
       ,rg,pg,ng,nnp,ncomp,nspin,ndens,nsi,nkd &
       ,nb1,nb2,mgs,bspotfc,bspots,xctp,axetp &
       ,myrank,nproc,comm)
#ifdef ESM
    use esm_manage
#endif
    use util
    use fft_3d
    use iounit
    use converter
    use bsplpot
    use types
    use mpi
    use meshgrid
    use timeprof
    use exchange
    use xcf_types
    use dftd2
    use dftd3_sg
    implicit none
    integer:: ng,nnp,nspin,nsi,nkd
    integer,intent(in):: ncomp,ndens
    integer:: nband,nb1,nb2
    integer:: laprj
    type (meshgrid_struct):: mgs
    complex(kind=8):: lpt(-mgs%nvx:mgs%nvx,-mgs%nvy:mgs%nvy, &
         -mgs%nvz:mgs%nvz,nspin*ndens)
    real(kind=8):: rho(mgs%nrxyz,nspin*ndens),vxc(mgs%nrxyz,nspin*ndens), &
         crho(mgs%nrxyz,nspin*ndens)
    type(elec_energy):: eleng
    type(elec_senergy):: elseng
    real(kind=8):: qbtp(3)
    real(kind=8):: wgtk(nb1:nb2,mgs%nks,nspin)
    real(kind=8):: ee(nband,mgs%nks,nspin)
    complex(kind=8):: dd(laprj,laprj,nsi,nspin)
    real(kind=8):: asi(3,nsi),aa(3,3),force(3,nsi),dforce(3,nsi)
    integer:: kd(nsi)
    type (wfn_for_k):: dp(mgs%nks,nspin)
    integer,intent(in):: xctp
    integer,intent(in):: axetp
    integer:: myrank,nproc,comm
    !
    ! input ewald part of the force from for and dfor
    !
    real(kind=8):: for(3,nsi),dfor(3,nsi),forel(3,nsi),dforel(3,nsi)
    integer:: rg(3,3,ng),pg(3,ng)
    !
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    !
    real(kind=8):: va,bb(3,3)
    integer:: nks,nkmesh,nkfxa
    integer:: nrx,nry,nrz,nrxyz,nvx,nvy,nvz
!SOPTION CONTAREA(nkm,kg,kf,kqg,g_count,g_displ,gq_count,gq_displ)
    type (irdkpt_struct),pointer:: irdkpt(:)
    integer,pointer:: nkm(:),kg(:,:,:)
    integer,pointer:: kf(:,:),kqg(:,:)
    integer:: bkc,ekc
    integer,pointer:: g_count(:,:),g_displ(:,:),gq_count(:),gq_displ(:)
    !
!SOPTION CONTAREA(wf)
    complex(kind=8),pointer:: wf(:,:,:)
    real(kind=8),allocatable,dimension(:,:):: rhog,vxcg,ztmp
    real(kind=8),allocatable:: crhog(:,:,:)
    complex(kind=8),allocatable:: vqq(:,:,:,:,:)
    complex(kind=8),allocatable:: gm(:,:,:,:),pgm(:,:,:,:)
    real(kind=8),allocatable:: aff(:),vl(:,:),pcc(:,:),svl(:,:),spcc(:,:)
    real(kind=8),allocatable:: akg(:),qgk(:,:)
    real(kind=8),allocatable:: bzwk(:)
    real(kind=8),allocatable:: dvpj(:,:,:,:),dsvpj(:,:,:,:)

    integer:: ierr
    type (fft3_struct):: fs
    integer:: nrx2,nry2,nrz2
    integer:: ikd,is,iks,lpnum,maxr
    integer:: bkm,ekm
    real(kind=8):: pekin(0:6),tekin(0:6)
    integer:: bt,et,pnsi
    real(kind=8):: axcforce(3,nsi),axforce(3,nsi),daxforce(3,nsi),axeng,axseng(6)

    call tpf_begin(tpf_engdf)
    !
    va = cellvol(aa)
    call recp_uvec(aa,bb)
    !
    nks = mgs%nks
    nkmesh = mgs%nkmesh
    nkfxa = mgs%nkfxa
    nrx = mgs%nrx
    nry = mgs%nry
    nrz = mgs%nrz
    nrxyz = mgs%nrxyz
    nvx = mgs%nvx
    nvy = mgs%nvy
    nvz = mgs%nvz

    irdkpt => mgs%irdkpt
    nkm => mgs%nkm
    kg => mgs%kg
    kf => mgs%kf
    kqg => mgs%kqg

    bkc = mgs%bkc
    ekc = mgs%ekc
    g_count => mgs%g_count
    g_displ => mgs%g_displ
    gq_count => mgs%gq_count
    gq_displ => mgs%gq_displ

    nrx2=nrx*2
    nry2=nry*2
    nrz2=nrz*2
    !
    forel = 0.0d0
    !
    allocate(aff(nkfxa),stat=ierr)
    call chkalloc('vbfor','aff,etc.',ierr)
    !
    ! regenerate aff to deform the cell.
    call kf2aff(bb,nkfxa,kf,aff)
    !
    eleng%epotl=0.0d0
    elseng%sepotl=0.0d0
    !
    allocate(FFT_A1D_DEFSZ(rhog,nrxyz), FFT_A1D_DEFSZ(vxcg,nrxyz), FFT_A1D_DEFSZ(ztmp,nrxyz),stat=ierr)
    call chkalloc('vbfor','rhog, vxcg, tmp',ierr)
    call fft3_init(nrx2,nry2,nrz2,nrx2,nry2,nrz2,fs)
    !
    if (nspin .eq. 1) then
       FFT_A1D_RL(rhog,:) = rho(:,1)
       FFT_A1D_IM(rhog,:) = 0.0d0
       FFT_A1D_RL(vxcg,:) = vxc(:,1)
       FFT_A1D_IM(vxcg,:) = 0.0d0
    else if (nspin .eq. 2) then
       FFT_A1D_RL(rhog,:) = rho(:,1) + rho(:,2)
       FFT_A1D_IM(rhog,:) = 0.0d0
       !
       ! pcc contains total, namely not per a spin, partial core charge.
       !
       FFT_A1D_RL(vxcg,:) = 0.5d0*(vxc(:,1) + vxc(:,2))
       FFT_A1D_IM(vxcg,:) = 0.0d0
    else
       write(6,*) 'vbfor(): invalid value in nspin = ', nspin
       stop
    end if
    call fft3_fw(fs,rhog,ztmp)
    call fft3_fw(fs,vxcg,ztmp)
    !
    allocate(vl(nkfxa,nkd),pcc(nkfxa,nkd), &
         svl(nkfxa,nkd),spcc(nkfxa,nkd), stat=ierr)
    call chkalloc('vbfor','vl,pcc,svl,spcc',ierr)
    !
    call evlpot(nkd,bspotfc,bspots,nkfxa,aff,va,vl,pcc)
    call evslpot(nkd,bspotfc,bspots,nkfxa,aff,va,svl,spcc)
    call vblfor(nkfxa,vl,pcc,svl,spcc,kf,rhog,vxcg &
         ,forel,eleng%epotl,elseng%sepotl,bb,va,asi,kd,nkd,nsi,nrx2,nry2,nrz2 &
         ,myrank,nproc,comm)
#ifdef ESM
    call esm_vblfor_xTAPP(asi,rhog,forel,eleng%epotl,myrank,comm)
#endif
    !
    ! vblfor gives forel and epotl.
    !
    deallocate(vl,pcc,svl,spcc)
    deallocate(rhog,vxcg)

    allocate(FFT_A1M_DEFSZ(crhog,nrxyz,nspin*ndens),stat=ierr)
    call chkalloc('vbfor','crhog',ierr)
    FFT_A1M_RL(crhog,:,:) = crho(:,:)
    FFT_A1M_IM(crhog,:,:) = 0.0d0
    do is=1,nspin*ndens
       call fft3_fw(fs,FFT_A1M_RL(crhog,1,is),ztmp)
    end do
    call vblstress(nkfxa,kf,lpt,crhog &
         ,elseng%sepotl,va,nvx,nvy,nvz,nspin*ndens,nrx2,nry2,nrz2)
#ifdef INCONSISTENT_THREAD_COMP
    call tpf_begin(tpf_mpi)
    call MPI_BCAST(elseng%sepotl,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call tpf_end(tpf_mpi)
#endif
    call fft3_done(fs)
    deallocate(ztmp,crhog)
    deallocate(aff)

    ! write(6,*)
    ! write(6,'(a)') ' IN  LOCAL FORCE'
    ! do in=1,nsi
    ! write(6,'(1x,i3,'' ( '',3(e14.6,1x),'')'')') &
    !    in,forel(1,in),forel(2,in),forel(3,in)
    ! end do
    !
    lpnum = getlpnum(nkd,bspots)
    maxr = getmaxr(nkd,bspots)
    !
    allocate(bzwk(nks),stat=ierr)
    call chkalloc('vbfor','bzwk',ierr)
    !
    bzwk(:) = irdkpt(:)%ngir
    if (nspin .eq. 1) then
       bzwk = 2.0d0/ncomp*bzwk
    end if

    allocate(vqq(9,laprj,laprj,nsi,nspin),stat=ierr)
    call chkalloc('vbfor','vqq',ierr)

    allocate(qgk(4,bkc:ekc),stat=ierr)
    call chkalloc('vbfor','qgk',ierr)
    call kqg2qgk(bb,ekc-bkc+1,kqg,qgk)

    eleng%epotnl=0.0d0
    elseng%sepotnl=0.0d0
    eleng%ekin=0.0d0
    elseng%sekin=0.0d0
    ! MAKE \INT DR V_LOC*\FRAC{D Q_ij}{D R_\TAU} : VQQ
    call mkvqq(bkc,ekc,kqg,qgk,lpt,vqq,kd,va,asi, &
         laprj,nsi,nkd,nvx,nvy,nvz,nspin,ndens,bspotfc,bspots,comm)

    allocate(gm(10,laprj,nb2,nbk_gm_nsi), &
         pgm(10,laprj,nb2,nbk_gm_nsi),stat=ierr)
    call chkalloc('vbfor','gm,pgm',ierr)

    do iks=1,nks
       bkm = g_displ(myrank,iks)+1
       ekm = g_displ(myrank,iks)+g_count(myrank,iks)
       allocate(dvpj(bkm:ekm,maxr,lpnum,nkd), &
            dsvpj(bkm:ekm,maxr,lpnum,nkd),stat=ierr)
       call chkalloc('vbfor','vpj,svpj',ierr)
       allocate(akg(bkm:ekm),stat=ierr)
       call chkalloc('vbfor','akg',ierr)
       call kg2akg(bb,irdkpt(iks)%sk,ekm-bkm+1,kg(:,bkm:ekm,iks),akg)
       do ikd=1,nkd
          call evbetg(bspotfc,bspots(ikd),ekm-bkm+1,akg, &
               ekm-bkm+1,maxr,lpnum,dvpj(bkm,1,1,ikd))
          call evsbetg(bspotfc,bspots(ikd),ekm-bkm+1,akg, &
               ekm-bkm+1,maxr,lpnum,dsvpj(bkm,1,1,ikd))
       end do
       deallocate(akg)

       do is=1,nspin
          wf => dp(iks,is)%wf

          pekin = 0.0d0
          call kinet(irdkpt(iks)%sk,bb,kg(:,bkm:ekm,iks),wf,wgtk(nb1,iks,is) &
               ,bzwk(iks),nkmesh,pekin(0),pekin(1:6),ncomp,ekm-bkm+1,nb1,nb2,qbtp)
          call tpf_begin(tpf_mpi)
          call MPI_ALLREDUCE(pekin,tekin,7,MPI_DOUBLE_PRECISION, &
               MPI_SUM,comm,ierr)
          call tpf_end(tpf_mpi)
          eleng%ekin = eleng%ekin + tekin(0)
          elseng%sekin = elseng%sekin + tekin(1:6)

          !
          ! blocked calculation to control the size of gm
          !
          pgm = (0.0d0, 0.0d0)
          do bt = 1, nsi, nbk_gm_nsi
             et = min(nsi, bt + nbk_gm_nsi - 1)
             pnsi = et - bt + 1
             call mkgmf(ncomp,irdkpt(iks)%sk,bb,pgm,kg(:,bkm:ekm,iks),wf, &
                  dvpj,dsvpj, &
                  bkm,ekm,pnsi,asi(1,bt),kd(bt),nkd,lpnum,maxr,laprj,nb2,bspots)

             call tpf_begin(tpf_mpi)
             call MPI_ALLREDUCE(pgm,gm,10*laprj*nb2*pnsi, &
                  MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
             call tpf_end(tpf_mpi)

             call vbnlf(eleng%epotnl,elseng%sepotnl,forel(1,bt) &
                  ,bzwk(iks),wgtk(nb1,iks,is),va,dd(1,1,bt,is),gm &
                  ,vqq(1,1,1,bt,is),ee(1,iks,is),kd(bt) &
                  ,nkd,pnsi,laprj,nkmesh,nb1,nb2,bspots)
          end do
       end do
       deallocate(dvpj,dsvpj)
    end do
    deallocate(gm,pgm)
    deallocate(qgk,vqq)

    if (xcf_exch_factor(xctp) .ne. 0.0d0) then
       call potexch_init(0,mgs%nwx,mgs%nwy,mgs%nwz,-1)
       
       call engexch(va,bb,mgs,nspin,nks, &
            nb2,nb1,nb2,wgtk,bzwk, &
            xcf_exch_factor(xctp), &
            dp,eleng%exx,elseng%sexx,nproc,comm,-1)

       call potexch_done()
    else
       eleng%exx = 0.0d0
       elseng%sexx = 0.0d0
    end if

    deallocate(bzwk)

#ifdef INCONSISTENT_THREAD_COMP
    call tpf_begin(tpf_mpi)
    call MPI_BCAST(eleng%epotnl,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(elseng%sepotnl,6,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(forel,3*nsi,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call tpf_end(tpf_mpi)
#endif
    ! write(6,*)
    ! write(6,'(a)') ' IN  ELECT FORCE'
    ! do in=1,nsi
    !   write(6,'(1x,i3,'' ( '',3(e14.6,1x),'')'')') &
    !       in,forel(1,in),forel(2,in),forel(3,in)
    ! end do

    call symfor(forel,rg,pg,ng,nnp,kd,asi,nsi)
    call symfor(for  ,rg,pg,ng,nnp,kd,asi,nsi)
    !
    call formod(forel,dforel,bb,nsi)
    ! NENNOTAME
    call formod(for  ,dfor  ,bb,nsi)

    if (axetp.eq.1) then
       !
       ! DFT-D2
       !
       call fdftd2(xctp,nkd,bspots,aa,nsi,kd,asi,axeng,axseng,axcforce)
       call acart2cell(aa,nsi,axcforce,axforce)
       call symfor(axforce,rg,pg,ng,nnp,kd,asi,nsi)
       call formod(axforce,daxforce,bb,nsi)
    else if (axetp.eq.2) then
       !
       ! call DFT-D3 code by S. Grimme
       !
       ! call with zero dumping + DFT-D2
       !
       call fdftd3sg(2,xctp,nkd,bspots,aa,nsi,kd,asi,axeng,axseng,axcforce, &
            myrank,comm)
       call acart2cell(aa,nsi,axcforce,axforce)
       call symfor(axforce,rg,pg,ng,nnp,kd,asi,nsi)
       call formod(axforce,daxforce,bb,nsi)
    else if (axetp.eq.3) then
       !
       ! call DFT-D3 code by S. Grimme
       !
       ! call with B-J dumping + DFT-D3
       !
       call fdftd3sg(4,xctp,nkd,bspots,aa,nsi,kd,asi,axeng,axseng,axcforce, &
            myrank,comm)
       call acart2cell(aa,nsi,axcforce,axforce)
       call symfor(axforce,rg,pg,ng,nnp,kd,asi,nsi)
       call formod(axforce,daxforce,bb,nsi)
    else
       axforce = 0.0d0
       daxforce = 0.0d0
       axeng = 0.0d0
       axseng = 0.0d0
    end if
    !
    force = for + forel + axforce
    dforce = dfor + dforel + daxforce
    !
    eleng%epot = eleng%epotl + eleng%epotnl
    eleng%total = eleng%eht + eleng%exc + eleng%exx + eleng%epot + eleng%ekin
    eleng%total = eleng%total + axeng
    !
    elseng%total(1:3) = elseng%sekin(1:3) &
         + eleng%epotl + eleng%epotnl + eleng%eht - eleng%exc &
         + eleng%evxc(0) + eleng%evxc(1:3) &
         + elseng%sepotl(1:3) + elseng%sepotnl(1:3) + elseng%sexx(1:3)
    elseng%total(4:6) = elseng%sekin(4:6) &
         + eleng%evxc(4:6) &
         + elseng%sepotl(4:6) + elseng%sepotnl(4:6) + elseng%sexx(4:6)
    elseng%total(1:6) = elseng%total(1:6) + axseng

    call symstress(ng,rg,nnp,pg,aa,elseng%total)

#ifdef INCONSISTENT_THREAD_COMP
!
! These are required for Intel OpenMP.
    call tpf_begin(tpf_mpi)
    call MPI_BCAST(eleng%total,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(elseng%total,6,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(force,3*nsi,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(dforce,3*nsi,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call tpf_end(tpf_mpi)
#endif
    call tpf_end(tpf_engdf)
    return
  end subroutine vbfor

  subroutine vblfor(nkfxa,vl,pcc,svl,spcc,kf,rho,vxc &
       ,forel,epotl,sepotl,bb,va,asi,kd,nkd,nsi,nrx2,nry2,nrz2 &
       ,myrank,nproc,comm)
    use mpi
    use xmpi
    implicit none
    integer:: nkd, nsi, nkfxa, nrx2, nry2, nrz2
    real(kind=8):: vl(nkfxa,nkd),pcc(nkfxa,nkd)
    real(kind=8):: svl(nkfxa,nkd),spcc(nkfxa,nkd)
    integer:: kf(3,nkfxa)
    real(kind=8):: epotl, sepotl(6), bb(3,3), va
    real(kind=8):: FFT_A3D_DEFSZ(rho,nrx2,nry2,nrz2),FFT_A3D_DEFSZ(vxc,nrx2,nry2,nrz2)
    integer:: kd(nsi)
    real(kind=8):: asi(3,nsi),forel(3,nsi)
    integer:: myrank,nproc,comm
    !
    real(kind=8),parameter:: p2 = 3.14159265358979323846d0*2
    !     
    integer:: ikd, in, k, k1, k2, k3, l1, l2, l3
    real(kind=8):: t1, t2, vwk, s, ss
    !
    integer,allocatable:: gf_count(:),gf_displ(:)
    integer:: irnk,bkfxa,ekfxa,ierr
    real(kind=8),allocatable:: pforel(:,:),tforel(:,:)
    real(kind=8):: psepotl(0:6),tsepotl(0:6)

    real(kind=8):: x,y,z,rsq
    integer:: ist

    allocate(gf_count(0:nproc-1),gf_displ(0:nproc-1), &
         pforel(3,nsi),tforel(3,nsi),stat=ierr)
    call chkalloc('vblfor','gf_count,pforel,tforel',ierr)
!poption noparallel
    gf_count = nkfxa/nproc
!poption noparallel
    do irnk=0,mod(nkfxa,nproc)-1
       gf_count(irnk) = gf_count(irnk)+1
    end do
    gf_displ(0) = 0
    do irnk=0,nproc-2
       gf_displ(irnk+1) = gf_displ(irnk)+gf_count(irnk)
    end do
    bkfxa = gf_displ(myrank)+1
    ekfxa = gf_displ(myrank)+gf_count(myrank)

    if (bkfxa .eq. 1) then
       if ( (kf(1,1).ne.0).or.(kf(2,1).ne.0).or.(kf(3,1).ne.0) ) then
          write(6,*) 'invalid kf in vblfor: myrank = ',myrank
          write(6,*) 'kf = ',kf(1,1),kf(2,1),kf(3,1)
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end if
       ist = 1
    else
       ist = 0
    end if

    pforel = 0.0d0
    psepotl = 0.0d0
!$omp parallel do private(ikd,k,k1,k2,k3,l1,l2,l3,t1,t2,vwk,s,ss,x,y,z,rsq), &
!$omp reduction(+:psepotl)
!poption psum(psepotl)
    do in=1,nsi
       ikd=kd(in)
       if (ist.eq.1) then
          t1=FFT_A3D_RL(rho,1,1,1)
          vwk=vl(1,ikd)*va
          psepotl(0)=psepotl(0)+vwk*t1
       end if
       do k=bkfxa+ist,ekfxa
          k1=kf(1,k)
          k2=kf(2,k)
          k3=kf(3,k)
          l1=mod(nrx2+k1,nrx2)+1
          l2=mod(nry2+k2,nry2)+1
          l3=mod(nrz2+k3,nrz2)+1
          t1=FFT_A3D_RL(rho,l1,l2,l3)
          t2=FFT_A3D_IM(rho,l1,l2,l3)

          vwk=vl(k,ikd)*va
          s=(k1*asi(1,in)+k2*asi(2,in)+k3*asi(3,in))*p2
          ss=vwk*(t1*sin(s)+t2*cos(s))*p2
          psepotl(0)=psepotl(0)+vwk*(t1*cos(s)-t2*sin(s))
          pforel(1,in)=pforel(1,in)+k1*ss
          pforel(2,in)=pforel(2,in)+k2*ss
          pforel(3,in)=pforel(3,in)+k3*ss

          x = bb(1,1)*k1+bb(1,2)*k2+bb(1,3)*k3
          y = bb(2,1)*k1+bb(2,2)*k2+bb(2,3)*k3
          z = bb(3,1)*k1+bb(3,2)*k2+bb(3,3)*k3
          rsq = x*x+y*y+z*z
          vwk=svl(k,ikd)*va*(t1*cos(s)-t2*sin(s))
          psepotl(1)=psepotl(1)+vwk*x*x/rsq
          psepotl(2)=psepotl(2)+vwk*y*y/rsq
          psepotl(3)=psepotl(3)+vwk*z*z/rsq
          psepotl(4)=psepotl(4)+vwk*2.0d0*y*z/rsq
          psepotl(5)=psepotl(5)+vwk*2.0d0*z*x/rsq
          psepotl(6)=psepotl(6)+vwk*2.0d0*x*y/rsq
       end do
    end do
    ! force due to core charge
    !
!$omp parallel do private(ikd,k,k1,k2,k3,l1,l2,l3,t1,t2,vwk,s,ss,x,y,z,rsq), &
!$omp reduction(+:psepotl)
!poption psum(psepotl)
    do in=1,nsi
       ikd=kd(in)
       do k=bkfxa+ist,ekfxa
          k1=kf(1,k)
          k2=kf(2,k)
          k3=kf(3,k)
          l1=mod(nrx2+k1,nrx2)+1
          l2=mod(nry2+k2,nry2)+1
          l3=mod(nrz2+k3,nrz2)+1
          t1=FFT_A3D_RL(vxc,l1,l2,l3)
          t2=FFT_A3D_IM(vxc,l1,l2,l3)
          vwk=pcc(k,ikd)*va
          s=(k1*asi(1,in)+k2*asi(2,in)+k3*asi(3,in))*p2
          ss=vwk*(t1*sin(s)+t2*cos(s))*p2
          pforel(1,in)=pforel(1,in)+k1*ss
          pforel(2,in)=pforel(2,in)+k2*ss
          pforel(3,in)=pforel(3,in)+k3*ss

          x = bb(1,1)*k1+bb(1,2)*k2+bb(1,3)*k3
          y = bb(2,1)*k1+bb(2,2)*k2+bb(2,3)*k3
          z = bb(3,1)*k1+bb(3,2)*k2+bb(3,3)*k3
          rsq = x*x+y*y+z*z
          vwk=spcc(k,ikd)*va*(t1*cos(s)-t2*sin(s))
          psepotl(1)=psepotl(1)+vwk*x*x/rsq
          psepotl(2)=psepotl(2)+vwk*y*y/rsq
          psepotl(3)=psepotl(3)+vwk*z*z/rsq
          psepotl(4)=psepotl(4)+vwk*2.0d0*y*z/rsq
          psepotl(5)=psepotl(5)+vwk*2.0d0*z*x/rsq
          psepotl(6)=psepotl(6)+vwk*2.0d0*x*y/rsq
       end do
    end do

    ! stress due to Hartree energy
!$omp parallel do private(k1,k2,k3,l1,l2,l3,t1,t2,vwk,x,y,z,rsq), &
!$omp reduction(+:psepotl)
!poption psum(psepotl)
    do k=bkfxa+ist,ekfxa
       k1=kf(1,k)
       k2=kf(2,k)
       k3=kf(3,k)
       l1=mod(nrx2+k1,nrx2)+1
       l2=mod(nry2+k2,nry2)+1
       l3=mod(nrz2+k3,nrz2)+1
       t1=FFT_A3D_RL(rho,l1,l2,l3)
       t2=FFT_A3D_IM(rho,l1,l2,l3)
       x = bb(1,1)*k1+bb(1,2)*k2+bb(1,3)*k3
       y = bb(2,1)*k1+bb(2,2)*k2+bb(2,3)*k3
       z = bb(3,1)*k1+bb(3,2)*k2+bb(3,3)*k3
       rsq = x*x+y*y+z*z
       vwk = -2.0d0*p2*(t1*t1+t2*t2)/(rsq*rsq)*va
       psepotl(1)=psepotl(1)+vwk*x*x
       psepotl(2)=psepotl(2)+vwk*y*y
       psepotl(3)=psepotl(3)+vwk*z*z
       psepotl(4)=psepotl(4)+vwk*2.0d0*y*z
       psepotl(5)=psepotl(5)+vwk*2.0d0*z*x
       psepotl(6)=psepotl(6)+vwk*2.0d0*x*y
    end do
    !
    call xmpi_alldsum(pforel,tforel,3*nsi,comm,ierr)
    call xmpi_alldsum(psepotl,tsepotl,7,comm,ierr)
    forel = forel + tforel
    epotl = epotl + tsepotl(0)
    sepotl = sepotl + tsepotl(1:6)
    deallocate(gf_count,gf_displ,pforel,tforel)
    !
    return
  end subroutine vblfor

  subroutine vblstress(nkfxa,kf,lpt,crho &
       ,sepotl,va,nvx,nvy,nvz,nspin,nrx2,nry2,nrz2)
    implicit none
    integer:: nkfxa, nrx2, nry2, nrz2, nvx, nvy, nvz, nspin
    integer:: kf(3,nkfxa)
    real(kind=8):: sepotl(6), va
    complex(kind=8):: lpt(-nvx:nvx,-nvy:nvy,-nvz:nvz,nspin)
    real(kind=8):: FFT_A3M_DEFSZ(crho,nrx2,nry2,nrz2,nspin)
    !     
    integer:: k, k1, k2, k3, l1, l2, l3, is
    real(kind=8):: sum

    sum = 0.0d0
!poption loop(2)
    do is=1,nspin
!$omp parallel do private(k1,k2,k3,l1,l2,l3),reduction(+:sum)
       do k=1,nkfxa
          k1=kf(1,k)
          k2=kf(2,k)
          k3=kf(3,k)
          l1=mod(nrx2-k1,nrx2)+1
          l2=mod(nry2-k2,nry2)+1
          l3=mod(nrz2-k3,nrz2)+1
          sum=sum+dble(lpt(k1,k2,k3,is))*FFT_A3M_RL(crho,l1,l2,l3,is) &
               -aimag(lpt(k1,k2,k3,is))*FFT_A3M_IM(crho,l1,l2,l3,is)
       end do
    end do
    sepotl(1:3) = sepotl(1:3) + sum*va
    !
    return
  end subroutine vblstress

  subroutine symfor(forel,rg,pg,ng,nnp,kd,asi,nsi)
    implicit none
    integer nsi,kd(nsi)
    real(kind=8) forel(3,nsi)
    real(kind=8) asi(3,nsi)
    integer ng,nnp,rg(3,3,ng),pg(3,ng)

    real(kind=8) x(3),y(3)
    real(kind=8),allocatable:: dforel(:,:)
    integer:: it,i,ig,ikd,itmin,iflg,it1,itt
    real(kind=8):: t1,t2,t3,s1,s2,s3,f1,f2,f3,difmin,dif,cnst
    integer:: ierr
    !
    if (ng.eq.1) return
    !
    allocate(dforel(3,nsi),stat=ierr)
    call chkalloc('symfor','dforel',ierr)
    !
    cnst=1.0d0/dble(ng)
    do it=1,nsi
       do i=1,3
          dforel(i,it)=0.0d0
       end do
    end do
    !
    do it=1,nsi
       ikd=kd(it)
       s1=asi(1,it)
       s2=asi(2,it)
       s3=asi(3,it)
       do ig=1,ng
          ! CORRESPONDANCE CHECK    IT1 AND R^T{IT}
          t1=s1-pg(1,ig)/dble(nnp)
          t2=s2-pg(2,ig)/dble(nnp)
          t3=s3-pg(3,ig)/dble(nnp)
          x(1)=t1*rg(1,1,ig)+t2*rg(2,1,ig)+t3*rg(3,1,ig)
          x(2)=t1*rg(1,2,ig)+t2*rg(2,2,ig)+t3*rg(3,2,ig)
          x(3)=t1*rg(1,3,ig)+t2*rg(2,3,ig)+t3*rg(3,3,ig)
          !
          iflg=0
          difmin=1.0d0
          do it1=1,nsi
             if (kd(it1).eq.ikd) then
                y(1)=mod(abs(x(1)-asi(1,it1)),1.0d0)
                y(2)=mod(abs(x(2)-asi(2,it1)),1.0d0)
                y(3)=mod(abs(x(3)-asi(3,it1)),1.0d0)
                y(1)=min(y(1),1.0d0-y(1))
                y(2)=min(y(2),1.0d0-y(2))
                y(3)=min(y(3),1.0d0-y(3))
                dif=y(1)*y(1)+y(2)*y(2)+y(3)*y(3)
                !
                if (dif.lt.difmin) then
                   difmin=dif
                   itmin=it1
                end if
                if (dif.lt.1.0d-8 ) then
                   iflg=iflg+1
                   f1=forel(1,it1)
                   f2=forel(2,it1)
                   f3=forel(3,it1)
                   dforel(1,it)=dforel(1,it)+ &
                        cnst*(rg(1,1,ig)*f1+rg(1,2,ig)*f2+rg(1,3,ig)*f3)
                   dforel(2,it)=dforel(2,it)+ &
                        cnst*(rg(2,1,ig)*f1+rg(2,2,ig)*f2+rg(2,3,ig)*f3)
                   dforel(3,it)=dforel(3,it)+ &
                        cnst*(rg(3,1,ig)*f1+rg(3,2,ig)*f2+rg(3,3,ig)*f3)
                end if
             end if
          end do
          if (iflg.ne.1) then
             write(6,*)'ITMIN=',itmin,' DIFMIN=',difmin
             write(6,*)' SYMMETRY ERROR IFLG,IT,IG=',iflg,it,ig
             write(6,7500)(x(i),i=1,3)
7500         format(' RG*ASI=',3e20.12)
             do itt=1,nsi
                write(6,7700)itt,(asi(i,itt),i=1,3)
             end do
7700         format(' ASI(',i5,')=',3e20.12)
             stop ' SYMFOR '
          end if
       end do
    end do
    !
    do it=1,nsi
       do i=1,3
          forel(i,it)=dforel(i,it)
       end do
    end do
    deallocate(dforel)
    return
  end subroutine symfor

  subroutine symstress(ng,rg,nnp,pg,aa,sg)
    implicit none
    integer,intent(in):: ng,nnp
    integer,intent(in):: rg(3,3,ng),pg(3,ng)
    real(kind=8),intent(in):: aa(3,3)
    real(kind=8):: sg(6)

    real(kind=8),parameter:: pi = 3.1415926535897932385d0
    integer:: ig,i,j,m,k
    real(kind=8):: w,aai(3,3),ts(3,3),rs(3,3),tt(3,3)

    w=aa(1,1)*(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)) &
         +aa(2,1)*(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)) &
         +aa(3,1)*(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))

    aai(1,1)=(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3))/w
    aai(1,2)=(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3))/w
    aai(1,3)=(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))/w
    aai(2,1)=(aa(2,3)*aa(3,1)-aa(3,3)*aa(2,1))/w
    aai(2,2)=(aa(3,3)*aa(1,1)-aa(1,3)*aa(3,1))/w
    aai(2,3)=(aa(1,3)*aa(2,1)-aa(2,3)*aa(1,1))/w
    aai(3,1)=(aa(2,1)*aa(3,2)-aa(3,1)*aa(2,2))/w
    aai(3,2)=(aa(3,1)*aa(1,2)-aa(1,1)*aa(3,2))/w
    aai(3,3)=(aa(1,1)*aa(2,2)-aa(2,1)*aa(1,2))/w

    ts(1,1) = sg(1)
    ts(2,2) = sg(2)
    ts(3,3) = sg(3)
    ts(2,3) = sg(4)
    ts(3,2) = sg(4)
    ts(3,1) = sg(5)
    ts(1,3) = sg(5)
    ts(1,2) = sg(6)
    ts(2,1) = sg(6)

    rs = 0.0d0
    do ig=1,ng
       tt = 0.0d0
       do j=1,3
          do i=1,3
             do k=1,3
                do m=1,3
                   tt(i,j) = tt(i,j)+aai(m,i)*rg(m,k,ig)*aa(j,k)
                end do
             end do
          end do
       end do

       do j=1,3
          do i=1,3
             do k=1,3
                do m=1,3
                   rs(i,j) = rs(i,j)+tt(m,i)*ts(m,k)*tt(k,j)
                end do
             end do
          end do
       end do
    end do
    rs = rs/ng

    sg(1) = rs(1,1)
    sg(2) = rs(2,2)
    sg(3) = rs(3,3)
    sg(4) = rs(2,3)
    sg(5) = rs(3,1)
    sg(6) = rs(1,2)

    return
  end subroutine symstress

  subroutine kinet(sk,bb,kg,p,wgtk,bzwk,nkmesh,ekin,sekin, &
       ncomp,nkm,nb1,nb2,qbtp)
    implicit none
    integer:: nkm,nb1,nb2,kg(3,nkm),nkmesh
    integer:: ncomp
    real(kind=8):: sk(3),bb(3,3)
    complex(kind=8):: p(ncomp,nkm,nb2)
    real(kind=8):: wgtk(nb1:nb2),bzwk
    real(kind=8):: ekin,sekin(6),qbtp(3)
    !
    real(kind=8),parameter:: pi = 3.14159265358979323846d0
    !
    real(kind=8),allocatable:: aw(:),saw(:,:)
    integer:: err,ig,ib
    real(kind=8):: u,v,w,x,y,z,ee,see(6),kgsq,ew,sew(6)
    real(kind=8):: a,qsq,sgm
    real(kind=8):: fac
    !
    allocate(aw(nkm),saw(nkm,6),stat=err)
    call chkalloc('kinet','aw,ew',err)
    !
    a=qbtp(1)
    qsq=qbtp(2)*qbtp(2)
    sgm=qbtp(3)

!$omp parallel do private(u,v,w,x,y,z,kgsq,fac)
    do ig=1,nkm
       u=sk(1)+kg(1,ig)
       v=sk(2)+kg(2,ig)
       w=sk(3)+kg(3,ig)
       x=u*bb(1,1)+v*bb(1,2)+w*bb(1,3)
       y=u*bb(2,1)+v*bb(2,2)+w*bb(2,3)
       z=u*bb(3,1)+v*bb(3,2)+w*bb(3,3)
       kgsq=x*x+y*y+z*z
       aw(ig)=0.5d0*( kgsq + a*( 1.0d0 + erf(0.5d0*(kgsq-qsq)/sgm) ) )
       fac=( 1.0d0 + a/(sgm*sqrt(pi)) &
            *exp(-(0.5d0*(kgsq-qsq)/sgm)**2) )
       saw(ig,1)=x*x*fac
       saw(ig,2)=y*y*fac
       saw(ig,3)=z*z*fac
       saw(ig,4)=2.0d0*y*z*fac
       saw(ig,5)=2.0d0*z*x*fac
       saw(ig,6)=2.0d0*x*y*fac
    end do
    !
    ee=0.0d0
    see=0.0d0
    if (ncomp.eq.1) then
!$omp parallel do private(ig),reduction(+:ee,see)
!poption psum(see)
       do ib=1,nb1-1
          do ig=1,nkm
             fac = dble(p(1,ig,ib))**2 + aimag(p(1,ig,ib))**2
             ee=ee+aw(ig)*fac
             see=see+saw(ig,1:6)*fac
          end do
       end do
    else
!$omp parallel do private(ig),reduction(+:ee,see)
!poption psum(see)
       do ib=1,nb1-1
          do ig=1,nkm
             fac = dble(p(1,ig,ib))**2 + aimag(p(1,ig,ib))**2 &
                  + dble(p(2,ig,ib))**2 + aimag(p(2,ig,ib))**2
             ee=ee+aw(ig)*fac
             see=see+saw(ig,1:6)*fac
          end do
       end do
    end if
    ee=ee*bzwk
    see=see*bzwk
    !
    if (ncomp.eq.1) then
!$omp parallel do private(ig,sew,ew),reduction(+:ee,see)
!poption psum(see)
       do ib=nb1,nb2
          ew=0.0d0
          sew=0.0d0
          do ig=1,nkm
             fac = dble(p(1,ig,ib))**2 + aimag(p(1,ig,ib))**2
             ew=ew+aw(ig)*fac
             sew=sew+saw(ig,1:6)*fac
          end do
          ee=ee+ew*wgtk(ib)
          see=see+sew*wgtk(ib)
       end do
    else
!$omp parallel do private(ig,sew,ew),reduction(+:ee,see)
!poption psum(see)
       do ib=nb1,nb2
          ew=0.0d0
          sew=0.0d0
          do ig=1,nkm
             fac = dble(p(1,ig,ib))**2 + aimag(p(1,ig,ib))**2 &
                  + dble(p(2,ig,ib))**2 + aimag(p(2,ig,ib))**2
             ew=ew+aw(ig)*fac
             sew=sew+saw(ig,1:6)*fac
          end do
          ee=ee+ew*wgtk(ib)
          see=see+sew*wgtk(ib)
       end do
    end if
    !
    ekin=ekin+ee/dble(nkmesh)
    sekin=sekin+see/dble(nkmesh)
    deallocate(aw,saw)
    return
  end subroutine kinet

  subroutine mkvqq(bkc,ekc,kqg,qgk,vpot,vqq,kd,va,asi, &
       laprj,nsi,nkd,nvx,nvy,nvz,nspin,ndens,bspotfc,bspots,comm)
    use converter
    use bsplpot
    use mpi
    use timeprof
    implicit none
    !
    integer:: bkc,ekc,kqg(3,bkc:ekc)
    real(kind=8):: qgk(4,bkc:ekc)
    integer:: laprj,nsi,nkd,nvx,nvy,nvz,nspin,ndens
    complex(kind=8),intent(in):: vpot(-nvx:nvx,-nvy:nvy,-nvz:nvz,nspin*ndens)
    complex(kind=8):: vqq(9,laprj,laprj,nsi,nspin)
    integer,intent(in):: kd(nsi)
    real(kind=8),intent(in):: asi(3,nsi)
    real(kind=8),intent(in):: va
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    integer:: comm
    !
    complex(kind=8),allocatable:: qg(:,:),sqg(:,:,:)
    !
    integer:: k1, nob, naob, numk1
    integer:: itt, it, l1, i1, m1, l2, i2, m2, ip1, ip2, ierr
    integer:: nsikd, itk2it(nsi)
    complex(kind=8),allocatable:: pcvqq(:,:,:,:),cvqq(:,:,:,:)
    integer:: idens
    !
    call tpf_begin(tpf_mkvqq)

    if ((nspin.ne.1).and.(nspin.ne.2)) then
       write(6,'(1x,a,i3)') 'invalid spin = ',nspin
       stop
    end if
    !
    naob = 0
    do itt=1,nkd
       naob = max(naob,bspots(itt)%nob)
    end do
    allocate(pcvqq(9,naob*(naob+1)/2,nsi,nspin), &
         cvqq(9,naob*(naob+1)/2,nsi,nspin),stat=ierr)
    call chkalloc('mkvqq','pcvqq,cvqq',ierr)
    !
    pcvqq = 0.0d0
    !
    do idens=1,ndens
       do itt=1,nkd
          nob = bspots(itt)%nob
          allocate(qg(bkc:ekc,nob*(nob+1)/2), &
               sqg(bkc:ekc,6,nob*(nob+1)/2),stat=ierr)
          call chkalloc('mkvqq','qg,sqg',ierr)
          call evqg(bspotfc,bspots(itt),ekc-bkc+1,qgk,va,nob,idens,qg)
          call evsqg(bspotfc,bspots(itt),ekc-bkc+1,qgk,va,nob,idens,sqg)
          nsikd = 0
          do it=1, nsi
             if (kd(it).eq.itt) then
                nsikd = nsikd + 1
                itk2it(nsikd) = it
             end if

          end do

          numk1=0
          do ip1 = 1, bspots(itt)%nprj
             l1 = bspots(itt)%ip2l(ip1)
             i1 = bspots(itt)%ip2r(ip1)
             m1 = bspots(itt)%ip2m(ip1)
             do ip2 = 1, bspots(itt)%l2pt(l1)
                l2 = bspots(itt)%ip2l(ip2)
                i2 = bspots(itt)%ip2r(ip2)
                m2 = bspots(itt)%ip2m(ip2)
                if (.not.( (l1.eq.l2).and. &
                     ((i2.gt.i1) .or. ((i1.eq.i2).and.(m2.gt.m1))))) then
                   numk1=numk1+1
                end if
             end do
          end do

          if (nspin.eq.1) then
             call mkvqqcore(bkc,ekc,kqg,vpot(:,:,:,idens), &
                  naob,pcvqq(:,:,:,1),va,asi, &
                  nsi,nvx,nvy,nvz,bspots(itt)%nob,numk1,nsikd,itk2it, &
                  qg,sqg)
          else
             call mkvqqcore(bkc,ekc,kqg,vpot(:,:,:,1), &
                  naob,pcvqq(:,:,:,1),va,asi, &
                  nsi,nvx,nvy,nvz,bspots(itt)%nob,numk1,nsikd,itk2it, &
                  qg,sqg)
             call mkvqqcore(bkc,ekc,kqg,vpot(:,:,:,2), &
                  naob,pcvqq(:,:,:,2),va,asi, &
                  nsi,nvx,nvy,nvz,bspots(itt)%nob,numk1,nsikd,itk2it, &
                  qg,sqg)
          end if
          deallocate(qg,sqg)
       end do
    end do
    call tpf_begin(tpf_mpi)
    call MPI_ALLREDUCE(pcvqq,cvqq,9*naob*(naob+1)/2*nsi*nspin, &
         MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
    call tpf_end(tpf_mpi)

    do it=1,nsi
       itt = kd(it)
       k1=0
       do ip1 = 1,bspots(itt)%nprj
          l1 = bspots(itt)%ip2l(ip1)
          i1 = bspots(itt)%ip2r(ip1)
          m1 = bspots(itt)%ip2m(ip1)
          do ip2 = 1,bspots(itt)%l2pt(l1)
             l2 = bspots(itt)%ip2l(ip2)
             i2 = bspots(itt)%ip2r(ip2)
             m2 = bspots(itt)%ip2m(ip2)
             if (.not.( (l1.eq.l2) .and. &
                  ((i2.gt.i1) .or. ((i1.eq.i2).and.(m2.gt.m1))))) then
                k1=k1+1
                if (ip1.eq.ip2) then
                   vqq(:,ip1,ip1,it,1) = dble(cvqq(:,k1,it,1))
                   if (nspin.eq.2) then
                      vqq(:,ip1,ip1,it,2) = dble(cvqq(:,k1,it,2))
                   end if
                else
                   vqq(:,ip1,ip2,it,1) = cvqq(:,k1,it,1)
                   vqq(:,ip2,ip1,it,1) = conjg(cvqq(:,k1,it,1))
                   if (nspin.eq.2) then
                      vqq(:,ip1,ip2,it,2) = cvqq(:,k1,it,2)
                      vqq(:,ip2,ip1,it,2) = conjg(cvqq(:,k1,it,2))
                   end if
                end if
             end if
          end do
       end do
    end do

    deallocate(pcvqq,cvqq)
    
    call tpf_end(tpf_mkvqq)
    return
  end subroutine mkvqq

  subroutine mkvqqcore(bkc,ekc,kqg,vpot,naob,cvqq,va,asi, &
       nsi,nvx,nvy,nvz,nob,numk1,nsikd,itk2it,qg,sqg)
    use bsplpot
#ifdef _OPENMP
    use omp_lib
#endif
    implicit none
    !
    integer,intent(in):: bkc,ekc,kqg(3,bkc:ekc)
    integer,intent(in):: nsi,nvx,nvy,nvz,naob,nob,numk1
    complex(kind=8),intent(in):: vpot(-nvx:nvx,-nvy:nvy,-nvz:nvz)
    complex(kind=8):: cvqq(9,naob*(naob+1)/2,nsi)
    real(kind=8),intent(in):: asi(3,nsi)
    real(kind=8),intent(in):: va
    integer,intent(in):: nsikd, itk2it(nsi)
    complex(kind=8),intent(in):: qg(bkc:ekc,nob*(nob+1)/2)
    complex(kind=8),intent(in):: sqg(bkc:ekc,6,nob*(nob+1)/2)
    !
    real(kind=8),parameter:: pi2 = 6.283185307179586d0
    integer,parameter:: blkkc = 100, blkthr = 200
    complex(kind=8),parameter:: ci = (0.0d0, 1.0d0)
    !
    complex(kind=8):: cwrk(bkc:ekc)
    !
    integer:: k1, k2, k3, kk1, kk2, kk3
    real(kind=8):: t1, t2, t3, ph
    complex(kind=8):: temp
    integer:: it, ig
    integer:: itk
    integer:: bbkc,bekc
    logical:: ifblock
    !
#ifdef _OPENMP
    if (ekc-bkc+1 .gt. omp_get_max_threads()*blkthr) then
       ifblock = .true.
    else
       ifblock = .false.
    end if
#else
    ifblock = .true.
#endif
    !
    if (ifblock) then
!$omp parallel do private(k1,k2,k3,kk1,kk2,kk3,t1,t2,t3,temp,ph), &
!$omp private(it,ig,itk,bbkc,bekc),reduction(+:cvqq)
       do bbkc=bkc,ekc,blkkc
          bekc=min(bbkc+blkkc-1,ekc)
          do itk=1,nsikd
             it=itk2it(itk)
             t1=asi(1,it)
             t2=asi(2,it)
             t3=asi(3,it)
             do ig=bbkc,bekc
                k1=-kqg(1,ig)
                k2=-kqg(2,ig)
                k3=-kqg(3,ig)
                ph=pi2*(t1*k1+t2*k2+t3*k3)
                cwrk(ig)=dcmplx(cos(ph),sin(ph))*vpot(k1,k2,k3)
             end do
             do k1=1,numk1
                do ig=bbkc,bekc
                   kk1=kqg(1,ig)
                   kk2=kqg(2,ig)
                   kk3=kqg(3,ig)
                   temp = -ci*pi2*qg(ig,k1)*cwrk(ig)*va
                   cvqq(1,k1,it)= cvqq(1,k1,it)+dble(kk1)*temp
                   cvqq(2,k1,it)= cvqq(2,k1,it)+dble(kk2)*temp
                   cvqq(3,k1,it)= cvqq(3,k1,it)+dble(kk3)*temp
                   cvqq(4:9,k1,it)= cvqq(4:9,k1,it) &
                        + sqg(ig,1:6,k1)*cwrk(ig)*va
                end do
             end do
          end do
       end do
    else
!$omp parallel do private(k1,k2,k3,kk1,kk2,kk3,t1,t2,t3,temp,ph), &
!$omp private(it,ig,itk,cwrk)
       do itk=1,nsikd
          it=itk2it(itk)
          t1=asi(1,it)
          t2=asi(2,it)
          t3=asi(3,it)
          do ig=bkc,ekc
             k1=-kqg(1,ig)
             k2=-kqg(2,ig)
             k3=-kqg(3,ig)
             ph=pi2*(t1*k1+t2*k2+t3*k3)
             cwrk(ig)=dcmplx(cos(ph),sin(ph))*vpot(k1,k2,k3)
          end do
          do k1=1,numk1
             do ig=bkc,ekc
                kk1=kqg(1,ig)
                kk2=kqg(2,ig)
                kk3=kqg(3,ig)
                temp = -ci*pi2*qg(ig,k1)*cwrk(ig)*va
                cvqq(1,k1,it)= cvqq(1,k1,it)+dble(kk1)*temp
                cvqq(2,k1,it)= cvqq(2,k1,it)+dble(kk2)*temp
                cvqq(3,k1,it)= cvqq(3,k1,it)+dble(kk3)*temp
                cvqq(4:9,k1,it)= cvqq(4:9,k1,it) &
                     +dble(sqg(ig,1:6,k1)*cwrk(ig))*va
             end do
          end do
       end do
    end if
    
    return
  end subroutine mkvqqcore

  subroutine mkgmf(ncomp,sk,bb,gm,kg,p,vpj,svpj,bkm,ekm, &
       nsi,asi,kd,nkd,lpnum,maxr,laprj,nb2,bspots)
    use util
    use bsplpot
    use arch_engdffr
    use timeprof
    implicit none
    integer,intent(in):: ncomp
    real(kind=8),intent(in):: sk(3),bb(3,3)
    integer,intent(in):: bkm,ekm,nsi,nkd,lpnum,maxr,laprj,nb2
    real(kind=8),intent(in):: vpj(bkm:ekm,maxr,lpnum,nkd)
    real(kind=8),intent(in):: svpj(bkm:ekm,maxr,lpnum,nkd)
    real(kind=8),intent(in):: asi(3,nsi)
    integer,intent(in):: kd(nsi)
    integer,intent(in):: kg(3,bkm:ekm)
    complex(kind=8),intent(in):: p(ncomp,bkm:ekm,nb2)
    complex(kind=8):: gm(10,laprj,nb2,nsi)
    type (bspot_for_atm):: bspots(nkd)
    !
    complex(kind=8),allocatable:: ylm(:,:,:)
    complex(kind=8),allocatable:: ylms(:,:,:,:)
    real(kind=8),allocatable:: rwrk(:,:)
    complex(kind=8),allocatable:: rwrk2(:,:,:,:)
    !
    integer:: ip
    integer,allocatable:: it2h(:)
    !
    integer:: ig,m,it,ikd,err
    integer:: bt,et,bbkm,bekm
    real(kind=8):: x,y,z,r,rsq
    complex(kind=8):: rtemp(ncomp)
    integer:: ist
    integer:: lmax

    call tpf_begin(tpf_mkgmf)

    allocate(it2h(nsi+1),stat=err)
    call chkalloc('mkgmf','it2h',err)

    ip = 0
    do it=1,nsi
       ikd = kd(it)
       it2h(it) = ip
       ip = ip + bspots(ikd)%nprj
    end do
    it2h(nsi+1) = ip

    lmax = getlmax(nkd,bspots)
    allocate(ylm(ncomp,bkm:ekm,ncomp*lmax**2), &
         rwrk(6,bkm:ekm),rwrk2(ncomp,6,bkm:ekm,ncomp*lmax**2), &
         ylms(ncomp,bkm:ekm,3,ncomp*lmax**2), stat=err)
    call chkalloc('mkgmf','ylm,ylms,rwrk',err)
    call getylb(ekm-bkm+1,sk,kg,bb,ylm,lmax,ncomp)
    call getyls(ekm-bkm+1,sk,kg,bb,ylms,lmax,ncomp)
    !
    x = bb(1,1)*(sk(1)+kg(1,bkm)) + bb(1,2)*(sk(2)+kg(2,bkm)) &
         + bb(1,3)*(sk(3)+kg(3,bkm))
    y = bb(2,1)*(sk(1)+kg(1,bkm)) + bb(2,2)*(sk(2)+kg(2,bkm)) &
         + bb(2,3)*(sk(3)+kg(3,bkm))
    z = bb(3,1)*(sk(1)+kg(1,bkm)) + bb(3,2)*(sk(2)+kg(2,bkm)) &
         + bb(3,3)*(sk(3)+kg(3,bkm))
    rsq = x*x + y*y + z*z
    if (rsq .eq. 0.0d0) then
       ist = 1
    else
       ist = 0
    end if

!$omp parallel do private(x,y,z,r,rsq)
    do ig=bkm+ist,ekm
       x = bb(1,1)*(sk(1)+kg(1,ig)) + bb(1,2)*(sk(2)+kg(2,ig)) &
            + bb(1,3)*(sk(3)+kg(3,ig))
       y = bb(2,1)*(sk(1)+kg(1,ig)) + bb(2,2)*(sk(2)+kg(2,ig)) &
            + bb(2,3)*(sk(3)+kg(3,ig))
       z = bb(3,1)*(sk(1)+kg(1,ig)) + bb(3,2)*(sk(2)+kg(2,ig)) &
            + bb(3,3)*(sk(3)+kg(3,ig))
       rsq = x*x + y*y + z*z
       r = sqrt(rsq)

       rwrk(1,ig) = x
       rwrk(2,ig) = y
       rwrk(3,ig) = z
       rwrk(4,ig) = r
       rwrk(5,ig) = rsq
    end do

!$omp parallel do private(m,ig,x,y,z,r,rsq,rtemp)
    do m=1,ncomp*lmax**2
       do ig=bkm+ist,ekm
          x = rwrk(1,ig)
          y = rwrk(2,ig)
          z = rwrk(3,ig)
          r = rwrk(4,ig)
          rsq = rwrk(5,ig)

          rtemp=(x/r)*ylms(:,ig,1,m)+(y/r)*ylms(:,ig,2,m) &
               +(z/r)*ylms(:,ig,3,m)

          rwrk2(:,1,ig,m) = ylms(:,ig,1,m)*x/r - rtemp*x*x/rsq
          rwrk2(:,2,ig,m) = ylms(:,ig,2,m)*y/r - rtemp*y*y/rsq
          rwrk2(:,3,ig,m) = ylms(:,ig,3,m)*z/r - rtemp*z*z/rsq

          rwrk2(:,4,ig,m) = ylms(:,ig,3,m)*y/r + ylms(:,ig,2,m)*z/r &
               - rtemp*2.0d0*y*z/rsq
          rwrk2(:,5,ig,m) = ylms(:,ig,1,m)*z/r + ylms(:,ig,3,m)*x/r &
               - rtemp*2.0d0*z*x/rsq
          rwrk2(:,6,ig,m) = ylms(:,ig,2,m)*x/r + ylms(:,ig,1,m)*y/r &
               - rtemp*2.0d0*x*y/rsq
       end do
    end do

!$omp parallel do private(ig,x,y,z,r,rsq)
    do ig=bkm+ist,ekm
       x = rwrk(1,ig)
       y = rwrk(2,ig)
       z = rwrk(3,ig)
       r = rwrk(4,ig)
       rsq = rwrk(5,ig)

       rwrk(1,ig) = x*x/rsq
       rwrk(2,ig) = y*y/rsq
       rwrk(3,ig) = z*z/rsq
       rwrk(4,ig) = 2.0d0*y*z/rsq
       rwrk(5,ig) = 2.0d0*z*x/rsq
       rwrk(6,ig) = 2.0d0*x*y/rsq
    end do

    deallocate(ylms)

    gm = 0.0d0

    bt = 1
    do while (bt .le. nsi)
       et = bt
       do while (et .lt. nsi .and. it2h(et)-it2h(bt)+1 .lt. blkprj)
          et = et + 1
       end do
       bbkm = bkm
       do while (bbkm .le. ekm)
          bekm = min(bbkm+blkpw,ekm)
          call mkgmfcore(ncomp,sk,bb,gm,kg,p,vpj,svpj,bkm,ekm, &
               nsi,asi,kd,nkd,lpnum,maxr,lmax,laprj,nb2,bspots, &
               it2h,bt,et,bbkm,bekm,ylm,rwrk,rwrk2)
          bbkm = bekm + 1
       end do
       bt = et + 1
    end do
    !
    deallocate(ylm,rwrk,rwrk2,it2h)

    call tpf_end(tpf_mkgmf)
    return
  end subroutine mkgmf

  subroutine mkgmfcore(ncomp,sk,bb,gm,kg,p,vpj,svpj,bkm,ekm, &
       nsi,asi,kd,nkd,lpnum,maxr,lmax,laprj,nb2,bspots, &
       it2h,bt,et,bbkm,bekm,ylm,rwrk,rwrk2)
    use util
    use bsplpot
    implicit none
    integer,intent(in):: ncomp
    real(kind=8),intent(in):: sk(3),bb(3,3)
    integer,intent(in):: bkm,ekm,nsi,nkd,lpnum,maxr,lmax,laprj,nb2
    real(kind=8),intent(in):: vpj(bkm:ekm,maxr,lpnum,nkd)
    real(kind=8),intent(in):: svpj(bkm:ekm,maxr,lpnum,nkd)
    real(kind=8),intent(in):: asi(3,nsi)
    integer,intent(in):: kd(nsi)
    integer,intent(in):: kg(3,bkm:ekm)
    complex(kind=8),intent(in):: p(ncomp,bkm:ekm,nb2)
    complex(kind=8):: gm(10,laprj,nb2,nsi)
    type (bspot_for_atm):: bspots(nkd)
    !
    integer:: it2h(nsi+1),bt,et,bbkm,bekm
    complex(kind=8):: ylm(ncomp,bkm:ekm,ncomp*lmax**2)
    real(kind=8):: rwrk(6,bkm:ekm)
    complex(kind=8):: rwrk2(ncomp,6,bkm:ekm,ncomp*lmax**2)
    !
    complex(kind=8),parameter:: ci = (0.0d0,1.0d0)
    real(kind=8),parameter:: pi2 = 3.1415926535897932385d0*2.0d0
    !
    integer:: ig,il,i1,m,ib,it,ikd,err
    real(kind=8):: ph
    complex(kind=8):: ctemp,ctemp2,ctemp3
    real(kind=8):: x,y,z,rsq
    integer:: ist
    integer:: ip, ipa, bnprj, bnkm
    integer:: ic
    !
    complex(kind=8):: cwrk(bbkm:bekm)

    complex(kind=8),allocatable:: gmt(:,:,:)
    complex(kind=8),allocatable:: zt(:,:,:,:)
    !
    x = bb(1,1)*(sk(1)+kg(1,bbkm)) + bb(1,2)*(sk(2)+kg(2,bbkm)) &
         + bb(1,3)*(sk(3)+kg(3,bbkm))
    y = bb(2,1)*(sk(1)+kg(1,bbkm)) + bb(2,2)*(sk(2)+kg(2,bbkm)) &
         + bb(2,3)*(sk(3)+kg(3,bbkm))
    z = bb(3,1)*(sk(1)+kg(1,bbkm)) + bb(3,2)*(sk(2)+kg(2,bbkm)) &
         + bb(3,3)*(sk(3)+kg(3,bbkm))
    rsq = x*x + y*y + z*z
    if (rsq .eq. 0.0d0) then
       ist = 1
    else
       ist = 0
    end if

    allocate(zt(ncomp,bbkm:bekm,10,it2h(bt)+1:it2h(et+1)), stat=err)
    call chkalloc('mkgmfcore','zt',err)
    
!$omp parallel do private(it,ig,il,i1,m,ip,ikd,ph,ctemp,ctemp2,ctemp3), &
!$omp private(ipa,cwrk,ic),schedule(static,1)
!POPTION TLOCAL(cwrk),CYCLIC
    do it=bt,et
       ikd = kd(it)
       do ig=bbkm,bekm
          ph=pi2*(asi(1,it)*kg(1,ig)+asi(2,it)*kg(2,ig) &
               +asi(3,it)*kg(3,ig))
          cwrk(ig)=dcmplx(cos(ph),-sin(ph))
       end do
       do ipa = 1, bspots(ikd)%nprj
          il = bspots(ikd)%ip2l(ipa)
          i1 = bspots(ikd)%ip2r(ipa)
          m = bspots(ikd)%ip2m(ipa)
          ctemp2 = (-ci)**bspots(ikd)%l2lq(il)
          ip = it2h(it) + ipa
          if (ist.eq.1) then
             ig = bbkm
             do ic=1,ncomp
                ctemp = cwrk(ig)*ylm(ic,ig,m)*vpj(ig,i1,il,ikd)*ctemp2
                zt(ic,ig,1:3,ip) = -ci*pi2*kg(1:3,ig)*ctemp
                zt(ic,ig,4,ip) = ctemp
                zt(ic,ig,5:10,ip) = 0.0d0
             end do
          end if
          do ig=bbkm+ist,bekm
             do ic=1,ncomp
                ctemp = cwrk(ig)*ylm(ic,ig,m)*vpj(ig,i1,il,ikd)*ctemp2
                zt(ic,ig,1:3,ip) = -ci*pi2*kg(1:3,ig)*ctemp
                zt(ic,ig,4,ip) = ctemp

                ctemp=cwrk(ig)*ylm(ic,ig,m)*svpj(ig,i1,il,ikd)*ctemp2
                ctemp3=cwrk(ig)*vpj(ig,i1,il,ikd)*ctemp2
                zt(ic,ig,5:10,ip) = ctemp*rwrk(1:6,ig) &
                     + ctemp3*rwrk2(ic,1:6,ig,m)
             end do
          end do
       end do
    end do

    allocate(gmt(10,it2h(bt)+1:it2h(et+1),nb2), stat=err)
    call chkalloc('mkgmfcore','gmt',err)

    bnprj = it2h(et+1)-it2h(bt)
    bnkm = ncomp*(bekm-bbkm+1)

    call zgemm('C','N',10*bnprj,nb2,bnkm,(1.0d0,0.0d0),zt,bnkm, &
         p(1,bbkm,1),ncomp*(ekm-bkm+1),(0.0d0,0.0d0),gmt,10*bnprj)

!$omp parallel do private(it,il,i1,m,ib,ikd,ip,ipa)
    do it=bt,et
       ikd = kd(it)
       do ipa = 1,bspots(ikd)%nprj
          il = bspots(ikd)%ip2l(ipa)
          i1 = bspots(ikd)%ip2r(ipa)
          m = bspots(ikd)%ip2m(ipa)
          ip = it2h(it) + ipa
          do ib=1,nb2
             gm(1:10,ipa,ib,it)= gm(1:10,ipa,ib,it) + gmt(1:10,ip,ib)
          end do
       end do
    end do

    deallocate(gmt,zt)
    !
    return
  end subroutine mkgmfcore

  subroutine vbnlf(epotnl,sepotnl,forel,bzwk,wgtk &
       ,va,dd,gm,vqq,ee,kd &
       ,nkd,nsi,laprj,nkmesh,nb1,nb2,bspots)
    use bsplpot
    implicit none
    integer,intent(in):: nkd,nsi,laprj, &
         nkmesh,nb1,nb2
    !
    real(kind=8):: epotnl,sepotnl(6)
    real(kind=8):: forel(3,nsi)
    !
    integer,intent(in):: kd(nsi)
    real(kind=8),intent(in):: bzwk
    real(kind=8),intent(in):: va,wgtk(nb1:nb2)
    real(kind=8),intent(in):: ee(nb2)
    complex(kind=8),intent(in):: dd(laprj,laprj,nsi)
    complex(kind=8),intent(in):: gm(10,laprj,nb2,nsi)
    complex(kind=8),intent(in):: vqq(9,laprj,laprj,nsi)
    type (bspot_for_atm):: bspots(nkd)
    !
    real(kind=8):: ew,sew(6),forelw(3)
    !
    complex(kind=8):: temp,atemp
    real(kind=8):: bxxk
    integer:: it,ikd,ib,l1,i1,m1,l2,i2,m2,ip1,ip2
    !
    ! WRITE(6,*)'EE(',IB,')=',EE(IB)
    !
!$omp parallel do private(ew,temp,atemp,bxxk,ikd,ib,l1,i1,m1,l2,i2,m2), &
!$omp private(ip1,ip2,sew,forelw),reduction(+:forel,sepotnl,epotnl)
!POPTION PARALLEL,TLOCAL(forelw,ew,sew)
!POPTION PSUM(epotnl,sepotnl,forel)
    do it=1,nsi
       ikd=kd(it)
       do ib=1,nb2
          forelw = 0.0d0
          ew = 0.0d0
          sew = 0.0d0
          do ip1 = 1,bspots(ikd)%nprj
             l1 = bspots(ikd)%ip2l(ip1)
             i1 = bspots(ikd)%ip2r(ip1)
             m1 = bspots(ikd)%ip2m(ip1)
             do ip2 = 1,ip1-1
                l2 = bspots(ikd)%ip2l(ip2)
                i2 = bspots(ikd)%ip2r(ip2)
                m2 = bspots(ikd)%ip2m(ip2)
                atemp=conjg(gm(4,ip1,ib,it))*gm(4,ip2,ib,it)/va
                if (m1.eq.m2) then
                   temp=dd(ip1,ip2,it) &
                        -ee(ib)*bspots(ikd)%qq(i1,i2,l1)
                   !
                   ! both the i1,i2 and i2,i1 case.
                   ! because ip1 != ip2 and m1 == m2, i1 != i2.
                   !
                   ew=ew + 2.0d0*dble(atemp)*bspots(ikd)%ddi(i1,i2,l1)
                   sew(1:3)=sew(1:3) - 2.0d0*ee(ib) &
                        *bspots(ikd)%qq(i1,i2,l1)*dble(atemp)
                else
                   temp=dd(ip1,ip2,it)
                end if
                temp=temp/va

                forelw(1)=forelw(1) - 2.0d0*dble( &
                        conjg(gm(4,ip1,ib,it))*temp*gm(1,ip2,ib,it) &
                       +conjg(gm(1,ip1,ib,it))*temp*gm(4,ip2,ib,it) ) &
                     - 2.0d0*dble(vqq(1,ip1,ip2,it)*atemp)
                forelw(2)=forelw(2) - 2.0d0*dble( &
                        conjg(gm(4,ip1,ib,it))*temp*gm(2,ip2,ib,it) &
                       +conjg(gm(2,ip1,ib,it))*temp*gm(4,ip2,ib,it) ) &
                     - 2.0d0*dble(vqq(2,ip1,ip2,it)*atemp)
                forelw(3)=forelw(3) - 2.0d0*dble( &
                        conjg(gm(4,ip1,ib,it))*temp*gm(3,ip2,ib,it) &
                       +conjg(gm(3,ip1,ib,it))*temp*gm(4,ip2,ib,it) ) &
                     - 2.0d0*dble(vqq(3,ip1,ip2,it)*atemp)

                sew(1:6)=sew(1:6) + 2.0d0*dble( &
                        conjg(gm(4,ip1,ib,it))*temp*gm(5:10,ip2,ib,it) &
                       +conjg(gm(5:10,ip1,ib,it))*temp*gm(4,ip2,ib,it) ) &
                     + 2.0d0*dble(vqq(4:9,ip1,ip2,it)*atemp)
             end do
             !
             ! diagonal part: ip1 = ip2
             !
             atemp=conjg(gm(4,ip1,ib,it))*gm(4,ip1,ib,it)/va
             temp=dd(ip1,ip1,it) &
                  -ee(ib)*bspots(ikd)%qq(i1,i1,l1)
             ew=ew + dble(atemp)*bspots(ikd)%ddi(i1,i1,l1)
             sew(1:3)=sew(1:3) - ee(ib) &
                  *bspots(ikd)%qq(i1,i1,l1)*dble(atemp)
             temp=temp/va

             forelw(1)=forelw(1) - 2.0d0*dble(temp) &
                  *dble(conjg(gm(4,ip1,ib,it))*gm(1,ip1,ib,it)) &
                  - dble(vqq(1,ip1,ip1,it))*dble(atemp)
             forelw(2)=forelw(2) - 2.0d0*dble(temp) &
                  *dble(conjg(gm(4,ip1,ib,it))*gm(2,ip1,ib,it)) &
                  - dble(vqq(2,ip1,ip1,it))*dble(atemp)
             forelw(3)=forelw(3) - 2.0d0*dble(temp) &
                  *dble(conjg(gm(4,ip1,ib,it))*gm(3,ip1,ib,it)) &
                  - dble(vqq(3,ip1,ip1,it))*dble(atemp)

             sew(1:6)=sew(1:6) + 2.0d0*dble(temp) &
                  *dble(conjg(gm(4,ip1,ib,it))*gm(5:10,ip1,ib,it)) &
                  + dble(vqq(4:9,ip1,ip1,it))*dble(atemp)
          end do
          !
          if (ib.lt.nb1) then
             bxxk=bzwk/dble(nkmesh)
          else
             bxxk=wgtk(ib)/dble(nkmesh)
          end if
          forel(1,it)=forel(1,it)+forelw(1)*bxxk
          forel(2,it)=forel(2,it)+forelw(2)*bxxk
          forel(3,it)=forel(3,it)+forelw(3)*bxxk
          epotnl     =epotnl     +ew       *bxxk
          sepotnl(1:6)=sepotnl(1:6) + sew(1:6)*bxxk
       end do
       !
    end do
    !
    return
  end subroutine vbnlf

end module engdffr
