! Copyright (c) 2012, Minoru Otani <minoru.otani@aist.go.jp>
! 
! Permission is hereby granted, free of charge, to any person 
! obtaining a copy of this software and associated documentation 
! files (the "Software"), to deal in the Software without restriction, 
! including without limitation the rights to use, copy, modify, merge, 
! publish, distribute, sublicense, and/or sell copies of the Software, 
! and to permit persons to whom the Software is furnished to do so, 
! subject to the following conditions:
 
! The above copyright notice and this permission notice shall be 
! included in all copies or substantial portions of the Software.
 
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
! OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
! NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
! HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
! WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
! OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
! DEALINGS IN THE SOFTWARE.
module m_2d
use global_esm
use m_deb
!ANDO
use esm_util
!ANDO
contains
!-------initialize---------------------
subroutine esm_2d_initialize(cginp,var2d)
  use esm_util,only:lat_points
  implicit none
  integer                ::err
  real(kind=8),parameter ::gsafe=1.2d0
  type(esm_2d),intent(out) ::var2d
  type(cg2esm),intent(in)  ::cginp
if (esm%normal .eq. 'X') then
  var2d%n0=cginp%nrx2
  var2d%n1=cginp%nry2
  var2d%n2=cginp%nrz2
  var2d%L =cginp%aa(1,1)
  var2d%Lg=cginp%bb(1,1)
  var2d%sa=cginp%aa(2,2)*cginp%aa(3,3)-cginp%aa(3,2)*cginp%aa(2,3)
  var2d%sb=cginp%bb(2,2)*cginp%bb(3,3)-cginp%bb(3,2)*cginp%bb(2,3)
  var2d%aa_2d(1:2,1:2)=cginp%aa(2:3,2:3)
  var2d%bb_2d(1:2,1:2)=cginp%bb(2:3,2:3)
else if (esm%normal .eq. 'Z') then
  var2d%n0=cginp%nrz2
  var2d%n1=cginp%nrx2
  var2d%n2=cginp%nry2
  var2d%L =cginp%aa(3,3)
  var2d%Lg=cginp%bb(3,3)
  var2d%sa=cginp%aa(1,1)*cginp%aa(2,2)-cginp%aa(2,1)*cginp%aa(1,2)
  var2d%sb=cginp%bb(1,1)*cginp%bb(2,2)-cginp%bb(2,1)*cginp%bb(1,2)
!added by ANDO
  var2d%sa=abs(var2d%sa)
  var2d%sb=abs(var2d%sb)
!end ANDO
  var2d%aa_2d(1:2,1:2)=cginp%aa(1:2,1:2)
  var2d%bb_2d(1:2,1:2)=cginp%bb(1:2,1:2)
endif
  var2d%naf=int(lat_points(cginp%qf,var2d%sb)*gsafe)
  allocate(var2d%kkg_2d(2,var2d%naf),var2d%akk_2d(var2d%naf) &
          ,var2d%asi_z(cginp%nsi),var2d%asi_2d(2,cginp%nsi) &
          ,stat=err)
  call chkalloc('esm_2d','kkg,akk,asi_z,asi_2d',err)
  return
end subroutine esm_2d_initialize

!-------make gp vector-----------------
subroutine esm_2d_make_nkf_2d_akk_2d(cginp,var2d)
  implicit none
  type(cg2esm),intent(in)     ::cginp
  type(esm_2d),intent(inout)  ::var2d
  !local
  integer      ::ix,iy,i
  integer      ::ind(var2d%naf),kg_2d(2,var2d%naf)
  real(kind=8) ::qqf,cx,cy,s
  !
  qqf=cginp%qf*cginp%qf
  var2d%nkf_2d = 0
  do ix=-var2d%n1,var2d%n1
    do iy=-var2d%n2,var2d%n2
      cx=ix*var2d%bb_2d(1,1)+iy*var2d%bb_2d(1,2)
      cy=ix*var2d%bb_2d(2,1)+iy*var2d%bb_2d(2,2)
      s=cx*cx+cy*cy
      if(s.le.qqf)then
        var2d%nkf_2d=var2d%nkf_2d+1
       if(var2d%nkf_2d > var2d%naf) then
           stop' nkf_2d>naf!! something wrong check gsafe'
        end if
        var2d%akk_2d(var2d%nkf_2d)=s
        var2d%kkg_2d(1,var2d%nkf_2d)=ix
        var2d%kkg_2d(2,var2d%nkf_2d)=iy
      endif
    enddo
  enddo
  !
  ! heap sort
  !
  call esm_hsort(var2d%nkf_2d,var2d%akk_2d,ind)
  do i=1,var2d%nkf_2d
    kg_2d(:,i)=var2d%kkg_2d(:,i)
  enddo
  do i=1,var2d%nkf_2d
    var2d%kkg_2d(:,i)=kg_2d(:,ind(i))
  enddo
  ! set smoothing parameter
  esm%nz_l=var2d%n0/2+1+esm%nfit
  esm%nz_r=var2d%n0/2+1-esm%nfit
  esm%z_l=dble(var2d%n0/2+1+esm%nfit-1)*var2d%L/dble(var2d%n0)-var2d%L
  esm%z_r=dble(var2d%n0/2+1-esm%nfit-1)*var2d%L/dble(var2d%n0)
  return
end subroutine esm_2d_make_nkf_2d_akk_2d

!--------Make local potential (long range part)-------
subroutine m_2d_make_long_range_part(cginp,var2d,v,vg,comm,mssg)
  use fft_1d
  use m_loc
  use mpi
  use timeprof

  implicit none
  type(esm_2d),intent(in)  ::var2d
  type(cg2esm),intent(in)  ::cginp
  integer,intent(in)       ::comm,mssg
  complex(kind=8),intent(inout)::v(cginp%nrx2,cginp%nry2,cginp%nrz2),vg(var2d%n0)

  !local
!!!!!0307!!!!!!!!
  ! FOR MPI
  integer,allocatable         :: gf_count(:),gf_displ(:)
  integer                     :: nproc, myrank, ierr, irnk, bkf_2d, ekf_2d, ist
  complex(kind=8),allocatable :: pv(:,:,:),tv(:,:,:)
!!!!!!!!!!!!!!!!!!

  integer ::ig,k1,k2,ks1,ks2,err
  !!! FOR FFT !!!
  type(fft1_struct) ::fs
  real(kind=8),allocatable ::fftw1(:,:),fftw2(:,:)
  allocate(fftw1(var2d%n0,2),fftw2(var2d%n0,2),stat=err)
  call chkalloc('fft_pot','fftw1,fftw2',err)
  call fft1_init(var2d%n0,fs)
  !!!!!!!!!!!!!!!

  allocate(pv(cginp%nrx2,cginp%nry2,cginp%nrz2),tv(cginp%nrx2,cginp%nry2,cginp%nrz2))

  call MPI_COMM_SIZE(comm,nproc,ierr)
  call MPI_COMM_RANK(comm,myrank,ierr)

  allocate(gf_count(0:nproc-1),gf_displ(0:nproc-1),stat=ierr)
  call chkalloc('esm_2d','gf_count, gf_displ',ierr)
!poption noparallel
  gf_count = var2d%nkf_2d/nproc
!poption noparallel
  do irnk=0,mod(var2d%nkf_2d,nproc)-1
     gf_count(irnk) = gf_count(irnk)+1
  end do
  gf_displ(0) = 0
  do irnk=0,nproc-2
     gf_displ(irnk+1) = gf_displ(irnk)+gf_count(irnk)
  end do
  bkf_2d = gf_displ(myrank)+1
  ekf_2d = gf_displ(myrank)+gf_count(myrank)

  if(bkf_2d == 1) bkf_2d = 2
  
  !initialize
  pv=0.d0;  tv=0.d0


  !---gp = 0 case
  call m_loc_local_potential_gp0(cginp%nsi &
                                ,cginp%zv &
                                ,cginp%agauss &
                                ,cginp%bgauss &
                                ,esm%ngau &
                                ,esm%z1 &
                                ,esm%z_r &
                                ,esm%z_l &
                                ,esm%nz_r &
                                ,esm%nz_l &
                                ,esm%v0 &
                                ,var2d%n0 &
                                ,var2d%L &
                                ,var2d%sa &
                                ,var2d%asi_z &
                                ,vg,fftw1,fftw2,fs) ! return v(g)
if (esm%normal .eq. 'X') then
  v(1:var2d%n0,1,1)=vg(1:var2d%n0)
else if (esm%normal .eq. 'Z') then
  v(1,1,1:var2d%n0)=vg(1:var2d%n0)
endif

!---gp!= 0 case
  do ig= bkf_2d, ekf_2d
    k1=var2d%kkg_2d(1,ig)
    k2=var2d%kkg_2d(2,ig)
    ks1=mod(var2d%n1+k1,var2d%n1)+1
    ks2=mod(var2d%n2+k2,var2d%n2)+1
    call m_loc_local_potential(cginp%nsi &
                              ,cginp%zv &
                              ,cginp%agauss &
                              ,cginp%bgauss &
                              ,esm%ngau &
                              ,esm%z1 &
                              ,var2d%n0 &
                              ,var2d%L &
                              ,var2d%sa &
                              ,var2d%asi_z &
                              ,var2d%asi_2d &
                              ,var2d%akk_2d(ig) &
                              ,k1,k2,vg,fftw1,fftw2,fs,mssg) ! return v(g)

if (esm%normal .eq. 'X') then
    pv(1:var2d%n0,ks1,ks2)=vg(1:var2d%n0)
else if (esm%normal .eq. 'Z') then
    pv(ks1,ks2,1:var2d%n0)=vg(1:var2d%n0)
endif
 enddo

  call tpf_begin(tpf_mpi)
  call MPI_ALLREDUCE(pv, tv, size(pv), MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
  call tpf_end(tpf_mpi)

!!$omp parallel do
  do ig= 2, var2d%nkf_2d
     k1=var2d%kkg_2d(1,ig)
     k2=var2d%kkg_2d(2,ig)
     ks1=mod(var2d%n1+k1,var2d%n1)+1
     ks2=mod(var2d%n2+k2,var2d%n2)+1
if (esm%normal .eq. 'X') then
     v(1:var2d%n0,ks1,ks2)=tv(1:var2d%n0,ks1,ks2)
else if (esm%normal .eq. 'Z') then
     v(ks1,ks2,1:var2d%n0)=tv(ks1,ks2,1:var2d%n0)
endif
  enddo

  call fft1_done(fs)
  deallocate(fftw1,fftw2,pv,tv)
  return
end subroutine m_2d_make_long_range_part

subroutine m_2d_make_force_long_range_part(cginp,var2d,rho,rg &
                                          ,epotl,w_f,f_z,f_2d,comm,mssg)
  use fft_1d
  use m_loc
  use mpi
  use timeprof
  
  implicit none
  type(esm_2d),intent(in)    ::var2d
  type(cg2esm),intent(in)    ::cginp
  integer,intent(in)         ::comm,mssg
  real(kind=8),intent(inout) ::f_z(cginp%nsi),f_2d(2,cginp%nsi),epotl
  complex(kind=8),intent(in) ::rho(cginp%nrx2,cginp%nry2,cginp%nrz2)
  complex(kind=8),intent(inout)::rg(var2d%n0),w_f(var2d%n0,3)
  !local
  integer ::ig,k1,k2,ks1,ks2,err
!!!!!0307!!!!!!!!
  ! FOR MPI
  integer,allocatable:: gf_count(:),gf_displ(:)
  integer:: nproc, myrank, ierr, irnk, bkf_2d, ekf_2d, ist
  real(kind=8),allocatable      :: pf_z(:),tf_z(:),pf_2d(:,:),tf_2d(:,:)
  real(kind=8) ::           pepotl, tepotl  
!!!!!!!!!!!!!!!!!!
  !!! FOR FFT !!!
  type(fft1_struct) ::fs
  real(kind=8),allocatable ::fftw1(:,:),fftw2(:,:)
  allocate(fftw1(var2d%n0,2),fftw2(var2d%n0,2),stat=err)
  call chkalloc('fft_pot','fftw1,fftw2',err)
  call fft1_init(var2d%n0,fs)
  !!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!

  call MPI_COMM_SIZE(comm,nproc,ierr)
  call MPI_COMM_RANK(comm,myrank,ierr)

  allocate(gf_count(0:nproc-1),gf_displ(0:nproc-1),stat=ierr)
  allocate(pf_z(cginp%nsi),tf_z(cginp%nsi),&
                        pf_2d(2,cginp%nsi),tf_2d(2,cginp%nsi))
                      
  call chkalloc('esm_2d','gf_count, gf_displ',ierr)

!poption noparallel
  gf_count = var2d%nkf_2d/nproc
!poption noparallel
  do irnk=0,mod(var2d%nkf_2d,nproc)-1
     gf_count(irnk) = gf_count(irnk)+1
  end do
  gf_displ(0) = 0
  do irnk=0,nproc-2
     gf_displ(irnk+1) = gf_displ(irnk)+gf_count(irnk)
  end do
  bkf_2d = gf_displ(myrank)+1
  ekf_2d = gf_displ(myrank)+gf_count(myrank)

  if(bkf_2d == 1) bkf_2d = 2

  pf_2d = 0.0d0
  pf_z  = 0.0d0

!---gp = 0 case
if (esm%normal .eq. 'X') then
  rg(1:var2d%n0)=rho(1:var2d%n0,1,1)
else if (esm%normal .eq. 'Z') then
  rg(1:var2d%n0)=rho(1,1,1:var2d%n0)
endif

  call m_loc_local_force_gp0(cginp%nsi &
                            ,cginp%zv &
                            ,cginp%va &
                            ,cginp%agauss &
                            ,cginp%bgauss &
                            ,esm%ngau &
                            ,esm%z1 &
                            ,esm%z_r &
                            ,esm%z_l &
                            ,esm%nz_r &
                            ,esm%nz_l &
                            ,esm%v0 &
                            ,var2d%n0 &
                            ,var2d%L &
                            ,var2d%Lg &
                            ,var2d%sa &
                            ,var2d%asi_z &
                            ,rg,w_f(1,1),f_z,epotl,fftw1,fftw2,fs)


!---gp!= 0 case
  do ig=bkf_2d, ekf_2d
     k1=var2d%kkg_2d(1,ig)
     k2=var2d%kkg_2d(2,ig)
     ks1=mod(var2d%n1+k1,var2d%n1)+1
     ks2=mod(var2d%n2+k2,var2d%n2)+1
if (esm%normal .eq. 'X') then
     rg(1:var2d%n0)=rho(1:var2d%n0,ks1,ks2)
else if (esm%normal .eq. 'Z') then
     rg(1:var2d%n0)=rho(ks1,ks2,1:var2d%n0)
endif
     call m_loc_local_force(cginp%nsi &
                          ,cginp%zv &
                          ,cginp%va &
                          ,cginp%agauss &
                          ,cginp%bgauss &
                          ,esm%ngau &
                          ,esm%z1 &
                          ,var2d%n0 &
                          ,var2d%L &
                          ,var2d%Lg &
                          ,var2d%sa &
                          ,var2d%asi_z &
                          ,var2d%asi_2d &
                          ,var2d%akk_2d(ig) &
                          ,k1,k2,rg,w_f,pf_z,pf_2d,pepotl,fftw1,fftw2,fs,mssg)

  enddo
  call tpf_begin(tpf_mpi)
  call MPI_ALLREDUCE(pf_z,  tf_z,  cginp%nsi,MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
  call MPI_ALLREDUCE(pf_2d, tf_2d, 2*cginp%nsi,MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
  call MPI_ALLREDUCE(pepotl,tepotl,1 ,MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
  call tpf_end(tpf_mpi)

  f_z = f_z + tf_z
  f_2d = f_2d + tf_2d
  epotl = epotl + tepotl
  
  call fft1_done(fs)
  deallocate(fftw1,fftw2,pf_z,tf_z,pf_2d,tf_2d)
  return
end subroutine m_2d_make_force_long_range_part

!--------Calculate ion-ion part (ewald method)---------s
subroutine m_2d_ewald(cginp,var2d,ewald,f_g_z,f_g_2d,f_r_z,f_r_2d,mssg)
  use ewald_esm
  implicit none
  integer,intent(in) :: mssg
  type(esm_2d),intent(in)  ::var2d
  type(cg2esm),intent(in)  ::cginp
  real(kind=8),intent(out) ::f_g_z(cginp%nsi),f_g_2d(2,cginp%nsi) &
                            ,f_r_z(cginp%nsi),f_r_2d(2,cginp%nsi) &
                            ,ewald
  !local
  real(kind=8) ::g,qgb,qrb,ewg,ewr
  !
  call m_ewald_init_q_qgb_qrb(var2d%sa,g,qgb,qrb)
  if(mssg .gt. 0) then
     write(6,'(A)')'%%%% Ewald method information %%%%'
     write(6,'(A,F10.5,A,F10.5)')' QGB=',qgb,' QRB=',qrb
     write(6,'(A,F10.5)')'   G=',g
  endif
  call m_ewald_gsum(cginp%nsi &
                   ,cginp%zv &
                   ,var2d%L &
                   ,var2d%Lg &
                   ,var2d%bb_2d &
                   ,var2d%sa &
                   ,var2d%sb &
                   ,var2d%asi_z &
                   ,var2d%asi_2d &
                   ,esm%z1 &
                   ,esm%v0 &
                   ,g,qgb,ewg,f_g_z,f_g_2d,mssg)
  call m_ewald_rsum(cginp%nsi &
                   ,cginp%zv &
                   ,var2d%L &
                   ,var2d%aa_2d &
                   ,var2d%Lg &
                   ,var2d%sa &
                   ,var2d%asi_z &
                   ,var2d%asi_2d &
                   ,g,qrb,ewr,f_r_z,f_r_2d,mssg)
  !
  ewald=ewg+ewr
  write(6,'(A,F18.10,A,F18.10,A,F18.10)')' EWALD=',ewald,' EWG=',ewg &
                                        ,' EWR=',ewr
  return
end subroutine m_2d_ewald

!--------Make hartree potential-------
subroutine m_2d_make_hartree_part(cginp,var2d,eht,v,rg,vg,comm)
  use fft_1d
  use m_pot
  use mpi
  use timeprof

  implicit none
  type(esm_2d),intent(in)  ::var2d
  type(cg2esm),intent(in)  ::cginp
  integer, intent(in)      ::comm
  real(kind=8),intent(out) ::eht
  complex(kind=8)          ::rg(var2d%n0),vg(var2d%n0) &
                            ,v(cginp%nrx2,cginp%nry2,cginp%nrz2)
  !local
  integer ::ig,k1,k2,ks1,ks2,err

!!!!!0307!!!!!!!!
  ! FOR MPI
  integer,allocatable:: gf_count(:),gf_displ(:)
  integer:: nproc, myrank, ierr, irnk, bkf_2d, ekf_2d, ist
  real(8)           :: peht, teht       
  complex(kind=8),allocatable   :: pv(:,:,:),tv(:,:,:)

  !!! FOR FFT !!!
  type(fft1_struct) ::fs
  real(kind=8),allocatable ::fftw1(:,:),fftw2(:,:)
  allocate(fftw1(var2d%n0,2),fftw2(var2d%n0,2),stat=err)
  call chkalloc('fft_pot','fftw1,fftw2',err)
  call fft1_init(var2d%n0,fs)

!!!!!!!!!!!!!!!!!!
  call MPI_COMM_SIZE(comm,nproc,ierr)
  call MPI_COMM_RANK(comm,myrank,ierr)

  allocate(gf_count(0:nproc-1),gf_displ(0:nproc-1),stat=ierr)
  allocate(pv(cginp%nrx2,cginp%nry2,cginp%nrz2),tv(cginp%nrx2,cginp%nry2,cginp%nrz2))
  call chkalloc('esm_2d','gf_count, gf_displ',ierr)

!poption noparallel
  gf_count = var2d%nkf_2d/nproc
!poption noparallel
  do irnk=0,mod(var2d%nkf_2d,nproc)-1
     gf_count(irnk) = gf_count(irnk)+1
  end do
  gf_displ(0) = 0
  do irnk=0,nproc-2
     gf_displ(irnk+1) = gf_displ(irnk)+gf_count(irnk)
  end do
  bkf_2d = gf_displ(myrank)+1
  ekf_2d = gf_displ(myrank)+gf_count(myrank)

  if(bkf_2d == 1) bkf_2d = 2
  !!!!!!!!!!!!!!!
  !initialize
  eht=0.d0; teht = 0.d0; peht=0.d0
  pv=0.d0;  tv=0.d0

  !---gp = 0 case
if (esm%normal .eq. 'X') then
  rg(1:var2d%n0)=v(1:var2d%n0,1,1)
else if (esm%normal .eq. 'Z') then
  rg(1:var2d%n0)=v(1,1,1:var2d%n0)
endif

  call m_pot_solve_Poisson_equation_gp0(cginp%va &
                                       ,esm%z1 &
                                       ,esm%z_r &
                                       ,esm%z_l &
                                       ,esm%nz_r &
                                       ,esm%nz_l &
                                       ,var2d%n0 &
                                       ,var2d%L &
                                       ,eht,vg,rg,fftw1,fftw2,fs) ! return v(g)

if (esm%normal .eq. 'X') then
  v(1:var2d%n0,1,1)=vg(1:var2d%n0)
else if (esm%normal .eq. 'Z') then
  v(1,1,1:var2d%n0)=vg(1:var2d%n0)
endif

!---gp!= 0 case
  do ig=bkf_2d, ekf_2d
     k1=var2d%kkg_2d(1,ig)
     k2=var2d%kkg_2d(2,ig)
     ks1=mod(var2d%n1+k1,var2d%n1)+1
     ks2=mod(var2d%n2+k2,var2d%n2)+1
if (esm%normal .eq. 'X') then
     rg(1:var2d%n0)=v(1:var2d%n0,ks1,ks2)
else if (esm%normal .eq. 'Z') then
     rg(1:var2d%n0)=v(ks1,ks2,1:var2d%n0)
endif
     call m_pot_solve_Poisson_equation(cginp%va &
                                     ,esm%z1 &
                                     ,var2d%n0 &
                                     ,var2d%L &
                                     ,var2d%akk_2d(ig) &
                                     ,peht,vg,rg,fftw1,fftw2,fs) ! return v(g)
if (esm%normal .eq. 'X') then
     pv(1:var2d%n0,ks1,ks2)=vg(1:var2d%n0)
else if (esm%normal .eq. 'Z') then
     pv(ks1,ks2,1:var2d%n0)=vg(1:var2d%n0)
endif
  enddo

  call tpf_begin(tpf_mpi)
  call MPI_ALLREDUCE(pv, tv, size(pv), MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
  call MPI_ALLREDUCE(peht,teht,1 ,MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
  call tpf_end(tpf_mpi)

!!$omp parallel do
  do ig= 2, var2d%nkf_2d
     k1=var2d%kkg_2d(1,ig)
     k2=var2d%kkg_2d(2,ig)
     ks1=mod(var2d%n1+k1,var2d%n1)+1
     ks2=mod(var2d%n2+k2,var2d%n2)+1
if (esm%normal .eq. 'X') then
     v(1:var2d%n0,ks1,ks2)=tv(1:var2d%n0,ks1,ks2)
else if (esm%normal .eq. 'Z') then
     v(ks1,ks2,1:var2d%n0)=tv(ks1,ks2,1:var2d%n0)
endif
  enddo
  eht = eht + teht

!  eht=-eht
!      ^ for TAPP
  call fft1_done(fs)
  deallocate(fftw1,fftw2,pv,tv)
  return
end subroutine m_2d_make_hartree_part
end module m_2d
