! Copyright (c) 2012, Minoru Otani <minoru.otani@aist.go.jp>
! 
! Permission is hereby granted, free of charge, to any person 
! obtaining a copy of this software and associated documentation 
! files (the "Software"), to deal in the Software without restriction, 
! including without limitation the rights to use, copy, modify, merge, 
! publish, distribute, sublicense, and/or sell copies of the Software, 
! and to permit persons to whom the Software is furnished to do so, 
! subject to the following conditions:
 
! The above copyright notice and this permission notice shall be 
! included in all copies or substantial portions of the Software.
 
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
! OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
! NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
! HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
! WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
! OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
! DEALINGS IN THE SOFTWARE.
#ifdef HPUX
!     -- HP needs the following to link erf in /lib/libm.a
$alias derf = 'erf' (%val)
$alias derfc = 'erfc' (%val)
#endif
#ifdef CRAY
#define derf bberf
#endif
#ifdef SUN
#define derf d_erf
#endif
#ifdef INTEL
#define derf bberf
#endif
module ewald_esm
  use global_esm
contains
!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!-------         calculate g & qrb &qgb         --------
!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
subroutine m_ewald_init_q_qgb_qrb(sa,g,qgb,qrb)
implicit none
real(kind=8),intent(in)  ::sa
real(kind=8),intent(out) ::g,qgb,qrb
!local
real(kind=8)           ::pi
real(kind=8),parameter ::epsg= 1.d-15 &
                        ,epsr= 1.d-15 &
                        ,safe= 1.2d0
pi=4.d0*atan(1.d0)
g=sqrt(pi/sa)*(log(epsr)/log(epsg))**(0.25d0)
qgb=sqrt(-4.d0*g**2*log(epsg))
qrb=sqrt(-log(epsr))/g
qgb=qgb*safe
qrb=qrb*safe
end subroutine m_ewald_init_q_qgb_qrb
!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!-------         calculate gsum                 --------
!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
subroutine m_ewald_gsum(nsi,zv,L,Lg,b2d,sa,sb,asi_z,asi_2d,z1,v0 &
                       ,g,qgb,ewg,f_g_z,f_g_2d, mssg)
use esm_util,only:lat_points,ordg0_2d
implicit none
integer,intent(in)       ::nsi, mssg
real(kind=8),intent(in)  ::b2d(2,2),L,Lg,asi_z(nsi),asi_2d(2,nsi) &
                          ,zv(nsi),sa,sb,z1,v0,g,qgb
real(kind=8),intent(out) ::ewg,f_g_z(nsi),f_g_2d(2,nsi)
!local
integer                  ::nam,nkg,nex,ney,it1,it2
integer,allocatable      ::kg(:,:)
real(kind=8),allocatable ::akg(:)
real(kind=8),parameter   ::gsafe=1.2d0    !for qgb,qrb
!
f_g_z(:)=0.d0; f_g_2d(:,:)=0.d0; ewg=0.d0
nam=int(lat_points(qgb,sb)*gsafe)
allocate(kg(2,nam),akg(nam))
call ordg0_2d(nam,qgb,kg,akg,b2d,nkg,nex,ney)
!
if (esm%normal .eq. 'X') then
if (mssg .gt. -1) then
   write(6,'(A,I5,A,I4,A,I4,A,I4)')' NKG=',nkg,'    NEY,NEZ=',nex,',',ney
endif
else if (esm%normal .eq. 'Z') then
if (mssg .gt. -1) then
   write(6,'(A,I5,A,I4,A,I4,A,I4)')' NKG=',nkg,'    NEX,NEY=',nex,',',ney
end if
endif
!
do it1=1,nsi
  do it2=1,nsi
    call gsume(it1,it2)
  enddo
enddo
if (esm%bctype .eq. 4) then
!!!!!!!!!!Add energy and force contribution of BC4!!!!!!!!!!
if (mssg .gt. -1) then
   write(6,'(A,F10.5)')'!Energy and force contribution of BC4 is included in EWALD'
end if
call add_bc4_energy_force
!!!!!!!!!!ewg includes rho_i*V''+1/2*C*V^2!!!!!!!!!!
endif
!
deallocate(kg,akg)
contains
  subroutine add_bc4_energy_force
  implicit none
  !local
  integer::it
  real(kind=8)::pi2,pi8,z
  pi2=2.d0*4.d0*atan(1.d0)
  pi8=8.d0*4.d0*atan(1.d0)
  do it=1,nsi
    !AA-->a.u.
    z=asi_z(it)
    if(z.gt.0.5d0)z=z-1.d0
    z=z*L
    ewg=ewg-0.5d0*zv(it)*v0*(1.d0-z/z1)
    !****************************************************************
    f_g_z(it)=f_g_z(it)+0.5d0*zv(it)*v0/z1*pi2/Lg
    !                                      ^^^^^^
    !__mo__20080425__
    !z component of the force is a.u. unit. To adapt with other 
    !components, it is divided by Lg and multiplied by pi2
    !$$$sign of the z-component force is problematic!!!
    !****************************************************************
  enddo
  ewg=ewg+0.5d0*v0**2*sa/pi8/z1
  return
  end subroutine add_bc4_energy_force

  subroutine gsume(it1,it2)
  implicit none
  integer,intent(in) ::it1,it2
  integer            ::ik
  real(kind=8)       ::z,zp,derf,derfc &
                      ,tt,ff,cc1,cc2,kk1,kk2,gp &
                      ,tt_for,kk1_for,kk2_for &
                      ,cx1_for,cx2_for &
                      ,cy1_for,cy2_for &
                      ,cz1_for,cz2_for
  real(kind=8)       ::pi,pi2
  pi =4.d0*atan(1.d0)
  pi2=2.d0*pi
  !
  !AA->a.u.
  z =asi_z(it1)
  zp=asi_z(it2)
  if(z .gt.0.5d0)z =z -1.d0
  if(zp.gt.0.5d0)zp=zp-1.d0
  z =z *L
  zp=zp*L
  !
  tt =zv(it1)*zv(it2)*pi2/sa
  kk1=(-(z-zp)*derf(g*(z-zp)) &
       -exp(-g**2*(z-zp)**2)/g/sqrt(pi))*0.5d0
if (esm%bctype .eq. 1) then
  kk2=0.d0
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
  kk2=(z1-z*zp/z1)*0.5d0
else if (esm%bctype .eq. 3) then
  kk2=(2.d0*z1-z-zp)*0.5d0
endif
  !******************************************************************
  tt_for=zv(it1)*zv(it2)*2.d0*pi2*pi2/sa
  !                               ^^^
  !surplus pi2: specifics on the TAPP (for formod)
  !******************************************************************
  kk1_for=0.5d0*derf(g*(z-zp))
if (esm%bctype .eq. 1) then
  kk2_for=0.d0
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
  kk2_for=-0.5d0*z/z1
else if (esm%bctype .eq. 3) then
  kk2_for=-0.5d0
endif
  !
  if(it1.eq.it2)then
    cc1=0.d0
    cc2=0.d0
    !
    cz1_for=0.d0
    cz2_for=0.d0
    do ik=2,nkg
      gp =sqrt(akg(ik))
      cc1=cc1+(exp(-gp*(z-zp))*derfc(gp/2.d0/g-g*(z-zp)) &
              +exp( gp*(z-zp))*derfc(gp/2.d0/g+g*(z-zp)))/4.d0/gp
if (esm%bctype .eq. 1) then
      cc2=0.d0
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
      cc2=cc2+(exp( gp*(z-zp-4.d0*z1))+exp(-gp*(z-zp+4.d0*z1)) &
              -exp( gp*(z+zp-2.d0*z1))-exp(-gp*(z+zp+2.d0*z1))) &
             /(1.d0-exp(-4.d0*gp*z1))/2.d0/gp
else if (esm%bctype .eq. 3) then
      cc2=cc2-exp(gp*(z+zp-2.d0*z1))/2.d0/gp
endif
      !
      cz1_for=0.d0 ! bare coulomb part is always 0, if it1=it2
if (esm%bctype .eq. 1) then
      cz2_for=0.d0
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
      cz2_for=cz2_for-(exp(gp*(z-zp-4.d0*z1))-exp(-gp*(z-zp+4.d0*z1)) &
                      +exp(gp*(z+zp-2.d0*z1))-exp(-gp*(z+zp+2.d0*z1))) &
                       /(1.d0-exp(-4.d0*gp*z1))/2.d0
else if (esm%bctype .eq. 3) then
      cz2_for=cz2_for-exp(gp*(z+zp-2.d0*z1))/2.d0
endif
    enddo
    ewg=ewg+tt*(kk1+kk2+cc1+cc2)
    !
    !****************************************************************
    f_g_z(it2)=f_g_z(it2) &
              +tt_for*(cz1_for+cz2_for+kk1_for+kk2_for)/Lg
    !                                                   ^^
    !z component of the force is a.u. unit. To adapt with other 
    !components, it is divided by Lg
    !****************************************************************
    goto 1
  endif
  !
  if(it1.gt.it2)then
    cc1=0.d0
    cc2=0.d0
    cx1_for=0.d0
    cy1_for=0.d0
    cz1_for=0.d0
    cx2_for=0.d0
    cy2_for=0.d0
    cz2_for=0.d0
    do ik=2,nkg
      gp =sqrt(akg(ik))
      ff =(kg(1,ik)*(asi_2d(1,it1)-asi_2d(1,it2)) &
         + kg(2,ik)*(asi_2d(2,it1)-asi_2d(2,it2))) * pi2
      cc1=cc1+cos(ff)*(exp(-gp*(z-zp))*derfc(gp/2.d0/g-g*(z-zp)) &
                      +exp( gp*(z-zp))*derfc(gp/2.d0/g+g*(z-zp)))/4.d0/gp
if (esm%bctype .eq. 1) then
      cc2=0.d0
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
      cc2=cc2+cos(ff)*(exp(gp*(z-zp-4.d0*z1))+exp(-gp*(z-zp+4.d0*z1)) &
                      -exp(gp*(z+zp-2.d0*z1))-exp(-gp*(z+zp+2.d0*z1))) &
                      /(1.d0-exp(-4.d0*gp*z1))/2.d0/gp
else if (esm%bctype .eq. 3) then
      cc2=cc2-cos(ff)*exp(gp*(z+zp-2.d0*z1))/2.d0/gp
endif
      !
      cx1_for=cx1_for+sin(ff) &
                     *(exp(-gp*(z-zp))*derfc(gp/2.d0/g-g*(z-zp)) &
                      +exp( gp*(z-zp))*derfc(gp/2.d0/g+g*(z-zp))) &
                     /4.d0/gp*kg(1,ik)
      cy1_for=cy1_for+sin(ff) &
                     *(exp(-gp*(z-zp))*derfc(gp/2.d0/g-g*(z-zp)) &
                      +exp( gp*(z-zp))*derfc(gp/2.d0/g+g*(z-zp))) &
                     /4.d0/gp*kg(2,ik)
      cz1_for=cz1_for+cos(ff) &
                     *(exp(-gp*(z-zp))*derfc(gp/2.d0/g-g*(z-zp)) &
                      -exp( gp*(z-zp))*derfc(gp/2.d0/g+g*(z-zp)))/4.d0
if (esm%bctype .eq. 1) then
      cx2_for=0.d0
      cy2_for=0.d0
      cz2_for=0.d0
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
      cx2_for=cx2_for+sin(ff)*(exp( gp*(z-zp-4.d0*z1)) &
                              +exp(-gp*(z-zp+4.d0*z1)) &
                              -exp( gp*(z+zp-2.d0*z1)) &
                              -exp(-gp*(z+zp+2.d0*z1))) &
                             /(1.d0-exp(-4.d0*gp*z1))/2.d0/gp*kg(1,ik)
      cx2_for=cx2_for+sin(ff)*(exp( gp*(z-zp-4.d0*z1)) &
                              +exp(-gp*(z-zp+4.d0*z1)) &
                              -exp( gp*(z+zp-2.d0*z1)) &
                              -exp(-gp*(z+zp+2.d0*z1))) &
                             /(1.d0-exp(-4.d0*gp*z1))/2.d0/gp*kg(2,ik)
      cz2_for=cz2_for-cos(ff)*(exp( gp*(z-zp-4.d0*z1)) &
                              -exp(-gp*(z-zp+4.d0*z1)) &
                              +exp( gp*(z+zp-2.d0*z1)) &
                              -exp(-gp*(z+zp+2.d0*z1))) &
                             /(1.d0-exp(-4.d0*gp*z1))/2.d0
else if (esm%bctype .eq. 3) then
      cx2_for=cx2_for-sin(ff)*exp(gp*(z+zp-2.d0*z1))/2.d0/gp*kg(1,ik)
      cy2_for=cy2_for-sin(ff)*exp(gp*(z+zp-2.d0*z1))/2.d0/gp*kg(2,ik)
      cz2_for=cz2_for-cos(ff)*exp(gp*(z+zp-2.d0*z1))/2.d0
endif
    enddo
    ewg=ewg+tt*(kk1+kk2+cc1+cc2)*2.d0
                               !^^^^^^for it1<it2 case
    !
    !*****************************************************************
    f_g_2d(1,it2)=f_g_2d(1,it2)+tt_for*(cx1_for+cx2_for)
    f_g_2d(2,it2)=f_g_2d(2,it2)+tt_for*(cy1_for+cy2_for)
    f_g_z   (it2)=f_g_z(it2) &
                 +tt_for*(cz1_for+cz2_for+kk1_for+kk2_for)/Lg
    !                                                      ^^
    !z component of the force is a.u. unit. To adapt with other 
    !components, it is divided by Lg
    !*****************************************************************
    goto 1
  endif
  !
  if(it1.lt.it2)then
    cx1_for=0.d0
    cy1_for=0.d0
    cz1_for=0.d0
    cx2_for=0.d0
    cy2_for=0.d0
    cz2_for=0.d0
    do ik=2,nkg
      gp=sqrt(akg(ik))
      ff=(kg(1,ik)*(asi_2d(1,it1)-asi_2d(1,it2)) &
        + kg(2,ik)*(asi_2d(2,it1)-asi_2d(2,it2)))*pi2
      !
      cx1_for=cx1_for+sin(ff) &
                     *(exp(-gp*(z-zp))*derfc(gp/2.d0/g-g*(z-zp)) &
                      +exp( gp*(z-zp))*derfc(gp/2.d0/g+g*(z-zp))) &
                     /4.d0/gp*kg(1,ik)
      cy1_for=cy1_for+sin(ff) &
                     *(exp(-gp*(z-zp))*derfc(gp/2.d0/g-g*(z-zp)) &
                      +exp( gp*(z-zp))*derfc(gp/2.d0/g+g*(z-zp))) &
                     /4.d0/gp*kg(2,ik)
      cz1_for=cz1_for+cos(ff) &
                     *(exp(-gp*(z-zp))*derfc(gp/2.d0/g-g*(z-zp)) &
                      -exp( gp*(z-zp))*derfc(gp/2.d0/g+g*(z-zp)))/4.d0
if (esm%bctype .eq. 1) then
      cx2_for=0.d0
      cy2_for=0.d0
      cz2_for=0.d0
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
      cx2_for=cx2_for+sin(ff)*(exp( gp*(z-zp-4.d0*z1)) &
                              +exp(-gp*(z-zp+4.d0*z1)) &
                              -exp( gp*(z+zp-2.d0*z1)) &
                              -exp(-gp*(z+zp+2.d0*z1))) &
                             /(1.d0-exp(-4.d0*gp*z1))/2.d0/gp*kg(1,ik)
      cy2_for=cy2_for+sin(ff)*(exp( gp*(z-zp-4.d0*z1)) &
                              +exp(-gp*(z-zp+4.d0*z1)) &
                              -exp( gp*(z+zp-2.d0*z1)) &
                              -exp(-gp*(z+zp+2.d0*z1))) &
                             /(1.d0-exp(-4.d0*gp*z1))/2.d0/gp*kg(2,ik)
      cz2_for=cz2_for-cos(ff)*(exp( gp*(z-zp-4.d0*z1)) &
                              -exp(-gp*(z-zp+4.d0*z1)) &
                              +exp( gp*(z+zp-2.d0*z1)) &
                              -exp(-gp*(z+zp+2.d0*z1))) &
                             /(1.d0-exp(-4.d0*gp*z1))/2.d0
else if (esm%bctype .eq. 3) then
      cx2_for=cx2_for-sin(ff)*exp(gp*(z+zp-2.d0*z1))/2.d0/gp*kg(1,ik)
      cy2_for=cy2_for-sin(ff)*exp(gp*(z+zp-2.d0*z1))/2.d0/gp*kg(2,ik)
      cz2_for=cz2_for-cos(ff)*exp(gp*(z+zp-2.d0*z1))/2.d0
endif
    enddo
    !*****************************************************************
    f_g_2d(1,it2)=f_g_2d(1,it2)+tt_for*(cx1_for+cx2_for)
    f_g_2d(2,it2)=f_g_2d(2,it2)+tt_for*(cy1_for+cy2_for)
    f_g_z   (it2)=f_g_z(it2) &
                 +tt_for*(cz1_for+cz2_for+kk1_for+kk2_for)/Lg
    !                                                      ^^
    !z component of the force is a.u. unit. To adapt with other 
    !components, it is divided by Lg
    !*****************************************************************
    goto 1
  endif
  !
1 return
  end subroutine gsume

end subroutine m_ewald_gsum
!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!-------         calculate rsum                 --------
!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
subroutine m_ewald_rsum(nsi,zv,L,a2d,Lg,sa,asi_z,asi_2d,g,qrb,ewr &
                       ,f_r_z,f_r_2d,mssg)
use esm_util,only:lat_points,ordg0_2d
implicit none
integer,intent(in) ::nsi,mssg
real(kind=8),intent(in)  ::a2d(2,2),L,Lg,asi_z(nsi),asi_2d(2,nsi) &
                          ,zv(nsi),sa,g,qrb
real(kind=8),intent(out) ::ewr,f_r_z(nsi),f_r_2d(2,nsi)
!local
integer                  ::nam,nkr,nex,ney,it1,it2
integer,allocatable      ::kr(:,:)
real(kind=8),allocatable ::akr(:)
real(kind=8),parameter   ::gsafe=1.2d0    !for qgb,qrb
!
f_r_z(:)=0.d0; f_r_2d(:,:)=0.d0; ewr=0.d0
nam=int(lat_points(qrb,sa)*gsafe)
allocate(kr(2,nam),akr(nam))
call ordg0_2d(nam,qrb,kr,akr,a2d,nkr,nex,ney)
!
if (esm%normal .eq. 'X') then
if (mssg .gt. -1) then
   write(6,'(A,I5,A,I4,A,I4,A,I4)')' NKR=',nkr,'    NEY,NEZ=',nex,',',ney
end if
else if (esm%normal .eq. 'Z') then
if (mssg .gt. -1) then
   write(6,'(A,I5,A,I4,A,I4,A,I4)')' NKR=',nkr,'    NEX,NEY=',nex,',',ney
end if
endif
!
do it1=1,nsi
  do it2=1,nsi
    if(it1.eq.it2)then
      call rsume0(it1,it2)
    else
      call rsume1(it1,it2)
    endif
  enddo
enddo
!
deallocate(kr,akr)
contains
  subroutine rsume0(it1,it2)
  implicit none
  integer,intent(in) ::it1,it2
  integer            ::ir
  real(kind=8)       ::derfc
  real(kind=8)       ::dd,tt,ff,cc,kk,a,b,x,y,z,zp
  real(kind=8)       ::pi
  pi=4.d0*atan(1.d0)
  !AA->a.u.
  z =asi_z(it1)
  zp=asi_z(it2)
  if(z .gt.0.5d0)z =z -1.d0
  if(zp.gt.0.5d0)zp=zp-1.d0
  z =z *L
  zp=zp*L
  !
  tt=zv(it1)*zv(it2)
  cc=0.d0
  do ir=2,nkr
    a=(dble(kr(1,ir))+asi_2d(1,it1)-asi_2d(1,it2))
    b=(dble(kr(2,ir))+asi_2d(2,it1)-asi_2d(2,it2))
    x=a*a2d(1,1)+b*a2d(1,2)
    y=a*a2d(2,1)+b*a2d(2,2)
    ff=sqrt(x*x+y*y+(z-zp)**2)
    cc=cc+derfc(g*ff)/ff*0.5d0
  enddo
  ewr=ewr+tt*(cc-g/sqrt(pi))
  return
  end subroutine rsume0

  subroutine rsume1(it1,it2)
  implicit none
  integer,intent(in) ::it1,it2
  integer            ::ir
  real(kind=8)       ::derfc,cc_for,xa,xb
  real(kind=8)       ::gg,tt,ff,cc,kk,a,b,x,y,z,zp &
                      ,cz1_for,cz2_for,cx1_for &
                      ,cx2_for,cy1_for,cy2_for
  real(kind=8)       ::pi,pi2
  pi =4.d0*atan(1.d0)
  pi2=2.d0*pi
  !AA->a.u.
  if(it1.gt.it2)then
    z =asi_z(it1)
    zp=asi_z(it2)
    if(z .gt.0.5d0)z =z -1.d0
    if(zp.gt.0.5d0)zp=zp-1.d0
    z =z *L
    zp=zp*L
    !
    tt=zv(it1)*zv(it2)
    cc=0.d0
    !
    cx1_for=0.d0
    cy1_for=0.d0
    cz1_for=0.d0
    cx2_for=0.d0
    cy2_for=0.d0
    cz2_for=0.d0
    do ir=1,nkr
      a =(dble(kr(1,ir))+asi_2d(1,it1)-asi_2d(1,it2))
      b =(dble(kr(2,ir))+asi_2d(2,it1)-asi_2d(2,it2))
      x=a*a2d(1,1)+b*a2d(1,2)
      y=a*a2d(2,1)+b*a2d(2,2)
      xa=x*a2d(1,1)+y*a2d(2,1)
      xb=x*a2d(1,2)+y*a2d(2,2)
      ff=sqrt(x*x+y*y+(z-zp)**2)
      cc=cc+derfc(g*ff)/ff*0.5d0
      !
      cx1_for=cx1_for+derfc(g*ff)/(ff*ff*ff)*xa
      cy1_for=cy1_for+derfc(g*ff)/(ff*ff*ff)*xb
      cz1_for=cz1_for+derfc(g*ff)/(ff*ff*ff)*(z-zp)
      cx2_for=cx2_for+2.d0 * g*exp(-g*g*ff*ff) & 
                     /sqrt(pi)/(ff*ff)*xa
      cy2_for=cy2_for+2.d0 * g*exp(-g*g*ff*ff) & 
                     /sqrt(pi)/(ff*ff)*xb
      cz2_for=cz2_for+2.d0 * g*exp(-g*g*ff*ff) & 
                     /sqrt(pi)/(ff*ff)*(z-zp)
    enddo
    ewr=ewr+tt*cc*2.d0
    !*****************************************************************
    f_r_2d(1,it2)=f_r_2d(1,it2)+tt*(cx1_for+cx2_for)
    f_r_2d(2,it2)=f_r_2d(2,it2)+tt*(cy1_for+cy2_for)
    f_r_z   (it2)=f_r_z(it2)+tt*(cz1_for+cz2_for)*pi2/Lg
    !                                             ^^^^^^
    !z component of the force is a.u. unit. To adapt with other 
    !components, it is divided by Lg and multiplied by pi2
    !*****************************************************************
    !
  endif
  !
  if(it1.lt.it2)then
    z =asi_z(it1)
    zp=asi_z(it2)
    if(z .gt.0.5d0)z =z -1.d0
    if(zp.gt.0.5d0)zp=zp-1.d0
    z =z *L
    zp=zp*L
    !
    tt=zv(it1)*zv(it2)
    !
    cx1_for=0.d0
    cy1_for=0.d0
    cz1_for=0.d0
    cx2_for=0.d0
    cy2_for=0.d0
    cz2_for=0.d0
    do ir=1,nkr
      a =(dble(kr(1,ir))+asi_2d(1,it1)-asi_2d(1,it2))
      b =(dble(kr(2,ir))+asi_2d(2,it1)-asi_2d(2,it2))
      x=a*a2d(1,1)+b*a2d(1,2)
      y=a*a2d(2,1)+b*a2d(2,2)
      xa=x*a2d(1,1)+y*a2d(2,1)
      xb=x*a2d(1,2)+y*a2d(2,2)
      ff=sqrt(x*x+y*y+(z-zp)**2)
      cc=cc+derfc(g*ff)/ff*0.5d0
      !
      cx1_for=cx1_for+derfc(g*ff)/(ff*ff*ff)*xa
      cy1_for=cy1_for+derfc(g*ff)/(ff*ff*ff)*xb
      cz1_for=cz1_for+derfc(g*ff)/(ff*ff*ff)*(z-zp)
      cx2_for=cx2_for+2.d0 * g*exp(-g*g*ff*ff) & 
                     /sqrt(pi)/(ff*ff)*xa
      cy2_for=cy2_for+2.d0 * g*exp(-g*g*ff*ff) & 
                     /sqrt(pi)/(ff*ff)*xb
      cz2_for=cz2_for+2.d0 * g*exp(-g*g*ff*ff) & 
                     /sqrt(pi)/(ff*ff)*(z-zp)
    enddo
    !*****************************************************************
    f_r_2d(1,it2)=f_r_2d(1,it2)+tt*(cx1_for+cx2_for)
    f_r_2d(2,it2)=f_r_2d(2,it2)+tt*(cy1_for+cy2_for)
    f_r_z   (it2)=f_r_z(it2)+tt*(cz1_for+cz2_for)*pi2/Lg
    !                                             ^^^^^^
    !x component of the force is a.u. unit. To adapt with other 
    !components, it is divided by Lg and multiplied by pi2
    !*****************************************************************
    !
  endif
  !
  return
  end subroutine rsume1
end subroutine m_ewald_rsum
end module ewald_esm
