!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2017 Yoshihide Yoshimoto
!
#include "fftgrid.h"

module lpot_init
contains
  subroutine loadlpt(nspin,lpt,nvx,nvy,nvz,spaxsi,mssg)
    use iounit
    use mpi
    implicit none
    integer:: nspin,nvx,nvy,nvz,mssg
    real(kind=8),intent(in):: spaxsi(3)
    complex(kind=8):: lpt(-nvx:nvx,-nvy:nvy,-nvz:nvz,nspin)
    !
    character(len=8),parameter:: magic = 'TAPP:lpt'
    !
    complex(kind=8),allocatable:: lptin(:,:,:,:)
    real(kind=8):: bbi(3,3)
    integer:: nspini, nvxi, nvyi, nvzi
    integer:: i, j, k
    integer:: is
    integer:: ierr
    character(len=8):: sign
    real(kind=8):: pol, spaxs(3)

    call xrewind(nflpt)
    read(nflpt,iostat=ierr) sign
    if (ierr.ne.0) then
       write(6,'(1x,a)') 'error in opening local potential'
       call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
       stop
    end if
    if (sign .ne. magic) then
       write(6,*)'bad initial local potential was specified.'
       call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
       stop
    end if
    read(nflpt) bbi, nspini
    if (mssg.gt.-1) then
       write(6,*)'bbi = '
       write(6,'(e13.6, '', '', e13.6, '', '', e13.6)') (bbi(i,1),i=1,3)
       write(6,'(e13.6, '', '', e13.6, '', '', e13.6)') (bbi(i,2),i=1,3)
       write(6,'(e13.6, '', '', e13.6, '', '', e13.6)') (bbi(i,3),i=1,3)
       write(6,'(1x,a,i2)') 'nspini = ',nspini
    end if
    read(nflpt) nvxi, nvyi, nvzi
    if (mssg.gt.-1) then
       write(6,*)'nv[xyz]i =', nvxi, nvyi, nvzi
    end if
    allocate(lptin(-nvxi:nvxi,-nvyi:nvyi,-nvzi:nvzi,nspini), stat=ierr)
    call chkalloc('initlpt','lptin',ierr)
    read(nflpt) lptin
    lpt = (0.0d0, 0.0d0)
    if (nspini .eq. 1) then
       do k=-min(nvz,nvzi),min(nvz,nvzi)
          do j=-min(nvy,nvyi),min(nvy,nvyi)
             do i=-min(nvx,nvxi),min(nvx,nvxi)
                lpt(i,j,k,1) = lptin(i,j,k,1)
             end do
          end do
       end do
       if (nspin.eq.2) then
          lpt(:,:,:,2) = lpt(:,:,:,1)
       else if (nspin.eq.4) then
          lpt(:,:,:,2:4) = 0.0d0
       end if
    else if (nspini.eq.2) then
       if (nspin.le.2) then
          do is=1,nspin
             do k=-min(nvz,nvzi),min(nvz,nvzi)
                do j=-min(nvy,nvyi),min(nvy,nvyi)
                   do i=-min(nvx,nvxi),min(nvx,nvxi)
                      lpt(i,j,k,is) = lptin(i,j,k,is)
                   end do
                end do
             end do
          end do
       else
          if (maxval(abs(spaxsi)).gt.0.0d0) then
             spaxs = spaxsi
          else
             spaxs(1) = 0.0d0
             spaxs(2) = 0.0d0
             spaxs(3) = 1.0d0
          end if
          do k=-min(nvz,nvzi),min(nvz,nvzi)
             do j=-min(nvy,nvyi),min(nvy,nvyi)
                do i=-min(nvx,nvxi),min(nvx,nvxi)
                   lpt(i,j,k,1) = 0.5d0*(lptin(i,j,k,1)+lptin(i,j,k,2))
                   pol = 0.5d0*(lptin(i,j,k,1)-lptin(i,j,k,2))
                   lpt(i,j,k,2) = spaxs(1)*pol
                   lpt(i,j,k,3) = spaxs(2)*pol
                   lpt(i,j,k,4) = spaxs(3)*pol
                end do
             end do
          end do
       end if
    else
       if (nspin.eq.1 .or. nspin.eq.4) then
          do is=1,nspin
             do k=-min(nvz,nvzi),min(nvz,nvzi)
                do j=-min(nvy,nvyi),min(nvy,nvyi)
                   do i=-min(nvx,nvxi),min(nvx,nvxi)
                      lpt(i,j,k,is) = lptin(i,j,k,is)
                   end do
                end do
             end do
          end do
       else if (nspin.eq.2) then
          do k=-min(nvz,nvzi),min(nvz,nvzi)
             do j=-min(nvy,nvyi),min(nvy,nvyi)
                do i=-min(nvx,nvxi),min(nvx,nvxi)
                   pol = lptin(i,j,k,2)**2+lptin(i,j,k,3)**2+lptin(i,j,k,4)**2
                   pol = sqrt(pol)
                   lpt(i,j,k,1) = lptin(i,j,k,1)+0.5d0*pol
                   lpt(i,j,k,2) = lptin(i,j,k,1)-0.5d0*pol
                end do
             end do
          end do
       end if
    end if
    deallocate(lptin)
    if (mssg.gt.-1) then
       write(6,*)'inputed the initial local potential.'
    end if
    return
  end subroutine loadlpt

  subroutine initlpt(nspin,ndens,spaxs,lpt,aa,nkd,kd,asi,nsi, &
       iopt,udirchr,mgs,bspotfc,bspots,xctp,mssg,comm)
    use util
    use bsplpot
    use ioniclpot
    use atomicchrg
    use potihxc
    use meshgrid
    use converter
    use mpi
    implicit none
    integer:: nspin,ndens,nkd,nsi,iopt,mssg,comm
    real(kind=8),pointer:: udirchr(:,:)
    real(kind=8):: spaxs(3)
    integer,intent(in):: xctp
    type (meshgrid_struct):: mgs
    real(kind=8):: aa(3,3)
    real(kind=8):: asi(3,nsi)
    integer:: kd(nsi)
    complex(kind=8):: lpt(-mgs%nvx:mgs%nvx,-mgs%nvy:mgs%nvy, &
         -mgs%nvz:mgs%nvz,nspin)
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    !
    real(kind=8),allocatable:: pcchr(:,:,:)
    real(kind=8),allocatable:: rho(:,:,:,:), ztmp(:,:,:,:,:)
    complex(kind=8),allocatable:: lption(:)
    integer:: is
    integer:: nrx2, nry2, nrz2
    integer:: ig, k1, k2, k3
    real(kind=8),allocatable:: aff(:)
    real(kind=8):: eht,exa,exc,evxc(0:6)
    real(kind=8):: bb(3,3), va
    integer:: myrank,ierr

    integer:: nkfxa,nvx,nvy,nvz,nvxyz
!soption contarea(kf)
    integer,pointer:: kf(:,:)
    real(kind=8):: vpol(3),vapol(3),aspol
    integer:: nrxyz

    va = cellvol(aa)
    call recp_uvec(aa,bb)

    nkfxa = mgs%nkfxa
    nvx = mgs%nvx
    nvy = mgs%nvy
    nvz = mgs%nvz
    nvxyz = mgs%nvxyz
    kf => mgs%kf

    if (iopt .eq. 1) then
       call MPI_COMM_RANK(comm,myrank,ierr)
       if (myrank.eq.0) then
          call loadlpt(nspin*ndens,lpt,nvx,nvy,nvz,spaxs,mssg)
       end if
       call MPI_BCAST(lpt,nvxyz*nspin*ndens,MPI_DOUBLE_COMPLEX,0,comm,ierr)
       return
    end if

    allocate(aff(nkfxa),stat=ierr)
    call chkalloc('initlpt','aff',ierr)
    call kf2aff(bb,nkfxa,kf,aff)

    nrx2 = 2*algn235(nvx+1,1)
    nry2 = 2*algn235(nvy+1,1)
    nrz2 = 2*algn235(nvz+1,1)

    allocate(lption(nkfxa), pcchr(nrx2,nry2,nrz2), stat=ierr)
    call chkalloc('initlpt','lption, pcchr',ierr)
    call mklption(lption,pcchr,nrx2,nry2,nrz2, &
         nkd,kd,asi,nsi,va,nkfxa,aff,kf,bspotfc,bspots,comm,mssg)
    if (iopt .ne. 2) then
       ! from the ionic local potential.
       lpt = 0.0d0
       do ig=1,nkfxa
          k1 = kf(1,ig)
          k2 = kf(2,ig)
          k3 = kf(3,ig)
!POPTION LOOP(2)
          do is=1,nspin
             lpt(k1,k2,k3,is) = lption(ig)
          end do
       end do
       if (mssg.gt.-1) then
          write(6,*) 'Initial local potential from ionic local potential'
       end if
       deallocate(lption, pcchr)
       return
    end if

    ! from the atomic charge of solps.
    allocate(rho(nrx2,nry2,nrz2,nspin*ndens),stat=ierr)
    call chkalloc('initlpt','rho',ierr)

    allocate(FFT_A3M_DEFSZ(ztmp,nrx2,nry2,nrz2,nspin*ndens),stat=ierr)
    call chkalloc('initlpt','ztmp',ierr)
    call mkatmchrg(bspotfc,nkd,bspots,nkfxa,kf,aff,va, &
         nsi,asi,kd,nrx2,nry2,nrz2,nspin,ndens,udirchr,ztmp)

    rho(:,:,:,:) = FFT_A3M_RL(ztmp,:,:,:,:)
    deallocate(ztmp)

    if (ndens.eq.4) then
       nrxyz = nrx2*nry2*nrz2
       call calchr2(rho(:,:,:,2),nrxyz,va,vpol(1),vapol(1))
       call calchr2(rho(:,:,:,3),nrxyz,va,vpol(2),vapol(2))
       call calchr2(rho(:,:,:,4),nrxyz,va,vpol(3),vapol(3))
       call calpol(rho(:,:,:,2:4),nrxyz,va,aspol)
       if (mssg.gt.-1) then
          write(6,'(1x,a,3(1x,f15.8))') 'init spin pol vector =', vpol
          write(6,'(1x,a,3(1x,f15.8))') &
               'init abs spin pol vector =', vapol
          write(6,'(1x,a,f15.8)') 'init abs spin polarization = ', aspol
       end if
    end if

    ! pass rho(:,:,:,1)
    call ypotp(nspin,ndens,spaxs,pcchr,rho,lpt,va,bb,eht,exa,exc &
         ,evxc,aff,kf,lption &
         ,nvx,nvy,nvz,nvxyz,nrx2,nry2,nrz2,nkfxa,xctp,mssg,comm)
    if (mssg.gt.-1) then
       write(6,*) 'Initial local potential from atomic charge'
    end if
    deallocate(aff)
    deallocate(rho)
    deallocate(lption,pcchr)
    return
  end subroutine initlpt
end module lpot_init
