!
! Copyright (c) 2000-2017 Yoshihide Yoshimoto
!
#include "fftgrid.h"

module moldyn
  private kinengnucl,tryrhowfn,zerotfor,sternpsi
  private debugscf
contains
  !
  ! Velocity Velret code with Berendsen Thermostat
  ! REF: DL_POLY_4 User Manual, I.T. Todorov and W. Smith
  !
  subroutine mdvelver(deltm,imdstp,emdstp,imtrw,casi,cavl,cforce &
       ,syskbt,tauinv &
       ,iexit,iopt5,lpt,rho,vxcr,pby,dp &
       ,qbtp,ismode,ekbt,nband,laprj,nsi,nkd,rg,pg,ng,nnp &
       ,znext,dspin,ncomp,nspin,ndens,nw,nb1,nb2 &
       ,aa,kd,mass,ee,wgtk,dd &
       ,scfcs,mgs,bspotfc,bspots,ifinvsym,axetp &
       ,udirchr,msgcycl,myrank,nproc,comm)
    use types
    use bsplpot
    use iounit
    use scfloop
    use engdffr
    use util
    use meshgrid
    use ewald_sum
    implicit none
    integer:: imdstp,emdstp,imtrw
    integer:: iexit,iopt5
    integer:: ismode,nband,laprj,nsi,nkd,ng,nnp
    integer:: ncomp,nspin,ndens,nw,nb1,nb2
    integer:: msgcycl
    !      
    type (scfctrl_struct):: scfcs
    type (meshgrid_struct):: mgs
    !
    real(kind=8):: znext, dspin
    real(kind=8):: deltm
    real(kind=8):: ekbt
    !
    real(kind=8):: rho(mgs%nrxyz,nspin*ndens),vxcr(mgs%nrxyz,nspin*ndens)
    complex(kind=8):: pby(laprj,nsi,nw,mgs%nks,nspin)
    complex(kind=8):: lpt(mgs%nvxyz,nspin*ndens)
    !
    real(kind=8):: aa(3,3)
    real(kind=8),dimension(3,nsi):: casi,cavl,cforce
    real(kind=8),intent(in):: syskbt,tauinv
    integer:: kd(nsi)
    real(kind=8):: ee(nband,mgs%nks,nspin)
    real(kind=8):: wgtk(nb1:nb2,mgs%nks,nspin)
    real(kind=8):: qbtp(3)
    complex(kind=8):: dd(laprj,laprj,nsi,nspin)
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    real(kind=8):: mass(nkd)
    integer:: rg(3,3,ng),pg(3,ng)
    type (wfn_for_k):: dp(mgs%nks,nspin)
    logical:: ifinvsym
    integer:: axetp
    real(kind=8),pointer:: udirchr(:,:)
    integer:: myrank,nproc,comm
    !
    real(kind=8):: scfexit(nscfexit)
    real(kind=8):: etotal,setotal(6),settr
    real(kind=8):: ewald,sewald(6)
    real(kind=8):: ef(nspin),eetrp
    real(kind=8):: eknuc
    type(elec_energy):: eleng
    type(elec_senergy):: elseng
    !
    real(kind=8),dimension(:,:),allocatable:: &
         force,dforce,for,dfor,forel,dforel
    real(kind=8),dimension(:,:),allocatable:: asir,pcavl,pcforce
    real(kind=8),allocatable:: pe(:,:,:)
    real(kind=8),allocatable:: crho(:,:)

    real(kind=8):: va
    integer:: istep,iex,it,ikd,ib,is,ierr
    real(kind=8):: brdsn_sigma, brdsn_chi

    integer:: mssg

    integer:: nks,nkr,nrxyz

    integer:: icycl,bkm,ekm,iks
    type (wfn_for_k):: pdp(mgs%nks,nspin),odp(mgs%nks,nspin)
    real(kind=8),allocatable:: prho(:,:),orho(:,:)
    real(kind=8),dimension(:,:),allocatable:: pcasi,ocasi,acasi

    va = cellvol(aa)
    nks = mgs%nks
    nkr = mgs%nkr
    nrxyz = mgs%nrxyz

    brdsn_sigma = (3.0d0*nsi-3.0d0)*0.5d0*syskbt

    allocate(pe(nkr,nb1:nb2,nspin), &
         asir(3,nsi),pcavl(3,nsi),pcforce(3,nsi), &
         force(3,nsi),dforce(3,nsi), &
         for(3,nsi),dfor(3,nsi),forel(3,nsi),dforel(3,nsi), &
         crho(nrxyz,nspin*ndens),stat=ierr)
    call chkalloc('mdvelver','tmp',ierr)

    allocate(prho(nrxyz,nspin*ndens),orho(nrxyz,nspin*ndens),stat=ierr)
    call chkalloc('mdvelver','tmp 2',ierr)
    do is=1,nspin
       do iks=1,nks
          bkm = mgs%g_displ(myrank,iks)+1
          ekm = mgs%g_displ(myrank,iks)+mgs%g_count(myrank,iks)
          allocate(pdp(iks,is)%wf(ncomp,bkm:ekm,nw), &
               odp(iks,is)%wf(ncomp,bkm:ekm,nw),stat=ierr)
          call chkalloc('mdvelver','tmp 3',ierr)
       end do
    end do
    allocate(pcasi(3,nsi),ocasi(3,nsi),acasi(3,nsi),stat=ierr)
    call chkalloc('mdvelver','tmp 4',ierr)

    icycl = 0

    if (iopt5 .ne. 1) then
       if (myrank.eq.0) then
          mssg = 0
       else
          mssg = -2
       end if
       !
       call cart2cellr(aa,nsi,casi,asir)

       call ewalds(asir,aa,nsi,nkd,kd &
            ,for,dfor,ewald,sewald,bspots,myrank,nproc,comm,mssg)
       call prntime('EWALDS ',mssg)
       call vbscf(dp,rho,pby,lpt,vxcr,crho &
            ,qbtp,ismode,iex,scfexit,ekbt,ewald,eleng &
            ,ef,eetrp,nband,laprj,nsi,nkd,ng &
            ,znext,dspin,ncomp,nspin,ndens,nw,nb1,nb2 &
            ,aa,asir,kd,ee,pe,wgtk,dd &
            ,scfcs,mgs,bspotfc,bspots,ifinvsym &
            ,mssg,myrank,nproc,comm)
       !
       if ( (iex .eq. -1) .or. (iex .eq. -2) ) then
          if (mssg.gt.-1) then
             if (iex .eq. -2) then
                write(6,*) 'vbscf has exited because of uptime.'
             else
                write(6,*) 'vbscf does not converged'
                call debugscf(nband,nks,nspin,ee,nw,nb1,nb2,wgtk, &
                     nkr,pe,nrxyz,rho,va)
             end if
          end if

          imdstp = 0

          if (iex .eq. -2) then
             iexit = -2
          else
             iexit = -1
          end if

          deallocate(prho,orho)
          do is=1,nspin
             do iks=1,nks
                deallocate(pdp(iks,is)%wf,odp(iks,is)%wf)
             end do
          end do

          deallocate(pe,asir,pcavl,pcforce,force,dforce, &
               for,dfor,forel,dforel,crho, &
               pcasi,ocasi,acasi)

          return
       else if (iex .le. -3) then
          write(6,*) 'unknown result code from vbscf'
          write(6,*) 'iex = ',iex
          call MPI_FINALIZE(ierr)
          stop
       end if
       if (mssg.gt.-1) then
          write(6,'(1x,a)') 'IB, PE(IB,1) = '
          do ib=nb1,nb2
             write(6,'(1x,i4,1x,e15.7,1x,e15.7)') &
                  ib, (pe(1,ib,is),is=1,nspin)
          end do
       end if
       !
       call vbfor(dp,lpt,rho,vxcr,crho &
            ,wgtk,eleng,forel,dforel,for,dfor,force,dforce &
            ,qbtp,elseng &
            ,aa,asir,kd,dd,ee,nband,laprj &
            ,rg,pg,ng,nnp,ncomp,nspin,ndens,nsi,nkd &
            ,nb1,nb2,mgs,bspotfc,bspots,scfcs%xctp,axetp &
            ,myrank,nproc,comm)
       etotal=ewald + eleng%total
       setotal=sewald + elseng%total
       settr=setotal(1) + setotal(2) + setotal(3)
       call cell2cart(aa,nsi,dforce,cforce)
       !
       ! notify vbscf() to use charge from tryrhowfn()
       !
       scfcs%ifchscf = 2

       eknuc = 0.0d0
       do it=1,nsi
          ikd = kd(it)
          eknuc = eknuc + mass(ikd)*( cavl(1,it)*cavl(1,it) &
               + cavl(2,it)*cavl(2,it) + cavl(3,it)*cavl(3,it) )
       end do
       eknuc = 0.5d0*eknuc
       !
       if (mssg.gt.-1) then
          write(6,'(1x,a,i5)') 'INITIAL STEP'
          write(6,'(1x,a,f10.2)') 'TIME FORCE = ', xetime(1)
          write(6,'(1x,a,e19.10,1x,e19.10)') &
               'FERMI     ENERGY = ', (ef(is),is=1,nspin)
          write(6,'(1x,a,e19.10)') 'EWALD            = ', ewald
          call printeleng(eleng)
          write(6,'(1x,a,e19.10)') 'TOTAL     ENERGY = ', etotal
          write(6,'(1x,a,e19.10)') 'NUCL-KIN  ENERGY = ', eknuc
       end if

       if (myrank.eq.0) then
          write(NFMDLOG) .false., 0.0d0
          write(NFMDLOG) etotal, 0.0d0   ! etotal, order
          write(NFMDLOG) etotal          ! enthalpy
          write(NFMDLOG) eknuc
          write(NFMDLOG) va              ! sysvol
          write(NFMDLOG) 0.0d0           ! stern_cons
          write(NFMDLOG) (/0.0d0,0.0d0,0.0d0/), (/0.0d0,0.0d0,0.0d0/) ! totmom, ctomas
          write(NFMDLOG) 0.0d0, 0.0d0, 0.0d0   ! zeta,eta,sigma
          write(NFMDLOG) kd,casi
          write(NFMDLOG) -settr/(3.0d0*va), setotal
          write(NFMDLOG) cforce
          write(NFMDLOG) cavl
       end if
    end if

    do istep = imdstp, emdstp
       !
       ! save for interruption
       !
       pcavl = cavl
       pcforce = cforce

       !
       ! save for interruption and tryrhowfn
       !
       acasi = ocasi
       ocasi = pcasi
       pcasi = casi

       do it=1,nsi
          ikd = kd(it)
          cavl(:,it) = pcavl(:,it) + 0.5d0/mass(ikd)*deltm*cforce(:,it)
          casi(:,it) = pcasi(:,it) + cavl(:,it)*deltm
       end do

       if (myrank.eq.0 .and. mod(istep,msgcycl).eq.0) then
          mssg = 0
       else
          mssg = -2
       end if

       if (scfcs%ifchscf.eq.2) then
          call tryrhowfn(nsi,casi,pcasi,ocasi,acasi,nspin,ndens,dp,pdp,odp, &
               rho,prho,orho,va,va, &
               kd,aa,bspotfc,nkd,mgs,bspots,udirchr,icycl,imtrw)
       end if

       call cart2cellr(aa,nsi,casi,asir)

       call ewalds(asir,aa,nsi,nkd,kd &
            ,for,dfor,ewald,sewald,bspots,myrank,nproc,comm,mssg)
       call prntime('EWALDS ',mssg)
       call vbscf(dp,rho,pby,lpt,vxcr,crho &
            ,qbtp,ismode,iex,scfexit,ekbt,ewald,eleng &
            ,ef,eetrp,nband,laprj,nsi,nkd,ng &
            ,znext,dspin,ncomp,nspin,ndens,nw,nb1,nb2 &
            ,aa,asir,kd,ee,pe,wgtk,dd &
            ,scfcs,mgs,bspotfc,bspots,ifinvsym &
            ,mssg,myrank,nproc,comm)
       !
       if ( (iex .eq. -1) .or. (iex .eq. -2) ) then
          if (mssg.gt.-1) then
             if (iex .eq. -2) then
                write(6,*) 'vbscf has exited because of uptime.'
             else
                write(6,*) 'vbscf does not converged'
                call debugscf(nband,nks,nspin,ee,nw,nb1,nb2,wgtk, &
                     nkr,pe,nrxyz,rho,va)
             end if
          end if

          !
          ! revert to the entry
          !
          cavl = pcavl
          cforce = pcforce
          casi = pcasi

          imdstp = istep

          if (iex .eq. -2) then
             iexit = -2
          else
             iexit = -1
          end if

          deallocate(prho,orho)
          do is=1,nspin
             do iks=1,nks
                deallocate(pdp(iks,is)%wf,odp(iks,is)%wf)
             end do
          end do

          deallocate(pe,asir,pcavl,pcforce,force,dforce, &
               for,dfor,forel,dforel,crho, &
               pcasi,ocasi,acasi)

          return
       else if (iex .le. -3) then
          write(6,*) 'unknown result code from vbscf'
          write(6,*) 'iex = ',iex
          call MPI_FINALIZE(ierr)
          stop
       end if
       if (mssg.gt.-1) then
          write(6,'(1x,a)') 'IB, PE(IB,1) = '
          do ib=nb1,nb2
             write(6,'(1x,i4,1x,e15.7,1x,e15.7)') &
                  ib, (pe(1,ib,is),is=1,nspin)
          end do
       end if
       !
       call vbfor(dp,lpt,rho,vxcr,crho &
            ,wgtk,eleng,forel,dforel,for,dfor,force,dforce &
            ,qbtp,elseng &
            ,aa,asir,kd,dd,ee,nband,laprj &
            ,rg,pg,ng,nnp,ncomp,nspin,ndens,nsi,nkd &
            ,nb1,nb2,mgs,bspotfc,bspots,scfcs%xctp,axetp &
            ,myrank,nproc,comm)
       etotal=ewald + eleng%total
       setotal=sewald + elseng%total
       settr=setotal(1) + setotal(2) + setotal(3)
       call cell2cart(aa,nsi,dforce,cforce)
       !
       ! notify vbscf() to use charge from tryrhowfn()
       !
       scfcs%ifchscf = 2
       !
       eknuc = 0.0d0
       do it=1,nsi
          ikd = kd(it)
          cavl(:,it) = cavl(:,it) + 0.5d0/mass(ikd)*deltm*cforce(:,it)

          eknuc = eknuc + mass(ikd)*( cavl(1,it)*cavl(1,it) &
               + cavl(2,it)*cavl(2,it) + cavl(3,it)*cavl(3,it) )
       end do
       eknuc = 0.5d0*eknuc

       !
       ! Berendsen Thermostat
       ! REF: DL_POLY_4 User Manual, I.T. Todorov and W. Smith
       !
       if (tauinv.gt.0.0d0) then
          brdsn_chi = sqrt(1.0d0 + deltm*tauinv*(brdsn_sigma/eknuc-1.0d0))
          do it=1,nsi
             cavl(:,it) = cavl(:,it)*brdsn_chi
          end do
          eknuc = eknuc*(brdsn_chi**2)
       end if

       !
       if (mssg.gt.-1) then
          write(6,'(1x,a,i5)') 'STEP = ', istep
          write(6,'(1x,a,f10.2)') 'TIME FORCE = ', xetime(1)
          write(6,'(1x,a,e19.10,1x,e19.10)') &
               'FERMI     ENERGY = ', (ef(is),is=1,nspin)
          write(6,'(1x,a,e19.10)') 'EWALD            = ', ewald
          call printeleng(eleng)
          write(6,'(1x,a,e19.10)') 'TOTAL     ENERGY = ', etotal
          write(6,'(1x,a,e19.10)') 'NUCL-KIN  ENERGY = ', eknuc
       end if

       if (myrank.eq.0) then
          write(NFMDLOG) .false., istep*deltm
          write(NFMDLOG) etotal, 0.0d0   ! etotal, order
          write(NFMDLOG) etotal          ! enthalpy
          write(NFMDLOG) eknuc
          write(NFMDLOG) va              ! sysvol
          write(NFMDLOG) 0.0d0           ! stern_cons
          write(NFMDLOG) (/0.0d0,0.0d0,0.0d0/), (/0.0d0,0.0d0,0.0d0/) ! totmom, ctomas
          write(NFMDLOG) 0.0d0, 0.0d0, 0.0d0   ! zeta,eta,sigma
          write(NFMDLOG) kd,casi
          write(NFMDLOG) -settr/(3.0d0*va), setotal
          write(NFMDLOG) cforce
          write(NFMDLOG) cavl
       end if
    end do
    imdstp = istep
    iexit = 0

    deallocate(prho,orho)
    do is=1,nspin
       do iks=1,nks
          deallocate(pdp(iks,is)%wf,odp(iks,is)%wf)
       end do
    end do

    deallocate(pe,asir,pcavl,pcforce,force,dforce, &
         for,dfor,forel,dforel,crho, &
         pcasi,ocasi,acasi)
    !
    return
  end subroutine mdvelver
  !
  ! REF: H.A. Stern, J. Compt. Chem, Vol. 25, No. 5, 749-761 (2004)
  !
  subroutine mdnpt_st(deltm,imdstp,emdstp,imtrw,syskbt,syspext &
       ,omzeta,ometa,casi,cavl,sysvol &
       ,zeta,eta,sigma,fzeta,feta,cforce,mbmts &
       ,iexit,iopt5,lpt,rho,vxcr,pby,dp &
       ,qbtp,ismode,ekbt,nband,laprj,nsi,nkd &
       ,rg,pg,ng,nnp,znext,dspin,ncomp,nspin,ndens,nw,nb1,nb2 &
       ,aa0,kd,mass,ee,wgtk,dd &
       ,scfcs,mgs,bspotfc,bspots,ifinvsym,axetp &
       ,udirchr,msgcycl,myrank,nproc,comm)
    use types
    use bsplpot
    use iounit
    use scfloop
    use engdffr
    use util
    use mpi
    use mdmuca
    use meshgrid
    use ewald_sum
    implicit none
    integer:: imdstp,emdstp,imtrw
    integer:: iexit,iopt5
    integer:: ismode,nband,laprj,nsi,nkd,ng,nnp
    integer:: ncomp,nspin,ndens,nw,nb1,nb2
    integer:: msgcycl
    
    type (scfctrl_struct):: scfcs
    type (meshgrid_struct):: mgs
    !      
    real(kind=8):: znext, dspin
    real(kind=8):: deltm
    real(kind=8):: ekbt
    !
    real(kind=8):: rho(mgs%nrxyz,nspin*ndens),vxcr(mgs%nrxyz,nspin*ndens)
    complex(kind=8):: pby(laprj,nsi,nw,mgs%nks,nspin)
    complex(kind=8):: lpt(mgs%nvxyz,nspin*ndens)
    !
    real(kind=8):: aa0(3,3)
    real(kind=8):: syskbt, syspext
    real(kind=8),dimension(3,nsi):: casi,cavl,cforce
    real(kind=8):: sysvol,zeta,eta,sigma,fzeta,feta
    integer:: kd(nsi)
    real(kind=8):: ee(nband,mgs%nks,nspin)
    real(kind=8):: wgtk(nb1:nb2,mgs%nks,nspin)
    real(kind=8):: qbtp(3)
    complex(kind=8):: dd(laprj,laprj,nsi,nspin)
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    real(kind=8),intent(in):: mass(nkd),omzeta,ometa
    integer:: rg(3,3,ng),pg(3,ng)
    type (wfn_for_k):: dp(mgs%nks,nspin)
    logical:: ifinvsym
    integer:: axetp
    real(kind=8),pointer:: udirchr(:,:)
    integer:: myrank,nproc,comm
    !
    type (mbmt_struct):: mbmts
    !
    real(kind=8):: scfexit(nscfexit)
    real(kind=8):: etotal,setotal(6),settr
    real(kind=8):: ewald,sewald(6)
    real(kind=8):: ef(nspin),eetrp
    real(kind=8):: eknuc
    type(elec_energy):: eleng
    type(elec_senergy):: elseng
    real(kind=8):: tfzeta,tfeta
    integer:: nstern,npstern
    real(kind=8):: tzeta,teta,tsigma,tsysvol
    !
    real(kind=8),dimension(:,:),allocatable:: &
         force,dforce,for,dfor,forel,dforel, &
         asi,pcasi,ocasi,acasi,tcasi,tcavl,camo,tcamo,tcforce
    real(kind=8),allocatable:: pe(:,:,:)
    real(kind=8),allocatable:: crho(:,:)

    integer:: istep,iex,it,ib,is,iks,ierr

    real(kind=8):: aa(3,3),va0,sysscl,psysvol
    real(kind=8):: totmom(3),ctomas(3),totmass
    real(kind=8):: cszeta,cseta,omeoz

    integer:: icycl,bkm,ekm
    type (wfn_for_k):: pdp(mgs%nks,nspin),odp(mgs%nks,nspin)
    real(kind=8),allocatable:: prho(:,:),orho(:,:)

    real(kind=8):: kbtdels

    integer:: mssg

    integer:: nks,nkr,nrxyz

    nks = mgs%nks
    nkr = mgs%nkr
    nrxyz = mgs%nrxyz

    allocate(pe(nkr,nb1:nb2,nspin), &
         force(3,nsi),dforce(3,nsi), &
         for(3,nsi),dfor(3,nsi),forel(3,nsi),dforel(3,nsi), &
         asi(3,nsi),pcasi(3,nsi),ocasi(3,nsi),acasi(3,nsi),tcasi(3,nsi), &
         tcavl(3,nsi),camo(3,nsi),tcamo(3,nsi),tcforce(3,nsi), &
         crho(nrxyz,nspin*ndens),stat=ierr)
    call chkalloc('mdnpt_st','tmp',ierr)

    allocate(prho(nrxyz,nspin*ndens),orho(nrxyz,nspin*ndens),stat=ierr)
    call chkalloc('mdnpt_st','tmp 2',ierr)
    do is=1,nspin
       do iks=1,nks
          bkm = mgs%g_displ(myrank,iks)+1
          ekm = mgs%g_displ(myrank,iks)+mgs%g_count(myrank,iks)
          allocate(pdp(iks,is)%wf(ncomp,bkm:ekm,nw), &
               odp(iks,is)%wf(ncomp,bkm:ekm,nw),stat=ierr)
          call chkalloc('mdnpt_st','tmp 3',ierr)
       end do
    end do

    va0 = cellvol(aa0)
    sysscl = (sysvol/va0)**(1.0d0/3.0d0)
    aa = aa0*sysscl

    icycl = 0

    totmass = 0.0d0
    ctomas = 0.0d0
    totmom = 0.0d0
    do it=1,nsi
       camo(:,it) = mass(kd(it))*cavl(:,it)
       totmass = totmass + mass(kd(it))
       ctomas = ctomas + mass(kd(it))*casi(:,it)
       totmom = totmom + camo(:,it)
    end do
    ctomas = ctomas/totmass

    if (maxval(abs(ctomas)).gt.1.0d-9) then
       if (imdstp.eq.0 .or. maxval(abs(ctomas)).gt.1.0d0) then
          write(6,'(1x,a)') 'center of mass must be zero'
          if (myrank.eq.0) then
             write(6,'(1x,a,3(1x,e18.11))') 'CM = ', ctomas
          end if
          call MPI_FINALIZE(ierr)
          stop
       end if
    end if
    if (maxval(abs(totmom)).gt.1.0d-9) then
       if (imdstp.eq.0 .or. maxval(abs(totmom)).gt.1.0d0) then
          write(6,'(1x,a)') 'total momentum must be zero'
          if (myrank.eq.0) then
             write(6,'(1x,a,3(1x,e18.11))') 'total P = ', totmom
          end if
          call MPI_FINALIZE(ierr)
          stop
       end if
    end if

    if (omzeta.le.0.0d0 .and. zeta.ne.0.0d0) then
       write(6,'(1x,a)') 'zeta must be 0 whem omzeta = 0'
       call MPI_FINALIZE(ierr)
       stop
    end if
    if (ometa.le.0.0d0 .and. eta.ne.0.0d0) then
       write(6,'(1x,a)') 'eta must be 0 whem ometa = 0'
       call MPI_FINALIZE(ierr)
       stop
    end if

    if (omzeta.gt.0.0d0) then
       omeoz = ometa/omzeta
    else
       omeoz = 0.0d0
    end if
    
    nstern = 3*nsi-3
    npstern = 3*nsi

    if (iopt5 .eq. 0) then
       if (myrank.eq.0) then
          mssg = 0
       else
          mssg = -2
       end if
       !
       ! use reduced asi within an unit cell for ewald calc.
       !
       call cart2cellr(aa,nsi,casi,asi)

       call ewalds(asi,aa,nsi,nkd,kd &
            ,for,dfor,ewald,sewald,bspots,myrank,nproc,comm,mssg)
       call prntime('EWALDS ',mssg)
       call vbscf(dp,rho,pby,lpt,vxcr,crho &
            ,qbtp,ismode,iex,scfexit,ekbt,ewald,eleng &
            ,ef,eetrp,nband,laprj,nsi,nkd,ng &
            ,znext,dspin,ncomp,nspin,ndens,nw,nb1,nb2 &
            ,aa,asi,kd,ee,pe,wgtk,dd &
            ,scfcs,mgs,bspotfc,bspots,ifinvsym &
            ,mssg,myrank,nproc,comm)
       !
       if ( iex .eq. -1 ) then
          if (myrank.eq.0) then
             write(6,'(1x,a)') 'vbscf does not converged'
             call debugscf(nband,nks,nspin,ee,nw,nb1,nb2,wgtk, &
                  nkr,pe,nrxyz,rho,sysvol)
          end if

          imdstp = 0
          if (iex .eq. -2) then
             iexit = -2
          else
             iexit = -1
          end if

          deallocate(prho,orho)
          do is=1,nspin
             do iks=1,nks
                deallocate(pdp(iks,is)%wf,odp(iks,is)%wf)
             end do
          end do

          deallocate(pe,force,dforce,for,dfor,forel,dforel, &
               asi,pcasi,ocasi,acasi,tcasi,tcavl,camo,tcamo,tcforce,crho)
          return
       else if ( iex .lt. 0 ) then
          write(6,'(1x,a)') 'invalid result code from vbscf'
          write(6,'(1x,a,i3)') 'iex = ',iex
          call MPI_FINALIZE(ierr)
          stop
       end if
       if (mssg.gt.-1) then
          write(6,'(1x,a)') 'IB, PE(IB,1) = '
          do ib=nb1,nb2
             write(6,'(1x,i4,1x,e15.7,1x,e15.7)') &
                  ib, (pe(1,ib,is),is=1,nspin)
          end do
       end if
       !
       call vbfor(dp,lpt,rho,vxcr,crho &
            ,wgtk,eleng,forel,dforel,for,dfor,force,dforce &
            ,qbtp,elseng & 
            ,aa,asi,kd,dd,ee,nband,laprj &
            ,rg,pg,ng,nnp,ncomp,nspin,ndens,nsi,nkd &
            ,nb1,nb2,mgs,bspotfc,bspots,scfcs%xctp,axetp & 
            ,myrank,nproc,comm)
       etotal=ewald + eleng%total
       setotal=sewald + elseng%total
       settr=setotal(1) + setotal(2) + setotal(3)
       !
       if (mssg.gt.-1) then
          write(6,'(1x,a,i5)') 'INITIAL STEP'
          write(6,'(1x,a,f10.2)') 'TIME FORCE = ', xetime(1)
          write(6,'(1x,a,e19.10,1x,e19.10)') &
               'FERMI     ENERGY = ', (ef(is),is=1,nspin)
          write(6,'(1x,a,e19.10)') 'EWALD            = ', ewald
          call printeleng(eleng)
          write(6,'(1x,a,e19.10)') 'TOTAL     ENERGY = ', etotal
       end if
       !
       call cell2cart(aa,nsi,dforce,cforce)
       !
       ! zero clear total force for virial
       !
       call zerotfor(nsi,cforce)
       !
       ! save for tryrhowfn
       !
       pcasi = casi
       psysvol = sysvol

       ! call mbmtforce(mbmts,etotal,sysvol,nsi,syskbt,cforce,settr,kbtdels)

       eknuc = 0.0d0
       do it=1,nsi
          eknuc = eknuc + ( camo(1,it)*camo(1,it) + camo(2,it)*camo(2,it) &
               + camo(3,it)*camo(3,it) ) / mass(kd(it))
       end do
       fzeta = eknuc/(nstern*syskbt) - 1.0d0

       feta = ( settr - 3*syspext*sysvol )/( nstern*syskbt )

       scfcs%ifchscf = 2
    end if

    do istep = imdstp, emdstp
       !
       ! save for exit by uptime
       !
       tcasi = casi
       tcamo = camo
       tcavl = cavl
       tsysvol = sysvol
       tzeta = zeta
       teta = eta
       tsigma = sigma
       tfzeta = fzeta
       tfeta = feta
       tcforce = cforce

       if (myrank.eq.0 .and. mod(istep,msgcycl).eq.0) then
          mssg = 0
       else
          mssg = -2
       end if

       zeta = zeta + omzeta**2*fzeta*deltm*0.5d0
       sigma = sigma + zeta*deltm*0.5d0

       call sternpsi(nsi,camo,cforce,-zeta,deltm*0.5d0)

       eta = eta + (ometa**2*feta + npstern*(omeoz*zeta)**2)*deltm*0.5d0

       sysvol = exp(3*eta*deltm)*sysvol
       zeta = exp(-npstern*eta*deltm)*zeta

       do it=1,nsi
          cavl(:,it) = camo(:,it)/mass(kd(it))
       end do
       call sternpsi(nsi,casi,cavl,eta,deltm)

       sysscl = (sysvol/va0)**(1.0d0/3.0d0)
       aa = aa0*sysscl
       !
       ! use reduced asi within an unit cell for ewald calc.
       !
       call cart2cellr(aa,nsi,casi,asi)

       if (scfcs%ifchscf.eq.2) then
          call tryrhowfn(nsi,casi,pcasi,ocasi,acasi,nspin,ndens,dp,pdp,odp, &
               rho,prho,orho,sysvol,psysvol, &
               kd,aa0,bspotfc,nkd,mgs,bspots,udirchr,icycl,imtrw)
       end if

       call ewalds(asi,aa,nsi,nkd,kd &
            ,for,dfor,ewald,sewald,bspots,myrank,nproc,comm,mssg)
       call prntime('EWALDS ',mssg)
       call vbscf(dp,rho,pby,lpt,vxcr,crho &
            ,qbtp,ismode,iex,scfexit,ekbt,ewald,eleng &
            ,ef,eetrp,nband,laprj,nsi,nkd,ng &
            ,znext,dspin,ncomp,nspin,ndens,nw,nb1,nb2 &
            ,aa,asi,kd,ee,pe,wgtk,dd &
            ,scfcs,mgs,bspotfc,bspots,ifinvsym &
            ,mssg,myrank,nproc,comm)
       !
       if ( (iex .eq. -1) .or. (iex .eq. -2) ) then
          if (myrank.eq.0) then
             if (iex .eq. -2) then
                write(6,'(1x,a)') 'vbscf has exited because of uptime.'
             else
                write(6,'(1x,a)') 'vbscf does not converged'
                call debugscf(nband,nks,nspin,ee,nw,nb1,nb2,wgtk, &
                     nkr,pe,nrxyz,rho,sysvol)
             end if
          end if
          !
          ! revert to the loop entry
          !
          casi = tcasi
          camo = tcamo
          cavl = tcavl
          sysvol = tsysvol
          zeta = tzeta
          eta = teta
          sigma = tsigma
          fzeta = tfzeta
          feta = tfeta
          cforce = tcforce

          sysscl = (sysvol/va0)**(1.0d0/3.0d0)
          aa = aa0*sysscl

          imdstp = istep
          if (iex .eq. -2) then
             iexit = -2
          else
             iexit = -1
          end if

          deallocate(prho,orho)
          do is=1,nspin
             do iks=1,nks
                deallocate(pdp(iks,is)%wf,odp(iks,is)%wf)
             end do
          end do

          deallocate(pe,force,dforce,for,dfor,forel,dforel, &
               asi,pcasi,ocasi,acasi,tcasi,tcavl,camo,tcamo,tcforce,crho)
          return
       else if (iex .le. -3) then
          write(6,'(1x,a)') 'unknown result code from vbscf'
          write(6,'(1x,a,i3)') 'iex = ',iex
          call MPI_FINALIZE(ierr)
          stop
       end if
       if (mssg.gt.-1) then
          write(6,'(1x,a)') 'IB, PE(IB,1) = '
          do ib=nb1,nb2
             write(6,'(1x,i4,1x,e15.7,1x,e15.7)') &
                  ib, (pe(1,ib,is),is=1,nspin)
          end do
       end if
       !
       call vbfor(dp,lpt,rho,vxcr,crho &
            ,wgtk,eleng,forel,dforel,for,dfor,force,dforce &
            ,qbtp,elseng &
            ,aa,asi,kd,dd,ee,nband,laprj &
            ,rg,pg,ng,nnp,ncomp,nspin,ndens,nsi,nkd &
            ,nb1,nb2,mgs,bspotfc,bspots,scfcs%xctp,axetp & 
            ,myrank,nproc,comm)
       etotal=ewald + eleng%total
       setotal=sewald + elseng%total
       settr=setotal(1) + setotal(2) + setotal(3)
       !
       if (mssg.gt.-1) then
          write(6,'(1x,a,i5)') 'STEP = ', istep
          write(6,'(1x,a,f10.2)') 'TIME FORCE = ', xetime(1)
          write(6,'(1x,a,e19.10,1x,e19.10)') &
               'FERMI     ENERGY = ', (ef(is),is=1,nspin)
          write(6,'(1x,a,e19.10)') 'EWALD            = ', ewald
          call printeleng(eleng)
          write(6,'(1x,a,e19.10)') 'TOTAL     ENERGY = ', etotal
       end if
       !
       call cell2cart(aa,nsi,dforce,cforce)
       !
       ! zero clear total force for virial
       !
       call zerotfor(nsi,cforce)
       !
       ! notify vbscf() to use charge from tryrhowfn()
       !
       scfcs%ifchscf = 2
       !
       ! save for tryrhowfn
       !
       acasi = ocasi
       ocasi = pcasi
       pcasi = casi
       psysvol = sysvol

       ! call mbmtforce(mbmts,etotal,sysvol,nsi,syskbt,cforce,settr,kbtdels)

       feta = ( settr - 3*syspext*sysvol )/( nstern*syskbt )

       sigma = sigma + zeta*deltm*0.5d0

       call sternpsi(nsi,camo,cforce,-zeta,deltm*0.5d0)
       do it=1,nsi
          cavl(:,it) = camo(:,it)/mass(kd(it))
       end do

       eta = eta + (ometa**2*feta + npstern*(omeoz*zeta)**2)*deltm*0.5d0

       ctomas = 0.0d0
       totmom = 0.0d0
       do it=1,nsi
          totmom = totmom + camo(:,it)
          ctomas = ctomas + mass(kd(it))*casi(:,it)
       end do
       ctomas = ctomas/totmass

       eknuc = 0.0d0
       do it=1,nsi
          eknuc = eknuc + ( camo(1,it)*camo(1,it) + camo(2,it)*camo(2,it) &
               + camo(3,it)*camo(3,it) ) / mass(kd(it))
       end do
       fzeta = eknuc/(nstern*syskbt) - 1.0d0
       eknuc = 0.5d0*eknuc

       zeta = zeta + omzeta**2*fzeta*deltm*0.5d0

       if (myrank.eq.0) then
          write(NFMDLOG) .false., istep*deltm
          write(NFMDLOG) etotal, 0.0d0 ! etotal, order
          write(NFMDLOG) etotal + syspext*sysvol

          write(NFMDLOG) eknuc
          write(NFMDLOG) sysvol
          if (omzeta.gt.0.0d0) then
             cszeta = 0.5d0*(zeta/omzeta)**2
          else
             cszeta = 0.0d0
          end if
          if (ometa.gt.0.0d0) then
             cseta = 0.5d0*(eta/ometa)**2
          else
             cseta = 0.0d0
          end if
          write(NFMDLOG) &
               etotal + kbtdels + eknuc + syspext*sysvol &
               + nstern*syskbt*( cszeta + cseta + sigma )
          write(NFMDLOG) totmom, ctomas
          write(NFMDLOG) zeta, eta, sigma
          write(NFMDLOG) kd,casi
          write(NFMDLOG) -settr/(3.0d0*sysvol), setotal
          write(NFMDLOG) cforce
          write(NFMDLOG) cavl
       end if
    end do
    imdstp = istep
    iexit = 0

    deallocate(prho,orho)
    do is=1,nspin
       do iks=1,nks
          deallocate(pdp(iks,is)%wf,odp(iks,is)%wf)
       end do
    end do

    deallocate(pe,force,dforce,for,dfor,forel,dforel, &
         asi,pcasi,ocasi,acasi,tcasi,tcavl,camo,tcamo,tcforce,crho)
    !
    return
  end subroutine mdnpt_st
  !
  ! \psi := \exp(bt)x + a/b( \exp(bt) - 1 )
  !      = \exp(bt)x + a*2/b\sinh(bt/2)\exp(bt/2)
  ! x := \psi
  !
  subroutine sternpsi(natm,x,a,b,t)
    implicit none
    integer:: natm
    real(kind=8):: x(3,natm)
    real(kind=8),intent(in):: a(3,natm)
    real(kind=8),intent(in):: b,t

    integer:: it
    real(kind=8):: psi(3),bt,bth

    if (b.ne.0.0d0) then
       bt = b*t
       bth = 0.5d0*bt
       do it=1,natm
          psi = exp(bt)*x(:,it) + a(:,it)*sinh(bth)*2.0d0*exp(bth)/b
          x(:,it) = psi
       end do
    else
       do it=1,natm
          x(:,it) = x(:,it) + a(:,it)*t
       end do
    end if

    return
  end subroutine sternpsi

  real(kind=8) function kinengnucl(aa,nsi,kd,avl,nkd,mass)
    implicit none
    integer,intent(in):: nsi,kd(nsi),nkd
    real(kind=8),intent(in):: aa(3,3),avl(3,nsi),mass(nkd)

    real(kind=8):: cvel(3),eknuc
    integer:: it, i, j

    eknuc = 0.0d0
    do it=1,nsi
       cvel = 0.0d0
       do j=1,3
          do i=1,3
             cvel(i) = cvel(i) + aa(i,j)*avl(j,it)
          end do
       end do
       eknuc = eknuc + 0.5d0*mass(kd(it)) &
            *(cvel(1)*cvel(1) + cvel(2)*cvel(2) + cvel(3)*cvel(3))
    end do
    kinengnucl = eknuc
    return
  end function kinengnucl

  subroutine setkbt_vel(syskbt,cgf,nkd,mass, &
       nsi,kd,iseed,cavl)
    use util
    implicit none
    real(kind=8):: syskbt
    integer,intent(in):: cgf
    integer:: nkd,nsi,kd(nsi),iseed
    real(kind=8):: mass(nkd),cavl(3,nsi)

    integer:: it
    real(kind=8):: denom,scale
    real(kind=8):: mom(3),sd

    do it=1,nsi
       sd = sqrt(syskbt/mass(kd(it)))
       cavl(1,it) = rndgauss(sd,iseed)
       cavl(2,it) = rndgauss(sd,iseed)
       cavl(3,it) = rndgauss(sd,iseed)
    end do
    mom = 0.0d0
    do it=1,nsi
       mom(1) = mom(1)+cavl(1,it)*mass(kd(it))
       mom(2) = mom(2)+cavl(2,it)*mass(kd(it))
       mom(3) = mom(3)+cavl(3,it)*mass(kd(it))
    end do
    mom = mom/nsi
    do it=1,nsi
       cavl(1,it) = cavl(1,it)-mom(1)/mass(kd(it))
       cavl(2,it) = cavl(2,it)-mom(2)/mass(kd(it))
       cavl(3,it) = cavl(3,it)-mom(3)/mass(kd(it))
    end do

    denom = 0.0d0
    do it=1,nsi
       denom = denom + mass(kd(it))*( cavl(1,it)*cavl(1,it) &
            + cavl(2,it)*cavl(2,it) + cavl(3,it)*cavl(3,it) )
    end do

    scale = sqrt(((3*nsi-cgf)*syskbt)/denom)
    write(6,'(1x,a,e10.3)') 'setkbt_vel: scale = ', scale

    cavl = scale*cavl

    return
  end subroutine setkbt_vel

  !
  ! REF: T.A. Arias, M.C. Payne, J.D. Joannopoulos,
  !        Phys. Rev. B 45 (1992) 1538
  !      Dario Alfe, Comp. Phys. Comm., 118 (1999) 31
  !
  subroutine tryrhowfn(nsi,ncasi,casi,pcasi,ocasi,nspin,ndens,dp,pdp,odp, &
       rho,prho,orho,newsysvol,sysvol, &
       kd,aa0,bspotfc,nkd,mgs,bspots,udirchr,icycl,imtrw)
    use types
    use mpi
    use iounit
    use bsplpot
    use atomicchrg
    use util
    use converter
    use meshgrid
    implicit none
    integer,intent(in):: nsi,nspin,ndens
    type (meshgrid_struct):: mgs
    real(kind=8),dimension(3,nsi),intent(in):: ncasi,casi,pcasi,ocasi
    type (wfn_for_k),dimension(mgs%nks,nspin):: dp,pdp,odp
    real(kind=8),dimension(mgs%nrxyz,nspin*ndens):: rho,prho,orho
    real(kind=8),intent(in):: newsysvol,sysvol
    integer,intent(in):: kd(nsi),nkd
    real(kind=8),intent(in):: aa0(3,3)
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    real(kind=8),pointer::udirchr(:,:)
    integer:: icycl,imtrw

    real(kind=8):: alpha,beta,a11,a22,a21,b1,b2,det
    real(kind=8):: nd(3),d(3),pd(3)
    integer:: it,iks,is,i,ig,ib,istat
    real(kind=8):: va0,aa(3,3),naa(3,3),bb0(3,3),bb(3,3),nbb(3,3),sysscl

!soption contarea(wf,pwf,owf)
    complex(kind=8),pointer,dimension(:,:,:):: wf,pwf,owf
    real(kind=8):: twf(2),trho

    real(kind=8),allocatable:: atmrho(:,:,:),tasi(:,:)

    integer:: nks,nkfxa,nrxyz
    integer:: nrx,nry,nrz
!soption contarea(kf)
    integer,pointer:: kf(:,:)
    integer:: ncomp

    nks = mgs%nks
    nkfxa = mgs%nkfxa
    nrxyz = mgs%nrxyz
    nrx = mgs%nrx
    nry = mgs%nry
    nrz = mgs%nrz
    kf => mgs%kf

    allocate(FFT_A1M_DEFSZ(atmrho,nrxyz,nspin*ndens), stat=istat)
    call chkalloc('tryrhowfn','atmrho',istat)

    if (imtrw.eq.2) then
       allocate(tasi(3,nsi), stat=istat)
       call chkalloc('tryrhowfn','tasi',istat)
    end if

    va0 = cellvol(aa0)
    call recp_uvec(aa0,bb0)
    sysscl = (sysvol/va0)**(1.0d0/3.0d0)
    aa = aa0*sysscl
    bb = bb0/sysscl
    sysscl = (newsysvol/va0)**(1.0d0/3.0d0)
    naa = aa0*sysscl
    nbb = bb0/sysscl

    if (imtrw.eq.2) then
       call cart2cell(aa,nsi,casi,tasi)
       call mkatmchrgkf(bspotfc,nkd,bspots,nkfxa,aa,kf, &
            nsi,tasi,kd,nrx*2,nry*2,nrz*2,nspin,ndens,udirchr,atmrho)
    else
       atmrho = 0.0d0
    end if

    !
    ! renormalize rho and atmrho for cell deformation
    !
    rho = sysvol*rho
    FFT_A1M_RL(atmrho,:,:) = sysvol*FFT_A1M_RL(atmrho,:,:)

    if (icycl.le.1 .or. (imtrw.eq.0)) then
!poption noparallel
       do is=1,nspin
!poption noparallel
          do iks=1,nks
             wf => dp(iks,is)%wf
             pwf => pdp(iks,is)%wf
             owf => odp(iks,is)%wf

             owf(:,:,:) = pwf(:,:,:)
             pwf(:,:,:) = wf(:,:,:)
          end do
       end do
       orho = prho
       prho = rho - FFT_A1M_RL(atmrho,:,:)
       rho = (1.0d0/newsysvol)*rho
       deallocate(atmrho)
       if (imtrw.eq.2) then
          deallocate(tasi)
       end if
       icycl = icycl + 1
       return
    end if

    a11 = 0.0d0
    a22 = 0.0d0
    a21 = 0.0d0
    b1 = 0.0d0
    b2 = 0.0d0
    do it=1,nsi
       nd = ncasi(:,it) - casi(:,it)
       d = casi(:,it) - pcasi(:,it)
       pd = pcasi(:,it) - ocasi(:,it)
       a11 = a11 + d(1)*d(1) + d(2)*d(2) + d(3)*d(3)
       a22 = a22 + pd(1)*pd(1) + pd(2)*pd(2) + pd(3)*pd(3)
       a21 = a21 + d(1)*pd(1) + d(2)*pd(2) + d(3)*pd(3)
       b1 = b1 + nd(1)*d(1) + nd(2)*d(2) + nd(3)*d(3)
       b2 = b2 + nd(1)*pd(1) + nd(2)*pd(2) + nd(3)*pd(3)
    end do
    det = a11*a22 - a21*a21
    !
    ! Check if d and pd are linear dependent.
    ! This is the case when we compute a diatomic molecule.
    !
    if (abs(det/(a11*a11+a22*a22)) .lt. 1.0d-8) then
       beta = 0.0d0
       alpha = b1/a11
    else
       alpha = (a22*b1 - a21*b2)/det
       beta = (-a21*b1 + a11*b2)/det
    end if

    ncomp = size(dp(1,1)%wf,1)
    do is=1,nspin
       do iks=1,nks
          wf => dp(iks,is)%wf
          pwf => pdp(iks,is)%wf
          owf => odp(iks,is)%wf
!poption parallel
          do ib=1,size(wf,3)
             do ig=lbound(wf,2),ubound(wf,2)
                twf(:ncomp) = wf(:,ig,ib)
                wf(:,ig,ib) = wf(:,ig,ib) &
                     + alpha*( wf(:,ig,ib) - pwf(:,ig,ib) ) &
                     + beta*( pwf(:,ig,ib) - owf(:,ig,ib) )
                owf(:,ig,ib) = pwf(:,ig,ib)
                pwf(:,ig,ib) = twf(:ncomp)
             end do
          end do
       end do
    end do

    do is=1,nspin*ndens
       do i=1,nrxyz
          rho(i,is) = rho(i,is) - FFT_A1M_RL(atmrho,i,is)
          trho = rho(i,is)
          rho(i,is) = rho(i,is) + alpha*(rho(i,is) - prho(i,is)) &
               + beta*(prho(i,is) - orho(i,is))
          orho(i,is) = prho(i,is)
          prho(i,is) = trho
       end do
    end do

    if (imtrw.eq.2) then
       call cart2cell(naa,nsi,ncasi,tasi)
       call mkatmchrgkf(bspotfc,nkd,bspots,nkfxa,naa,kf, &
            nsi,tasi,kd,nrx*2,nry*2,nrz*2,nspin,ndens,udirchr,atmrho)
    else
       atmrho = 0.0d0
    end if
    !
    ! renormalize atmrho for cell deformation
    !
    FFT_A1M_RL(atmrho,:,:) = newsysvol*FFT_A1M_RL(atmrho,:,:)

    rho(:,:) = rho(:,:) + FFT_A1M_RL(atmrho,:,:)

    deallocate(atmrho)
    if (imtrw.eq.2) then
       deallocate(tasi)
    end if

    !
    ! restore the normalization of rho
    !
    rho(:,:) = (1.0d0/newsysvol)*rho(:,:)

    icycl = icycl + 1

    return
  end subroutine tryrhowfn

  subroutine debugscf(nband,nks,nspin,ee,nw,nb1,nb2,wgtk,nkr,pe,nrxyz,rho,va)
    implicit none
    integer:: nks,nband,nw,nb1,nb2,nspin,nrxyz
    real(kind=8):: ee(nband,nks,nspin)
    real(kind=8):: wgtk(nb1:nb2,nks,nspin)
    integer:: nkr
    real(kind=8):: pe(nkr,nb1:nb2,nspin)
    real(kind=8):: rho(nrxyz,nspin),va

    integer:: iks,is,ib

    write(6,'(/1x,a)') 'eigen energy'
    do is=1,nspin
       do iks=1,nks
          write(6,'(1x,a,1x,i4,1x,i2)') 'iks,is =',iks,is
          write(6,'(5(1x,e14.7))') (ee(ib,iks,is),ib=1,nw)
       end do
       write(6,*)
    end do
    write(6,'(/1x,a)') 'IB, PE(IB,1) = '
    do ib=nb1,nb2
       write(6,'(1x,i4,1x,e14.7,1x,e14.7)') &
            ib, (pe(1,ib,is),is=1,nspin)
    end do
    write(6,'(/1x,a)') 'weight for k'
    do is=1,nspin
       do iks=1,nks
          write(6,'(1x,a,1x,i4,1x,i2)') 'iks,is =',iks,is
          write(6,'(5(1x,e14.7))') (wgtk(ib,iks,is),ib=nb1,nb2)
       end do
       write(6,*)
    end do
    write(6,'(/1x,a)') 'charge'
    do is=1,nspin
       write(6,'(1x,a,1x,i1,1x,e14.7)') 'is, rho =',is,sum(rho(:,is))*va/nrxyz
    end do
    
  end subroutine debugscf

  subroutine zerotfor(nsi,force)
    implicit none
    integer:: nsi
    real(kind=8):: force(3,nsi)

    real(kind=8):: tforce(3)
    integer:: it

    tforce = 0.0d0
    do it=1,nsi
       tforce = tforce + force(:,it)
    end do
    tforce = tforce/nsi

    do it=1,nsi
       force(:,it) = force(:,it) - tforce
    end do
    return
  end subroutine zerotfor

  subroutine printeleng(eleng)
    use types
    implicit none
    type(elec_energy):: eleng
    write(6,'(1x,a,e19.10)') 'HARTREE   ENERGY = ', eleng%eht
    write(6,'(1x,a,e19.10,a,e19.10)') &
         'EXCHANGE  ENERGY = ', eleng%exc, ' EXA = ', eleng%exa
    write(6,'(1x,a,e19.10,a,e19.10,a,e19.10)') &
         'POTENTIAL ENERGY = ', eleng%epot, ' LOCAL PART = ', eleng%epotl, &
         ' NONLOCAL PART = ', eleng%epotnl
    write(6,'(1x,a,e19.10)') 'KINETIC   ENERGY = ', eleng%ekin
    return
  end subroutine printeleng

  subroutine mkatmchrgkf(bspotfc,nkd,bspots,nkfxa,aa,kf, &
       nsi,asi,kd,nrx2,nry2,nrz2,nspin,ndens,udirchr,atmrho)
    use bsplpot
    use converter
    use atomicchrg
    use util
    implicit none
    integer,intent(in):: nkd,nkfxa,kf(3,nkfxa),nsi,kd(nsi)
    integer,intent(in):: nrx2,nry2,nrz2,nspin,ndens
    real(kind=8):: aa(3,3),asi(3,nsi)
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    real(kind=8),pointer:: udirchr(:,:)
    real(kind=8),intent(out):: FFT_A1M_DEFSZ(atmrho,nrx2*nry2*nrz2,nspin*ndens)

    real(kind=8):: va,bb(3,3)
    real(kind=8),allocatable:: aff(:)
    integer:: ierr

    va = cellvol(aa)
    call recp_uvec(aa,bb)
    allocate(aff(nkfxa),stat=ierr)
    call chkalloc('mkatmchrgkf','aff',ierr)
    call kf2aff(bb,nkfxa,kf,aff) 
    call mkatmchrg(bspotfc,nkd,bspots,nkfxa,kf,aff,va, &
         nsi,asi,kd,nrx2,nry2,nrz2,nspin,ndens,udirchr,atmrho)
    deallocate(aff)
    return
  end subroutine mkatmchrgkf
end module moldyn
