!
! Copyright (c) 2014 Yoshihide Yoshimoto
!
program pe2dos
  use subr_tetra
  implicit none

  real(kind=8),allocatable:: pe(:,:,:)
  real(kind=8),allocatable:: emesh(:,:,:,:,:)
  integer,allocatable:: kz(:,:,:),nsy(:)
  real(kind=8),allocatable:: e(:,:,:)

  integer,parameter:: nfpe = 51
  real(kind=8),parameter:: pi = 3.14159265358979323846d0

  integer:: ndx,ndy,ndz,nsmpl,mode
  integer:: nbd1,nbd2,nspin,nkr,nng
  character(len=256):: cbuf
  integer:: ie,is,nis,ix,iy,iz,ib,iis
  real(kind=8):: emax,emin,ef,de,fac,s1,s2
  integer:: ierr

  call getarg(1, cbuf)
  open(nfpe, file=cbuf, status='old', action='read', &
       form='unformatted', iostat=ierr)
  if (ierr.ne.0) then
     write(6,*) 'error in opening '//trim(cbuf)
     stop
  end if

  call getarg(2, cbuf)
  read(cbuf,*) ndx
  call getarg(3, cbuf)
  read(cbuf,*) ndy
  call getarg(4, cbuf)
  read(cbuf,*) ndz
  call getarg(5, cbuf)
  read(cbuf,*) nsmpl
  call getarg(6, cbuf)
  read(cbuf,*) mode
  write(6,'(a,3(1x,i4),1x,i8,1x,i2)') &
       ' # ndx,ndy,ndz,nsmpl,mode = ',ndx,ndy,ndz,nsmpl,mode

  read(nfpe) nbd1,nbd2,nkr,nng
  write(6,'(a,4(1x,i6))') ' # nbd1,nbd2,nkr,nng =',nbd1,nbd2,nkr,nng

  allocate(kz(3,nkr,nng),nsy(nkr))

  !
  ! try nspin = 2
  !
  nspin = 2
  allocate(pe(nbd1:nbd2,nkr,nspin))
  read(nfpe,iostat=ierr) pe,kz,nsy
  if (ierr.ne.0) then
     !
     ! try nspin = 1
     !
     deallocate(pe)
     nspin = 1
     allocate(pe(nbd1:nbd2,nkr,nspin))
     backspace(nfpe)
     read(nfpe,iostat=ierr) pe,kz,nsy
     if (ierr.ne.0) then
        write(6,*) 'read error for NFPE'
        stop
     end if
  end if
  write(6,'(a,1x,i2)') ' # nspin =',nspin

  allocate(emesh(-ndx:ndx,-ndy:ndy,0:ndz,nbd1:nbd2,nspin))

  call meshe(pe(:,:,1),emesh(:,:,:,:,1), &
       kz,nsy,ndx,ndy,ndz,nkr,nbd1,nbd2,nng)
  if (nspin.eq.2) then
     call meshe(pe(:,:,2),emesh(:,:,:,:,2), &
          kz,nsy,ndx,ndy,ndz,nkr,nbd1,nbd2,nng)
  end if
  emax = maxval(emesh)
  emin = minval(emesh)
  write(6,'(a,2(1x,e18.11))') ' # emin, emax =', emin, emax

  nis = 0
  do iz = 0, ndz-1
     do iy = -ndy,ndy-1
        do ix = -ndx,ndx-1
           nis = nis + 1
        end do
     end do
  end do
  nis = nis*(nbd2-nbd1+1)

  allocate(e(nis,8,nspin))

  do iis=1,nspin
     is = 0
     do ib = nbd1,nbd2
        do iz = 0, ndz-1
           do iy = -ndy,ndy-1
              do ix = -ndx,ndx-1
                 is = is + 1
                 e(is,1,iis)=emesh(ix  ,iy,  iz,  ib,iis)
                 e(is,2,iis)=emesh(ix  ,iy+1,iz,  ib,iis)
                 e(is,3,iis)=emesh(ix  ,iy,  iz+1,ib,iis)
                 e(is,4,iis)=emesh(ix  ,iy+1,iz+1,ib,iis)
                 e(is,5,iis)=emesh(ix+1,iy,  iz,  ib,iis)
                 e(is,6,iis)=emesh(ix+1,iy+1,iz,  ib,iis)
                 e(is,7,iis)=emesh(ix+1,iy,  iz+1,ib,iis)
                 e(is,8,iis)=emesh(ix+1,iy+1,iz+1,ib,iis)
              end do
           end do
        end do
     end do
  end do

  if (mode.eq.0) then
     write(6,*) '# eng[a.u], integrated states'
  else
     write(6,*) '# eng[a.u], DOS[1/a.u]'
  end if

  if (nspin.eq.1) then
     fac = 2.0d0*2.0d0/dble(8*ndx*ndy*ndz)
  else
     fac = 2.0d0/dble(8*ndx*ndy*ndz)
  end if
  de = (emax-emin)/nsmpl
  do ie = 0, nsmpl
     ef = emin + de*ie
     call tetra0c(e(:,1,1),e(:,2,1),e(:,3,1),e(:,4,1), &
          e(:,5,1),e(:,6,1),e(:,7,1),e(:,8,1), &
          ef,s1,nis,mode)
     if (nspin.eq.2) then
        call tetra0c(e(:,1,2),e(:,2,2),e(:,3,2),e(:,4,2), &
             e(:,5,2),e(:,6,2),e(:,7,2),e(:,8,2), &
             ef,s2,nis,mode)
     end if

     if (nspin.eq.1) then
        write(6,'(1x,e18.11,1x,e18.11)') ef, fac*s1
     else
        write(6,'(3(1x,e18.11))') ef, fac*s1, fac*s2
     end if
  end do

  stop

end program pe2dos
