!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2016,2017 Yoshihide Yoshimoto
!
!     $Id: potihxc.F90,v 1.1.2.1 2013-02-11 02:43:04 yosimoto Exp $
!
!     from p8fy.F 1.35
!
#include "fftgrid.h"

module potihxc
  private
  public ypotp, potxc, rhovin2v, rhov2vin, rho4to2c, vxc2cto4

contains
  subroutine zmsh2lst(nkfxa,kf,nrx2,nry2,nrz2,ztmp,nvx,nvy,nvz,lpt)
    implicit none
    integer,intent(in):: nkfxa,kf(3,nkfxa),nrx2,nry2,nrz2,nvx,nvy,nvz
    real(kind=8),intent(in):: FFT_A3D_DEFSZ(ztmp,nrx2,nry2,nrz2)
    complex(kind=8):: lpt(-nvx:nvx,-nvy:nvy,-nvz:nvz)

    integer:: ig,i1,i2,i3,j1,j2,j3

!$omp parallel do private(ig,i1,i2,i3,j1,j2,j3)
    do ig=1,nkfxa
       i1 = kf(1,ig)
       i2 = kf(2,ig)
       i3 = kf(3,ig)
       j1 = modulo(i1,nrx2)+1 
       j2 = modulo(i2,nry2)+1
       j3 = modulo(i3,nrz2)+1
       lpt(i1,i2,i3) = lpt(i1,i2,i3) &
            + dcmplx(FFT_A3D_RL(ztmp,j1,j2,j3),FFT_A3D_IM(ztmp,j1,j2,j3))
    end do
  end subroutine zmsh2lst

  subroutine rho4to2c(nrxyz,spaxs,rho4,rho2c)
    implicit none
    integer,intent(in):: nrxyz
    real(kind=8),intent(in):: rho4(nrxyz,4)
    real(kind=8),intent(in):: spaxs(3)
    real(kind=8):: rho2c(nrxyz,2)
    integer:: i
    real(kind=8):: pol,rtmp,vtmp(3)

    if (maxval(abs(spaxs)).eq.0.0d0) then
!$omp parallel do private(i,rtmp,vtmp,pol)
       do i=1,nrxyz
          rtmp = maxval(abs(rho4(i,2:4)))
          if (rtmp.gt.0.0d0) then
             vtmp = rho4(i,2:4)/rtmp
             pol = rtmp*sqrt(vtmp(1)**2+vtmp(2)**2+vtmp(3)**2)
          else
             pol = 0.0d0
          end if
          rho2c(i,1) = 0.5d0*rho4(i,1)+0.5d0*pol
          rho2c(i,2) = 0.5d0*rho4(i,1)-0.5d0*pol
       end do
    else
!$omp parallel do private(i,pol)
       do i=1,nrxyz
          pol = rho4(i,2)*spaxs(1)+rho4(i,3)*spaxs(2)+rho4(i,4)*spaxs(3)
          rho2c(i,1) = 0.5d0*rho4(i,1)+0.5d0*pol
          rho2c(i,2) = 0.5d0*rho4(i,1)-0.5d0*pol
       end do
    end if
    return
  end subroutine rho4to2c

  subroutine vxc2cto4(nrxyz,spaxs,vxc2c,rho4,vxc4)
    implicit none
    integer,intent(in):: nrxyz
    real(kind=8),intent(in):: spaxs(3)
    real(kind=8),intent(in):: vxc2c(nrxyz,2),rho4(nrxyz,4)
    real(kind=8):: vxc4(nrxyz,4)
    integer:: i
    real(kind=8):: pol,rtmp,vtmp(3)

    if (maxval(abs(spaxs)).eq.0.0d0) then
!$omp parallel do private(i,rtmp,vtmp,pol)
       do i=1,nrxyz
          rtmp = maxval(abs(rho4(i,2:4)))
          if (rtmp.gt.0.0d0) then
             vtmp = rho4(i,2:4)/rtmp
             pol = rtmp*sqrt(vtmp(1)**2+vtmp(2)**2+vtmp(3)**2)
          else
             pol = 0.0d0
          end if
          vxc4(i,1) = 0.5d0*(vxc2c(i,1)+vxc2c(i,2))
          rtmp = 0.5d0*(vxc2c(i,1)-vxc2c(i,2))
          if (pol.ne.0.0d0) then
             vxc4(i,2) = rtmp*(rho4(i,2)/pol)
             vxc4(i,3) = rtmp*(rho4(i,3)/pol)
             vxc4(i,4) = rtmp*(rho4(i,4)/pol)
          else
             vxc4(i,2:4) = 0.0d0
          end if
       end do
    else
!$omp parallel do private(i,rtmp)
       do i=1,nrxyz
          vxc4(i,1) = 0.5d0*(vxc2c(i,1)+vxc2c(i,2))
          rtmp = 0.5d0*(vxc2c(i,1)-vxc2c(i,2))
          vxc4(i,2:4) = rtmp*spaxs(1:3)
       end do
    end if
    return
  end subroutine vxc2cto4

  subroutine ypotp(nspin,ndens,spaxs,rhoc,rhov,lpt,va,bb,eht,exa,exc &
       ,evxc,aff,kf,lption &
       ,nvx,nvy,nvz,nvxyz,nrx2,nry2,nrz2,nkfxa,xctp,mssg,comm)
    use fft_3d
    implicit none
    integer,intent(in):: nspin,ndens
    integer,intent(in):: nvx,nvy,nvz,nvxyz,nrx2,nry2,nrz2
    integer,intent(in):: nkfxa,mssg,comm
    integer,intent(in):: xctp
    real(kind=8),intent(in):: spaxs(3)
    real(kind=8),intent(in):: va,bb(3,3)
    real(kind=8):: eht,exa,exc,evxc(0:6)
    real(kind=8),intent(in):: rhoc(nrx2*nry2*nrz2)
    real(kind=8):: rhov(nrx2*nry2*nrz2,nspin*ndens)
    complex(kind=8),intent(in):: lption(nkfxa)
    real(kind=8),intent(in):: aff(nkfxa)
    integer,intent(in):: kf(3,nkfxa)
    complex(kind=8),intent(out):: lpt(nvxyz,nspin*ndens)

    real(kind=8),allocatable:: rhov2(:,:),tvxc4(:,:)
    real(kind=8),allocatable:: ztmp(:,:),wtmp(:,:)
    integer:: i,nrxyz
    type (fft3_struct):: fs

    nrxyz = nrx2*nry2*nrz2
    if (ndens.eq.1) then
       call ypotp_core(nspin,rhoc,rhov,lpt,va,bb,eht,exa,exc &
            ,evxc,aff,kf,lption &
            ,nvx,nvy,nvz,nvxyz,nrx2,nry2,nrz2,nkfxa,xctp,mssg,comm)
       if (nspin.eq.2) then
          lpt(:,2) = lpt(:,1)
       end if

       allocate(FFT_A1D_DEFSZ(ztmp,nrxyz),FFT_A1D_DEFSZ(wtmp,nrxyz))
       call fft3_init(nrx2,nry2,nrz2,nrx2,nry2,nrz2,fs)
       ztmp= 0.0d0
       do i=1,nspin
          FFT_A1D_RL(ztmp,:) = rhov(:,i)
          FFT_A1D_IM(ztmp,:) = 0.0d0
          call fft3_fw(fs,ztmp,wtmp)
          call zmsh2lst(nkfxa,kf,nrx2,nry2,nrz2,ztmp,nvx,nvy,nvz,lpt(:,i))
       end do
       call fft3_done(fs)
       deallocate(ztmp,wtmp)
    else
       allocate(rhov2(nrxyz,2))
       call rho4to2c(nrxyz,spaxs,rhov,rhov2)
       call ypotp_core(2,rhoc,rhov2,lpt,va,bb,eht,exa,exc &
            ,evxc,aff,kf,lption &
            ,nvx,nvy,nvz,nvxyz,nrx2,nry2,nrz2,nkfxa,xctp,mssg,comm)
       !
       ! lpt(:,1) contains the scalar part of the local potential.
       ! rhov2 contains the Vxc(R).
       !
       allocate(tvxc4(nrxyz,4))
       call vxc2cto4(nrxyz,spaxs,rhov2,rhov,tvxc4)
       deallocate(rhov2)
       allocate(FFT_A1D_DEFSZ(ztmp,nrxyz),FFT_A1D_DEFSZ(wtmp,nrxyz))
       call fft3_init(nrx2,nry2,nrz2,nrx2,nry2,nrz2,fs)
       do i=2,4
          lpt(:,i) = 0.0d0
       end do
       ztmp = 0.0d0
       do i=1,4
          FFT_A1D_RL(ztmp,:) = tvxc4(:,i)
          FFT_A1D_IM(ztmp,:) = 0.0d0
          call fft3_fw(fs,ztmp,wtmp)
          call zmsh2lst(nkfxa,kf,nrx2,nry2,nrz2,ztmp,nvx,nvy,nvz,lpt(:,i))
       end do
       call fft3_done(fs)
       deallocate(ztmp,wtmp)
       !
       ! we return tvxc4 via rhov
       rhov = tvxc4
       deallocate(tvxc4)
    end if
    return
  end subroutine ypotp

  subroutine ypotp_core(nspin,rhoc,rhovin,lptih,va,bb,eht,exa,exc &
       ,evxc,aff,kf,lption &
       ,nvx,nvy,nvz,nvxyz,nrx2,nry2,nrz2,nkfxa,xctp,mssg,comm)
#ifdef ESM
    use esm_manage
#endif
    use fft_3d
    use timeprof
    use xcf_types
    use mpi
    implicit none
    integer,intent(in):: nspin
    integer,intent(in):: nvx,nvy,nvz,nvxyz,nrx2,nry2,nrz2
    integer,intent(in):: nkfxa,mssg,comm
    integer,intent(in):: xctp
    real(kind=8),intent(in):: va,bb(3,3)
    real(kind=8):: eht,exa,exc,evxc(0:6)
    real(kind=8),intent(in):: rhoc(nrx2,nry2,nrz2)
    real(kind=8):: rhovin(nrx2,nry2,nrz2,nspin)
    complex(kind=8),intent(in):: lption(nkfxa)
    real(kind=8),intent(in):: aff(nkfxa)
    integer,intent(in):: kf(3,nkfxa)
    complex(kind=8),intent(out):: lptih(nvxyz)
    !
    real(kind=8),allocatable:: rhov(:,:)
    real(kind=8):: ztmp(:,:),wtmp(:,:)
    complex(kind=8):: hrt(:)
    allocatable ztmp,wtmp,hrt
    type (fft3_struct):: fs
    integer:: ierr
    integer:: nlx,nly,nlz,nlxyz
    integer:: is

#ifdef ESM
! for mesh convert
    real(kind=8),allocatable::rho_tmp_3d(:,:,:,:)
    real(kind=8),parameter:: pi = 3.1415926535897932385d0
#endif
    call tpf_begin(tpf_pihxc)
    !
    nlx = FFT_PAD_NX(nrx2)
    nly = FFT_PAD_NY(nry2)
    nlz = FFT_PAD_NZ(nrz2)

    nlxyz=nlx*nly*nlz
    allocate(rhov(nlxyz,nspin),hrt(nkfxa), &
         FFT_A1D_DEFSZ(ztmp,nlxyz),FFT_A1D_DEFSZ(wtmp,nlxyz),stat=ierr)
    call chkalloc('ypotp1','tmp',ierr)
    !
#ifdef ESM
    if (esm%enable) then
       allocate(rho_tmp_3d(nlx,nly,nlz,2),stat=ierr)
       call chkalloc('ypotp1','rho_tmp_3d',ierr)
    end if
#endif
    call fft3_init(nrx2,nry2,nrz2,nlx,nly,nlz,fs)
    !
    do is=1,nspin
       call rhovin2v(nrx2,nry2,nrz2,nlx,nly,nlz, &
            rhovin(1,1,1,is),rhov(1,is))
    end do
    !
    ! ztmp <- Rho_v(R)
    !
    if (nspin .eq. 1) then
       FFT_A1D_RL(ztmp,:)=rhov(:,1)
       FFT_A1D_IM(ztmp,:)=0.0d0
    else if (nspin .eq. 2) then
       FFT_A1D_RL(ztmp,:)=rhov(:,1) + rhov(:,2)
       FFT_A1D_IM(ztmp,:)=0.0d0
    else
       write(6,*) 'invalid value in nspin = ', nspin
       stop
    end if
    !
    ! generate rhov(G)
    !
    call fft3_fw(fs,ztmp,wtmp)
#ifdef ESM
    if (esm%enable) then
!___ESM__s__
       ! make hartree potential by esm method
       rho_tmp_3d(:,:,:,1) = reshape(FFT_A1D_RL(ztmp,:),(/nlx,nly,nlz/))
       rho_tmp_3d(:,:,:,2) = reshape(FFT_A1D_IM(ztmp,:),(/nlx,nly,nlz/))
       !
       call esm_yhtgen(nlx,nly,nlz,rho_tmp_3d,eht,comm, mssg)
    else
#endif
    !
    ! generate Hartree part of the local potential V_h(G)
    ! Output :: hrt(nkfxa) eht
    !
    call yhtgen(ztmp,aff,kf,hrt,va,eht,nkfxa,nrx2,nry2,nrz2, &
         nlx,nly,nlz)
#ifdef ESM
    end if
#endif
#ifdef INCONSISTENT_THREAD_COMP
    call tpf_begin(tpf_mpi)
    call MPI_BCAST(eht,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call tpf_end(tpf_mpi)
#endif
    deallocate(ztmp,wtmp)

    call potxc(va,bb,nspin,fs,rhoc,rhov,exa,exc &
       ,evxc,nkfxa,kf,nrx2,nry2,nrz2,nlx,nly,nlz,xctp,mssg,comm)
    !skimu PCC
    ! pass Vxc(R) to vbscf through rhovin
    ! to save Vxc(R) at the end of SCF Loop.
    !skimu PCC
    do is=1,nspin
       call rhov2vin(nrx2,nry2,nrz2,nlx,nly,nlz, &
            rhov(1,is),rhovin(1,1,1,is))
    end do
    !
    allocate(FFT_A1D_DEFSZ(ztmp,nlxyz),FFT_A1D_DEFSZ(wtmp,nlxyz),stat=ierr)
    call chkalloc('ypotp1','ztmp,wtmp',ierr)
    !
    !B-ANNO
    ! ytotpt: lptih = lption + hrt + ztmp
    ! lptih contains local potential on 
    ! (-nvx:nvx,-nvy:nvy,-nvz:nvz) reciprocal mesh.
    !E-ANNO
#ifdef ESM
    if (esm%enable) then
       FFT_A1D_RL(ztmp,:) = reshape(rho_tmp_3d(:,:,:,1),(/nlxyz/))
       FFT_A1D_IM(ztmp,:) = reshape(rho_tmp_3d(:,:,:,2),(/nlxyz/))
       hrt=(0.0,0.0)
    else
       ztmp = 0.0d0
    end if
#else
    ztmp = 0.0d0
#endif
    call ytotpt(lption,hrt,ztmp,lptih,kf,nvx,nvy,nvz, &
         nrx2,nry2,nrz2,nlx,nly,nlz,nkfxa)
#ifdef ESM
    if (esm%enable) then
       deallocate(rho_tmp_3d)
    end if
#endif
    call fft3_done(fs)
    !
    deallocate(ztmp,wtmp,hrt,rhov)

    call tpf_end(tpf_pihxc)
    return
  end subroutine ypotp_core

  subroutine potxc(va,bb,nspin,fs,rhoc,rhov,exa,exc &
       ,evxc,nkfxa,kf,nrx2,nry2,nrz2,nlx,nly,nlz,xctp,mssg,comm)
    use fft_3d
    use pot_gga_wb
    use pot_ldaxc
    use xcf_types
    use timeprof
    use mpi
    implicit none
    real(kind=8),intent(in):: va,bb(3,3)
    type (fft3_struct):: fs
    integer,intent(in):: nspin
    integer,intent(in):: nkfxa,kf(3,nkfxa)
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz
    integer,intent(in):: mssg,comm
    integer,intent(in):: xctp
    real(kind=8):: exa,exc,evxc(0:6)
    real(kind=8),intent(in):: rhoc(nrx2,nry2,nrz2)
    real(kind=8):: rhov(nlx*nly*nlz,nspin)
    !
    integer:: nlxyz
    real(kind=8),allocatable:: rhot(:,:,:,:)
    integer:: ierr
    integer:: is
    !
    nlxyz=nlx*nly*nlz

    allocate(rhot(nlx,nly,nlz,nspin),stat=ierr)
    call chkalloc('potxc','rhot',ierr)
    !skimu PCC
    !
    ! rhot: Rho_v(R) + (1/nspin)*Rho_c(R)
    !
    call rhovpc2t(nspin,nrx2,nry2,nrz2,nlx,nly,nlz, &
         rhoc,rhov,rhot)
    !
    !B-ANNO
    !     Exchange correlation potential
    ! In
    !     rhov : charge derived from band wave function
    !     rhot : Rho_v + Rho_c
    ! Out
    !     rhov : Vxc(i) 
    !
    !           /
    !     EXC = \ exc(rho(1,2)) * rho(1,2) dr^3
    !           /
    !                 /
    !     EXA = EXC - \ Vxc(rho(1,2)) * rho(1,1) dr^3
    !                 /
    !E-ANNO
    if (xcf_is_gga(xctp).eq.1) then
       !
       ! at this point
       ! rhot contains Rho_v(R) + Rho_c(R)
       ! rhov contains Rho_v(R)
       !
       ! rhot and rhov must be padded by zero
       !
       call ggawb(nspin,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz,nkfxa,kf, &
            rhot,rhov,va,bb,exa,exc,evxc,fs,xctp,mssg,comm)
       !
       ! at this point
       ! rhov contains Vxc(R)
       ! rhot destructed
       !
    else
       !
       select case(xctp)
       case(xcf_capz)
          !
          ! enforce charge to be positive.
          !     
          do is=1,nspin
             call poschr(rhot(1,1,1,is),nlxyz)
          end do
          if (mssg.gt.0) then
             write(6,*)' POSITIVE CHARGE FORCED IN YPOTP'
             write(6,600) sum(rhot)*va/(nrx2*nry2*nrz2)
          end if
600       format(' ',' CORRECTED TOTAL CHARGE=',f25.20)
          call calder(nspin,nrx2,nry2,nrz2,nlx,nly,nlz, &
               rhov,rhot,va,exa,exc,evxc(0))
          do is=1,nspin 
             call pad_zero(nrx2,nry2,nrz2,nlx,nly,nlz,rhov(1,is))
          end do
       case(xcf_ldapw92)
          !
          ! at this point
          ! rhot contains Rho_v(R) + Rho_c(R)
          ! rhov contains Rho_v(R)
          !
          ! rhot and rhov must be padded by zero
          !
          call ldaxc(nspin,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz,nkfxa,kf, &
               rhot,rhov,va,bb,exa,exc,evxc,fs,xctp,mssg,comm)
          !
          ! at this point
          ! rhov contains Vxc(R)
          ! rhot destructed
          !
       case default
          if (mssg.gt.0) then
             write(6,*) 'ypotp(): unsupported value in xctp = ', xctp
          end if
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end select
#ifdef INCONSISTENT_THREAD_COMP
       call tpf_begin(tpf_mpi)
       call MPI_BCAST(exa,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(exc,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(evxc(0),1,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call tpf_end(tpf_mpi)
#endif
       evxc(1:6) = 0.0d0

       !B-TODO
       ! Xalpha and Wigner does not supported.
       !E-TODO
       ! CALL XALPHA(RHOV,RHOT,VA,EXA,EXC,NLXYZ)
       ! CALL WIGNER(RHOV,RHOT,VA,EXA,EXC,NLXYZ)
       !
    end if
    !
    deallocate(rhot)

    return
  end subroutine potxc

  subroutine pad_zero(nx,ny,nz,lx,ly,lz,dat)
    implicit none
    integer,intent(in):: nx,ny,nz,lx,ly,lz
    real(kind=8):: dat(lx,ly,lz)

    integer:: j
    do j=nz+1,lz
       dat(:,:,j) = 0.0d0
    end do
    do j=nx+1,lx
       dat(j,:,:) = 0.0d0
    end do
    do j=ny+1,ly
       dat(:,j,:) = 0.0d0
    end do
    return
  end subroutine pad_zero

  subroutine rhovin2v(nrx2,nry2,nrz2,nlx,nly,nlz, &
       rhovin,rhov)
    implicit none
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz
    real(kind=8):: rhovin(nrx2,nry2,nrz2)
    real(kind=8):: rhov(nlx,nly,nlz)

    rhov(:nrx2,:nry2,:nrz2) = rhovin
    call pad_zero(nrx2,nry2,nrz2,nlx,nly,nlz,rhov)
    return
  end subroutine rhovin2v

  subroutine rhovpc2t(nspin,nrx2,nry2,nrz2,nlx,nly,nlz, &
       rhoc,rhov,rhot)
    implicit none
    integer,intent(in):: nspin,nrx2,nry2,nrz2,nlx,nly,nlz
    real(kind=8):: rhot(nlx,nly,nlz,nspin)
    real(kind=8):: rhoc(nrx2,nry2,nrz2)
    real(kind=8):: rhov(nlx,nly,nlz,nspin)

    integer:: is

    do is=1,nspin
       rhot(:nrx2,:nry2,:nrz2,is) = &
            rhov(:nrx2,:nry2,:nrz2,is) + (1.0d0/nspin)*rhoc
       call pad_zero(nrx2,nry2,nrz2,nlx,nly,nlz,rhot(1,1,1,is))
    end do
    return
  end subroutine rhovpc2t

  subroutine rhov2vin(nrx2,nry2,nrz2,nlx,nly,nlz, &
       rhov,rhovin)
    implicit none
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz
    real(kind=8):: rhov(nlx,nly,nlz)
    real(kind=8):: rhovin(nrx2,nry2,nrz2)

    rhovin = rhov(:nrx2,:nry2,:nrz2)
    return
  end subroutine rhov2vin

  subroutine poschr(rho,nrxyz)
    implicit none
    integer:: nrxyz
    real(kind=8):: rho(nrxyz)
    real(kind=8),parameter:: eps = 1.0d-27
    integer:: i
    do i=1,nrxyz
       rho(i)=max(rho(i),eps)
    end do
    return
  end subroutine poschr

  subroutine yhtgen(crho,aff,kf,hrt,va,eht,nkfxa,nrx2,nry2,nrz2, &
       nlx,nly,nlz)
    implicit none
    integer,intent(in):: nkfxa,nrx2,nry2,nrz2,nlx,nly,nlz
    real(kind=8),intent(in):: aff(nkfxa),va
    real(kind=8),intent(out):: eht
    integer,intent(in):: kf(3,nkfxa)
    real(kind=8),intent(in):: FFT_A3D_DEFSZ(crho,nlx,nly,nlz)
    complex(kind=8),intent(out):: hrt(nkfxa)
    !
    real(kind=8),parameter:: pi = 3.1415926535897932385d0
    real(kind=8):: fac
    integer:: k,k1,k2,k3
    !
    hrt(1) = (0.0d0, 0.0d0)
    eht = 0.0d0
!$omp parallel do private(fac,k1,k2,k3), &
!$omp reduction(+:eht)
    do k=2,nkfxa
       k1=kf(1,k)
       k2=kf(2,k)
       k3=kf(3,k)
       k1=mod(nrx2+k1,nrx2)+1
       k2=mod(nry2+k2,nry2)+1
       k3=mod(nrz2+k3,nrz2)+1
       fac = 4.0d0*pi/aff(k)
       eht = eht + fac*(FFT_A3D_RL(crho,k1,k2,k3)*FFT_A3D_RL(crho,k1,k2,k3) &
            + FFT_A3D_IM(crho,k1,k2,k3)*FFT_A3D_IM(crho,k1,k2,k3))
       hrt(k) = fac*dcmplx(FFT_A3D_RL(crho,k1,k2,k3),FFT_A3D_IM(crho,k1,k2,k3))
    end do
    eht=eht*va*0.5d0
    return
  end subroutine yhtgen

  ! REF: J.P. Perdew and A. Zunger, Phys. Rev. B 23, 5048 (1981)
  !
  ! Ceperley-Alder, Perdew-Zunger
  !
  subroutine calder(nspin,nrx2,nry2,nrz2,nlx,nly,nlz, &
       rho,rhop,va,exa,exc,evxc)
    ! rho: valence charge
    ! rhop : valence charge + partial core charge.
    implicit none
    integer nspin,nrx2,nry2,nrz2,nlx,nly,nlz
    double precision rho(nlx,nly,nlz,nspin)
    double precision rhop(nlx,nly,nlz,nspin)
    double precision va,exa,exc,evxc
    double precision au,bu,cu,du,ap,bp,cp,dp
    double precision gamu,bet1u,bet2u,gamp,bet1p,bet2p
    double precision alpha,third,pi
    data au,bu,cu,du/0.0311d0, -0.048d0, 0.0020d0, -0.0116d0/
    data ap,bp,cp,dp/0.01555d0, -0.0269d0, 0.0007d0, -0.0048d0/
    data gamu,bet1u,bet2u/-0.1423d0, 1.0529d0, 0.3334d0/
    data gamp,bet1p,bet2p/-0.0843d0, 1.3981d0, 0.2611d0/
    !
    ! \alpha = (\frac{9}{4\pi^2})^(1/3)
    !
    data alpha/0.6108870577d0/
    data third/0.333333333333333333333d0/
    parameter(pi=3.14159265358979323846d0)
    !
    integer ix,iy,iz
    double precision fac,rs,rssq,rsln,vx,ex
    double precision rhopt,vx1,vx2,rs1,rs2,ex1,ex2,vc1,vc2
    double precision zeta,ecu,ecp,vcu,vcp,fzeta,dfz
    !
    fac=(3.0d0/4.0d0/pi)**third
    exa=0.0d0
    exc=0.0d0
    evxc=0.0d0
    !
    if (nspin .eq. 1) then
!$omp parallel do private(ix,iy,iz,rs,rssq,rsln,vx,ex,ecu,vcu), &
!$omp reduction(+:evxc,exc,exa),collapse(3)
       do iz=1,nrz2
          do iy=1,nry2
             do ix=1,nrx2
                rs=fac/(rhop(ix,iy,iz,1)**third)
                rssq=sqrt(rs)
                if (rs.ge.1.0d0) then
                   ecu=gamu/(1.0d0+bet1u*rssq+bet2u*rs)
                   vcu=ecu*ecu/gamu* &
                        (1.0d0+7.0d0/6.0d0*bet1u*rssq+4.0d0/3.0d0*bet2u*rs)
                else
                   rsln=log(rs)
                   ecu=au*rsln+bu+cu*rs*rsln+du*rs
                   vcu=au*rsln+(bu-au/3.0d0)+2.0d0/3.0d0*cu*rs*rsln+ &
                        (2.0d0*du-cu)*rs/3.0d0
                end if
                vx=-alpha/rs
                ex=0.75d0*vx
                exc=exc+rhop(ix,iy,iz,1)*(ex+ecu)
                exa=exa+rhop(ix,iy,iz,1)*(ex+ecu) &
                     -rho(ix,iy,iz,1)*(vx+vcu)
                evxc=evxc+rhop(ix,iy,iz,1)*(vx+vcu)
                rho(ix,iy,iz,1)=vx+vcu
             end do
          end do
       end do
    else if (nspin .eq. 2) then
!$omp parallel do private(ix,iy,iz,rs,rssq,rsln,rhopt,vx1,vx2,ex1,ex2), &
!$omp private(vc1,vc2,zeta,ecu,ecp,vcu,vcp,fzeta,dfz,rs1,rs2), &
!$omp reduction(+:evxc,exc,exa),collapse(3)
       do iz=1,nrz2
          do iy=1,nry2
             do ix=1,nrx2
                rhopt=rhop(ix,iy,iz,1)+rhop(ix,iy,iz,2)
                rs=fac/(rhopt**third)
                rssq=sqrt(rs)
                if (rs.ge.1.0d0) then
                   ecu=gamu/(1.0d0+bet1u*rssq+bet2u*rs)
                   ecp=gamp/(1.0d0+bet1p*rssq+bet2p*rs)
                   vcu=ecu*ecu/gamu* &
                        (1.0d0+7.0d0/6.0d0*bet1u*rssq+4.0d0/3.0d0*bet2u*rs)
                   vcp=ecp*ecp/gamp* &
                        (1.0d0+7.0d0/6.0d0*bet1p*rssq+4.0d0/3.0d0*bet2p*rs)
                else
                   rsln=log(rs)
                   ecu=au*rsln+bu+cu*rs*rsln+du*rs
                   ecp=ap*rsln+bp+cp*rs*rsln+dp*rs
                   vcu=au*rsln+(bu-au/3.0d0)+2.0d0/3.0d0*cu*rs*rsln+ &
                        (2.0d0*du-cu)*rs/3.0d0
                   vcp=ap*rsln+(bp-ap/3.0d0)+2.0d0/3.0d0*cp*rs*rsln+ &
                        (2.0d0*dp-cp)*rs/3.0d0
                end if
                rs1=fac/((2.0d0*rhop(ix,iy,iz,1))**third)
                rs2=fac/((2.0d0*rhop(ix,iy,iz,2))**third)
                vx1=-alpha/rs1
                vx2=-alpha/rs2
                ex1=0.75d0*vx1
                ex2=0.75d0*vx2
                zeta=(rhop(ix,iy,iz,1)-rhop(ix,iy,iz,2))/rhopt
                fzeta=( (1.0d0+zeta)**(4.0d0/3.0d0) &
                     + (1.0d0-zeta)**(4.0d0/3.0d0) - 2.0d0 ) / &
                     ( 2.0d0**(4.0d0/3.0d0) - 2.0d0 )
                exc=exc+rhop(ix,iy,iz,1)*ex1+rhop(ix,iy,iz,2)*ex2 &
                     +rhopt*(ecu + fzeta*(ecp-ecu))
                dfz=(4.0d0/3.0d0)/(2.0d0**(4.0d0/3.0d0) - 2.0d0)* &
                     ( (1.0d0+zeta)**(1.0d0/3.0d0) &
                     - (1.0d0-zeta)**(1.0d0/3.0d0) )
                vc1=vcu+fzeta*(vcp-vcu)+(ecp-ecu)*(1.0d0-zeta)*dfz
                vc2=vcu+fzeta*(vcp-vcu)+(ecp-ecu)*(-1.0d0-zeta)*dfz
                exa=exa+rhop(ix,iy,iz,1)*ex1+rhop(ix,iy,iz,2)*ex2 &
                     +rhopt*(ecu + fzeta*(ecp-ecu)) &
                     - ( rho(ix,iy,iz,1)*(vx1+vc1) &
                     +rho(ix,iy,iz,2)*(vx2+vc2) )
                evxc=evxc+rhop(ix,iy,iz,1)*(vx1+vc1) &
                     +rhop(ix,iy,iz,2)*(vx2+vc2)
                rho(ix,iy,iz,1)=vx1+vc1
                rho(ix,iy,iz,2)=vx2+vc2
             end do
          end do
       end do
    end if
    exa=exa*va/dble(nrx2*nry2*nrz2)
    exc=exc*va/dble(nrx2*nry2*nrz2)
    evxc=evxc*va/dble(nrx2*nry2*nrz2)
    return
  end subroutine calder

  subroutine xalpha(rho,va,exa,exc,nrxyz)
    implicit real*8(a-h,o-z)
    dimension rho(nrxyz)
    data alpha,third/-2.95423506552808957d0,0.333333333333333333333d0/
    calpha=alpha*0.7d0*.5d0
    exa=0.0d0
    exc=0.0d0
!$omp parallel do private(a,b),reduction(+:exc,exa)
    do i=1,nrxyz
       a=calpha*rho(i)**third
       b=0.75d0*a
       exa=exa+(b-a)*rho(i)
       exc=exc+b*rho(i)
       rho(i)=a
    end do
    exa=exa*va/dble(nrxyz)
    exc=exc*va/dble(nrxyz)
    return
  end subroutine xalpha

  subroutine wigner(rho,va,exa,exc,nrxyz)
    implicit real*8(a-h,o-z)
    dimension rho(nrxyz)
    data zb,zc,w3/7.8d0,0.88d0,0.333333333333333333d0/
    data p8/25.1327412287183459d0/
    zy=3.0d0*(3.0d0/p8)**w3
    za=(6.0d0/p8)**w3
    exa=0.0d0
    exc=0.0d0
!$omp parallel do private(a,b),reduction(+:exc,exa)
    do i=1,nrxyz
       a=-4.0d0*w3*zy*rho(i)**w3 &
            -(4.0d0*w3*za*zc*rho(i)**w3+zb*zc*rho(i)**(2.0d0*w3)) &
            /(za+zb*rho(i)**w3)/(za+zb*rho(i)**w3)
       b=-zy*rho(i)**w3 &
            -zc/(za*rho(i)**(-w3)+zb)
       a=a*.5d0
       b=b*.5d0
       exa=exa+(b-a)*rho(i)
       exc=exc+b*rho(i)
       rho(i)=a
    end do
    exa=exa*va/dble(nrxyz)
    exc=exc*va/dble(nrxyz)
    return
  end subroutine wigner

  subroutine ytotpt(cpoti,cpoth,cpot1,cpot2,kf,nvx,nvy,nvz &
       ,nrx2,nry2,nrz2,nlx,nly,nlz,nkfxa)
    implicit none
    integer,intent(in):: nvx,nvy,nvz,nrx2,nry2,nrz2,nkfxa
    integer,intent(in):: nlx,nly,nlz
    complex(kind=8),intent(in):: cpoti(nkfxa),cpoth(nkfxa)
    real(kind=8),intent(in):: FFT_A3D_DEFSZ(cpot1,nlx,nly,nlz)
    complex(kind=8),intent(out):: cpot2(-nvx:nvx,-nvy:nvy,-nvz:nvz)
    integer:: kf(3,nkfxa)
    !
    integer:: k,k1,k2,k3,l1,l2,l3
    !
    cpot2 = (0.0d0, 0.0d0)
!$omp parallel do private(k1,k2,k3,l1,l2,l3)
!OCL NOVREC(cpot2)
!CDIR NODEP(cpot2)
!VOPTION INDEP(cpot2)
!POPTION INDEP(cpot2)
    do k=1,nkfxa
       k1=kf(1,k)
       k2=kf(2,k)
       k3=kf(3,k)
       l1=mod(nrx2+k1,nrx2)+1
       l2=mod(nry2+k2,nry2)+1
       l3=mod(nrz2+k3,nrz2)+1
       cpot2(k1,k2,k3) = cpoti(k)+cpoth(k) &
            +dcmplx(FFT_A3D_RL(cpot1,l1,l2,l3),FFT_A3D_IM(cpot1,l1,l2,l3))
    end do
    return
  end subroutine ytotpt
end module potihxc
