!
! Copyright (c) 2000-2013,2015 Yoshihide Yoshimoto
!
#include "fftgrid.h"

module prclpt
contains
  subroutine prclptldos(nspin, nrx, nry, nrz, nvx, nvy, nvz, &
       nkfxa, aff, kf, trylpt, newlpt, lption, efldosin, &
       pcchr, rhov, aa, bb, va, nwk, idcwk, ng, nrcell, xctp, qpc, mssg, comm)
    use util
    use poisneqn
    use fft_3d
    use symchrg
    use potihxc
    implicit none
    integer,intent(in):: nspin, nrx, nry, nrz, nvx, nvy, nvz, nkfxa
    real(kind=8),intent(in):: aff(nkfxa)
    integer,intent(in):: kf(3,nkfxa)
    complex(kind=8),intent(in):: trylpt(-nvx:nvx, -nvy:nvy, -nvz:nvz, nspin)
    complex(kind=8):: newlpt(-nvx:nvx, -nvy:nvy, -nvz:nvz, nspin)
    complex(kind=8),intent(in):: lption(nkfxa)
    real(kind=8),intent(in):: pcchr(nrx*2, nry*2, nrz*2)
    real(kind=8),dimension(nrx*2, nry*2, nrz*2, nspin),intent(in):: &
         efldosin, rhov
    real(kind=8),intent(in):: aa(3,3), bb(3,3), va
    integer,intent(in):: nwk(8*nrx*nry*nrz,3), ng, nrcell
    integer,intent(in):: idcwk(nrcell,2)
    integer,intent(in):: xctp
    real(kind=8),intent(in):: qpc(4)
    integer,intent(in):: mssg, comm

    real(kind=8),parameter:: pi = 3.1415926535897932385d0

    integer ig, i1, i2, i3, k1, k2, k3, j1, j2, j3, is, ierr
    real(kind=8):: fac, gr
    real(kind=8):: dum1, dum2, dum3, dum4(0:6)
    integer:: nvxyz, nrxyz
    integer:: npxa, iig
    real(kind=8):: qpc1, qpc2, qpca, qpcb

    integer:: npx2, npy2, npz2, nlpx, nlpy, nlpz, nlrx, nlry, nlrz
    real(kind=8),dimension(:,:,:,:),allocatable:: ptmp1, ptmp2, ztmp1, ztmp2
    real(kind=8),dimension(:,:,:),allocatable:: srct, phi, efldost
    real(kind=8),dimension(:,:,:,:),allocatable:: rhovprc, efldos, src
    real(kind=8),allocatable:: winpfac(:)
    type(fft3_struct):: fs
    type(fft3_struct):: fsp
    complex(kind=8),allocatable:: tmplpt(:,:,:,:)
    complex(kind=8):: dlpt
    integer,allocatable:: kfp(:,:)

    real(kind=8),parameter:: adum(3) = (/0.0d0, 0.0d0, 0.0d0/)

    qpc1 = qpc(1)
    qpc2 = qpc(2)
    qpca = qpc(3)
    qpcb = qpc(4)

    nrxyz = 8*nrx*nry*nrz
    if (mssg.gt.0) then
       write(6,'(1x, a, e11.3)') 'DOS(E_f) =', sum(efldosin)/nrxyz
    end if

    npxa = 0
    npx2 = 0
    npy2 = 0
    npz2 = 0
    do ig = 1,nkfxa
       if ( sqrt(aff(ig)).le.qpcb ) then
          npxa = npxa + 1
          npx2 = max(npx2,abs(kf(1,ig)))
          npy2 = max(npy2,abs(kf(2,ig)))
          npz2 = max(npz2,abs(kf(3,ig)))
       end if
    end do
    npx2 = 2*algn235(npx2,1)
    npy2 = 2*algn235(npy2,1)
    npz2 = 2*algn235(npz2,1)
    nlrx = FFT_PAD_NX(nrx*2)
    nlry = FFT_PAD_NY(nry*2)
    nlrz = FFT_PAD_NZ(nrz*2)
    nlpx = FFT_PAD_NX(npx2)
    nlpy = FFT_PAD_NY(npy2)
    nlpz = FFT_PAD_NZ(npz2)
    allocate(kfp(3,npxa),winpfac(npxa), stat=ierr)
    call chkalloc('prclpt','tmp',ierr)
    iig = 0
    do ig = 1,nkfxa
       gr = sqrt(aff(ig))
       if ( gr .le. qpcb ) then
          iig = iig + 1
          kfp(1,iig) = kf(1,ig)
          kfp(2,iig) = kf(2,ig)
          kfp(3,iig) = kf(3,ig)
          if (gr .le. qpca) then
             winpfac(iig) = 1.0d0
          else if (gr .le. qpcb) then
             winpfac(iig) = 0.5d0*( cos(pi/(qpcb-qpca)*(gr-qpca)) + 1.0d0 )
          end if
       end if
    end do

    call fft3_init(npx2, npy2, npz2, nlpx, nlpy, nlpz, fsp)
    call fft3_init(nrx*2, nry*2, nrz*2, nlrx, nlry, nlrz, fs)

    allocate(FFT_A3D_DEFSZ(ptmp1, nlpx, nlpy, nlpz), FFT_A3D_DEFSZ(ptmp2, nlpx, nlpy, nlpz), &
         FFT_A3D_DEFSZ(ztmp1, nlrx, nlry, nlrz),  FFT_A3D_DEFSZ(ztmp2, nlrx, nlry, nlrz), &
         efldos(npx2, npy2, npz2, nspin), stat=ierr)
    call chkalloc('prclpt','ptmp,ztmp,efldos',ierr)

    do is=1,nspin
       FFT_A3D_RL(ztmp1,:nrx*2,:nry*2,:nrz*2) = efldosin(:,:,:,is)
       FFT_A3D_IM(ztmp1,:,:,:) = 0.0d0
       call fft3_fw(fs, ztmp1, ztmp2)
       ptmp1 = 0.0d0
!cdir nodep(ptmp1)
!voption indep(ptmp1)
!vocl loop,disjoint(ptmp1)
       do ig = 1, npxa
          i1 = kfp(1,ig)
          i2 = kfp(2,ig)
          i3 = kfp(3,ig)
          k1 = modulo(i1, npx2) + 1
          k2 = modulo(i2, npy2) + 1
          k3 = modulo(i3, npz2) + 1
          j1 = modulo(i1, nrx*2) + 1
          j2 = modulo(i2, nry*2) + 1
          j3 = modulo(i3, nrz*2) + 1
          FFT_A3D_RL(ptmp1,k1,k2,k3) = winpfac(ig)*FFT_A3D_RL(ztmp1,j1,j2,j3)
          FFT_A3D_IM(ptmp1,k1,k2,k3) = winpfac(ig)*FFT_A3D_IM(ztmp1,j1,j2,j3)
       end do
       call fft3_bw(fsp, ptmp1, ptmp2)
       efldos(:,:,:,is) = FFT_A3D_RL(ptmp1,:npx2,:npy2,:npz2)
    end do

    allocate(srct(npx2, npy2, npz2), efldost(npx2, npy2, npz2), &
         src(npx2, npy2, npz2, nspin), stat=ierr)
    call chkalloc('prclpt','srct, efldost, src',ierr)
    if (nspin .eq. 1) then
       efldost = efldos(:,:,:,1)
    else if (nspin .eq. 2) then 
       efldost = efldos(:,:,:,1) + efldos(:,:,:,2)
    else
       write(6,*) 'invalid value in nspin = ', nspin
       stop
    end if
    if (mssg.gt.0) then
       write(6,'(1x,a,e11.3)') 'Min(LDOS) =', minval(efldost)
    end if

    srct = 0.0d0
    do is=1,nspin
       ptmp1 = 0.0d0
!cdir nodep(ptmp1)
!voption indep(ptmp1)
!vocl loop,disjoint(ptmp1)
       do ig = 2, npxa
          i1 = kfp(1,ig)
          i2 = kfp(2,ig)
          i3 = kfp(3,ig)
          k1 = modulo(i1, npx2) + 1
          k2 = modulo(i2, npy2) + 1
          k3 = modulo(i3, npz2) + 1
          dlpt = trylpt(i1,i2,i3,is) - newlpt(i1,i2,i3,is)
          FFT_A3D_RL(ptmp1,k1,k2,k3) = winpfac(ig)*dble(dlpt)
          FFT_A3D_IM(ptmp1,k1,k2,k3) = winpfac(ig)*aimag(dlpt)
       end do

       call fft3_bw(fsp, ptmp1, ptmp2)

       src(:,:,:,is) = - FFT_A3D_RL(ptmp1,:npx2,:npy2,:npz2)*efldos(:,:,:,is)
       srct = srct + src(:,:,:,is)
    end do
    deallocate(winpfac)

    allocate(phi(npx2, npy2, npz2), stat=ierr)
    call chkalloc('prclpt','tmp 2',ierr)
    !
    !     set the initial solution of phi.
    !
    phi = srct
    call poisson(aa, npx2, npy2, npz2, &
         efldost, srct, phi, 1.0d-10, mssg)

    deallocate(efldost)

    allocate(rhovprc(nrx*2, nry*2, nrz*2, nspin), stat=ierr)
    call chkalloc('prclpt','tmp 3',ierr)

    do is=1,nspin
       FFT_A3D_RL(ptmp1,:npx2,:npy2,:npz2) = - phi*efldos(:,:,:,is) + src(:,:,:,is)
       FFT_A3D_IM(ptmp1,:,:,:) = 0.0d0
       call fft3_fw(fsp, ptmp1, ptmp2)

       ztmp1 = 0.0d0
       do ig = 1, npxa
          i1 = kfp(1,ig)
          i2 = kfp(2,ig)
          i3 = kfp(3,ig)
          k1 = modulo(i1, npx2) + 1
          k2 = modulo(i2, npy2) + 1
          k3 = modulo(i3, npz2) + 1
          j1 = modulo(i1, nrx*2) + 1
          j2 = modulo(i2, nry*2) + 1
          j3 = modulo(i3, nrz*2) + 1
          FFT_A3D_RL(ztmp1,j1,j2,j3) = FFT_A3D_RL(ptmp1,k1,k2,k3)
          FFT_A3D_IM(ztmp1,j1,j2,j3) = FFT_A3D_IM(ptmp1,k1,k2,k3)
       end do
       call fft3_bw(fs, ztmp1, ztmp2)

       !rhovprc = rhov - ( - phi*efldos + src )
       rhovprc(:,:,:,is) = rhov(:,:,:,is) - FFT_A3D_RL(ztmp1,:nrx*2,:nry*2,:nrz*2)
    end do
    call fft3_done(fs)
    call fft3_done(fsp)
    deallocate(ptmp1,ptmp2,ztmp1,ztmp2,srct,phi,efldos,src,kfp)

    ! Since the supplied symmetry list can be incomplete,
    ! the above dlpt sampling can be incompatible with not also
    ! the omitted symmetries but also the listed ones.
    ! Therefore, we must enforce symmetry on ztmp1.
    do is=1,nspin
       call ysymch(rhovprc(1,1,1,is),nwk,idcwk, &
            nrx*2,nry*2,nrz*2,nrxyz,ng,nrcell)
    end do

    allocate(tmplpt(-nvx:nvx, -nvy:nvy, -nvz:nvz, 2), stat=ierr)
    call chkalloc('prclpt','tmp 4',ierr)

    nvxyz = (2*nvx+1)*(2*nvy+1)*(2*nvz+1)
    call ypotp(nspin,1,adum,pcchr,rhovprc,tmplpt,va,bb,dum1,dum2,dum3 &
         ,dum4,aff,kf,lption &
         ,nvx,nvy,nvz,nvxyz,nrx*2,nry*2,nrz*2,nkfxa,xctp,mssg,comm)

    ! Since nv[xyz] can be larger than nr[xyz],
    ! we have to use kf.
    do ig = 1, nkfxa
       gr = sqrt(aff(ig))
       if (gr .le. qpc1) then
          fac = 1.0d0
       else if (gr .le. qpc2) then
          fac = 0.5d0*( cos(pi/(qpc2-qpc1)*(gr-qpc1)) + 1.0d0 )
       else
          cycle
       end if
       i1 = kf(1,ig)
       i2 = kf(2,ig)
       i3 = kf(3,ig)
       newlpt(i1,i2,i3,1) = fac*tmplpt(i1,i2,i3,1) &
            + (1.0d0 - fac)*newlpt(i1,i2,i3,1)
       if (nspin .eq. 2) then
          newlpt(i1,i2,i3,2) = fac*tmplpt(i1,i2,i3,2) &
               + (1.0d0 - fac)*newlpt(i1,i2,i3,2)
       end if
    end do

    deallocate(tmplpt,rhovprc)

    return
  end subroutine prclptldos

  logical function pfbq(nb1, nb2, nkr, pe, eps)
    implicit none
    integer,intent(in):: nb1, nb2, nkr
    real(kind=8),intent(in):: pe(nkr,nb1:nb2), eps

    integer:: ib
    logical:: exist

    exist = .false.

    do ib = nb1, nb2
       if ( (abs(pe(1, ib)) .gt. eps) .and. &
            (abs(pe(1, ib)) .lt. 1.0d0-eps) ) then
          exist = .true.
          exit
       end if
    end do
    pfbq = exist
    return
  end function pfbq
end module prclpt
