!
! Copyright (c) 2000-2013,2015 Yoshihide Yoshimoto
!
module orthg_seavec
contains
subroutine sorthcl(bkm,ekm,n,nwc,p,sp,cl,comm)
  use mpi
  use arch_sorthcl
  use timeprof
  implicit none
  integer,intent(in):: bkm,ekm,n,nwc
  complex(kind=8),intent(in):: p(bkm:ekm,n),sp(bkm:ekm,n)
  complex(kind=8):: cl(bkm:ekm,nwc)
  integer,intent(in):: comm

  integer:: ierr
  complex(kind=8),allocatable:: ovlp(:,:),povlp(:,:)

  if (n .eq. 0) then
     return
  end if

  call tpf_begin(tpf_sot)
  !
  !     orthonormalization of cl
  !
  allocate(ovlp(n,nwc),povlp(n,nwc),stat=ierr)
  call chkalloc('sorthcl','ovlp,povlp',ierr)

  call ovlmat(bkm,ekm,n,nwc,sp,cl,povlp)

  call tpf_begin(tpf_mpi)
  call MPI_ALLREDUCE(povlp,ovlp,n*nwc, &
       MPI_DOUBLE_COMPLEX,MPI_SUM,comm,ierr)
  call tpf_end(tpf_mpi)

  call subovlpt(bkm,ekm,n,nwc,ovlp,p,cl)

  deallocate(ovlp,povlp)

  call tpf_end(tpf_sot)
  return
end subroutine sorthcl

subroutine sorthcli(bkm,ekm,n,nwc,p,sp,cl,comm)
  use mpi
  use arch_sorthcli
  use timeprof
  implicit none
  integer,intent(in):: bkm,ekm,n,nwc
  real(kind=8),intent(in):: p(bkm:ekm,n),sp(bkm:ekm,n)
  real(kind=8):: cl(bkm:ekm,nwc)
  integer,intent(in):: comm

  integer:: ierr
  real(kind=8),allocatable:: ovlp(:,:),povlp(:,:)

  if (n .eq. 0) then
     return
  end if

  call tpf_begin(tpf_sot)
  !
  !     orthonormalization of cl
  !
  allocate(ovlp(n,nwc),povlp(n,nwc),stat=ierr)
  call chkalloc('sorthcli','ovlp,povlp',ierr)

  call ovlmat(bkm,ekm,n,nwc,sp,cl,povlp)

  call tpf_begin(tpf_mpi)
  call MPI_ALLREDUCE(povlp,ovlp,n*nwc, &
       MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
  call tpf_end(tpf_mpi)

  call subovlpt(bkm,ekm,n,nwc,ovlp,p,cl)

  deallocate(ovlp,povlp)

  call tpf_end(tpf_sot)
  return
end subroutine sorthcli

subroutine sorthclg(bkm,ekm,n,nwc,p,sp,cl,comm)
  use mpi
  use arch_sorthclg
  use timeprof
  implicit none
  integer,intent(in):: bkm,ekm,n,nwc
  complex(kind=8),intent(in):: p(bkm:ekm,n),sp(bkm:ekm,n)
  complex(kind=8):: cl(bkm:ekm,nwc)
  integer,intent(in):: comm

  integer:: ierr
  real(kind=8),allocatable:: ovlp(:,:),povlp(:,:)

  if (n .eq. 0) then
     return
  end if

  call tpf_begin(tpf_sot)
  !
  !     orthonormalization of cl
  !
  allocate(ovlp(n,nwc),povlp(n,nwc),stat=ierr)
  call chkalloc('sorthcl','ovlp,povlp',ierr)

  call ovlmat(bkm,ekm,n,nwc,sp,cl,povlp)

  call tpf_begin(tpf_mpi)
  call MPI_ALLREDUCE(povlp,ovlp,n*nwc, &
       MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
  call tpf_end(tpf_mpi)

  call subovlpt(bkm,ekm,n,nwc,ovlp,p,cl)

  deallocate(ovlp,povlp)

  call tpf_end(tpf_sot)
  return
end subroutine sorthclg
end module orthg_seavec
