!
! Copyright (c) 2013 Yoshihide Yoshimoto
! Copyright (c) 1983-2000,2013 TAPP Consortium
!
module stropt_cnstr
  private:: formdg, hyplc
  type optcons_struct
     integer:: mode
     integer:: nmkd
     real(kind=8),pointer:: tim(:,:,:)
     integer,pointer:: mkd(:)
     real(kind=8),pointer:: vnormal(:,:)
  end type optcons_struct
  character(len=*),parameter:: magic_tappstroptconstr = '# str_opt_constr data'
contains
  subroutine dealloc_optcons(optcons)
    implicit none
    type (optcons_struct):: optcons

    deallocate(optcons%tim,optcons%mkd,optcons%vnormal)
    return
  end subroutine dealloc_optcons
  !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
  subroutine read_tappstroptconstr(iin,nsi,mode,optcons,mssg,info)
    !----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+
    implicit none

    integer, intent(in) :: iin, nsi, mode
    type (optcons_struct):: optcons

    integer, intent(in) :: mssg
    integer :: info

    integer :: i, j, it
    integer :: imkd, nset, nmsi
    integer :: tatm(nsi)

    integer:: nmkd
    integer,pointer:: mkd(:)
    real(kind=8),pointer:: tim(:,:,:)

    optcons%mode = mode

    read(iin,*) optcons%nmkd
    nmkd = optcons%nmkd
    allocate(optcons%tim(3,3,nmkd),optcons%mkd(nsi),optcons%vnormal(3,nsi))
    mkd => optcons%mkd
    tim => optcons%tim

    tim(:,:,1) = 0.0d0
    tim(1,1,1) = 1.0d0
    tim(2,2,1) = 1.0d0
    tim(3,3,1) = 1.0d0
    do imkd = 2, nmkd
       do i = 1, 3
          read(iin,*)(tim(j,i,imkd),j=1,3)
       end do
    end do

    ! mkd = 1 is the default
    !
    mkd = 1
    
    read(iin,*) nset
    do i = 1, nset
       read(iin,*) imkd, nmsi
       if ( (imkd.lt.1) .or. (imkd.gt.nmkd) ) then
          if (mssg.gt.-1) write(6,*) 'invalid constaint # ', imkd
          info = 1
       end if
       read(iin,*) (tatm(j),j=1,nmsi)
       do j=1,nmsi
          if ( tatm(j).lt.1 .or. tatm(j).gt.nsi ) then
             if (mssg.gt.-1) write(6,*) 'invalid atom index ', tatm(j)
             info = 1
          end if
          mkd(tatm(j)) = imkd
       end do
    end do

    if (mssg.gt.-1) then
       write(6,'(a)') magic_tappstroptconstr
       write(6,'(1x,i3)') nmkd
       do imkd = 2, nmkd
          do i = 1, 3
             write(6,'(3(1x,e18.11))') (tim(j,i,imkd),j=1,3)
          end do
       end do

       ! mkd = 1 is the default
       !
       nset = 0
       do imkd=2,nmkd
          nmsi = 0
          do it = 1, nsi
             if (mkd(it).eq.imkd) nmsi = nmsi + 1
          end do
          if (nmsi.gt.0) nset = nset + 1
       end do
       write(6,'(1x,i5)') nset

       do imkd=2,nmkd
          nmsi = 0
          do it = 1, nsi
             if (mkd(it).eq.imkd) nmsi = nmsi + 1
          end do
          if (nmsi.gt.0) then
             write(6,'(1x,i5,1x,i5)') imkd, nmsi
             do it = 1, nsi
                if (mkd(it).eq.imkd) then
                   write(6,'(1x,i5)',advance='no') it
                end if
             end do
             write(6,'(/)',advance='no')
          end if
       end do
    end if

    if (optcons%mode.eq.1 .or. optcons%mode.eq.2) then
       write(6,*)
       do it = 1, nsi
          read(iin,*)(optcons%vnormal(i,it),i=1,3)
       end do
       if (mssg.gt.-1) then
          do it = 1, nsi
             write(6,'(3(1x,e18.11))') (optcons%vnormal(i,it),i=1,3)
          end do
       end if
    end if
  end subroutine read_tappstroptconstr

  subroutine dforce_cnstr(aa,nsi,dforce,dform,optcons)
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    integer,intent(in):: nsi
    real(kind=8):: dforce(3,nsi),dform(3,nsi)
    type(optcons_struct):: optcons

    call formdg(dforce,dform,optcons%tim,optcons%nmkd,optcons%mkd,nsi)
    call hyplc(dform,nsi,optcons%mode,optcons%vnormal,aa)

    return
  end subroutine dforce_cnstr

  subroutine formdg(dforce,dform,tim,nmkd,mkd,nsi)
    implicit none
    integer:: nsi,nmkd
    real(kind=8):: dforce(3,nsi),dform(3,nsi)
    real(kind=8):: tim(3,3,nmkd)
    integer:: mkd(nsi)

    integer:: it,i,j,imkd

    do it=1,nsi
       do i=1,3
          dform(i,it)=0.0d0
       end do
    end do
    do it=1,nsi
       imkd=mkd(it)
       do i=1,3
          do j=1,3
             dform(i,it)=dform(i,it)+tim(i,j,imkd)*dforce(j,it)
          end do
       end do
    end do
    return
  end subroutine formdg

  subroutine hyplc(dform,nsi,ifhpc,vnormal,aa)
    !  Hyper Plane Constraint method and Force Inversion method
    !  for saddle point search.
    !  2004.10.30 implemented in tapp-3.0.b56 by J.Yamauchi
    !  2013.5.9 ported to xtapp-20130425 by Y.Suwa
    !  2013.5.20 imported and modified by Y. Yoshimoto
    !---------------------------------------------------------------------
    !-- calc. displacement of each atom in the next step according to HPC
    !   or FI method. This is done by nullifying displacements along the
    !   direction given by vnormal (HPC) or inverting them (FI).
    !   This routine is called only if ifhpc<>0.
    !input
    !  dforce(3,nsi): force acting on each atom
    !  tim(3,3,nmkd): inverse mass tensor
    !  nmkd: num. of defined mass tensor
    !  mkd(nsi): mass kind number of each atom
    !  nsi: num. of atoms
    !  ifhpc: 1=HPC, 2=FI, 0=no saddle(causes error in this routine)
    !  vnormal(3,nsi): displacement vector from initial to final.
    !  aa(3,3): unit cell vector in real space
    !output
    !  dform(3,nsi): displacement of each atom in the next step.
    !---------------------------------------------------------------------
    implicit none
    integer,intent(in):: nsi
    real(kind=8),intent(out):: dform(3,nsi)
    integer,intent(in):: ifhpc
    real(kind=8),intent(in):: vnormal(3,nsi),aa(3,3)
    !local
    integer:: it,j
    real(kind=8):: dnorm,proj,cc(3,3),v(3),u(3)
    !
    ! Hyperplane constraint or Force Inversion
    if ((ifhpc.eq.1).or.(ifhpc.eq.2)) then
      cc=matmul(transpose(aa),aa)
      dnorm=0.0d0
      proj=0.0d0
      do it=1,nsi
        v(:)=0.0d0
        u(:)=0.0d0
        do j=1,3
          v(:)=v(:)+cc(:,j)*vnormal(j,it)
          u(:)=u(:)+cc(:,j)*dform(j,it)
        end do
        dnorm=dnorm+dot_product(vnormal(:,it),v(:))
        proj=proj+dot_product(vnormal(:,it),u(:))
      end do
      if (ifhpc.eq.1)then 
        ! Hyperplane constraint
        proj=proj/dnorm
      else if (ifhpc.eq.2) then
        ! Force Inversion
        proj=2.d0*proj/dnorm
      else
        write(6,*)'THIS CANNOT HAPPEN! ifhpc=',ifhpc
        stop'formdg'
      end if
      !
      dform(1:3,1:nsi)=dform(1:3,1:nsi)-proj*vnormal(1:3,1:nsi)
    end if
    !
    return
  end subroutine hyplc

end module stropt_cnstr
