!
! Copyright (c) 2012 TAPP Consortium
! Copyright (c) 2012-2017 Yoshihide Yoshimoto
!
! based on the development by K. Yoshizawa 2012-08
!
#ifdef NO_MPI
#define SYM_MODULE_NAME subr_rddata_nompi
#else
#define SYM_MODULE_NAME subr_rddata
#endif

module SYM_MODULE_NAME
  use subr_inputpef
contains

  subroutine abortinput(info,mssg)
#ifndef NO_MPI
    use mpi
#endif
    implicit none
    integer,intent(in):: info,mssg
    integer:: stat

    if (info.ne.0) then
       if (mssg.gt.-1) write(6,*) 'input error'
#ifndef NO_MPI
       call MPI_ABORT(MPI_COMM_WORLD,1,stat)
#endif
       stop
    end if
    return
  end subroutine abortinput

  subroutine seektomagic(iin, magic, mssg, info)
    implicit none
    integer:: iin
    character(len=*):: magic
    logical:: flg
    character(len=1024):: buf
    integer,intent(in):: mssg
    integer:: info

    if (mssg.gt.-1) write(6,*)
    ! write(6,'(1x,a)') 'reading section ' // magic

    rewind(iin)
    flg = .true.
    do while (flg)
       read(iin,'(a)',END=999) buf
       if (trim(buf).eq.magic) then
          flg = .false.
       end if
    end do
    return

999 continue
    if (mssg.gt.-1) write(6,*) 'not found: ', magic
    info = 1
  end subroutine seektomagic

  subroutine calcmesh(bb,ng,pg,nnp,qf,qm,nhx,nhy,nhz,nrx,nry,nrz, &
       ngx,ngy,ngz,qwf,nwx,nwy,nwz,ntt,nrxyz,ngxyz,mssg)
    use util
    implicit none
    real(kind=8),intent(in):: bb(3,3),qf,qm
    integer,intent(in):: ng,pg(3,ng),nnp
    integer:: nhx,nhy,nhz,nrx,nry,nrz,ngx,ngy,ngz,nwx,nwy,nwz
    integer:: ntt,nrxyz,ngxyz
    real(kind=8):: qwf
    integer,intent(in):: mssg

    real(kind=8):: h(3)
    integer:: gctr(3)
    integer:: i

    call LatticeInterval(bb,h)
    nhx = int(qm/h(1)) + 1
    nhy = int(qm/h(2)) + 1
    nhz = int(qm/h(3)) + 1

    call grtcomtrns(ng,pg,nnp,gctr)
    do i=1,3
       if (mod(gctr(i),2) .eq. 0) then
          gctr(i) = gctr(i)/2
       end if
    end do
    nrx = algn235(int(qf/h(1)) + 1,gctr(1))
    nry = algn235(int(qf/h(2)) + 1,gctr(2))
    nrz = algn235(int(qf/h(3)) + 1,gctr(3))

    ngx = nrx + 1
    ngy = nry + 1
    ngz = nrz + 1

    qwf = min(2.0d0*qm + 0.1d0, qf)

    nwx = algn235(int(qwf/h(1)) + 1,1)
    nwy = algn235(int(qwf/h(2)) + 1,1)
    nwz = algn235(int(qwf/h(3)) + 1,1)

    ntt = nrx*nry/grtcomdiv(nrx,nry)
    ntt = nrz*ntt/grtcomdiv(nrz,ntt)

    nrxyz = 8*nrx*nry*nrz
    ngxyz = (2*ngx+1)*(2*ngy+1)*(2*ngz+1)

    if (mssg.gt.-1) then
       write(6,100) nrx,nry,nrz,ntt,nrxyz
       write(6,200) ngx,ngy,ngz,ngxyz
       write(6,300) nhx,nhy,nhz
100    format(' NRX=',i5,' NRY=',i5,' NRZ=',i5,' NTT=',i5,' NRXYZ=',i9)
200    format(' NGX=',i5,' NGY=',i5,' NGZ=',i5,' NGXYZ=',i9)
300    format(' NHX=',i5,' NHY=',i5,' NHZ=',i5)

       write(6, '('' nwx='',i5,'' nwy='',i5, '' nwz='',i5)') nwx,nwy,nwz
    end if
    return
  end subroutine calcmesh

  subroutine latgen(ax,bx,aa,bb,va,vb)
    implicit none
    real(kind=8):: ax,bx,aa(3,3),bb(3,3),va,vb
    real(kind=8),parameter:: p2 = 6.28318530717958648d0

    real(kind=8):: v,w

    aa = aa*ax

    v=aa(1,1)*(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)) &
          + aa(2,1)*(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)) &
          + aa(3,1)*(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))
    va=abs(v)
    bx=p2/ax
    w=v/p2

    bb(1,1)=(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3))/w
    bb(2,1)=(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3))/w
    bb(3,1)=(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))/w
    bb(1,2)=(aa(2,3)*aa(3,1)-aa(3,3)*aa(2,1))/w
    bb(2,2)=(aa(3,3)*aa(1,1)-aa(1,3)*aa(3,1))/w
    bb(3,2)=(aa(1,3)*aa(2,1)-aa(2,3)*aa(1,1))/w
    bb(1,3)=(aa(2,1)*aa(3,2)-aa(3,1)*aa(2,2))/w
    bb(2,3)=(aa(3,1)*aa(1,2)-aa(1,1)*aa(3,2))/w
    bb(3,3)=(aa(1,1)*aa(2,2)-aa(2,1)*aa(1,2))/w
    v=bb(1,1)*(bb(2,2)*bb(3,3)-bb(3,2)*bb(2,3)) &
          +bb(2,1)*(bb(3,2)*bb(1,3)-bb(1,2)*bb(3,3)) &
          +bb(3,1)*(bb(1,2)*bb(2,3)-bb(2,2)*bb(1,3))
    vb=abs(v)

    return
  end subroutine latgen

  subroutine printlat(ax,bx,aa,bb,va,vb,mssg)
    implicit none
    real(kind=8),intent(in):: ax,bx,aa(3,3),bb(3,3),va,vb
    integer,intent(in):: mssg

    if (mssg.gt.-1) then
       write(6,100) aa
       write(6,200) bb
       write(6,300) va,vb
    end if
100 format(' AA...PRIMITIVE REAL-LATTICE VECTORS'/(' ',3e15.7))
200 format(' BB...PRIMITIVE RECIPROCAL LATTICE VECTORS'/(' ',3e15.7))
300 format(' VOLUME OF UNIT CELL'/' ',5x,'REAL SPACE...',e18.10,7x,'RECIPROCAL SPACE...',e18.10)
    !
    return
  end subroutine printlat

  subroutine rdinidata(ax,bx,aa,bb,va,vb,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp &
       ,nkd,nsi,kd,asi &
       ,nband,nspin,ndens,znext,dspin &
       ,nrx,nry,nrz,nrxyz,ntt,ngx,ngy,ngz,ngxyz,nhx,nhy,nhz &
       ,qwf,nwx,nwy,nwz &
       ,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode,ekbt &
       ,nf,mssg &
       ,qbsp,nbsf,fwqcom,xctp,zoin,znin)
    !
    use util
    use subr_inputparam
    use subr_inputsym
    use subr_inputatom
    use subr_inputkpoints
    use subr_filemap
    implicit none
    real(kind=8),intent(out):: ax,bx,va,vb,qf,rf,qm,qc,qwf,znext,dspin
    real(kind=8),intent(out):: qbtp(3)
    real(kind=8),intent(out):: ekbt
    integer,intent(in):: nsymq,nopt
    integer,intent(out):: ng,nsv,nnp,nkd,nsi,nband,nspin,ndens
    integer,intent(out):: nk,nw,nb1,nb2,nsk,ismode
    integer,intent(out):: idat(3)
    integer,intent(out):: nrx,nry,nrz,nrxyz,ntt,ngx,ngy,ngz,ngxyz,nwx,nwy,nwz
    integer,intent(out):: nhx,nhy,nhz
    integer,intent(out):: iopt(nopt)
    integer:: nf,mssg
    !
    real(kind=8),intent(out):: aa(3,3),bb(3,3)
    integer,intent(out):: rg(3,3,nsymq),pg(3,nsymq),srg(nsymq)
    integer,pointer:: kd(:),mmm(:,:,:)
    real(kind=8),pointer:: asi(:,:)

    real(kind=8):: qbsp
    integer:: nbsf
    real(kind=8):: fwqcom
    integer:: xctp
    real(kind=8),pointer,intent(out):: zoin(:),znin(:)

    real(kind=8):: eps,eeps,beta,qexat(2)
    real(kind=8):: uptime
    integer:: ndiag0,ndiag1,niter0,niter1,itsb
    integer:: ndx,ndy,ndz
    real(kind=8):: thr_hybrid
    integer:: osqnrm
    integer:: axetp
    real(kind=8):: qprcnd(6)
    integer:: ncomp
    real(kind=8):: thr_noncol,spaxs(3)

    integer:: info

    if (mssg.gt.-1) write(6,'(/1x,a)') 'BEGIN input'

    info = 0

    ! namelist input

    call seektomagic(nf,magic_filemap,mssg,info)
    call read_filemap(nf,mssg,info)

    call seektomagic(nf,magic_inputparam,mssg,info)
    call read_tappinput(nf,ax,aa,ncomp,nspin,ndens,ismode,qf,qm,qc,beta, &
         qbtp,nkd,nsi,nw,znext,dspin,nopt,iopt,eps,eeps,ekbt, &
         uptime,niter0,niter1,itsb,ndiag0,ndiag1, &
         qbsp,nbsf,fwqcom,xctp,qexat,thr_hybrid,osqnrm,axetp, &
         qprcnd,thr_noncol,spaxs,mssg,info)
    nband = nw

    call seektomagic(nf,magic_tappsym,mssg,info)
    call read_tappsym(nf,ng,nsv,nnp,rg,pg,srg,nsymq,mssg,info)

    call seektomagic(nf,magic_tappatom,mssg,info)
    call read_tappatom(nf,nkd,zoin,znin,nsi,kd,asi,mssg,info)

    call seektomagic(nf,magic_tappkpoints,mssg,info)
    call read_tappkpoints(nf,nw,ax,aa,nb1,nb2,ekbt, &
         rf,ndx,ndy,ndz,nk,nsk,mmm,idat, &
         mssg,info)

    if (ndens.eq.4 .and. ndx.gt.0 .and. ndx.gt.0 .and. ndx.gt.0) then
       info = 1
       if (mssg.gt.-1) then
          write(6,'(1x,a)') 'COS fit mode is not supported ' &
               //' for non-collinear calculation.'
       end if
    end if

    call abortinput(info,mssg)

    if (mssg.gt.-1) write(6,'(/1x,a/)') 'END input'

    call latgen(ax,bx,aa,bb,va,vb)
    call printlat(ax,bx,aa,bb,va,vb,mssg)

    call calcmesh(bb,ng,pg,nnp,qf,qm,nhx,nhy,nhz,nrx,nry,nrz, &
         ngx,ngy,ngz,qwf,nwx,nwy,nwz,ntt,nrxyz,ngxyz,mssg)

    return
  end subroutine rdinidata

  subroutine rdpefdata(aa,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp,qexat &
       ,nkd,nsi,kd,asi &
       ,nband,ncomp,nspin,ndens,spaxs,znext,dspin &
       ,nrx,nry,nrz,nrxyz,nhx,nhy,nhz &
       ,qwf,nwx,nwy,nwz &
       ,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode &
       ,itsb,ndiag0,uptime &
       ,nbk,nkpt,ak,nkfi,ifbunp,ifrho,nbba1,nbba2,nsph,iaxis,rzero &
       ,rmin,rmax,ndiv,deps,nf,mssg)
    !
    use util
    use subr_inputparam
    use subr_inputsym
    use subr_inputatom
    use subr_inputkpoints
    use subr_filemap
    implicit none
    real(kind=8),intent(out):: qf,rf,qm,qc,qwf,znext,dspin
    real(kind=8),intent(out):: qbtp(3),qexat(2)
    real(kind=8),intent(out):: deps
    real(kind=8),intent(out):: uptime
    integer,intent(in):: nsymq,nopt
    integer,intent(out):: ng,nsv,nnp,nkd,nsi,nband,ncomp,nspin,ndens
    real(kind=8),intent(out):: spaxs(3)
    integer,intent(out):: nk,nw,nb1,nb2,nsk,ismode
    integer,intent(out):: idat(3)
    integer,intent(out):: ndiag0
    integer,intent(out):: nrx,nry,nrz,nrxyz,nwx,nwy,nwz
    integer,intent(out):: nhx,nhy,nhz
    integer,intent(out):: iopt(nopt)
    integer,intent(out):: ifbunp,ifrho,nbk
    integer,pointer,intent(out):: nkfi(:)
    real(kind=8),pointer,intent(out):: ak(:,:)
    character(len=nkptl),pointer:: nkpt(:)
    integer,pointer,intent(out):: ndiv(:),iaxis(:)
    real(kind=8),pointer:: rzero(:,:),rmin(:),rmax(:)
    integer,intent(out):: nbba1,nbba2,nsph
    integer:: nf,mssg
    !
    real(kind=8),intent(out):: aa(3,3)
    integer,intent(out):: rg(3,3,nsymq),pg(3,nsymq),srg(nsymq)
    integer,pointer:: kd(:),mmm(:,:,:)
    real(kind=8),pointer:: asi(:,:)

    real(kind=8):: ax,bx,va,vb,bb(3,3)
    real(kind=8):: beta,eps,eeps,ekbt
    integer:: ntt,ngx,ngy,ngz,ngxyz
    integer:: ndx,ndy,ndz
    integer:: ndiag1,niter0,niter1,itsb
    real(kind=8):: qbsp
    integer:: nbsf
    real(kind=8):: fwqcom
    integer:: xctp
    real(kind=8):: thr_hybrid
    integer:: osqnrm
    integer:: axetp
    real(kind=8):: qprcnd(6)
    real(kind=8),pointer:: zoin(:),znin(:)
    real(kind=8):: thr_noncol
    integer:: info

    if (mssg.gt.-1) write(6,'(/1x,a)') 'BEGIN input'

    info = 0

    ! namelist input

    call seektomagic(nf,magic_filemap,mssg,info)
    call read_filemap(nf,mssg,info)

    call seektomagic(nf,magic_inputparam,mssg,info)
    call read_tappinput(nf,ax,aa,ncomp,nspin,ndens,ismode,qf,qm,qc,beta, &
         qbtp,nkd,nsi,nw,znext,dspin,nopt,iopt,eps,eeps,ekbt, &
         uptime,niter0,niter1,itsb,ndiag0,ndiag1, &
         qbsp,nbsf,fwqcom,xctp,qexat,thr_hybrid,osqnrm,axetp, &
         qprcnd,thr_noncol,spaxs,mssg,info)
    nband = nw

    if (iopt(3).ne.1) then
       if (mssg.gt.-1) write(6,*) 'initial_lpt must be 1'
       info = 1
    end if

    call seektomagic(nf,magic_tappsym,mssg,info)
    call read_tappsym(nf,ng,nsv,nnp,rg,pg,srg,nsymq,mssg,info)

    call seektomagic(nf,magic_tappatom,mssg,info)
    call read_tappatom(nf,nkd,zoin,znin,nsi,kd,asi,mssg,info)
    deallocate(zoin,znin)

    call seektomagic(nf,magic_tappkpoints,mssg,info)
    call read_tappkpoints(nf,nw,ax,aa,nb1,nb2,ekbt, &
         rf,ndx,ndy,ndz,nk,nsk,mmm,idat, &
         mssg,info)

    call seektomagic(nf,magic_tapppef,mssg,info)
    call read_tapppef(nf,nbk,nkpt,ak,nkfi,nw,nb2,itsb,ndiag0,deps, &
          ifbunp,ifrho,nbba1,nbba2,iaxis,ndiv,rzero, &
          rmin,rmax,nsph,mssg,info)

    call abortinput(info,mssg)

    if (mssg.gt.-1) write(6,'(/1x,a/)') 'END input'

    call latgen(ax,bx,aa,bb,va,vb)
    call printlat(ax,bx,aa,bb,va,vb,mssg)

    call calcmesh(bb,ng,pg,nnp,qf,qm,nhx,nhy,nhz,nrx,nry,nrz, &
         ngx,ngy,ngz,qwf,nwx,nwy,nwz,ntt,nrxyz,ngxyz,mssg)

    return
  end subroutine rdpefdata

  subroutine rdpefcosdata(aa,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp &
       ,nkd,nsi,kd,asi &
       ,nband,nspin,znext,dspin &
       ,nrx,nry,nrz,nrxyz &
       ,qwf,nwx,nwy,nwz &
       ,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode &
       ,uptime &
       ,nbk,nkpt,ak,nkfi,ifbunp,ifrho,nbba1,nbba2,nsph,iaxis,rzero &
       ,rmin,rmax,ndiv,deps,nf,mssg)
    !
    use util
    use subr_inputparam
    use subr_inputsym
    use subr_inputatom
    use subr_inputkpoints
    use subr_filemap
    implicit none
    real(kind=8),intent(out):: qf,rf,qm,qc,qwf,znext,dspin
    real(kind=8),intent(out):: qbtp(3)
    real(kind=8),intent(out):: deps
    real(kind=8),intent(out):: uptime
    integer,intent(in):: nsymq,nopt
    integer,intent(out):: ng,nsv,nnp,nkd,nsi,nband,nspin
    integer,intent(out):: nk,nw,nb1,nb2,nsk,ismode
    integer,intent(out):: idat(3)
    integer,intent(out):: nrx,nry,nrz,nrxyz,nwx,nwy,nwz
    integer,intent(out):: iopt(nopt)
    integer,intent(out):: ifbunp,ifrho,nbk
    integer,pointer,intent(out):: nkfi(:)
    real(kind=8),pointer,intent(out):: ak(:,:)
    character(len=nkptl),pointer:: nkpt(:)
    integer,pointer,intent(out):: ndiv(:),iaxis(:)
    real(kind=8),pointer:: rzero(:,:),rmin(:),rmax(:)
    integer,intent(out):: nbba1,nbba2,nsph
    integer:: nf,mssg
    !
    real(kind=8):: aa(3,3)
    integer,intent(out):: rg(3,3,nsymq),pg(3,nsymq),srg(nsymq)
    integer,pointer:: kd(:),mmm(:,:,:)
    real(kind=8),pointer:: asi(:,:)

    real(kind=8):: ax,bx,va,vb,bb(3,3)
    real(kind=8):: beta,eps,eeps,ekbt,qexat(2)
    integer:: ntt,ngx,ngy,ngz,ngxyz
    integer:: ndx,ndy,ndz,nhx,nhy,nhz
    integer:: ndiag0,ndiag1,niter0,niter1,itsb
    real(kind=8):: qbsp
    integer:: nbsf
    real(kind=8):: fwqcom
    integer:: xctp
    real(kind=8):: thr_hybrid
    integer:: osqnrm
    integer:: axetp
    real(kind=8):: qprcnd(6)
    real(kind=8),pointer:: zoin(:),znin(:)
    integer:: ncomp,ndens
    real(kind=8):: thr_noncol,spaxs(3)
    integer:: info

    if (mssg.gt.-1) write(6,'(/1x,a)') 'BEGIN input'

    info = 0

    ! namelist input

    call seektomagic(nf,magic_filemap,mssg,info)
    call read_filemap(nf,mssg,info)

    call seektomagic(nf,magic_inputparam,mssg,info)
    call read_tappinput(nf,ax,aa,ncomp,nspin,ndens,ismode,qf,qm,qc,beta, &
         qbtp,nkd,nsi,nw,znext,dspin,nopt,iopt,eps,eeps,ekbt, &
         uptime,niter0,niter1,itsb,ndiag0,ndiag1, &
         qbsp,nbsf,fwqcom,xctp,qexat,thr_hybrid,osqnrm,axetp, &
         qprcnd,thr_noncol,spaxs,mssg,info)
    nband = nw

    call seektomagic(nf,magic_tappsym,mssg,info)
    call read_tappsym(nf,ng,nsv,nnp,rg,pg,srg,nsymq,mssg,info)

    call seektomagic(nf,magic_tappatom,mssg,info)
    call read_tappatom(nf,nkd,zoin,znin,nsi,kd,asi,mssg,info)
    deallocate(zoin,znin)

    call seektomagic(nf,magic_tappkpoints,mssg,info)
    call read_tappkpoints(nf,nw,ax,aa,nb1,nb2,ekbt, &
         rf,ndx,ndy,ndz,nk,nsk,mmm,idat, &
         mssg,info)

    call seektomagic(nf,magic_tapppef,mssg,info)
    call read_tapppef(nf,nbk,nkpt,ak,nkfi,nw,nb2,itsb,ndiag0,deps, &
          ifbunp,ifrho,nbba1,nbba2,iaxis,ndiv,rzero, &
          rmin,rmax,nsph,mssg,info)

    if (info.ne.0) then
       if (mssg.gt.-1) write(6,*) 'input error'
       stop
    end if

    if (mssg.gt.-1) write(6,'(/1x,a/)') 'END input'

    call latgen(ax,bx,aa,bb,va,vb)
    call printlat(ax,bx,aa,bb,va,vb,mssg)

    call calcmesh(bb,ng,pg,nnp,qf,qm,nhx,nhy,nhz,nrx,nry,nrz, &
         ngx,ngy,ngz,qwf,nwx,nwy,nwz,ntt,nrxyz,ngxyz,mssg)

    return
  end subroutine rdpefcosdata

  subroutine rdwfn2chgdata(aa,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp &
       ,nkd,nsi,kd,asi &
       ,nband,ncomp,nspin,ndens,znext,dspin &
       ,nrx,nry,nrz,nrxyz &
       ,qwf,nwx,nwy,nwz &
       ,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode &
       ,uptime,ifbunp,ifrho,nbba1,nbba2,nsph,iaxis,rzero &
       ,rmin,rmax,ndiv,itwav,nf,mssg)
    !
    use util
    use subr_inputparam
    use subr_inputsym
    use subr_inputatom
    use subr_inputkpoints
    use subr_inputwfn2chg
    use subr_filemap
    implicit none
    real(kind=8),intent(out):: qf,rf,qm,qc,qwf,znext,dspin
    real(kind=8),intent(out):: qbtp(3)
    real(kind=8),intent(out):: uptime
    integer,intent(in):: nsymq,nopt
    integer,intent(out):: ng,nsv,nnp,nkd,nsi,nband,ncomp,nspin,ndens
    integer,intent(out):: nk,nw,nb1,nb2,nsk,ismode
    integer,intent(out):: idat(3)
    integer,intent(out):: nrx,nry,nrz,nrxyz,nwx,nwy,nwz
    integer,intent(out):: iopt(nopt)
    integer,intent(out):: ifbunp,ifrho
    integer,pointer,intent(out):: ndiv(:),iaxis(:)
    real(kind=8),pointer:: rzero(:,:),rmin(:),rmax(:)
    integer,intent(out):: nbba1,nbba2,nsph,itwav
    integer:: nf,mssg
    !
    real(kind=8),intent(out):: aa(3,3)
    integer,intent(out):: rg(3,3,nsymq),pg(3,nsymq),srg(nsymq)
    integer,pointer:: kd(:),mmm(:,:,:)
    real(kind=8),pointer:: asi(:,:)

    real(kind=8):: ax,bx,va,vb,bb(3,3)
    real(kind=8):: beta,eps,eeps,qexat(2),ekbt
    integer:: ntt,ngx,ngy,ngz,ngxyz
    integer:: nhx,nhy,nhz,ndx,ndy,ndz
    integer:: ndiag0,ndiag1,niter0,niter1,itsb
    real(kind=8):: qbsp
    integer:: nbsf
    real(kind=8):: fwqcom
    integer:: xctp
    real(kind=8):: thr_hybrid
    integer:: osqnrm
    integer:: axetp
    real(kind=8):: qprcnd(6)
    real(kind=8),pointer:: zoin(:),znin(:)
    real(kind=8):: thr_noncol,spaxs(3)
    integer:: info

    if (mssg.gt.-1) write(6,'(/1x,a)') 'BEGIN input'

    info = 0

    ! namelist input

    call seektomagic(nf,magic_filemap,mssg,info)
    call read_filemap(nf,mssg,info)

    call seektomagic(nf,magic_inputparam,mssg,info)
    call read_tappinput(nf,ax,aa,ncomp,nspin,ndens,ismode,qf,qm,qc,beta, &
         qbtp,nkd,nsi,nw,znext,dspin,nopt,iopt,eps,eeps,ekbt, &
         uptime,niter0,niter1,itsb,ndiag0,ndiag1, &
         qbsp,nbsf,fwqcom,xctp,qexat,thr_hybrid,osqnrm,axetp, &
         qprcnd,thr_noncol,spaxs,mssg,info)
    nband = nw

    call seektomagic(nf,magic_tappsym,mssg,info)
    call read_tappsym(nf,ng,nsv,nnp,rg,pg,srg,nsymq,mssg,info)

    call seektomagic(nf,magic_tappatom,mssg,info)
    call read_tappatom(nf,nkd,zoin,znin,nsi,kd,asi,mssg,info)
    deallocate(zoin,znin)

    call seektomagic(nf,magic_tappkpoints,mssg,info)
    call read_tappkpoints(nf,nw,ax,aa,nb1,nb2,ekbt, &
         rf,ndx,ndy,ndz,nk,nsk,mmm,idat,&
         mssg,info)

    call seektomagic(nf,magic_tappwfn2chg,mssg,info)
    call read_tappwfn2chg(nf,nw,nsi, &
          ifbunp,ifrho,nbba1,nbba2,iaxis,ndiv,rzero, &
          rmin,rmax,nsph,itwav,mssg,info)

    call abortinput(info,mssg)

    if (mssg.gt.-1) write(6,'(/1x,a/)') 'END input'

    call latgen(ax,bx,aa,bb,va,vb)
    call printlat(ax,bx,aa,bb,va,vb,mssg)

    call calcmesh(bb,ng,pg,nnp,qf,qm,nhx,nhy,nhz,nrx,nry,nrz, &
         ngx,ngy,ngz,qwf,nwx,nwy,nwz,ntt,nrxyz,ngxyz,mssg)

    return
  end subroutine rdwfn2chgdata

  subroutine rdstmdata(aa,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp,qexat &
       ,nkd,nsi,kd,asi &
       ,nband,ncomp,nspin,ndens,spaxs,znext,dspin &
       ,nrx,nry,nrz,nrxyz &
       ,qwf,nwx,nwy,nwz &
       ,ndx,ndy,ndz,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode,ekbt &
       ,itsb,ndiag0 &
       ,uptime,nbias,eferm,vs,nf,mssg)
    !
    use util
    use subr_inputparam
    use subr_inputsym
    use subr_inputatom
    use subr_inputkpoints
    use subr_inputstm
    use subr_filemap
    implicit none
    real(kind=8),intent(out):: qf,rf,qm,qc,qwf,znext,dspin
    real(kind=8),intent(out):: qbtp(3),qexat(2)
    real(kind=8),intent(out):: ekbt
    real(kind=8),intent(out):: uptime
    integer,intent(in):: nsymq,nopt
    integer,intent(out):: ng,nsv,nnp,nkd,nsi,nband,ncomp,nspin,ndens
    real(kind=8),intent(out):: spaxs(3)
    integer,intent(out):: nk,nw,nb1,nb2,nsk,ismode
    integer,intent(out):: idat(3)
    integer,intent(out):: ndiag0,itsb
    integer,intent(out):: nrx,nry,nrz,nrxyz,nwx,nwy,nwz
    integer,intent(out):: ndx,ndy,ndz
    integer,intent(out):: iopt(nopt)
    integer,intent(out) :: nbias
    real(8),intent(out) :: eferm
    real(8),pointer,intent(out) :: vs(:)

    integer:: nf,mssg
    !
    real(kind=8),intent(out):: aa(3,3)
    integer,intent(out):: rg(3,3,nsymq),pg(3,nsymq),srg(nsymq)
    integer,pointer:: kd(:),mmm(:,:,:)
    real(kind=8),pointer:: asi(:,:)

    real(kind=8):: ax,bx,va,vb,bb(3,3)
    real(kind=8):: beta,eps,eeps
    integer:: ntt,ngx,ngy,ngz,ngxyz,nhx,nhy,nhz
    integer:: ndiag1,niter0,niter1
    real(kind=8):: qbsp
    integer:: nbsf
    real(kind=8):: fwqcom
    integer:: xctp
    real(kind=8):: thr_hybrid
    integer:: osqnrm
    integer:: axetp
    real(kind=8):: qprcnd(6)
    real(kind=8),pointer:: zoin(:),znin(:)
    real(kind=8):: thr_noncol

    integer:: info

    if (mssg.gt.-1) write(6,'(/1x,a)') 'BEGIN input'

    info = 0

    ! namelist input

    call seektomagic(nf,magic_filemap,mssg,info)
    call read_filemap(nf,mssg,info)

    call seektomagic(nf,magic_inputparam,mssg,info)
    call read_tappinput(nf,ax,aa,ncomp,nspin,ndens,ismode,qf,qm,qc,beta, &
         qbtp,nkd,nsi,nw,znext,dspin,nopt,iopt,eps,eeps,ekbt, &
         uptime,niter0,niter1,itsb,ndiag0,ndiag1, &
         qbsp,nbsf,fwqcom,xctp,qexat,thr_hybrid,osqnrm,axetp, &
         qprcnd,thr_noncol,spaxs,mssg,info)
    nband = nw

    if (iopt(3).ne.1) then
       if (mssg.gt.-1) write(6,*) 'initial_lpt must be 1'
       info = 1
    end if

    call seektomagic(nf,magic_tappsym,mssg,info)
    call read_tappsym(nf,ng,nsv,nnp,rg,pg,srg,nsymq,mssg,info)

    call seektomagic(nf,magic_tappatom,mssg,info)
    call read_tappatom(nf,nkd,zoin,znin,nsi,kd,asi,mssg,info)
    deallocate(zoin,znin)

    call seektomagic(nf,magic_tappkpoints,mssg,info)
    call read_tappkpoints(nf,nw,ax,aa,nb1,nb2,ekbt, &
         rf,ndx,ndy,ndz,nk,nsk,mmm,idat, &
         mssg,info)

    call seektomagic(nf,magic_tappstm,mssg,info)
    call read_tappstm(nf,nbias,eferm,vs,mssg,info)

    call abortinput(info,mssg)

    if (mssg.gt.-1) write(6,'(/1x,a/)') 'END input'

    call latgen(ax,bx,aa,bb,va,vb)
    call printlat(ax,bx,aa,bb,va,vb,mssg)

    call calcmesh(bb,ng,pg,nnp,qf,qm,nhx,nhy,nhz,nrx,nry,nrz, &
         ngx,ngy,ngz,qwf,nwx,nwy,nwz,ntt,nrxyz,ngxyz,mssg)

    return
  end subroutine rdstmdata

  subroutine rdwannierdata(aa,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp &
       ,nkd,nsi,kd,asi &
       ,nband,ncomp,nspin,ndens,znext,dspin &
       ,nrx,nry,nrz,nrxyz &
       ,qwf,nwx,nwy,nwz &
       ,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode,icell,ax,wan_params &
       ,uptime,nf,mssg)
    !
    use util
    use subr_inputparam
    use subr_inputsym
    use subr_inputatom
    use subr_inputkpoints
    use subr_inputwannier
    use subr_wannier
    use subr_filemap
    implicit none
    real(kind=8),intent(out):: qf,rf,qm,qc,qwf,znext,dspin
    real(kind=8),intent(out):: qbtp(3)
    real(kind=8),intent(out):: uptime
    integer,intent(in):: nsymq,nopt
    integer,intent(out):: ng,nsv,nnp,nkd,nsi,nband,nspin
    integer,intent(out):: ncomp,ndens
    integer,intent(out):: nk,nw,nb1,nb2,nsk,ismode
    integer,intent(out):: idat(3)
    integer,intent(out):: nrx,nry,nrz,nrxyz,nwx,nwy,nwz
    integer,intent(out):: iopt(nopt)
    integer:: nf,mssg
    !
    real(kind=8),intent(out):: aa(3,3)
    integer,intent(out):: rg(3,3,nsymq),pg(3,nsymq),srg(nsymq)
    integer,pointer:: kd(:),mmm(:,:,:)
    real(kind=8),pointer:: asi(:,:)

    real(kind=8):: ax,bx,va,vb,bb(3,3)
    real(kind=8):: beta,eps,eeps,qexat(2),ekbt
    integer:: ntt,ngx,ngy,ngz,ngxyz
    integer:: nhx,nhy,nhz,ndx,ndy,ndz
    integer:: ndiag0,ndiag1,niter0,niter1,itsb
    real(kind=8):: qbsp
    integer:: nbsf
    real(kind=8):: fwqcom
    integer:: xctp
    real(kind=8):: thr_hybrid
    integer:: osqnrm
    integer:: axetp
    real(kind=8):: qprcnd(6)
    real(kind=8),pointer:: zoin(:),znin(:)
    integer:: icell
    type (wannier_params_struct):: wan_params
    real(kind=8):: thr_noncol,spaxs(3)

    integer:: info
    if (mssg.gt.-1) write(6,'(/1x,a)') 'BEGIN input'

    info = 0

    ! namelist input

    call seektomagic(nf,magic_filemap,mssg,info)
    call read_filemap(nf,mssg,info)

    call seektomagic(nf,magic_inputparam,mssg,info)
    call read_tappinput(nf,ax,aa,ncomp,nspin,ndens,ismode,qf,qm,qc,beta, &
         qbtp,nkd,nsi,nw,znext,dspin,nopt,iopt,eps,eeps,ekbt, &
         uptime,niter0,niter1,itsb,ndiag0,ndiag1, &
         qbsp,nbsf,fwqcom,xctp,qexat,thr_hybrid,osqnrm,axetp, &
         qprcnd,thr_noncol,spaxs,mssg,info)
    nband = nw

    call seektomagic(nf,magic_tappsym,mssg,info)
    call read_tappsym(nf,ng,nsv,nnp,rg,pg,srg,nsymq,mssg,info)

    call seektomagic(nf,magic_tappatom,mssg,info)
    call read_tappatom(nf,nkd,zoin,znin,nsi,kd,asi,mssg,info)
    deallocate(zoin,znin)

    call seektomagic(nf,magic_tappkpoints,mssg,info)
    call read_tappkpoints(nf,nw,ax,aa,nb1,nb2,ekbt, &
         rf,ndx,ndy,ndz,nk,nsk,mmm,idat,&
         mssg,info)

    call seektomagic(nf,magic_tappwannier,mssg,info)
    call read_tappwannier(nf,nspin,nw,aa,nkd,icell,wan_params, &
         mssg,info)

    call abortinput(info,mssg)

    if (mssg.gt.-1) write(6,'(/1x,a/)') 'END input'

    call latgen(ax,bx,aa,bb,va,vb)
    call printlat(ax,bx,aa,bb,va,vb,mssg)

    call calcmesh(bb,ng,pg,nnp,qf,qm,nhx,nhy,nhz,nrx,nry,nrz, &
         ngx,ngy,ngz,qwf,nwx,nwy,nwz,ntt,nrxyz,ngxyz,mssg)

    return
  end subroutine rdwannierdata
end module SYM_MODULE_NAME
