!
! Copyright (c) 2012-2013,2016,2017 Yoshihide Yoshimoto
!
module subr_kpoint
  type irdkpt_struct
     integer:: nsg,ngir
     integer,pointer:: irrg(:),sgrg(:)
     integer:: kp(3)
     real(kind=8):: sk(3)
  end type irdkpt_struct
contains
  function towsc(bb,nk,kp)
    implicit none
    integer,dimension(3):: towsc
    real(kind=8),intent(in):: bb(3,3)
    integer,intent(in):: nk,kp(3)

    integer:: isft, tkp(3), kpws(3), ip, ipws
    real(kind=8):: sk(3),csk(3),a,akpws,eps

    integer,parameter:: sft(3,8) = reshape( &
         (/   0, 0, 0, -1, 0, 0,  0,-1, 0, &
             -1,-1, 0,  0, 0,-1, -1, 0,-1, &
              0,-1,-1, -1,-1,-1 /), (/ 3, 8/) )

    if (minval(kp).lt.0 .or. maxval(kp).ge.nk) then
       write(6,*) 'towsc(): range error of kp'
       stop
    end if

    eps = sum(abs(bb))/dble(nk)*1.0d-8

    akpws = huge(1.0d0)
    do isft = 1,8
       tkp =  kp + nk*sft(:,isft)
       if (minval(tkp) .lt. -nk .or. maxval(tkp).ge.nk) then
          write(6,*) 'towsc(): range error of tkp'
          stop
       end if
          
       sk = dble(tkp)/dble(nk)
       csk = matmul(bb,sk)
       a = csk(1)*csk(1)+csk(2)*csk(2)+csk(3)*csk(3)

       if (abs(a-akpws).lt.eps) then
          ip = (sign(1,tkp(1))+1)/2 + sign(1,tkp(2))+1 &
               + 2*(sign(1,tkp(3))+1)
          ipws = (sign(1,kpws(1))+1)/2 + sign(1,kpws(2))+1 &
               + 2*(sign(1,kpws(3))+1)
          if (ip.gt.ipws) then
             kpws = tkp
             akpws = a
          else if (ip.eq.ipws) then
             write(6,*) 'equivalent kp in towsc'
             write(6,*) kpws
             stop
          end if
       else if (a .lt. akpws) then
          kpws = tkp
          akpws = a
       end if
    end do
    towsc = kpws
    return
  end function towsc

  subroutine calcirdkpt(bb,ng,rg,nnp,pg,srg,nk,nsk,mmm,nto,nir,irdkpt)
    implicit none

    real(kind=8),intent(in):: bb(3,3)
    integer,intent(in):: ng,rg(3,3,ng),nnp,pg(3,ng),srg(ng)
    integer,intent(in):: nk,nsk,mmm(3,2,nsk)
    integer:: nir,nto
    type (irdkpt_struct),pointer:: irdkpt(:)

    integer,allocatable:: mesh(:,:,:),meshws(:,:,:)
    integer:: isk,idx,idy,idz,isx,isy,isz,i1,i2,i3
    integer:: igir,isg,ig,iir,ntoc,kp(3),tkp(3)

    allocate(meshws(-nk:nk,-nk:nk,-nk:nk))
    allocate(mesh(0:nk-1,0:nk-1,0:nk-1))

    !
    ! tile the BZ 
    mesh = 0
    do isk=1,nsk
       idx = mmm(1,2,isk)
       idy = mmm(2,2,isk)
       idz = mmm(3,2,isk)

       isx = mmm(1,1,isk)
       isy = mmm(2,1,isk)
       isz = mmm(3,1,isk)

       do i3=-isz,isz,idz
          do i2=-isy,isy,idy
             do i1=-isx,isx,idx
                kp(1) = i1
                kp(2) = i2
                kp(3) = i3
                do ig=1,ng
                   !
                   tkp = matmul(rg(:,:,ig),kp)
                   tkp = srg(ig)*tkp
                   tkp = modulo(tkp,nk)
                   mesh(tkp(1),tkp(2),tkp(3)) = 1
                end do
             end do
          end do
       end do
    end do

    nto = sum(mesh)

    meshws = 0
    !
    ! count irreducible k points
    nir = 0 
    do i3=0,nk-1
       do i2=0,nk-1
          do i1=0,nk-1
             if (mesh(i1,i2,i3).eq.1) then
                nir = nir + 1 

                kp(1) = i1
                kp(2) = i2
                kp(3) = i3
                do ig=1,ng
                   tkp = matmul(rg(:,:,ig),kp)
                   tkp = srg(ig)*tkp
                   tkp = modulo(tkp,nk)
                   mesh(tkp(1),tkp(2),tkp(3)) = 2
                   !
                   ! convert tkp into the Wigner-Seitz cell
                   tkp = towsc(bb,nk,tkp)
                   meshws(tkp(1),tkp(2),tkp(3)) = 1
                end do
             end if
          end do
       end do
    end do

    deallocate(mesh)

    ntoc = sum(meshws)
    if (ntoc .ne. nto) then
       write(6,*) '# of points in mesh and meshws are differ'
       stop
    end if
    !
    ! inspect irreducible k points
    allocate(irdkpt(nir))
    iir = 0
    ntoc = 0
    do i3=nk,-nk,-1
       do i2=nk,-nk,-1
          do i1=nk,-nk,-1
             if (meshws(i1,i2,i3) .eq. 1) then
                iir = iir + 1

                kp(1) = i1
                kp(2) = i2
                kp(3) = i3
                irdkpt(iir)%nsg = 0
                irdkpt(iir)%ngir = 0
                irdkpt(iir)%kp = kp
                irdkpt(iir)%sk = dble(kp)/dble(nk)
                !
                ! count sub group and corresponding coset
                do ig=1,ng
                   !
                   tkp = matmul(rg(:,:,ig),kp)
                   tkp = srg(ig)*tkp
                   !
                   ! pull back k points on the boundary of Wigner Seitz cell
                   ! to the sampled position
                   tkp = towsc(bb,nk,modulo(tkp,nk))

                   if (meshws(tkp(1),tkp(2),tkp(3)) .eq. 1) then
                      meshws(tkp(1),tkp(2),tkp(3)) = 2
                      irdkpt(iir)%ngir = irdkpt(iir)%ngir + 1
                   else if (meshws(tkp(1),tkp(2),tkp(3)) .eq. 3) then
                      write(6,*) 'calcirdkpt: double counting'
                      stop
                   else if (meshws(tkp(1),tkp(2),tkp(3)) .eq. 0) then
                      write(6,*) 'calcirdkpt: not-marked point is counted'
                      stop
                   end if
                   if (tkp(1).eq.kp(1) .and. tkp(2).eq.kp(2) &
                        .and. tkp(3).eq.kp(3)) then
                      irdkpt(iir)%nsg = irdkpt(iir)%nsg + 1
                   end if
                end do
                if (ng .ne. irdkpt(iir)%nsg*irdkpt(iir)%ngir) then
                   write(6,*) 'calcirdkpt: # of sym. op. != nsg*ngir'
                   stop
                end if
                ntoc = ntoc + irdkpt(iir)%ngir

                !
                ! inspect sub group and corresponding coset
                allocate(irdkpt(iir)%sgrg(irdkpt(iir)%nsg), &
                     irdkpt(iir)%irrg(irdkpt(iir)%ngir))
                isg = 0
                igir = 0
                do ig=1,ng
                   !
                   tkp = matmul(rg(:,:,ig),kp)
                   tkp = srg(ig)*tkp
                   !
                   ! pull back k points on the boundary of Wigner Seitz cell
                   ! to the sampled position
                   tkp = towsc(bb,nk,modulo(tkp,nk))

                   if (meshws(tkp(1),tkp(2),tkp(3)) .eq. 2) then
                      meshws(tkp(1),tkp(2),tkp(3)) = 3
                      igir = igir + 1
                      irdkpt(iir)%irrg(igir) = ig
                   end if
                   if (tkp(1).eq.kp(1) .and. tkp(2).eq.kp(2) &
                        .and. tkp(3).eq.kp(3)) then
                      isg = isg + 1
                      irdkpt(iir)%sgrg(isg) = ig
                   end if
                end do
             end if
          end do
       end do
    end do
    if (ntoc .ne. nto) then
       write(6,*) 'calcirdkpt: nto != ntoc'
       stop
    end if

    ntoc = 0
    do i3=-nk,nk
       do i2=-nk,nk
          do i1=-nk,nk
             if (meshws(i1,i2,i3).eq.3) then
                ntoc = ntoc + 1
             else if (meshws(i1,i2,i3).ne.0) then
                write(6,*) 'calcirdkpt: meshws(i1,i2,i3) != 0,3'
                stop
             end if
          end do
       end do
    end do
    if (ntoc .ne. nto) then
       write(6,*) 'calcirdkpt: nto != ntoc'
       stop
    end if

    deallocate(meshws)
    return
  end subroutine calcirdkpt

  subroutine calcmapkpt(ng,rg,srg,nk,nir,irdkpt,nto,mapkpt)
    implicit none
    integer,intent(in):: ng,rg(3,3,ng),srg(ng),nk,nir,nto
    type (irdkpt_struct),pointer,intent(in):: irdkpt(:)
    integer:: mapkpt(5,-nk:nk,-nk:nk,-nk:nk)

    integer:: ik,ig,kp(3),tkp(3),ukp(3),i1,i2,i3,tnto
    integer:: irg

    tnto = 0
    mapkpt = 0
    do ik=1,nir
       do ig=1,irdkpt(ik)%ngir
          kp = irdkpt(ik)%kp
          irg = irdkpt(ik)%irrg(ig)
          tkp = matmul(rg(:,:,irg),kp)
          tkp = srg(irg)*tkp
          tnto = tnto + 1
          do i3 = -1,1
             do i2 = -1,1
                do i1 = -1,1
                   ukp(1) = tkp(1) + i1*nk
                   ukp(2) = tkp(2) + i2*nk
                   ukp(3) = tkp(3) + i3*nk
                   if (ukp(3).ge.-nk .and. ukp(3).le.nk .and. &
                        ukp(2).ge.-nk .and. ukp(2).le.nk .and. &
                        ukp(1).ge.-nk .and. ukp(1).le.nk) then
                      if ( mapkpt(1,ukp(1),ukp(2),ukp(3)).gt.0 ) then
                         write(6,*) 'duplicate mapkpt in calcmapkpt()'
                         stop
                      end if
                      mapkpt(1,ukp(1),ukp(2),ukp(3)) = ik
                      mapkpt(2,ukp(1),ukp(2),ukp(3)) = irg
                      mapkpt(3,ukp(1),ukp(2),ukp(3)) = i1
                      mapkpt(4,ukp(1),ukp(2),ukp(3)) = i2
                      mapkpt(5,ukp(1),ukp(2),ukp(3)) = i3
                   end if
                end do
             end do
          end do
       end do
    end do
    if (nto.ne.tnto) then
       write(6,*) 'invalid total number of k points in calcmapkpt()'
       stop
    end if

    return
  end subroutine calcmapkpt

  subroutine calcsklist(ng,rg,srg,nk,nir,irdkpt,nto,sklist,itype)
    implicit none
    integer,intent(in):: ng,rg(3,3,ng),srg(ng),nk,nir,nto,itype
    type (irdkpt_struct),pointer,intent(in):: irdkpt(:)
    real(kind=8):: sklist(3,nto)

    integer:: ik,ig,kp(3),tkp(3),tnto
    integer:: irg

    tnto = 0
    sklist = 0.0d0
    do ik=1,nir
       do ig=1,irdkpt(ik)%ngir
          kp = irdkpt(ik)%kp
          irg = irdkpt(ik)%irrg(ig)
          tkp = matmul(rg(:,:,irg),kp)
          tkp = srg(irg)*tkp
          tnto = tnto + 1
          if (itype.eq.0) then
             sklist(:,tnto) = dble(tkp)/nk
          else if (itype.eq.1) then
             sklist(:,tnto) = modulo(dble(tkp)/nk+0.5d0,1.0d0) - 0.5d0
          else
             write(6,*) 'calcsklist: unknown itype'
             stop
          end if
       end do
    end do
    if (nto.ne.tnto) then
       write(6,*) 'invalid total number of k points in calcmapkpt()'
       stop
    end if

    return
  end subroutine calcsklist

  subroutine sort_irdkpt(bb,nir,irdkpt,sirdkpt)
    use util
    implicit none
    real(kind=8):: bb(3,3)
    integer,intent(in):: nir
    type (irdkpt_struct),pointer:: irdkpt(:)
    type (irdkpt_struct),pointer:: sirdkpt(:)

    integer:: iir,ngir,nsg
    real(kind=8):: csk(3)
    real(kind=8):: sq_sk(nir)
    integer:: idx(nir),jir

    do iir = 1, nir
       csk = matmul(bb,irdkpt(iir)%sk)
       sq_sk(iir) = csk(1)*csk(1)+csk(2)*csk(2)+csk(3)*csk(3)
       idx = iir
    end do
    call hsort(nir, sq_sk, idx)
    allocate(sirdkpt(nir))

    do iir=1, nir
       jir = idx(iir)

       ngir = irdkpt(jir)%ngir
       nsg = irdkpt(jir)%nsg
       allocate(sirdkpt(iir)%irrg(ngir),sirdkpt(iir)%sgrg(nsg))
       
       sirdkpt(iir)%nsg = irdkpt(jir)%nsg
       sirdkpt(iir)%ngir = irdkpt(jir)%ngir
       sirdkpt(iir)%kp = irdkpt(jir)%kp
       sirdkpt(iir)%sk = irdkpt(jir)%sk
       sirdkpt(iir)%irrg(:) = irdkpt(jir)%irrg(:)
       sirdkpt(iir)%sgrg(:) = irdkpt(jir)%sgrg(:)
    end do
    return
  end subroutine sort_irdkpt

  subroutine dealloc_irdkpt(nir,irdkpt)
    implicit none
    integer,intent(in):: nir
    type (irdkpt_struct),pointer:: irdkpt(:)

    integer:: iir

    do iir=1,nir
       deallocate(irdkpt(iir)%irrg,irdkpt(iir)%sgrg)
    end do
    deallocate(irdkpt)

    return
  end subroutine dealloc_irdkpt

  subroutine store_irdkpt(nf,nto,nir,irdkpt)
    implicit none
    integer,intent(in):: nf,nto,nir
    type (irdkpt_struct),pointer,intent(in):: irdkpt(:)

    integer:: iir

    write(nf) nto,nir

    do iir=1,nir
       write(nf) irdkpt(iir)%nsg,irdkpt(iir)%ngir
       write(nf) irdkpt(iir)%sk, irdkpt(iir)%kp
       write(nf) irdkpt(iir)%sgrg,irdkpt(iir)%irrg
    end do

    return
  end subroutine store_irdkpt

  subroutine load_irdkpt(nf,nto,nir,irdkpt)
    implicit none
    integer,intent(in):: nf
    integer:: nto,nir
    type (irdkpt_struct),pointer:: irdkpt(:)

    integer:: iir,nsg,ngir

    read(nf) nto,nir

    allocate(irdkpt(nir))
    do iir=1,nir
       read(nf) irdkpt(iir)%nsg,irdkpt(iir)%ngir
       nsg = irdkpt(iir)%nsg
       ngir = irdkpt(iir)%ngir
       read(nf) irdkpt(iir)%sk, irdkpt(iir)%kp
       allocate(irdkpt(iir)%sgrg(nsg),irdkpt(iir)%irrg(ngir))
       read(nf) irdkpt(iir)%sgrg,irdkpt(iir)%irrg
    end do

    return
  end subroutine load_irdkpt

  subroutine calcrco(ng,rg,nnp,pg,srg,nir,irdkpt,nkr,kr,rco)
    implicit none
    integer,intent(in):: ng,rg(3,3,ng),nnp,pg(3,ng),srg(ng),nir,nkr,kr(3,nkr)
    type (irdkpt_struct):: irdkpt(nir)
    real(kind=8):: rco(nkr,nir)

    real(kind=8),parameter:: tpi=2.0d0*3.14159265358979323846d0
    real(kind=8):: ph,sk(3)
    integer:: ikr,iir,igir,ig

    do ikr = 1,nkr
       do iir = 1,nir
          rco(ikr,iir) = 0.0d0
          do igir=1,irdkpt(iir)%ngir
             ig = irdkpt(iir)%irrg(igir)
             sk = matmul(rg(:,:,ig),irdkpt(iir)%sk)
             sk = srg(ig)*sk

             ph = sk(1)*kr(1,ikr)+sk(2)*kr(2,ikr)+sk(3)*kr(3,ikr)

             rco(ikr,iir) = rco(ikr,iir)+cos(tpi*ph)
          end do
       end do
    end do
    return
  end subroutine calcrco

  subroutine meshk(bb,ng,rg,nnp,pg,srg,nk,nsk,mmm,nkr,kr,nks,nkmesh,irdkpt)
    use iounit
    implicit none
    real(kind=8),intent(in):: bb(3,3)
    integer,intent(in):: ng,nnp,nk,nkr,nsk
    integer,intent(in):: rg(3,3,ng), pg(3,ng), srg(ng), mmm(3,2,nsk), kr(3,nkr)
    integer:: nks,nkmesh
    type (irdkpt_struct),pointer:: irdkpt(:)

    real(kind=8),allocatable:: rco(:,:)

    integer:: iks,i

    call calcirdkpt(bb,ng,rg,nnp,pg,srg,nk,nsk,mmm,nkmesh,nks,irdkpt)

    write(6,*) 'NI=',nkmesh
    write(6,100) (iks,(irdkpt(iks)%sk(i),i=1,3), &
         dble(irdkpt(iks)%ngir)/dble(ng),iks=1,nks)
100 format(2(i4,3f8.4,2x,f8.4))

    allocate(rco(nkr,nks))
    call calcrco(ng,rg,nnp,pg,srg,nks,irdkpt,nkr,kr,rco)

    call xrewind(NFSAMPK)
    do iks=1,nks
       write(NFSAMPK) (irdkpt(iks)%sk(i),i=1,3),(rco(i,iks),i=1,nkr)
    end do
    deallocate(rco)

    return
  end subroutine meshk
end module subr_kpoint
