!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
module timeprof
  integer,parameter:: tpf_nspec = 27
  integer,parameter:: tpf_arch = 1, tpf_fftc = 2, tpf_ffts = 3
  integer,parameter:: tpf_mpi = 4, tpf_eig = 5, tpf_npj = 6
  integer,parameter:: tpf_rwf = 7, tpf_hpl = 8, tpf_sot = 9
  integer,parameter:: tpf_fftsc = 10, tpf_fio = 11
  integer,parameter:: tpf_bchrg = 12, tpf_cchrg = 13, tpf_mkdd = 14
  integer,parameter:: tpf_xtrap = 15, tpf_pihxc = 16, tpf_engdf = 17
  integer,parameter:: tpf_ewald = 18, tpf_dhml = 19, tpf_scgt = 20
  integer,parameter:: tpf_mkgmf = 21, tpf_mkvqq = 22, tpf_hplm = 23
  integer,parameter:: tpf_rwfm = 24, tpf_cexch = 25, tpf_potexch = 26
  integer,parameter:: tpf_exclkern = 27

  real(kind=8),private:: tmsum(tpf_nspec), tmbeg(tpf_nspec)
  integer,private:: cnt(tpf_nspec)
  logical,private:: ifstart(tpf_nspec)
  data tmsum/tpf_nspec*0.0d0/, tmbeg/tpf_nspec*0.0d0/, cnt/tpf_nspec*0/
  data ifstart/tpf_nspec*.false./

  logical,private:: tpf_actq = .false.

contains
  subroutine tpf_onoff(swtch)
    implicit none
    integer swtch

    if (swtch.eq.1) then
       tpf_actq = .true.
    else if (swtch.eq.0) then
       tpf_actq = .false.
    end if
  end subroutine tpf_onoff

  subroutine tpf_reset()
    implicit none
    tmsum = 0.0d0
    tmbeg = 0.0d0
    cnt = 0
    ifstart = .false.
    return
  end subroutine tpf_reset

  subroutine tpf_begin(spec)
    use util
    implicit none
    integer:: spec

    if (tpf_actq) then
       if (ifstart(spec)) then
          write(6,'(1x,a,i3)') 'tpf_begin: already started = ', spec
       end if
       tmbeg(spec) = xetime(1)
       ifstart(spec) = .true.
    end if
  end subroutine tpf_begin

  subroutine tpf_end(spec)
    use util
    implicit none
    integer:: spec

    if (tpf_actq) then
       if (.not. ifstart(spec)) then
          write(6,'(1x,a,i3)') 'tpf_end: not started = ', spec
       end if
       tmsum(spec) = tmsum(spec) + xetime(1) - tmbeg(spec)
       cnt(spec) = cnt(spec) + 1
       ifstart(spec) = .false.
    end if
  end subroutine tpf_end

  subroutine tpf_begin_hc(spec)
    use util
    implicit none
    integer:: spec

    if (tpf_actq) then
       if (ifstart(spec)) then
          write(6,'(1x,a,i3)') 'tpf_begin_hc: already started = ', spec
       end if
       tmbeg(spec) = xetime(1)
       ifstart(spec) = .true.
    end if
  end subroutine tpf_begin_hc

  subroutine tpf_end_hc(spec)
    use util
    implicit none
    integer:: spec

    if (tpf_actq) then
       if (.not. ifstart(spec)) then
          write(6,'(1x,a,i3)') 'tpf_end_hc: not started = ', spec
       end if
       tmsum(spec) = tmsum(spec) + xetime(1) - tmbeg(spec)
       ifstart(spec) = .false.
    end if
  end subroutine tpf_end_hc

  subroutine tpf_print()
    implicit none
    write(6,'(1x,a)') 'profile: time count'
100 format(1x,a,f12.6,1x,i6)
    write(6,100) 'arch = ', tmsum(tpf_arch), cnt(tpf_arch)
    write(6,100) 'fftc = ', tmsum(tpf_fftc), cnt(tpf_fftc)
    write(6,100) 'ffts = ', tmsum(tpf_ffts), cnt(tpf_ffts)
    write(6,100) 'mpi = ', tmsum(tpf_mpi), cnt(tpf_mpi)
    write(6,100) 'eig = ', tmsum(tpf_eig), cnt(tpf_eig)
    write(6,100) 'npj = ', tmsum(tpf_npj), cnt(tpf_npj)
    write(6,100) 'rwf = ', tmsum(tpf_rwf), cnt(tpf_rwf)
    write(6,100) 'hpl = ', tmsum(tpf_hpl), cnt(tpf_hpl)
    write(6,100) 'sot = ', tmsum(tpf_sot), cnt(tpf_sot)
    write(6,100) 'fftsc = ', tmsum(tpf_fftsc), cnt(tpf_fftsc)
    write(6,100) 'fio = ', tmsum(tpf_fio), cnt(tpf_fio)
    write(6,100) 'bchrg = ', tmsum(tpf_bchrg), cnt(tpf_bchrg)
    write(6,100) 'cchrg = ', tmsum(tpf_cchrg), cnt(tpf_cchrg)
    write(6,100) 'mkdd = ', tmsum(tpf_mkdd), cnt(tpf_mkdd)
    write(6,100) 'xtrap = ', tmsum(tpf_xtrap), cnt(tpf_xtrap)
    write(6,100) 'pihxc = ', tmsum(tpf_pihxc), cnt(tpf_pihxc)
    write(6,100) 'engdf = ', tmsum(tpf_engdf), cnt(tpf_engdf)
    write(6,100) 'ewald = ', tmsum(tpf_ewald), cnt(tpf_ewald)
    write(6,100) 'dhml = ', tmsum(tpf_dhml), cnt(tpf_dhml)
    write(6,100) 'scgt = ', tmsum(tpf_scgt), cnt(tpf_scgt)
    write(6,100) 'mkgmf = ', tmsum(tpf_mkgmf), cnt(tpf_mkgmf)
    write(6,100) 'mkvqq = ', tmsum(tpf_mkvqq), cnt(tpf_mkvqq)
    write(6,100) 'hplm = ', tmsum(tpf_hplm), cnt(tpf_hplm)
    write(6,100) 'rwfm = ', tmsum(tpf_rwfm), cnt(tpf_rwfm)
    write(6,100) 'cexch = ', tmsum(tpf_cexch), cnt(tpf_cexch)
    write(6,100) 'potexch = ', tmsum(tpf_potexch), cnt(tpf_potexch)
    write(6,100) 'exclkern = ', tmsum(tpf_exclkern), cnt(tpf_exclkern)
  end subroutine tpf_print
end module timeprof
