!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015-2017 Yoshihide Yoshimoto
!

#include "fftgrid.h"

module trace_band
contains
  !
  ! NFOUT :FILE # TO WRITE W.FN. DISTRIBUTION DATA
  subroutine vbband(aa,bb,va,nkd,nsi,asi,kd,ifinvsym,qm,qbtp &
       ,ncomp,nspin,ndens,spaxs,nw,laprj,nhx,nhy,nhz &
       ,nrx,nry,nrz,nvx,nvy,nvz,nwx,nwy,nwz,ndiag,itsb,eps &
       ,nfbnd,nfout,nfrok,nfwfk,ifbunp,ifrho,nbk,nkfi,ak,nbba1,nbba2 &
       ,nsph,ndiv,iaxis,rzero,rmin,rmax &
       ,vg,bkc,ekc,kqg,dd,bspotfc,bspots &
       ,xctp,mgs,dpf,nb1,nb2,wgtk,mssg,myrank,nproc,comm)
    use utilmpi
    use types
    use converter
    use bsplpot
    use wfnc_init
    use meshgrid
    use xcf_types
    use bandchrg
    use exchange

    implicit none

    real(kind=8),intent(in):: aa(3,3),bb(3,3),va
    integer,intent(in):: nkd,nsi,kd(nsi)
    real(kind=8),intent(in):: asi(3,nsi)
    integer,intent(in):: ncomp,nspin,ndens,nw
    real(kind=8),intent(in):: spaxs(3)
    integer,intent(in):: nwx,nwy,nwz,nhx,nhy,nhz,nrx,nry,nrz,nvx,nvy,nvz
    integer,intent(in):: laprj
    integer,intent(in):: ndiag,itsb
    real(kind=8),intent(in):: qm,qbtp(3),eps
    logical,intent(in):: ifinvsym

    integer,intent(in):: nfbnd,nfout,nfrok,nfwfk
    integer,intent(in):: ifbunp,ifrho
    integer,intent(in):: nbk,nkfi(nbk),nsph,nbba1,nbba2
    real(kind=8),intent(in):: ak(3,nbk+1)
    real(kind=8),intent(in):: rzero(3,nsph),rmin(nsph),rmax(nsph)
    integer,intent(in):: ndiv(nsph),iaxis(nsph)

    complex(kind=8),intent(in):: vg(-nvx:nvx,-nvy:nvy,-nvz:nvz,nspin*ndens)
    integer,intent(in):: bkc,ekc,kqg(3,bkc:ekc)
    complex(kind=8),intent(in):: dd(laprj,laprj,nsi,nspin)
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    integer,intent(in):: xctp
    type (meshgrid_struct):: mgs
    type (wfn_for_k):: dpf(mgs%nks,nspin)
    integer,intent(in):: nb1,nb2
    real(kind=8):: wgtk(nb1:nb2,mgs%nks,nspin)
    integer,intent(in):: mssg
    integer,intent(in):: myrank,nproc,comm

    real(kind=8):: savedsk(3),sk(3)
    real(kind=8),allocatable:: vr(:,:)
    integer:: savednkm
    integer:: ldvpj
    real(kind=8),allocatable:: dvpj(:)
    type(wfn_for_k),allocatable:: saveddp(:),dp(:)
!SOPTION CONTAREA(qg)
    complex(kind=8),pointer:: qg(:,:)
    type(qg_for_atm),allocatable:: dqgs(:,:)
    integer ierr
    integer nam
    real(kind=8),allocatable:: ee(:,:)
    real(kind=8),allocatable:: qgk(:,:),akk(:),akg(:)
    integer,allocatable:: kg(:,:),savedkg(:,:)
    integer:: lpnum,maxr
    integer:: ikd, nb, nke, nob, ivpj
    integer:: nlx, nly, nlz, nlxyz
    integer:: icount, nk, nkm, tnkm, nknl, is, idens
    integer:: iseed
    integer:: irnk
    integer:: bkm,ekm
    integer,allocatable:: g_count(:),g_displ(:)
    integer,allocatable:: saved_g_count(:),saved_g_displ(:)

    real(kind=8):: bexa,bexc,bevxc(0:6)
    real(kind=8),allocatable:: brho(:,:),bzwk(:)
    complex(kind=8),allocatable:: hybvg(:,:,:,:)
    !
    allocate(ee(nw,nspin),stat=ierr)
    call chkalloc('vbband','tmp',ierr)

    allocate(dqgs(nkd,ncomp**2),stat=ierr)
    call chkalloc('vbband','qgs',ierr)
    do idens=1,ncomp**2
       do ikd=1,nkd
          nob=bspots(ikd)%nob
          allocate(dqgs(ikd,idens)%qg(bkc:ekc,nob*(nob+1)/2),stat=ierr)
          call chkalloc('vbband','qg',ierr)
       end do
    end do
    allocate(qgk(4,bkc:ekc),stat=ierr)
    call chkalloc('vbband','qgk',ierr)
    call kqg2qgk(bb,ekc-bkc+1,kqg,qgk)
    do idens=1,ncomp**2
       do ikd=1,nkd
          qg => dqgs(ikd,idens)%qg
          nob=bspots(ikd)%nob
          call evqg(bspotfc,bspots(ikd),ekc-bkc+1,qgk,va,nob,idens,qg)
       end do
    end do
    deallocate(qgk)
    !
    nlx = FFT_PAD_NX(nwx*2)
    nly = FFT_PAD_NY(nwy*2)
    nlz = FFT_PAD_NZ(nwz*2)

    nlxyz=nlx*nly*nlz

    allocate(bzwk(mgs%nks), stat=ierr)
    call chkalloc('vbband','bzwk',ierr)
    bzwk(:) = mgs%rco(1,:)
    if (nspin .eq. 1) then
       bzwk = 2.0d0/ncomp*bzwk
    end if
    !
    allocate(hybvg(-nvx:nvx,-nvy:nvy,-nvz:nvz,nspin*ndens),stat=ierr)
    call chkalloc('vbband','hybvg',ierr)
    if (ncomp.eq.1 .and. xcf_exch_factor(xctp) .ne. 0.0d0) then
       allocate(brho(mgs%nrxyz,nspin*ndens),stat=ierr)
       call chkalloc('vbband','brho',ierr)
       call yps3(nspin,dpf,brho,bzwk,wgtk,va,aa,mgs%nkmesh &
            ,ncomp,ndens,mgs%nrx,mgs%nry,mgs%nrz,mgs%nrxyz &
            ,mgs%ng,mgs%nnp,mgs%rg,mgs%pg,mgs%srg &
            ,mgs%nrcell,mgs%nks,mgs%nkm,mgs%kg,mgs%ldkg &
            ,nb1,nb2,mgs%nwk,mgs%idcwk,mgs%irdkpt &
            ,(ifinvsym .or. xcf_exch_factor(xctp).ne.0.0d0) &
            ,1,myrank,nproc,mgs%g_count,mgs%g_displ,comm)
       call pothybrid(va,bb,nspin,ndens,spaxs,brho,vg,bexa,bexc &
            ,bevxc,mgs%nkfxa,mgs%kf,nrx*2,nry*2,nrz*2,nvx,nvy,nvz &
            ,hybvg,xctp,mssg,comm)
       if (mssg.gt.-1) write(6,*) 'local potential for a hybrid functional'
       deallocate(brho)
    else
       hybvg = vg
    end if

    allocate(vr(nlxyz,nspin*ndens),stat=ierr)
    call chkalloc('vbband','vr',ierr)
    if (ncomp.eq.1) then
       do is=1,nspin
          call vg2vr(hybvg(-nvx,-nvy,-nvz,is),vr(1,is), &
               nvx,nvy,nvz,nwx,nwy,nwz,nlx,nly,nlz)
       end do
    else
       do idens=1,ndens
          call vg2vr(hybvg(-nvx,-nvy,-nvz,idens),vr(1,idens), &
               nvx,nvy,nvz,nwx,nwy,nwz,nlx,nly,nlz)
       end do
    end if
    deallocate(hybvg)
    !
    ! BAND CALCULATION
    !
    icount=0
    do nb=1,nbk
       nke=nkfi(nb)
       if(nb.eq.nbk) nke=nke+1
       do nk=1,nke
          if (nkfi(nb).gt.0) then
             sk(1)=ak(1,nb)+(ak(1,nb+1)-ak(1,nb))*dble(nk-1)/dble(nkfi(nb))
             sk(2)=ak(2,nb)+(ak(2,nb+1)-ak(2,nb))*dble(nk-1)/dble(nkfi(nb))
             sk(3)=ak(3,nb)+(ak(3,nb+1)-ak(3,nb))*dble(nk-1)/dble(nkfi(nb))
          else
             sk(:)=ak(:,nb)
          end if
          !JY
          sk(1)=sk(1)-anint(sk(1))
          sk(2)=sk(2)-anint(sk(2))
          sk(3)=sk(3)-anint(sk(3))
          !
          ! ORDERING THE RECIPROCAL LATTICE VECTOR
          !   nam: maximum number of plane waves to expand wave functions
          !
          nam=8*nrx*nry*nrz
          allocate(akk(nam),kg(3,nam),stat=ierr)
          call chkalloc('vbband','ykarr3',ierr)
          !
          !
          ! make plane wave basis for wave function (nkm)
          !
          call ykarr3(akk,kg,sk,bb,nkm,nknl,qm,qm,nam,nhx,nhy,nhz)
          if (mssg.gt.-1) write(6,*) 'NKM=',nkm

          allocate(g_count(0:nproc-1),g_displ(0:nproc-1),stat=ierr)
          call chkalloc('vbband','g_count',ierr)

          if ( (sk(1).eq.0.0d0) .and. (sk(2).eq.0.0d0) &
               .and. (sk(3).eq.0.0d0) ) then
             tnkm = (nkm-1)/2
             g_count(:) = tnkm/nproc
             do irnk=0,mod(tnkm,nproc)-1
                g_count(irnk) = g_count(irnk)+1
             end do
             g_count(:) = 2*g_count(:)
             g_count(0) = g_count(0)+1
          else
             g_count(:) = nkm/nproc
             do irnk=0,mod(nkm,nproc)-1
                g_count(irnk) = g_count(irnk)+1
             end do
          end if
          g_displ(0) = 0
          do irnk=0,nproc-2
             g_displ(irnk+1) = g_displ(irnk)+g_count(irnk)
          end do
          bkm=g_displ(myrank)+1
          ekm=g_displ(myrank)+g_count(myrank)
          !
          lpnum = getlpnum(nkd,bspots)
          maxr = getmaxr(nkd,bspots)
          ldvpj=g_count(myrank)*maxr*lpnum*nkd
          allocate(dvpj(ldvpj),stat=ierr)
          call chkalloc('vbband','dvpj',ierr)
          allocate(akg(bkm:ekm),stat=ierr)
          call chkalloc('vbband','akg',ierr)

          call kg2akg(bb,sk,g_count(myrank),kg(1,bkm),akg)
          ivpj=1
          do ikd=1,nkd
             call evbetg(bspotfc,bspots(ikd),g_count(myrank),akg, &
                  g_count(myrank),maxr,lpnum,dvpj(ivpj))
             ivpj=ivpj+g_count(myrank)*maxr*lpnum
          end do
          deallocate(akg)

          if (mssg.gt.-1) write(6,101) sk,nkm
101       format(' SK=',3(1x,f8.5),' NKM=',i6)
          !
          icount=icount+1
          !
          allocate(dp(nspin),stat=ierr)
          call chkalloc('vbband','dp',ierr)
          do is=1,nspin
             allocate(dp(is)%wf(ncomp,bkm:ekm,nw),stat=ierr)
             call chkalloc('vbband','dp%wf',ierr)
          end do

          if (icount.eq.1) then
             !
             ! MAKE INITIAL RANDOM WAVE FN
             iseed = 456214531
             do is=1,nspin
                call initwfn1k(ncomp,nkm,nw,kg,akk,nhx,nhy,nhz,iseed, &
                     bkm,ekm,dp(is),comm)
             end do
          else
             do is=1,nspin
                !
                if (maxval(abs(sk)).eq.0.0d0) then
                   !
                   ! for gamma point, current implementation of copywfn()
                   ! is not appropriate to copy from other k points.
                   !
                   call initwfn1k(ncomp,nkm,nw,kg,akk,nhx,nhy,nhz,iseed, &
                        bkm,ekm,dp(is),comm)
                else
                   call copywfn(ncomp,nw,savedsk,savednkm,savedkg,saveddp(is)%wf, &
                        sk,nkm,kg,dp(is)%wf,nproc, &
                        saved_g_count,saved_g_displ, &
                        g_count,g_displ,ifinvsym,comm)
                   dp(is)%ifothnm = .false.
                end if
                deallocate(saveddp(is)%wf)
             end do
             deallocate(saveddp,savedkg,saved_g_count,saved_g_displ)
          end if
          deallocate(akk)
          !
          call eigebv(va,bb,nkd,nsi,asi,kd,ifinvsym,qbtp,ncomp,nspin,ndens,nw &
               ,lpnum,maxr,laprj,nrx,nry,nrz,nwx,nwy,nwz,nlx,nly,nlz,nlxyz &
               ,ndiag,itsb,eps &
               ,nfbnd,nfout,nfrok,nfwfk &
               ,ifbunp,ifrho,nbba1,nbba2,nsph,ndiv,iaxis &
               ,rzero,rmin,rmax,icount &
               ,sk,vr,nkm,bkm,ekm,kg,bkc,ekc,kqg,dqgs,dvpj,ldvpj,dp,dd,ee &
               ,bspots,xctp,mgs,dpf,nb1,nb2,bzwk,wgtk &
               ,mssg,myrank,nproc,g_count,g_displ,comm)

          deallocate(dvpj)

          allocate(saveddp(nspin),savedkg(3,nkm), &
               saved_g_count(0:nproc-1),saved_g_displ(0:nproc-1),stat=ierr)
          call chkalloc('vbband','saveddp',ierr)
          do is=1,nspin
             allocate(saveddp(is)%wf(ncomp,bkm:ekm,nw),stat=ierr)
             call chkalloc('vbband','saveddp%wf',ierr)
             saveddp(is)%wf = dp(is)%wf
          end do
          savedsk = sk
          savednkm = nkm
          savedkg(1:3,1:nkm) = kg(1:3,1:nkm)
          saved_g_count = g_count
          saved_g_displ = g_displ

          do is=1,nspin
             deallocate(dp(is)%wf)
          end do
          deallocate(dp)
          deallocate(g_count,g_displ)
          deallocate(kg)
          !
       end do
    end do
    do is=1,nspin
       deallocate(saveddp(is)%wf)
    end do
    deallocate(saveddp,savedkg,saved_g_count,saved_g_displ)
    deallocate(vr)
    deallocate(bzwk)
    do idens=1,ncomp**2
       do ikd=1,nkd
          deallocate(dqgs(ikd,idens)%qg)
       end do
    end do
    deallocate(dqgs)
    deallocate(ee)
    return
  end subroutine vbband

  subroutine eigebv(va,bb,nkd,nsi,asi,kd,ifinvsym,qbtp,ncomp,nspin,ndens,nw &
       ,lpnum,maxr,laprj,nrx,nry,nrz,nwx,nwy,nwz,nlx,nly,nlz,nlxyz &
       ,ndiag,itsb,eps &
       ,nfbnd,nfout,nfrok,nfwfk &
       ,ifbunp,ifrho,nbba1,nbba2,nsph,ndiv,iaxis &
       ,rzero,rmin,rmax,icount &
       ,sk,vr,nkm,bkm,ekm,kg,bkc,ekc,kqg,dqgs,dvpj,ldvpj,dp,dd,ee &
       ,bspots,xctp,mgs,dpf,nb1,nb2,bzwk,wgtk &
       ,mssg,myrank,nproc,g_count,g_displ,comm)
    !
    use diag_hml
    use distrib_wfn
    use types
    use bsplpot
    use meshgrid
    use mpi
    use util
    implicit none
    real(kind=8),intent(in):: bb(3,3),va
    integer,intent(in):: nfbnd,nfout,nfrok,nfwfk
    integer,intent(in):: ifbunp,ifrho
    integer,intent(in):: nkd,nsi,kd(nsi)
    real(kind=8),intent(in):: asi(3,nsi)
    integer,intent(in):: ncomp,ndens
    integer,intent(in):: nspin,nw
    integer,intent(in):: nwx,nwy,nwz,nlx,nly,nlz,nlxyz,nrx,nry,nrz
    integer,intent(in):: lpnum,maxr,laprj
    integer,intent(in):: ndiag,itsb
    real(kind=8),intent(in):: eps,qbtp(3)
    logical,intent(in):: ifinvsym

    real(kind=8),intent(in):: sk(3)
    integer,intent(in):: nkm,bkm,ekm,kg(3,nkm),bkc,ekc,kqg(3,bkc:ekc)
    complex(kind=8),intent(in):: dd(laprj,laprj,nsi,nspin)
    real(kind=8):: ee(nw,nspin)

    integer,intent(in):: icount
    integer,intent(in):: nsph,nbba1,nbba2
    real(kind=8),intent(in):: rzero(3,nsph),rmin(nsph),rmax(nsph)
    integer,intent(in):: ndiv(nsph),iaxis(nsph)

    real(kind=8),intent(in):: vr(nlxyz,nspin*ndens)
    integer,intent(in):: ldvpj
    real(kind=8),intent(in):: dvpj(ldvpj)
    type(qg_for_atm):: dqgs(nkd,ncomp**2)
    type(wfn_for_k):: dp(nspin)
    type(bspot_for_atm):: bspots(nkd)
    integer,intent(in):: xctp
    type (meshgrid_struct):: mgs
    type (wfn_for_k):: dpf(mgs%nks,nspin)
    integer,intent(in):: nb1,nb2
    real(kind=8):: bzwk(mgs%nks),wgtk(nb1:nb2,mgs%nks,nspin)
    integer,intent(in):: mssg
    integer,intent(in):: myrank,nproc,comm
    integer,intent(in):: g_count(0:nproc-1),g_displ(0:nproc-1)

    real(kind=8),allocatable:: out(:,:,:)
    complex(kind=8),allocatable:: pby(:,:,:,:)
    !
    ! CODATA 2014
    real(kind=8):: har
    parameter (har = 27.21138602d0)
    integer:: i,is,ib,isph,ierr
    integer:: it,ip
    real(kind=8):: clk,omax1,omax2,omax3
    !
    integer:: tifrho
    allocate(pby(laprj,nsi,nw,nspin),out(nw,3,nspin), stat=ierr)
    call chkalloc('eigebv','pby,out',ierr)
    !
    pby = (0.0d0, 0.0d0)
    !
    clk=xetime(1)
    if (mssg.gt.-1) write(6,*)'####  SCGDGL START=',clk
    !
    do is=1,nspin
       call scgdgl(ncomp,ndiag,itsb,nwx,nwy,nwz,nlx,nly,nlz,nlxyz, &
            nkm,bkm,ekm,nw, &
            -1,sk,kg,bb,ee(1,is),ndens,vr(1,is),out(1,1,is), &
            lpnum,maxr,dvpj,laprj,dd(1,1,1,is),.false.,eps, &
            ierr,va,asi,bspots,nsi,nkd,kd,qbtp,dp(is),pby(1,1,1,is), &
            ifinvsym, &
            xctp,dpf(:,is),mgs,nb1,nb2,bzwk,wgtk(nb1,1,is), &
            nproc,g_count,g_displ,comm)
    end do
    clk=xetime(1)
    if (mssg.gt.-1) write(6,*)'####  SCGDGL END  =',clk,'  RTN CODE=',ierr

    if (mssg.gt.-1) then
       write(6,*)'EIGEN VALUE'
       if (nspin.eq.2) write(6,'(1x,a)') 'spin = 1'
       write(6,600)ee(:,1)
       if (nspin.eq.2) then
          write(6,'(1x,a)') 'spin = 2'
          write(6,600)ee(:,2)
       end if
    end if
600 format(5e15.7)
    !
    ee = ee*har
    !
    ! CONV CHECK
    omax1=0.0d0
    omax2=0.0d0
    omax3=0.0d0
    do is=1,nspin
       do ib=1,nw
          omax1=max(omax1,out(ib,1,is))
          omax2=max(omax2,out(ib,2,is))
          omax3=max(omax3,out(ib,3,is))
       end do
    end do
    if (mssg.gt.-1) write(6,'(1x,a,3(1x,e11.4))') &
         'EIGEN, RESID, SIN',omax1,omax2,omax3
    !
    if (myrank.eq.0) then
       do is=1,nspin
          write(nfbnd,5040) ee(:,is)
       end do
    end if
5040 format(' ',5e15.7)
    if ( (ifbunp.eq.0) .and. (ifrho.eq.0) ) then
       deallocate(pby, out)
       return
    end if
    !     
    if (myrank.eq.0) then
       write(nfout,5110) icount,sk,nbba1,nbba2,nw
5110 format(i5,3e15.7,3i5,' # No. sk(3)')
       if (ifrho .eq. 1) then
          write(nfrok) sk
       else if (ifrho .eq. 2) then
          write(nfwfk) sk,nkm,ee(nbba1:nbba2,:)
          write(nfwfk) kg
       end if
    end if
    tifrho = ifrho
    do isph=1,nsph
       if (ifbunp .eq. 1) then
          !
          ! One-dimensional distribution
          if (myrank.eq.0) write(nfout,5120) iaxis(isph), &
               rmin(isph),rmax(isph),ndiv(isph)
5120      format(i5,2e15.7,i5)
       else if (ifbunp .eq. 2) then
          !
          ! Radial distribution around the point rzero(3)
          if (myrank.eq.0) write(nfout,5130) (rzero(i,isph),i=1,3), &
               rmin(isph),rmax(isph),ndiv(isph)
5130      format(5e14.6,i5)
       end if

       !
       do is=1,nspin
          call bunpu(ifbunp,tifrho,nfout,nfrok,nfwfk &
               ,ndiv(isph),rmin(isph),rmax(isph),iaxis(isph),rzero(1,isph) &
               ,nw,nbba1,nbba2,kg,nkm &
               ,kd,asi,nsi,nkd,ee(1,is),bb,va &
               ,laprj,nrx,nry,nrz,dp(is),pby &
               ,bkc,ekc,kqg,dqgs,bspots,ncomp &
               ,myrank,nproc,g_count,g_displ,comm)
       end do
       !
       ! XXX: output charge distribution only at the first cycle.
       !
       tifrho = 0
    end do

    if (ifrho .eq. 2) then
       if (myrank.eq.0) write(nfwfk) &
            ((((pby(ip,it,ib,is) &
            ,ip=1,bspots(kd(it))%nprj),it=1,nsi) &
            ,ib=nbba1,nbba2),is=1,nspin)
    end if

    clk=xetime(1)
    if (mssg.gt.-1) write(6,*)'####  BUNPU ',ifbunp,' END  =',clk
    !
    deallocate(pby,out)
    return
  end subroutine eigebv
end module trace_band
