!
! Copyright (c) 2003-2014 Kazuma Nakamura
! Copyright (c) 2014,2017 Yoshihide Yoshimoto
!
module wannier_util
contains
  function towsc(aa,nkb,irp)
    implicit none
    integer,dimension(3):: towsc
    real(kind=8),intent(in):: aa(3,3)
    integer,intent(in):: nkb(3),irp(3)

    integer:: isft, tirp(3), irpws(3)
    real(kind=8):: ctrp(3),a,arpws

    integer,parameter:: sft(3,8) = reshape( &
         (/   0, 0, 0, -1, 0, 0,  0,-1, 0, &
             -1,-1, 0,  0, 0,-1, -1, 0,-1, &
              0,-1,-1, -1,-1,-1 /), (/ 3, 8/) )

    if (minval(irp).lt.0 .or. maxval(irp-nkb).ge.0) then
       write(6,*) 'towsc(): range error of irp'
       stop
    end if

    arpws = huge(1.0d0)
    do isft = 1,8
       tirp =  irp + nkb(:)*sft(:,isft)
       if (minval(tirp+nkb) .lt. 0 .or. maxval(tirp-nkb).ge.0) then
          write(6,*) 'towsc(): range error of tirp'
          stop
       end if
          
       ctrp = matmul(aa,tirp)
       a = ctrp(1)*ctrp(1)+ctrp(2)*ctrp(2)+ctrp(3)*ctrp(3)

       if (a .lt. arpws) then
          irpws = tirp
          arpws = a
       end if
    end do
    towsc = irpws
    return
  end function towsc

  subroutine wan_itplband(aa,nkb,n_occ,H_MAT_R,sk,ee,wf,mssg,info)
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    integer,intent(in):: nkb(3),n_occ
    complex(kind=8),intent(in):: &
         H_MAT_R(n_occ,n_occ,0:nkb(1)-1,0:nkb(2)-1,0:nkb(3)-1)
    real(kind=8),intent(in):: sk(3)
    real(kind=8),intent(out):: ee(n_occ)
    complex(Kind=8),intent(out):: wf(n_occ,n_occ)
    integer,intent(in):: mssg
    integer,intent(out):: info

    real(kind=8):: ph
    complex(kind=8):: pf,ztmp
    real(kind=8),parameter:: tpi = 6.28318530717958648d0

    integer:: ia1,ia2,ia3,ib,jb
    integer:: irp(3),irpws(3)

    complex(kind=8),allocatable:: work(:)
    real(kind=8),allocatable:: rwork(:)
    integer:: lwork

    lwork = -1
    allocate(work(1),rwork(3*n_occ-2))
    call zheev('V','U',n_occ,wf,n_occ,ee, &
         work,lwork,rwork,info)
    lwork = work(1)
    deallocate(work)
    allocate(work(lwork))

    wf = 0.0D0

    if (mssg.eq.1) write(6,'(1x,a)') 'R, to_wiger_seitz_cell(R)'
    do ia3 = 0, nkb(3)-1
       irp(3) = ia3
       do ia2 = 0, nkb(2)-1
          irp(2) = ia2
          do ia1 = 0 , nkb(1)-1
             irp(1) = ia1
             irpws = towsc(aa,nkb,irp)
             if (mssg.eq.1) write(6,'(3(1x,i5),3x,3(1x,i5))') irp,irpws

             ph = tpi*(sk(1)*dble(irpws(1)) &
                  +sk(2)*dble(irpws(2))+sk(3)*dble(irpws(3)))

             pf = dcmplx(cos(ph),sin(ph))

             do ib = 1 , n_occ
                do jb = 1 , n_occ
                   wf(ib,jb) = wf(ib,jb) &
                        + H_MAT_R(ib,jb,ia1,ia2,ia3) * pf
                end do
             end do
          end do
       end do
    end do

    call zheev('V','U',n_occ,wf,n_occ,ee, &
         work,lwork,rwork,info)

    deallocate(work,rwork)

    return
  end subroutine wan_itplband
end module wannier_util
