!
! Copyright (c) 2012-2013 Yoshihide Yoshimoto
!
module b3lyp_xcgga
  private
  public b3lyp_exchcor,b3lyp_x_exchcor,b3lyp_exchcorsp,b3lyp_x_exchcorsp

  real(kind=8),parameter:: b3lyp_a0 = 0.2d0
  real(kind=8),parameter:: b3lyp_ax = 0.72d0
  real(kind=8),parameter:: b3lyp_ac = 0.81d0

  ! VWN3 RPA
  real(kind=8),parameter:: vwny0(2)=(/-0.409286d0, -0.743294d0/)
  real(kind=8),parameter:: vwnb(2)= (/13.0720d0,   20.1231d0/)  
  real(kind=8),parameter:: vwnc(2)= (/42.7198d0,  101.578d0/)
  real(kind=8),parameter:: vwnA(2)= (/0.0621814d0,  0.0310907d0/)

  ! VWN5
  !real(kind=8),parameter:: vwny0(2)=(/-0.10498d0,  -0.32500d0/)
  !real(kind=8),parameter:: vwnb(2)= (/3.72744d0,   7.06042d0/)
  !real(kind=8),parameter:: vwnc(2)= (/12.9352d0,   18.0578d0/)
  !real(kind=8),parameter:: vwnA(2)= (/0.0621814d0, 0.0310907d0/)
contains
  subroutine exb88(np, n, dn, ucex, ucexn, ucexdn, fla, fga)
    implicit none
    integer:: np
    real(kind=8),dimension(np):: n, dn, ucex, ucexn, ucexdn
    real(kind=8):: fla, fga
    !
    real(kind=8):: cx, beta, neps
    ! real(kind=8),parameter:: Pi=3.14159265358979323846d0
    ! parameter(cx=3^(4/3)/(2*(4*pi)**(1/3)))
    parameter(cx=0.930525736349100d0)
    parameter(beta=0.0042d0)
    parameter(neps=1.0d-5)
    integer:: i
    real(kind=8):: fx, x, xn, xdn, fxx, t, asx, rho
    !
!$omp parallel do private(x,xn,xdn,fx,fxx,t,asx,rho)
    do i=1,np
       if (n(i).le.0.0d0) then
          ucex(i) = 0.0d0
          ucexn(i) = 0.0d0
          ucexdn(i) = 0.0d0
          cycle
       end if
       t = exp(-n(i)/neps)
       rho = n(i)+neps*t
       x = dn(i)*rho**(-4.0d0/3.0d0)
       xn = -(4.0d0/3.0d0)*dn(i)*rho**(-7.0d0/3.0d0)*(1.0d0-t)
       xdn = rho**(-4.0d0/3.0d0)
       ! asx = asinh(x)
       asx = log(x+sqrt(x*x+1))
       t = 1.0d0 + 6.0d0*beta*x*asx
       fx = fla + fga/cx*beta*x*x/t
       fxx = fga/cx*( 2.0d0*beta*x/t &
            - beta*x*x*( 6.0d0*beta*asx+6.0d0*beta*x/sqrt(1+x*x) )/(t*t) )

       ucex(i) = -cx*n(i)**(4.0d0/3.0d0)*fx
       ucexn(i) = -(4.0d0/3.0d0)*cx*n(i)**(1.0d0/3.0d0)*fx &
            - cx*n(i)**(4.0d0/3.0d0)*fxx*xn
       ucexdn(i) = - cx*n(i)**(4.0d0/3.0d0)*fxx*xdn
    end do
    return
  end subroutine exb88

  subroutine vwn_cor(y0, b, c, A, np, n, epsc, vc)
    implicit none
    real(kind=8),intent(in):: y0, b, c, A
    integer,intent(in):: np
    real(kind=8),intent(in):: n(np)
    real(kind=8):: epsc(np), vc(np)
    !
    real(kind=8),parameter:: Pi=3.14159265358979323846d0
    real(kind=8),parameter:: neps=1.0d-8

    integer:: i
    real(kind=8):: rs, lcy, ucy, ucy0, rsmecrs, t1, t2, Q
    !
    Q = sqrt(4*c-b*b)
    ucy0 = y0*y0 + b*y0 + c

!$omp parallel do private(rs,lcy,ucy,rsmecrs,t1,t2)
    do i=1,np
       if (n(i).lt.neps) then
          epsc(i) = 0.0d0
          vc(i)= 0.0d0
          cycle
       end if
       rs = (3.0d0/(4.0d0*Pi*n(i)))**(1.0d0/3.0d0)
       lcy = sqrt(rs)
       ucy = lcy*lcy + b*lcy + c
       t1 = atan(Q/(2.0d0*lcy+b))
       t2 = log((lcy-y0)*(lcy-y0)/ucy)

       epsc(i) = 0.5d0*A*(log(lcy*lcy/ucy) + 2.0d0*b/Q*t1 &
            - b*y0/ucy0*(t2 + 2*(b+2*y0)/Q*t1))

       rsmecrs = lcy*A*(-b*y0*(ucy*(2*(lcy-y0)/ucy-(2*lcy+b)*(lcy-y0)&
            &**2/ucy**2)/(lcy-y0)**2-4*(2*y0+b)/((2*lcy+b)**2*(Q**2&
            &/(2*lcy+b)**2+1)))/ucy0-4*b/((2*lcy+b)**2*(Q**2/(2*lcy&
            &+b)**2+1))+(2*lcy/ucy-lcy**2*(2*lcy+b)/ucy**2)*ucy/lcy&
            &**2)/4.0d+0

       vc(i) = epsc(i) - 1.0d0/3.0d0*rsmecrs
    end do
    return
  end subroutine vwn_cor

  subroutine lyp_cor(np,n,dn,ec,ecn,ecdn)
    implicit none
    integer:: np
    real(kind=8),dimension(np),intent(in):: n, dn
    real(kind=8),dimension(np):: ec, ecn, ecdn
    !
    real(kind=8):: Pi,neps,a,b,c,d,cf
    parameter(Pi=3.14159265358979323846d0)
    parameter(neps=1.0d-8)
    parameter(a=0.04918d0,b=0.132d0,c=0.2533d0,d=0.349d0)
    parameter(cf=3.0d0/10.0d0*(3.0d0*Pi*Pi)**(2.0d0/3.0d0))
    !
    integer:: i
    real(kind=8):: rho,drho,delta,tmp1,tmp2

!$omp parallel do private(i,tmp1,tmp2,rho,drho,delta)
    do i=1,np
       if (n(i).lt.neps) then
          ec(i) = 0.0d0 
          ecn(i) = 0.0d0 
          ecdn(i) = 0.0d0
          cycle
       end if
       rho = n(i)
       drho = dn(i)
       delta = d*rho**((-1.0D+0)/3.0D+0)/(d*rho**((-1.0D+0)/3.0D+0)+1) &
            +c*rho**((-1.0D+0)/3.0D+0)
       tmp1 = exp(-c*rho**((-1.0D+0)/3.0D+0))
       tmp2 = d*rho**((-1.0D+0)/3.0D+0)+1
       ec(i) = -a*b*rho**((-1.1D+1)/3.0D+0)*(rho**2*(4*cf*rho**(8.0D&
            &+0/3.0D+0)+drho**2*(delta/1.8D+1+(-5.0D+0)/2.0D+0)/2.0D&
            &+0+drho**2*((-7.0D+0)*delta/1.8D+1+4.7D+1/1.8D+1)+drho&
            &**2*(-d*rho**((-1.0D+0)/3.0D+0)/tmp2-c*rho**((-1.0D+0)&
            &/3.0D+0)+11)/3.6D+1)/4.0D+0+(-1.1D+1)*drho**2*rho**2&
            &/2.4D+1)*tmp1/tmp2-a*rho/tmp2
       ecn(i) = 1.1D+1*a*b*rho**((-1.4D+1)/3.0D+0)*(rho**2*(4*cf*rho&
            &**(8.0D+0/3.0D+0)+drho**2*(delta/1.8D+1+(-5.0D+0)/2.0D&
            &+0)/2.0D+0+drho**2*((-7.0D+0)*delta/1.8D+1+4.7D+1/1.8D&
            &+1)+drho**2*(-d*rho**((-1.0D+0)/3.0D+0)/tmp2-c*rho**((&
            &-1.0D+0)/3.0D+0)+11)/3.6D+1)/4.0D+0+(-1.1D+1)*drho**2&
            &*rho**2/2.4D+1)*tmp1/(3.0D+0*tmp2)-a*b*c*(rho**2*(4*cf&
            &*rho**(8.0D+0/3.0D+0)+drho**2*(delta/1.8D+1+(-5.0D+0)&
            &/2.0D+0)/2.0D+0+drho**2*((-7.0D+0)*delta/1.8D+1+4.7D+1&
            &/1.8D+1)+drho**2*(-d*rho**((-1.0D+0)/3.0D+0)/tmp2-c*rho&
            &**((-1.0D+0)/3.0D+0)+11)/3.6D+1)/4.0D+0+(-1.1D+1)*drho&
            &**2*rho**2/2.4D+1)*tmp1/(tmp2*rho**5)/3.0D+0-a*b*d*(rho&
            &**2*(4*cf*rho**(8.0D+0/3.0D+0)+drho**2*(delta/1.8D+1+(&
            &-5.0D+0)/2.0D+0)/2.0D+0+drho**2*((-7.0D+0)*delta/1.8D+1&
            &+4.7D+1/1.8D+1)+drho**2*(-d*rho**((-1.0D+0)/3.0D+0)/tmp2&
            &-c*rho**((-1.0D+0)/3.0D+0)+11)/3.6D+1)/4.0D+0+(-1.1D+1)&
            &*drho**2*rho**2/2.4D+1)*tmp1/(tmp2**2*rho**5)/3.0D+0-a*b&
            &*rho**((-1.1D+1)/3.0D+0)*(rho*(4*cf*rho**(8.0D+0/3.0D+0)&
            &+drho**2*(delta/1.8D+1+(-5.0D+0)/2.0D+0)/2.0D+0+drho**2&
            &*((-7.0D+0)*delta/1.8D+1+4.7D+1/1.8D+1)+drho**2*(-d*rho&
            &**((-1.0D+0)/3.0D+0)/tmp2-c*rho**((-1.0D+0)/3.0D+0)+11)&
            &/3.6D+1)/2.0D+0+rho**2*(3.2D+1*cf*rho**(5.0D+0/3.0D+0)&
            &/3.0D+0+drho**2*(d*rho**((-4.0D+0)/3.0D+0)/tmp2/3.0D+0+c&
            &*rho**((-4.0D+0)/3.0D+0)/3.0D+0-d**2*rho**((-5.0D+0)&
            &/3.0D+0)/tmp2**2/3.0D+0)/3.6D+1+(-1.3D+1)*drho**2*(-d&
            &*rho**((-4.0D+0)/3.0D+0)/tmp2/3.0D+0-c*rho**((-4.0D+0)&
            &/3.0D+0)/3.0D+0+d**2*rho**((-5.0D+0)/3.0D+0)/tmp2**2&
            &/3.0D+0)/3.6D+1)/4.0D+0+(-1.1D+1)*drho**2*rho/1.2D+1)&
            &*tmp1/tmp2-a*d*rho**((-1.0D+0)/3.0D+0)/tmp2**2/3.0D+0-a&
            &/tmp2
       ecdn(i) = -a*b*rho**((-1.1D+1)/3.0D+0)*((drho*(delta/1.8D+1+(&
            &-5.0D+0)/2.0D+0)+2*drho*((-7.0D+0)*delta/1.8D+1+4.7D+1&
            &/1.8D+1)+drho*(-d*rho**((-1.0D+0)/3.0D+0)/tmp2-c*rho**((&
            &-1.0D+0)/3.0D+0)+11)/1.8D+1)*rho**2/4.0D+0+(-1.1D+1)&
            &*drho*rho**2/1.2D+1)*tmp1/tmp2
    end do

    return
  end subroutine lyp_cor

  subroutine lyp_corsp(np,pn,pdn,dn,ec,ecn,ecpdn,ecdn)
    implicit none
    integer:: np
    real(kind=8),dimension(np,2),intent(in):: pn, pdn
    real(kind=8),dimension(np),intent(in):: dn
    real(kind=8),dimension(np):: ec
    real(kind=8),dimension(np,2):: ecn, ecpdn
    real(kind=8),dimension(np):: ecdn
    !
    real(kind=8):: Pi,neps,a,b,c,d,cf
    parameter(Pi=3.14159265358979323846d0)
    parameter(neps=1.0d-8)
    parameter(a=0.04918d0,b=0.132d0,c=0.2533d0,d=0.349d0)
    parameter(cf=3.0d0/10.0d0*(3.0d0*Pi*Pi)**(2.0d0/3.0d0))
    !
    integer:: i
    real(kind=8):: rho,rhoa,rhob,drhoa,drhob,drho
    real(kind=8):: delta,tmp1,tmp2

!$omp parallel do private(i,rho,rhoa,rhob,drhoa,drhob,drho), &
!$omp private(delta,tmp1,tmp2)
    do i=1,np
       rho = pn(i,1)+pn(i,2)
       if (rho .le. neps) then
          ec(i) = 0.0d0
          ecn(i,1) = 0.0d0
          ecn(i,2) = 0.0d0
          ecdn(i) = 0.0d0
          ecpdn(i,1) = 0.0d0
          ecpdn(i,2) = 0.0d0
          cycle
       end if
       rhoa = pn(i,1)
       rhob = pn(i,2)
       drhoa = pdn(i,1)
       drhob = pdn(i,2)
       drho = dn(i)
       delta = d*rho**((-1.0D+0)/3.0D+0)/(d*rho**((-1.0D+0)/3.0D+0)+1) &
            +c*rho**((-1.0D+0)/3.0D+0)
       tmp1 = exp(-c*rho**((-1.0D+0)/3.0D+0))
       tmp2 = d*rho**((-1.0D+0)/3.0D+0)+1

       ec(i) = -a*b*rho**((-1.1D+1)/3.0D+0)*(rhoa*rhob*((drhob**2&
            &+drhoa**2)*(delta/1.8D+1+(-5.0D+0)/2.0D+0)+drho**2*((&
            &-7.0D+0)*delta/1.8D+1+4.7D+1/1.8D+1)+(drhob**2*rhob/rho&
            &+drhoa**2*rhoa/rho)*(-d*rho**((-1.0D+0)/3.0D+0)/tmp2-c&
            &*rho**((-1.0D+0)/3.0D+0)+11)/9.0D+0+2**(1.1D+1/3.0D+0)&
            &*cf*(rhob**(8.0D+0/3.0D+0)+rhoa**(8.0D+0/3.0D+0)))+(&
            &-2.0D+0)*drho**2*rho**2/3.0D+0+drhoa**2*(2.0D+0*rho**2&
            &/3.0D+0-rhob**2)+drhob**2*(2.0D+0*rho**2/3.0D+0-rhoa&
            &**2))*tmp1/tmp2-4*a*rhoa*rhob/(rho*tmp2)
       ecn(i,1) = 1.1D+1*a*b*rho**((-1.4D+1)/3.0D+0)*(rhoa*rhob&
            &*((drhob**2+drhoa**2)*(delta/1.8D+1+(-5.0D+0)/2.0D+0)&
            &+drho**2*((-7.0D+0)*delta/1.8D+1+4.7D+1/1.8D+1)+(drhob&
            &**2*rhob/rho+drhoa**2*rhoa/rho)*(-d*rho**((-1.0D+0)/3.0D&
            &+0)/tmp2-c*rho**((-1.0D+0)/3.0D+0)+11)/9.0D+0+2**(1.1D+1&
            &/3.0D+0)*cf*(rhob**(8.0D+0/3.0D+0)+rhoa**(8.0D+0/3.0D&
            &+0)))+(-2.0D+0)*drho**2*rho**2/3.0D+0+drhoa**2*(2.0D+0&
            &*rho**2/3.0D+0-rhob**2)+drhob**2*(2.0D+0*rho**2/3.0D+0&
            &-rhoa**2))*tmp1/(3.0D+0*tmp2)-a*b*c*(rhoa*rhob*((drhob&
            &**2+drhoa**2)*(delta/1.8D+1+(-5.0D+0)/2.0D+0)+drho**2*((&
            &-7.0D+0)*delta/1.8D+1+4.7D+1/1.8D+1)+(drhob**2*rhob/rho&
            &+drhoa**2*rhoa/rho)*(-d*rho**((-1.0D+0)/3.0D+0)/tmp2-c&
            &*rho**((-1.0D+0)/3.0D+0)+11)/9.0D+0+2**(1.1D+1/3.0D+0)&
            &*cf*(rhob**(8.0D+0/3.0D+0)+rhoa**(8.0D+0/3.0D+0)))+(&
            &-2.0D+0)*drho**2*rho**2/3.0D+0+drhoa**2*(2.0D+0*rho**2&
            &/3.0D+0-rhob**2)+drhob**2*(2.0D+0*rho**2/3.0D+0-rhoa&
            &**2))*tmp1/(rho**5*tmp2)/3.0D+0-a*b*d*(rhoa*rhob*((drhob&
            &**2+drhoa**2)*(delta/1.8D+1+(-5.0D+0)/2.0D+0)+drho**2*((&
            &-7.0D+0)*delta/1.8D+1+4.7D+1/1.8D+1)+(drhob**2*rhob/rho&
            &+drhoa**2*rhoa/rho)*(-d*rho**((-1.0D+0)/3.0D+0)/tmp2-c&
            &*rho**((-1.0D+0)/3.0D+0)+11)/9.0D+0+2**(1.1D+1/3.0D+0)&
            &*cf*(rhob**(8.0D+0/3.0D+0)+rhoa**(8.0D+0/3.0D+0)))+(&
            &-2.0D+0)*drho**2*rho**2/3.0D+0+drhoa**2*(2.0D+0*rho**2&
            &/3.0D+0-rhob**2)+drhob**2*(2.0D+0*rho**2/3.0D+0-rhoa&
            &**2))*tmp1/(rho**5*tmp2**2)/3.0D+0-a*b*rho**((-1.1D+1)&
            &/3.0D+0)*(rhob*((drhob**2+drhoa**2)*(delta/1.8D+1+(-5.0D&
            &+0)/2.0D+0)+drho**2*((-7.0D+0)*delta/1.8D+1+4.7D+1/1.8D&
            &+1)+(drhob**2*rhob/rho+drhoa**2*rhoa/rho)*(-d*rho**((&
            &-1.0D+0)/3.0D+0)/tmp2-c*rho**((-1.0D+0)/3.0D+0)+11)/9.0D&
            &+0+2**(1.1D+1/3.0D+0)*cf*(rhob**(8.0D+0/3.0D+0)+rhoa&
            &**(8.0D+0/3.0D+0)))+rhoa*rhob*((drhoa**2/rho-drhob**2&
            &*rhob/rho**2-drhoa**2*rhoa/rho**2)*(-d*rho**((-1.0D+0)&
            &/3.0D+0)/tmp2-c*rho**((-1.0D+0)/3.0D+0)+11)/9.0D+0&
            &+(drhob**2*rhob/rho+drhoa**2*rhoa/rho)*(d*rho**((-4.0D&
            &+0)/3.0D+0)/tmp2/3.0D+0-d**2*rho**((-5.0D+0)/3.0D+0)&
            &/tmp2**2/3.0D+0+c*rho**((-4.0D+0)/3.0D+0)/3.0D+0)/9.0D+0&
            &+(drhob**2+drhoa**2)*(-d*rho**((-4.0D+0)/3.0D+0)/tmp2&
            &/3.0D+0+d**2*rho**((-5.0D+0)/3.0D+0)/tmp2**2/3.0D+0-c&
            &*rho**((-4.0D+0)/3.0D+0)/3.0D+0)/1.8D+1+(-7.0D+0)*drho&
            &**2*(-d*rho**((-4.0D+0)/3.0D+0)/tmp2/3.0D+0+d**2*rho**((&
            &-5.0D+0)/3.0D+0)/tmp2**2/3.0D+0-c*rho**((-4.0D+0)/3.0D&
            &+0)/3.0D+0)/1.8D+1+2**(2.0D+1/3.0D+0)*cf*rhoa**(5.0D+0&
            &/3.0D+0)/3.0D+0)+drhob**2*(4.0D+0*rho/3.0D+0-2*rhoa)&
            &+4.0D+0*drhoa**2*rho/3.0D+0+(-4.0D+0)*drho**2*rho/3.0D&
            &+0)*tmp1/tmp2-4*a*rhob/(rho*tmp2)+4*a*rhoa*rhob/(rho**2&
            &*tmp2)+(-4.0D+0)*a*d*rhoa*rhob*rho**((-7.0D+0)/3.0D+0)&
            &/(3.0D+0*tmp2**2)
       ecpdn(i,1) = -a*b*rho**((-1.1D+1)/3.0D+0)*(rhoa*rhob*(2*drhoa&
            &*(delta/1.8D+1+(-5.0D+0)/2.0D+0)+2.0D+0*drhoa*rhoa*(-d&
            &*rho**((-1.0D+0)/3.0D+0)/tmp2-c*rho**((-1.0D+0)/3.0D+0)&
            &+11)/(9.0D+0*rho))+2*drhoa*(2.0D+0*rho**2/3.0D+0-rhob&
            &**2))*tmp1/tmp2
       ecdn(i) = -a*b*rho**((-1.1D+1)/3.0D+0)*(2*drho*rhoa*rhob*((&
            &-7.0D+0)*delta/1.8D+1+4.7D+1/1.8D+1)+(-4.0D+0)*drho*rho&
            &**2/3.0D+0)*tmp1/tmp2
       ecn(i,2) = 1.1D+1*a*b*rho**((-1.4D+1)/3.0D+0)*(rhoa*rhob&
            &*((drhob**2+drhoa**2)*(delta/1.8D+1+(-5.0D+0)/2.0D+0)&
            &+drho**2*((-7.0D+0)*delta/1.8D+1+4.7D+1/1.8D+1)+(drhob&
            &**2*rhob/rho+drhoa**2*rhoa/rho)*(-d*rho**((-1.0D+0)/3.0D&
            &+0)/tmp2-c*rho**((-1.0D+0)/3.0D+0)+11)/9.0D+0+2**(1.1D+1&
            &/3.0D+0)*cf*(rhob**(8.0D+0/3.0D+0)+rhoa**(8.0D+0/3.0D&
            &+0)))+(-2.0D+0)*drho**2*rho**2/3.0D+0+drhoa**2*(2.0D+0&
            &*rho**2/3.0D+0-rhob**2)+drhob**2*(2.0D+0*rho**2/3.0D+0&
            &-rhoa**2))*tmp1/(3.0D+0*tmp2)-a*b*c*(rhoa*rhob*((drhob&
            &**2+drhoa**2)*(delta/1.8D+1+(-5.0D+0)/2.0D+0)+drho**2*((&
            &-7.0D+0)*delta/1.8D+1+4.7D+1/1.8D+1)+(drhob**2*rhob/rho&
            &+drhoa**2*rhoa/rho)*(-d*rho**((-1.0D+0)/3.0D+0)/tmp2-c&
            &*rho**((-1.0D+0)/3.0D+0)+11)/9.0D+0+2**(1.1D+1/3.0D+0)&
            &*cf*(rhob**(8.0D+0/3.0D+0)+rhoa**(8.0D+0/3.0D+0)))+(&
            &-2.0D+0)*drho**2*rho**2/3.0D+0+drhoa**2*(2.0D+0*rho**2&
            &/3.0D+0-rhob**2)+drhob**2*(2.0D+0*rho**2/3.0D+0-rhoa&
            &**2))*tmp1/(rho**5*tmp2)/3.0D+0-a*b*d*(rhoa*rhob*((drhob&
            &**2+drhoa**2)*(delta/1.8D+1+(-5.0D+0)/2.0D+0)+drho**2*((&
            &-7.0D+0)*delta/1.8D+1+4.7D+1/1.8D+1)+(drhob**2*rhob/rho&
            &+drhoa**2*rhoa/rho)*(-d*rho**((-1.0D+0)/3.0D+0)/tmp2-c&
            &*rho**((-1.0D+0)/3.0D+0)+11)/9.0D+0+2**(1.1D+1/3.0D+0)&
            &*cf*(rhob**(8.0D+0/3.0D+0)+rhoa**(8.0D+0/3.0D+0)))+(&
            &-2.0D+0)*drho**2*rho**2/3.0D+0+drhoa**2*(2.0D+0*rho**2&
            &/3.0D+0-rhob**2)+drhob**2*(2.0D+0*rho**2/3.0D+0-rhoa&
            &**2))*tmp1/(rho**5*tmp2**2)/3.0D+0-a*b*rho**((-1.1D+1)&
            &/3.0D+0)*(rhoa*((drhob**2+drhoa**2)*(delta/1.8D+1+(-5.0D&
            &+0)/2.0D+0)+drho**2*((-7.0D+0)*delta/1.8D+1+4.7D+1/1.8D&
            &+1)+(drhob**2*rhob/rho+drhoa**2*rhoa/rho)*(-d*rho**((&
            &-1.0D+0)/3.0D+0)/tmp2-c*rho**((-1.0D+0)/3.0D+0)+11)/9.0D&
            &+0+2**(1.1D+1/3.0D+0)*cf*(rhob**(8.0D+0/3.0D+0)+rhoa&
            &**(8.0D+0/3.0D+0)))+rhoa*rhob*((drhob**2/rho-drhob**2&
            &*rhob/rho**2-drhoa**2*rhoa/rho**2)*(-d*rho**((-1.0D+0)&
            &/3.0D+0)/tmp2-c*rho**((-1.0D+0)/3.0D+0)+11)/9.0D+0&
            &+(drhob**2*rhob/rho+drhoa**2*rhoa/rho)*(d*rho**((-4.0D&
            &+0)/3.0D+0)/tmp2/3.0D+0-d**2*rho**((-5.0D+0)/3.0D+0)&
            &/tmp2**2/3.0D+0+c*rho**((-4.0D+0)/3.0D+0)/3.0D+0)/9.0D+0&
            &+(drhob**2+drhoa**2)*(-d*rho**((-4.0D+0)/3.0D+0)/tmp2&
            &/3.0D+0+d**2*rho**((-5.0D+0)/3.0D+0)/tmp2**2/3.0D+0-c&
            &*rho**((-4.0D+0)/3.0D+0)/3.0D+0)/1.8D+1+(-7.0D+0)*drho&
            &**2*(-d*rho**((-4.0D+0)/3.0D+0)/tmp2/3.0D+0+d**2*rho**((&
            &-5.0D+0)/3.0D+0)/tmp2**2/3.0D+0-c*rho**((-4.0D+0)/3.0D&
            &+0)/3.0D+0)/1.8D+1+2**(2.0D+1/3.0D+0)*cf*rhob**(5.0D+0&
            &/3.0D+0)/3.0D+0)+drhoa**2*(4.0D+0*rho/3.0D+0-2*rhob)&
            &+4.0D+0*drhob**2*rho/3.0D+0+(-4.0D+0)*drho**2*rho/3.0D&
            &+0)*tmp1/tmp2-4*a*rhoa/(rho*tmp2)+4*a*rhoa*rhob/(rho**2&
            &*tmp2)+(-4.0D+0)*a*d*rhoa*rhob*rho**((-7.0D+0)/3.0D+0)&
            &/(3.0D+0*tmp2**2)
       ecpdn(i,2) = -a*b*rho**((-1.1D+1)/3.0D+0)*(rhoa*rhob*(2*drhob&
            &*(delta/1.8D+1+(-5.0D+0)/2.0D+0)+2.0D+0*drhob*rhob*(-d&
            &*rho**((-1.0D+0)/3.0D+0)/tmp2-c*rho**((-1.0D+0)/3.0D+0)&
            &+11)/(9.0D+0*rho))+2*drhob*(2.0D+0*rho**2/3.0D+0-rhoa&
            &**2))*tmp1/tmp2
    end do

    return
  end subroutine lyp_corsp

  subroutine b3lyp_exchcor(np, n, dn, exc, fxcn, fxcdn, mssg)
    implicit none
    integer:: np
    real(kind=8),dimension(np),intent(in):: n, dn
    real(kind=8),dimension(np):: exc, fxcn, fxcdn
    integer,intent(in):: mssg
    !
    real(kind=8),dimension(:),allocatable:: na,dna,epscvwn,vcvwn,ex,exn,exdn
    real(kind=8),dimension(:),allocatable:: ec,ecn,ecdn
    real(kind=8),parameter:: neps=1.0d-8
    integer:: i
    !
    if (mssg.gt.0) then
       write(6,*) 'using B3LYP'
    end if
    allocate(na(np),dna(np),ec(np),ecn(np),ecdn(np), &
         ex(np),exn(np),exdn(np),epscvwn(np),vcvwn(np))
    !
    na = 0.5d0*n
    dna = 0.5d0*dn

    ! pothybrid() b3lyp -> 1.0
    call exb88(np,na,dna,ex,exn,exdn,1.0d0,b3lyp_ax)
    call vwn_cor(vwny0(1),vwnb(1),vwnc(1),vwnA(1), &
         np,n,epscvwn,vcvwn)
    call lyp_cor(np,n,dn,ec,ecn,ecdn)
    !
!$omp parallel do
    do i=1,np
       if (n(i).lt.neps) then
          exc(i) = 0.0d0
          fxcn(i) = 0.0d0
          fxcdn(i) = 0.0d0
          cycle
       end if
       exc(i) = ( 2.0d0*ex(i) + b3lyp_ac*ec(i) )/n(i) &
            + (1.0d0-b3lyp_ac)*epscvwn(i)
       fxcn(i) = exn(i) + b3lyp_ac*ecn(i) + (1.0d0-b3lyp_ac)*vcvwn(i)
       fxcdn(i) = exdn(i) + b3lyp_ac*ecdn(i)
    end do
    !
    deallocate(na,dna,epscvwn,vcvwn,ex,exn,exdn,ec,ecn,ecdn)
    return
  end subroutine b3lyp_exchcor

  subroutine b3lyp_x_exchcor(np, n, dn, exc, fxcn, fxcdn, mssg)
    implicit none
    integer:: np
    real(kind=8),dimension(np),intent(in):: n, dn
    real(kind=8),dimension(np):: exc, fxcn, fxcdn
    integer,intent(in):: mssg
    !
    real(kind=8),dimension(:),allocatable:: na,dna,ex,exn,exdn
    real(kind=8),parameter:: neps=1.0d-8
    integer:: i
    !
    if (mssg.gt.0) then
       write(6,*) 'using B3LYP'
    end if
    allocate(na(np),dna(np),ex(np),exn(np),exdn(np))
    !
    na = 0.5d0*n
    dna = 0.5d0*dn

    call exb88(np,na,dna,ex,exn,exdn,b3lyp_a0,0.0d0)
    !
!$omp parallel do
    do i=1,np
       if (n(i).lt.neps) then
          exc(i) = 0.0d0
          fxcn(i) = 0.0d0
          fxcdn(i) = 0.0d0
          cycle
       end if
       exc(i) = 2.0d0*ex(i)/n(i)
       fxcn(i) = exn(i)
       fxcdn(i) = exdn(i)
    end do
    !
    deallocate(na,dna,ex,exn,exdn)
    return
  end subroutine b3lyp_x_exchcor

  subroutine b3lyp_exchcorsp(np, pn, pdn, dn, exc, pfxcn, pfxcdn, &
       fxcdn, mssg)
    implicit none
    integer:: np
    real(kind=8),dimension(np,2),intent(in):: pn, pdn
    real(kind=8),intent(in):: dn(np)
    real(kind=8),intent(out):: exc
    real(kind=8):: fxcdn(np)
    real(kind=8),dimension(np,2):: pfxcn, pfxcdn
    integer,intent(in):: mssg
    !
    real(kind=8),dimension(:),allocatable:: n,ec
    real(kind=8),dimension(:),allocatable:: epscvwn1,vcvwn1,epscvwn2,vcvwn2
    real(kind=8),dimension(:,:),allocatable:: ex,exn,exdn
    real(kind=8),dimension(:,:),allocatable:: ecn,ecpdn
    integer:: i
    real(kind=8):: epscvwn,vcvwn,zeta,fz,dfz
    !
    if (mssg.gt.0) then
       write(6,*) 'using B3LYP'
    end if
    allocate(n(np),epscvwn1(np),vcvwn1(np),epscvwn2(np),vcvwn2(np), &
         ec(np),ecn(np,2),ecpdn(np,2), &
         ex(np,2),exn(np,2),exdn(np,2))
    !
    n = pn(:,1)+pn(:,2)

    ! pothybrid() a0 -> 1.0
    call exb88(np,pn(:,1),pdn(:,1),ex(:,1),exn(:,1),exdn(:,1),1.0d0,b3lyp_ax)
    call exb88(np,pn(:,2),pdn(:,2),ex(:,2),exn(:,2),exdn(:,2),1.0d0,b3lyp_ax)
    call vwn_cor(vwny0(1),vwnb(1),vwnc(1),vwnA(1), &
         np,n,epscvwn1,vcvwn1)
    call vwn_cor(vwny0(2),vwnb(2),vwnc(2),vwnA(2), &
         np,n,epscvwn2,vcvwn2)
    call lyp_corsp(np,pn,pdn,dn,ec,ecn,ecpdn,fxcdn)
    !
!$omp parallel do reduction(+:exc),private(zeta,fz,dfz,epscvwn,vcvwn)
    do i=1,np
       if (n(i).gt.0.0d0) then
          zeta = (pn(i,1)-pn(i,2))/n(i)
          zeta = min(1.0d0,zeta)
          zeta = max(-1.0d0,zeta)
       else
          zeta = 0.0d0
       end if
       fz = (-2.0d0 + (1.0d0-zeta)**(4.0d0/3.0d0) &
            + (1.0d0+zeta)**(4.0d0/3.0d0))/ &
            (-2.0d0 + 2.0d0**(4.0d0/3.0d0))
       dfz = (4.0d0/3.0d0)/(-2.0d0 + 2.0d0**(4.0d0/3.0d0))* &
            ( (1.0d0+zeta)**(1.0d0/3.0d0) &
            - (1.0d0-zeta)**(1.0d0/3.0d0) )

       epscvwn = epscvwn1(i) + (epscvwn2(i)-epscvwn1(i))*fz

       exc = exc + ex(i,1) + ex(i,2) + b3lyp_ac*ec(i) &
            + (1.0d0-b3lyp_ac)*epscvwn*n(i)

       vcvwn = vcvwn1(i) + fz*(vcvwn2(i)-vcvwn1(i)) &
            + (epscvwn2(i)-epscvwn1(i))*(1-zeta)*dfz
       pfxcn(i,1) = exn(i,1) + b3lyp_ac*ecn(i,1) + (1.0d0-b3lyp_ac)*vcvwn

       vcvwn = vcvwn1(i) + fz*(vcvwn2(i)-vcvwn1(i)) &
            + (epscvwn2(i)-epscvwn1(i))*(-1-zeta)*dfz
       pfxcn(i,2) = exn(i,2) + b3lyp_ac*ecn(i,2) + (1.0d0-b3lyp_ac)*vcvwn

       pfxcdn(i,1) = exdn(i,1) + b3lyp_ac*ecpdn(i,1)
       pfxcdn(i,2) = exdn(i,2) + b3lyp_ac*ecpdn(i,2)
       fxcdn(i) = b3lyp_ac*fxcdn(i)
    end do
    !
    deallocate(n,epscvwn1,vcvwn1,epscvwn2,vcvwn2,ex,exn,exdn,ec,ecn,ecpdn)
    return
  end subroutine b3lyp_exchcorsp

  subroutine b3lyp_x_exchcorsp(np, pn, pdn, dn, exc, pfxcn, pfxcdn, &
       fxcdn, mssg)
    implicit none
    integer:: np
    real(kind=8),dimension(np,2),intent(in):: pn, pdn
    real(kind=8),intent(in):: dn(np)
    real(kind=8),intent(out):: exc
    real(kind=8):: fxcdn(np)
    real(kind=8),dimension(np,2):: pfxcn, pfxcdn
    integer,intent(in):: mssg
    !
    real(kind=8),dimension(:),allocatable:: n
    real(kind=8),dimension(:,:),allocatable:: ex,exn,exdn
    integer:: i
    !
    if (mssg.gt.0) then
       write(6,*) 'using B3LYP'
    end if
    allocate(n(np),ex(np,2),exn(np,2),exdn(np,2))
    !
    n = pn(:,1)+pn(:,2)

    call exb88(np,pn(:,1),pdn(:,1),ex(:,1),exn(:,1),exdn(:,1), &
         b3lyp_a0,0.0d0)
    call exb88(np,pn(:,2),pdn(:,2),ex(:,2),exn(:,2),exdn(:,2), &
         b3lyp_a0,0.0d0)
    !
!$omp parallel do reduction(+:exc)
    do i=1,np
       exc = exc + ex(i,1) + ex(i,2)
       pfxcn(i,1) = exn(i,1)
       pfxcn(i,2) = exn(i,2)
       pfxcdn(i,1) = exdn(i,1)
       pfxcdn(i,2) = exdn(i,2)
       fxcdn(i) = 0.0d0
    end do
    !
    deallocate(n,ex,exn,exdn)
    return
  end subroutine b3lyp_x_exchcorsp

end module b3lyp_xcgga
